/* Automatically @generated by tree-sitter v0.25.10 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 1137
#define LARGE_STATE_COUNT 30
#define SYMBOL_COUNT 372
#define ALIAS_COUNT 0
#define TOKEN_COUNT 195
#define EXTERNAL_TOKEN_COUNT 0
#define FIELD_COUNT 47
#define MAX_ALIAS_SEQUENCE_LENGTH 7
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 131
#define SUPERTYPE_COUNT 8

enum ts_symbol_identifiers {
  anon_sym_module = 1,
  anon_sym_version = 2,
  anon_sym_is = 3,
  anon_sym_end = 4,
  anon_sym_import = 5,
  anon_sym_LBRACK = 6,
  anon_sym_RBRACK = 7,
  anon_sym_from = 8,
  anon_sym_COLON_COLON = 9,
  anon_sym_as = 10,
  sym_identifier = 11,
  anon_sym_COLON = 12,
  anon_sym_AT = 13,
  anon_sym_EQ = 14,
  anon_sym_assert = 15,
  sym_controlled_language_tag = 16,
  anon_sym_LPAREN = 17,
  anon_sym_RPAREN = 18,
  anon_sym_COMMA = 19,
  sym_reserved_self = 20,
  anon_sym_event = 21,
  anon_sym_with = 22,
  anon_sym_for = 23,
  anon_sym_def = 24,
  anon_sym_LBRACE = 25,
  anon_sym_RBRACE = 26,
  sym_quoted_string = 27,
  sym_language_tag = 28,
  anon_sym_LT = 29,
  aux_sym_iri_token1 = 30,
  anon_sym_GT = 31,
  anon_sym_POUND_LBRACK = 32,
  sym_hex_byte = 33,
  sym_double = 34,
  sym_decimal = 35,
  sym_integer = 36,
  sym_rational = 37,
  sym_unsigned = 38,
  anon_sym_true = 39,
  anon_sym_u22a4 = 40,
  anon_sym_false = 41,
  anon_sym_u22a5 = 42,
  sym_value_empty_sequence = 43,
  anon_sym_Thing = 44,
  anon_sym_Nothing = 45,
  anon_sym_real = 46,
  anon_sym_rational = 47,
  anon_sym_anyURI = 48,
  anon_sym_base64Binary = 49,
  anon_sym_boolean = 50,
  anon_sym_date = 51,
  anon_sym_dateTime = 52,
  anon_sym_decimal = 53,
  anon_sym_double = 54,
  anon_sym_duration = 55,
  anon_sym_float = 56,
  anon_sym_gDay = 57,
  anon_sym_gMonth = 58,
  anon_sym_gMonthDay = 59,
  anon_sym_gYearMonth = 60,
  anon_sym_gYear = 61,
  anon_sym_hexBinary = 62,
  anon_sym_string = 63,
  anon_sym_time = 64,
  anon_sym_dateTimeStamp = 65,
  anon_sym_dayTimeDuration = 66,
  anon_sym_yearMonthDuration = 67,
  anon_sym_integer = 68,
  anon_sym_long = 69,
  anon_sym_int = 70,
  anon_sym_short = 71,
  anon_sym_byte = 72,
  anon_sym_nonNegativeInteger = 73,
  anon_sym_positiveInteger = 74,
  anon_sym_unsignedLong = 75,
  anon_sym_unsignedInt = 76,
  anon_sym_unsigned = 77,
  anon_sym_unsignedShort = 78,
  anon_sym_unsignedByte = 79,
  anon_sym_nonPositiveInteger = 80,
  anon_sym_negativeInteger = 81,
  anon_sym_normalizedString = 82,
  anon_sym_token = 83,
  anon_sym_language = 84,
  anon_sym_binary = 85,
  anon_sym_iri = 86,
  anon_sym_without = 87,
  anon_sym_datatype = 88,
  sym_opaque = 89,
  anon_sym_length = 90,
  anon_sym_maxLength = 91,
  anon_sym_minLength = 92,
  anon_sym_fractionDigits = 93,
  anon_sym_totalDigits = 94,
  anon_sym_maxExclusive = 95,
  anon_sym_minExclusive = 96,
  anon_sym_maxInclusive = 97,
  anon_sym_minInclusive = 98,
  anon_sym_explicitTimezone = 99,
  anon_sym_required = 100,
  anon_sym_prohibited = 101,
  anon_sym_optional = 102,
  anon_sym_pattern = 103,
  sym_kw_is_fixed = 104,
  anon_sym_dimension = 105,
  anon_sym_parent = 106,
  anon_sym_source = 107,
  anon_sym_entity = 108,
  anon_sym_enum = 109,
  anon_sym_of = 110,
  anon_sym_metric = 111,
  anon_sym_group = 112,
  anon_sym_on = 113,
  anon_sym_ref = 114,
  anon_sym_property = 115,
  anon_sym_rdf = 116,
  anon_sym_a = 117,
  anon_sym_type = 118,
  anon_sym_structure = 119,
  anon_sym_class = 120,
  sym_wildcard = 121,
  anon_sym_union = 122,
  anon_sym_identity = 123,
  sym_unknown_type = 124,
  anon_sym_ordered = 125,
  anon_sym_unordered = 126,
  anon_sym_unique = 127,
  anon_sym_nonunique = 128,
  anon_sym_DOT_DOT = 129,
  anon_sym_SLASH_EQ = 130,
  anon_sym_u2260 = 131,
  sym_op_greater_than = 132,
  anon_sym_LT_EQ = 133,
  anon_sym_u2264 = 134,
  anon_sym_GT_EQ = 135,
  anon_sym_u2265 = 136,
  anon_sym_not = 137,
  anon_sym_u00ac = 138,
  anon_sym_and = 139,
  anon_sym_u2227 = 140,
  anon_sym_or = 141,
  anon_sym_u2228 = 142,
  anon_sym_xor = 143,
  anon_sym_u22bb = 144,
  anon_sym_implies = 145,
  anon_sym_EQ_EQ_GT = 146,
  anon_sym_u21d2 = 147,
  anon_sym_iff = 148,
  anon_sym_LT_EQ_EQ_GT = 149,
  anon_sym_u21d4 = 150,
  anon_sym_forall = 151,
  anon_sym_u2200 = 152,
  anon_sym_exists = 153,
  anon_sym_u2203 = 154,
  anon_sym_exists_BANG = 155,
  anon_sym_u2203_BANG = 156,
  anon_sym_u222a = 157,
  anon_sym_intersects = 158,
  anon_sym_u2229 = 159,
  anon_sym_subset = 160,
  anon_sym_u2282 = 161,
  anon_sym_subseteq = 162,
  anon_sym_u2286 = 163,
  anon_sym_supset = 164,
  anon_sym_u2283 = 165,
  anon_sym_supseteq = 166,
  anon_sym_u2287 = 167,
  anon_sym_complement = 168,
  anon_sym_u2216 = 169,
  anon_sym_product = 170,
  anon_sym_u2a09 = 171,
  anon_sym_in = 172,
  anon_sym_u2208 = 173,
  anon_sym_u2209 = 174,
  sym_seq_builder_separator = 175,
  anon_sym_STAR = 176,
  anon_sym_u00d7 = 177,
  anon_sym_SLASH = 178,
  anon_sym_u00f7 = 179,
  anon_sym_mod = 180,
  anon_sym_PERCENT = 181,
  anon_sym_PLUS = 182,
  sym_math_op_subtract = 183,
  anon_sym_DOT = 184,
  anon_sym_u2218 = 185,
  anon_sym_COLON_EQ = 186,
  anon_sym_u2254 = 187,
  anon_sym_DASH_GT = 188,
  anon_sym_u2192 = 189,
  anon_sym_LT_DASH = 190,
  anon_sym_u2190 = 191,
  sym_comment_aligned = 192,
  sym_comment_local = 193,
  sym_comment_page = 194,
  sym_module = 195,
  sym_module_version = 196,
  sym_module_body = 197,
  sym_import_statement = 198,
  sym_from_clause = 199,
  sym_module_path_absolute = 200,
  sym_module_path_relative = 201,
  sym__module_path_actual = 202,
  sym_member_import = 203,
  sym_module_import = 204,
  sym_qualified_identifier = 205,
  sym_identifier_reference = 206,
  sym_annotation = 207,
  sym_annotation_property = 208,
  sym_constraint = 209,
  sym_annotation_only_body = 210,
  sym_informal_constraint = 211,
  sym_formal_constraint = 212,
  sym_constraint_sentence = 213,
  sym_simple_sentence = 214,
  sym_atomic_sentence = 215,
  sym_equation = 216,
  sym_inequation = 217,
  sym_boolean_sentence = 218,
  sym_unary_boolean_sentence = 219,
  sym_binary_boolean_sentence = 220,
  sym_expression_sentence = 221,
  sym_set_expression_sentence = 222,
  sym_arithmetic_expression_sentence = 223,
  sym_quantified_sentence = 224,
  sym_quantified_variable_binding = 225,
  sym_quantified_variable = 226,
  sym_variable = 227,
  sym_term = 228,
  sym_functional_term = 229,
  sym_function_composition = 230,
  sym_predicate_value = 231,
  sym_sequence_of_predicate_values = 232,
  sym_reserved_event = 233,
  sym_sentence_with_environment = 234,
  sym_keyword_function_def = 235,
  sym_function_def = 236,
  sym_function_signature = 237,
  sym_function_parameter = 238,
  sym_function_type_reference = 239,
  sym_function_body = 240,
  sym_sequence_builder = 241,
  sym_value = 242,
  sym_simple_value = 243,
  sym_string = 244,
  sym_iri = 245,
  sym_binary = 246,
  sym_boolean = 247,
  sym_boolean_truth = 248,
  sym_boolean_falsity = 249,
  sym_value_constructor = 250,
  sym_mapping_value = 251,
  sym_sequence_of_values = 252,
  sym_builtin_types = 253,
  sym_definition = 254,
  sym_from_definition_clause = 255,
  sym_mixin_with_members = 256,
  sym_mixin_without_members = 257,
  sym_mixin_member = 258,
  sym_datatype_def = 259,
  sym_datatype_base_type_reference = 260,
  sym_datatype_type_restrictions = 261,
  sym_restriction_facet = 262,
  sym_length_restriction_facet = 263,
  sym_digit_restriction_facet = 264,
  sym_value_restriction_facet = 265,
  sym_restriction_value = 266,
  sym_tz_restriction_facet = 267,
  sym_tz_restriction_value = 268,
  sym_pattern_restriction_facet = 269,
  sym_dimension_def = 270,
  sym_dimension_body = 271,
  sym_dimension_parent = 272,
  sym_source_entity = 273,
  sym_entity_def = 274,
  sym_entity_body = 275,
  sym_enum_def = 276,
  sym_enum_body = 277,
  sym_event_def = 278,
  sym_event_body = 279,
  sym_metric_def = 280,
  sym_metric_group_def = 281,
  sym_metric_event_binding = 282,
  sym_metric_group_body = 283,
  sym_metric_ref = 284,
  sym_property_def = 285,
  sym_rdf_def = 286,
  sym_structure_def = 287,
  sym_structure_body = 288,
  sym_type_class_def = 289,
  sym_type_class_body = 290,
  sym_type_parameter = 291,
  sym_type_parameter_restriction = 292,
  sym_type_restriction_argument = 293,
  sym_union_def = 294,
  sym_union_body = 295,
  sym_entity_identity = 296,
  sym_member = 297,
  sym_member_def = 298,
  sym_annotation_member_def = 299,
  sym_property_ref = 300,
  sym_type_reference = 301,
  sym_mapping_type = 302,
  sym_cardinality_expression = 303,
  sym_cardinality_reference_expression = 304,
  sym_sequence_ordering = 305,
  sym_sequence_uniqueness = 306,
  sym_cardinality_range = 307,
  sym_value_variant = 308,
  sym_type_variant = 309,
  sym_op_equality = 310,
  sym_inequality_relation = 311,
  sym_op_inequality = 312,
  sym_op_less_than = 313,
  sym_op_less_than_or_equal = 314,
  sym_op_greater_than_or_equal = 315,
  sym_logical_op_negation = 316,
  sym_logical_connective = 317,
  sym_logical_conjunction = 318,
  sym_logical_disjunction = 319,
  sym_logical_exclusive_disjunction = 320,
  sym_logical_implication = 321,
  sym_logical_biconditional = 322,
  sym_logical_quantifier = 323,
  sym_logical_quantifier_universal = 324,
  sym_logical_quantifier_existential = 325,
  sym_logical_quantifier_existential_one = 326,
  sym_set_operator = 327,
  sym_set_op_union = 328,
  sym_set_op_intersection = 329,
  sym_set_op_subset = 330,
  sym_set_op_subset_or_equal = 331,
  sym_set_op_supset = 332,
  sym_set_op_supset_or_equal = 333,
  sym_set_op_complement = 334,
  sym_set_op_product = 335,
  sym_set_membership = 336,
  sym_set_op_is_member = 337,
  sym_set_op_is_not_member = 338,
  sym_math_operator = 339,
  sym_math_op_multiply = 340,
  sym_math_op_divide = 341,
  sym_math_op_modulo = 342,
  sym_math_op_add = 343,
  sym_type_op_combiner = 344,
  sym_line_comment = 345,
  aux_sym_module_body_repeat1 = 346,
  aux_sym_module_body_repeat2 = 347,
  aux_sym_module_body_repeat3 = 348,
  aux_sym_import_statement_repeat1 = 349,
  aux_sym__module_path_actual_repeat1 = 350,
  aux_sym__actual_arguments_repeat1 = 351,
  aux_sym_function_composition_repeat1 = 352,
  aux_sym_sequence_of_predicate_values_repeat1 = 353,
  aux_sym_sentence_with_environment_repeat1 = 354,
  aux_sym_function_signature_repeat1 = 355,
  aux_sym_sequence_builder_repeat1 = 356,
  aux_sym_binary_repeat1 = 357,
  aux_sym_sequence_of_values_repeat1 = 358,
  aux_sym__mixin_member_or_sequence_repeat1 = 359,
  aux_sym_datatype_type_restrictions_repeat1 = 360,
  aux_sym_pattern_restriction_facet_repeat1 = 361,
  aux_sym_dimension_body_repeat1 = 362,
  aux_sym_dimension_body_repeat2 = 363,
  aux_sym_enum_body_repeat1 = 364,
  aux_sym_metric_group_body_repeat1 = 365,
  aux_sym__rdf_types_repeat1 = 366,
  aux_sym_type_class_body_repeat1 = 367,
  aux_sym__class_type_parameters_repeat1 = 368,
  aux_sym_type_parameter_repeat1 = 369,
  aux_sym_type_parameter_restriction_repeat1 = 370,
  aux_sym_union_body_repeat1 = 371,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [anon_sym_module] = "module",
  [anon_sym_version] = "version",
  [anon_sym_is] = "is",
  [anon_sym_end] = "end",
  [anon_sym_import] = "import",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_from] = "from",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_as] = "as",
  [sym_identifier] = "identifier",
  [anon_sym_COLON] = ":",
  [anon_sym_AT] = "@",
  [anon_sym_EQ] = "=",
  [anon_sym_assert] = "assert",
  [sym_controlled_language_tag] = "controlled_language_tag",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_COMMA] = ",",
  [sym_reserved_self] = "reserved_self",
  [anon_sym_event] = "event",
  [anon_sym_with] = "with",
  [anon_sym_for] = "for",
  [anon_sym_def] = "def",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [sym_quoted_string] = "quoted_string",
  [sym_language_tag] = "language_tag",
  [anon_sym_LT] = "<",
  [aux_sym_iri_token1] = "iri_token1",
  [anon_sym_GT] = ">",
  [anon_sym_POUND_LBRACK] = "#[",
  [sym_hex_byte] = "hex_byte",
  [sym_double] = "double",
  [sym_decimal] = "decimal",
  [sym_integer] = "integer",
  [sym_rational] = "rational",
  [sym_unsigned] = "unsigned",
  [anon_sym_true] = "true",
  [anon_sym_u22a4] = "\u22a4",
  [anon_sym_false] = "false",
  [anon_sym_u22a5] = "\u22a5",
  [sym_value_empty_sequence] = "value_empty_sequence",
  [anon_sym_Thing] = "Thing",
  [anon_sym_Nothing] = "Nothing",
  [anon_sym_real] = "real",
  [anon_sym_rational] = "rational",
  [anon_sym_anyURI] = "anyURI",
  [anon_sym_base64Binary] = "base64Binary",
  [anon_sym_boolean] = "boolean",
  [anon_sym_date] = "date",
  [anon_sym_dateTime] = "dateTime",
  [anon_sym_decimal] = "decimal",
  [anon_sym_double] = "double",
  [anon_sym_duration] = "duration",
  [anon_sym_float] = "float",
  [anon_sym_gDay] = "gDay",
  [anon_sym_gMonth] = "gMonth",
  [anon_sym_gMonthDay] = "gMonthDay",
  [anon_sym_gYearMonth] = "gYearMonth",
  [anon_sym_gYear] = "gYear",
  [anon_sym_hexBinary] = "hexBinary",
  [anon_sym_string] = "string",
  [anon_sym_time] = "time",
  [anon_sym_dateTimeStamp] = "dateTimeStamp",
  [anon_sym_dayTimeDuration] = "dayTimeDuration",
  [anon_sym_yearMonthDuration] = "yearMonthDuration",
  [anon_sym_integer] = "integer",
  [anon_sym_long] = "long",
  [anon_sym_int] = "int",
  [anon_sym_short] = "short",
  [anon_sym_byte] = "byte",
  [anon_sym_nonNegativeInteger] = "nonNegativeInteger",
  [anon_sym_positiveInteger] = "positiveInteger",
  [anon_sym_unsignedLong] = "unsignedLong",
  [anon_sym_unsignedInt] = "unsignedInt",
  [anon_sym_unsigned] = "unsigned",
  [anon_sym_unsignedShort] = "unsignedShort",
  [anon_sym_unsignedByte] = "unsignedByte",
  [anon_sym_nonPositiveInteger] = "nonPositiveInteger",
  [anon_sym_negativeInteger] = "negativeInteger",
  [anon_sym_normalizedString] = "normalizedString",
  [anon_sym_token] = "token",
  [anon_sym_language] = "language",
  [anon_sym_binary] = "binary",
  [anon_sym_iri] = "iri",
  [anon_sym_without] = "without",
  [anon_sym_datatype] = "datatype",
  [sym_opaque] = "opaque",
  [anon_sym_length] = "length",
  [anon_sym_maxLength] = "maxLength",
  [anon_sym_minLength] = "minLength",
  [anon_sym_fractionDigits] = "fractionDigits",
  [anon_sym_totalDigits] = "totalDigits",
  [anon_sym_maxExclusive] = "maxExclusive",
  [anon_sym_minExclusive] = "minExclusive",
  [anon_sym_maxInclusive] = "maxInclusive",
  [anon_sym_minInclusive] = "minInclusive",
  [anon_sym_explicitTimezone] = "explicitTimezone",
  [anon_sym_required] = "required",
  [anon_sym_prohibited] = "prohibited",
  [anon_sym_optional] = "optional",
  [anon_sym_pattern] = "pattern",
  [sym_kw_is_fixed] = "kw_is_fixed",
  [anon_sym_dimension] = "dimension",
  [anon_sym_parent] = "parent",
  [anon_sym_source] = "source",
  [anon_sym_entity] = "entity",
  [anon_sym_enum] = "enum",
  [anon_sym_of] = "of",
  [anon_sym_metric] = "metric",
  [anon_sym_group] = "group",
  [anon_sym_on] = "on",
  [anon_sym_ref] = "ref",
  [anon_sym_property] = "property",
  [anon_sym_rdf] = "rdf",
  [anon_sym_a] = "a",
  [anon_sym_type] = "type",
  [anon_sym_structure] = "structure",
  [anon_sym_class] = "class",
  [sym_wildcard] = "wildcard",
  [anon_sym_union] = "union",
  [anon_sym_identity] = "identity",
  [sym_unknown_type] = "unknown_type",
  [anon_sym_ordered] = "ordered",
  [anon_sym_unordered] = "unordered",
  [anon_sym_unique] = "unique",
  [anon_sym_nonunique] = "nonunique",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_u2260] = "\u2260",
  [sym_op_greater_than] = "op_greater_than",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_u2264] = "\u2264",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_u2265] = "\u2265",
  [anon_sym_not] = "not",
  [anon_sym_u00ac] = "\u00ac",
  [anon_sym_and] = "and",
  [anon_sym_u2227] = "\u2227",
  [anon_sym_or] = "or",
  [anon_sym_u2228] = "\u2228",
  [anon_sym_xor] = "xor",
  [anon_sym_u22bb] = "\u22bb",
  [anon_sym_implies] = "implies",
  [anon_sym_EQ_EQ_GT] = "==>",
  [anon_sym_u21d2] = "\u21d2",
  [anon_sym_iff] = "iff",
  [anon_sym_LT_EQ_EQ_GT] = "<==>",
  [anon_sym_u21d4] = "\u21d4",
  [anon_sym_forall] = "forall",
  [anon_sym_u2200] = "\u2200",
  [anon_sym_exists] = "exists",
  [anon_sym_u2203] = "\u2203",
  [anon_sym_exists_BANG] = "exists!",
  [anon_sym_u2203_BANG] = "\u2203!",
  [anon_sym_u222a] = "\u222a",
  [anon_sym_intersects] = "intersects",
  [anon_sym_u2229] = "\u2229",
  [anon_sym_subset] = "subset",
  [anon_sym_u2282] = "\u2282",
  [anon_sym_subseteq] = "subseteq",
  [anon_sym_u2286] = "\u2286",
  [anon_sym_supset] = "supset",
  [anon_sym_u2283] = "\u2283",
  [anon_sym_supseteq] = "supseteq",
  [anon_sym_u2287] = "\u2287",
  [anon_sym_complement] = "complement",
  [anon_sym_u2216] = "\u2216",
  [anon_sym_product] = "product",
  [anon_sym_u2a09] = "\u2a09",
  [anon_sym_in] = "in",
  [anon_sym_u2208] = "\u2208",
  [anon_sym_u2209] = "\u2209",
  [sym_seq_builder_separator] = "seq_builder_separator",
  [anon_sym_STAR] = "*",
  [anon_sym_u00d7] = "\u00d7",
  [anon_sym_SLASH] = "/",
  [anon_sym_u00f7] = "\u00f7",
  [anon_sym_mod] = "mod",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PLUS] = "+",
  [sym_math_op_subtract] = "math_op_subtract",
  [anon_sym_DOT] = ".",
  [anon_sym_u2218] = "\u2218",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_u2254] = "\u2254",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_u2192] = "\u2192",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_u2190] = "\u2190",
  [sym_comment_aligned] = "comment_aligned",
  [sym_comment_local] = "comment_local",
  [sym_comment_page] = "comment_page",
  [sym_module] = "module",
  [sym_module_version] = "module_version",
  [sym_module_body] = "module_body",
  [sym_import_statement] = "import_statement",
  [sym_from_clause] = "from_clause",
  [sym_module_path_absolute] = "module_path_absolute",
  [sym_module_path_relative] = "module_path_relative",
  [sym__module_path_actual] = "_module_path_actual",
  [sym_member_import] = "member_import",
  [sym_module_import] = "module_import",
  [sym_qualified_identifier] = "qualified_identifier",
  [sym_identifier_reference] = "identifier_reference",
  [sym_annotation] = "annotation",
  [sym_annotation_property] = "annotation_property",
  [sym_constraint] = "constraint",
  [sym_annotation_only_body] = "annotation_only_body",
  [sym_informal_constraint] = "informal_constraint",
  [sym_formal_constraint] = "formal_constraint",
  [sym_constraint_sentence] = "constraint_sentence",
  [sym_simple_sentence] = "simple_sentence",
  [sym_atomic_sentence] = "atomic_sentence",
  [sym_equation] = "equation",
  [sym_inequation] = "inequation",
  [sym_boolean_sentence] = "boolean_sentence",
  [sym_unary_boolean_sentence] = "unary_boolean_sentence",
  [sym_binary_boolean_sentence] = "binary_boolean_sentence",
  [sym_expression_sentence] = "expression_sentence",
  [sym_set_expression_sentence] = "set_expression_sentence",
  [sym_arithmetic_expression_sentence] = "arithmetic_expression_sentence",
  [sym_quantified_sentence] = "quantified_sentence",
  [sym_quantified_variable_binding] = "quantified_variable_binding",
  [sym_quantified_variable] = "quantified_variable",
  [sym_variable] = "variable",
  [sym_term] = "term",
  [sym_functional_term] = "functional_term",
  [sym_function_composition] = "function_composition",
  [sym_predicate_value] = "predicate_value",
  [sym_sequence_of_predicate_values] = "sequence_of_predicate_values",
  [sym_reserved_event] = "reserved_event",
  [sym_sentence_with_environment] = "sentence_with_environment",
  [sym_keyword_function_def] = "keyword_function_def",
  [sym_function_def] = "function_def",
  [sym_function_signature] = "function_signature",
  [sym_function_parameter] = "function_parameter",
  [sym_function_type_reference] = "function_type_reference",
  [sym_function_body] = "function_body",
  [sym_sequence_builder] = "sequence_builder",
  [sym_value] = "value",
  [sym_simple_value] = "simple_value",
  [sym_string] = "string",
  [sym_iri] = "iri",
  [sym_binary] = "binary",
  [sym_boolean] = "boolean",
  [sym_boolean_truth] = "boolean_truth",
  [sym_boolean_falsity] = "boolean_falsity",
  [sym_value_constructor] = "value_constructor",
  [sym_mapping_value] = "mapping_value",
  [sym_sequence_of_values] = "sequence_of_values",
  [sym_builtin_types] = "builtin_types",
  [sym_definition] = "definition",
  [sym_from_definition_clause] = "from_definition_clause",
  [sym_mixin_with_members] = "mixin_with_members",
  [sym_mixin_without_members] = "mixin_without_members",
  [sym_mixin_member] = "mixin_member",
  [sym_datatype_def] = "datatype_def",
  [sym_datatype_base_type_reference] = "datatype_base_type_reference",
  [sym_datatype_type_restrictions] = "datatype_type_restrictions",
  [sym_restriction_facet] = "restriction_facet",
  [sym_length_restriction_facet] = "length_restriction_facet",
  [sym_digit_restriction_facet] = "digit_restriction_facet",
  [sym_value_restriction_facet] = "value_restriction_facet",
  [sym_restriction_value] = "restriction_value",
  [sym_tz_restriction_facet] = "tz_restriction_facet",
  [sym_tz_restriction_value] = "tz_restriction_value",
  [sym_pattern_restriction_facet] = "pattern_restriction_facet",
  [sym_dimension_def] = "dimension_def",
  [sym_dimension_body] = "dimension_body",
  [sym_dimension_parent] = "dimension_parent",
  [sym_source_entity] = "source_entity",
  [sym_entity_def] = "entity_def",
  [sym_entity_body] = "entity_body",
  [sym_enum_def] = "enum_def",
  [sym_enum_body] = "enum_body",
  [sym_event_def] = "event_def",
  [sym_event_body] = "event_body",
  [sym_metric_def] = "metric_def",
  [sym_metric_group_def] = "metric_group_def",
  [sym_metric_event_binding] = "metric_event_binding",
  [sym_metric_group_body] = "metric_group_body",
  [sym_metric_ref] = "metric_ref",
  [sym_property_def] = "property_def",
  [sym_rdf_def] = "rdf_def",
  [sym_structure_def] = "structure_def",
  [sym_structure_body] = "structure_body",
  [sym_type_class_def] = "type_class_def",
  [sym_type_class_body] = "type_class_body",
  [sym_type_parameter] = "type_parameter",
  [sym_type_parameter_restriction] = "type_parameter_restriction",
  [sym_type_restriction_argument] = "type_restriction_argument",
  [sym_union_def] = "union_def",
  [sym_union_body] = "union_body",
  [sym_entity_identity] = "entity_identity",
  [sym_member] = "member",
  [sym_member_def] = "member_def",
  [sym_annotation_member_def] = "annotation_member_def",
  [sym_property_ref] = "property_ref",
  [sym_type_reference] = "type_reference",
  [sym_mapping_type] = "mapping_type",
  [sym_cardinality_expression] = "cardinality_expression",
  [sym_cardinality_reference_expression] = "cardinality_reference_expression",
  [sym_sequence_ordering] = "sequence_ordering",
  [sym_sequence_uniqueness] = "sequence_uniqueness",
  [sym_cardinality_range] = "cardinality_range",
  [sym_value_variant] = "value_variant",
  [sym_type_variant] = "type_variant",
  [sym_op_equality] = "op_equality",
  [sym_inequality_relation] = "inequality_relation",
  [sym_op_inequality] = "op_inequality",
  [sym_op_less_than] = "op_less_than",
  [sym_op_less_than_or_equal] = "op_less_than_or_equal",
  [sym_op_greater_than_or_equal] = "op_greater_than_or_equal",
  [sym_logical_op_negation] = "logical_op_negation",
  [sym_logical_connective] = "logical_connective",
  [sym_logical_conjunction] = "logical_conjunction",
  [sym_logical_disjunction] = "logical_disjunction",
  [sym_logical_exclusive_disjunction] = "logical_exclusive_disjunction",
  [sym_logical_implication] = "logical_implication",
  [sym_logical_biconditional] = "logical_biconditional",
  [sym_logical_quantifier] = "logical_quantifier",
  [sym_logical_quantifier_universal] = "logical_quantifier_universal",
  [sym_logical_quantifier_existential] = "logical_quantifier_existential",
  [sym_logical_quantifier_existential_one] = "logical_quantifier_existential_one",
  [sym_set_operator] = "set_operator",
  [sym_set_op_union] = "set_op_union",
  [sym_set_op_intersection] = "set_op_intersection",
  [sym_set_op_subset] = "set_op_subset",
  [sym_set_op_subset_or_equal] = "set_op_subset_or_equal",
  [sym_set_op_supset] = "set_op_supset",
  [sym_set_op_supset_or_equal] = "set_op_supset_or_equal",
  [sym_set_op_complement] = "set_op_complement",
  [sym_set_op_product] = "set_op_product",
  [sym_set_membership] = "set_membership",
  [sym_set_op_is_member] = "set_op_is_member",
  [sym_set_op_is_not_member] = "set_op_is_not_member",
  [sym_math_operator] = "math_operator",
  [sym_math_op_multiply] = "math_op_multiply",
  [sym_math_op_divide] = "math_op_divide",
  [sym_math_op_modulo] = "math_op_modulo",
  [sym_math_op_add] = "math_op_add",
  [sym_type_op_combiner] = "type_op_combiner",
  [sym_line_comment] = "line_comment",
  [aux_sym_module_body_repeat1] = "module_body_repeat1",
  [aux_sym_module_body_repeat2] = "module_body_repeat2",
  [aux_sym_module_body_repeat3] = "module_body_repeat3",
  [aux_sym_import_statement_repeat1] = "import_statement_repeat1",
  [aux_sym__module_path_actual_repeat1] = "_module_path_actual_repeat1",
  [aux_sym__actual_arguments_repeat1] = "_actual_arguments_repeat1",
  [aux_sym_function_composition_repeat1] = "function_composition_repeat1",
  [aux_sym_sequence_of_predicate_values_repeat1] = "sequence_of_predicate_values_repeat1",
  [aux_sym_sentence_with_environment_repeat1] = "sentence_with_environment_repeat1",
  [aux_sym_function_signature_repeat1] = "function_signature_repeat1",
  [aux_sym_sequence_builder_repeat1] = "sequence_builder_repeat1",
  [aux_sym_binary_repeat1] = "binary_repeat1",
  [aux_sym_sequence_of_values_repeat1] = "sequence_of_values_repeat1",
  [aux_sym__mixin_member_or_sequence_repeat1] = "_mixin_member_or_sequence_repeat1",
  [aux_sym_datatype_type_restrictions_repeat1] = "datatype_type_restrictions_repeat1",
  [aux_sym_pattern_restriction_facet_repeat1] = "pattern_restriction_facet_repeat1",
  [aux_sym_dimension_body_repeat1] = "dimension_body_repeat1",
  [aux_sym_dimension_body_repeat2] = "dimension_body_repeat2",
  [aux_sym_enum_body_repeat1] = "enum_body_repeat1",
  [aux_sym_metric_group_body_repeat1] = "metric_group_body_repeat1",
  [aux_sym__rdf_types_repeat1] = "_rdf_types_repeat1",
  [aux_sym_type_class_body_repeat1] = "type_class_body_repeat1",
  [aux_sym__class_type_parameters_repeat1] = "_class_type_parameters_repeat1",
  [aux_sym_type_parameter_repeat1] = "type_parameter_repeat1",
  [aux_sym_type_parameter_restriction_repeat1] = "type_parameter_restriction_repeat1",
  [aux_sym_union_body_repeat1] = "union_body_repeat1",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [anon_sym_module] = anon_sym_module,
  [anon_sym_version] = anon_sym_version,
  [anon_sym_is] = anon_sym_is,
  [anon_sym_end] = anon_sym_end,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_from] = anon_sym_from,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_as] = anon_sym_as,
  [sym_identifier] = sym_identifier,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_assert] = anon_sym_assert,
  [sym_controlled_language_tag] = sym_controlled_language_tag,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [sym_reserved_self] = sym_reserved_self,
  [anon_sym_event] = anon_sym_event,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_def] = anon_sym_def,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [sym_quoted_string] = sym_quoted_string,
  [sym_language_tag] = sym_language_tag,
  [anon_sym_LT] = anon_sym_LT,
  [aux_sym_iri_token1] = aux_sym_iri_token1,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_POUND_LBRACK] = anon_sym_POUND_LBRACK,
  [sym_hex_byte] = sym_hex_byte,
  [sym_double] = sym_double,
  [sym_decimal] = sym_decimal,
  [sym_integer] = sym_integer,
  [sym_rational] = sym_rational,
  [sym_unsigned] = sym_unsigned,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_u22a4] = anon_sym_u22a4,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_u22a5] = anon_sym_u22a5,
  [sym_value_empty_sequence] = sym_value_empty_sequence,
  [anon_sym_Thing] = anon_sym_Thing,
  [anon_sym_Nothing] = anon_sym_Nothing,
  [anon_sym_real] = anon_sym_real,
  [anon_sym_rational] = anon_sym_rational,
  [anon_sym_anyURI] = anon_sym_anyURI,
  [anon_sym_base64Binary] = anon_sym_base64Binary,
  [anon_sym_boolean] = anon_sym_boolean,
  [anon_sym_date] = anon_sym_date,
  [anon_sym_dateTime] = anon_sym_dateTime,
  [anon_sym_decimal] = anon_sym_decimal,
  [anon_sym_double] = anon_sym_double,
  [anon_sym_duration] = anon_sym_duration,
  [anon_sym_float] = anon_sym_float,
  [anon_sym_gDay] = anon_sym_gDay,
  [anon_sym_gMonth] = anon_sym_gMonth,
  [anon_sym_gMonthDay] = anon_sym_gMonthDay,
  [anon_sym_gYearMonth] = anon_sym_gYearMonth,
  [anon_sym_gYear] = anon_sym_gYear,
  [anon_sym_hexBinary] = anon_sym_hexBinary,
  [anon_sym_string] = anon_sym_string,
  [anon_sym_time] = anon_sym_time,
  [anon_sym_dateTimeStamp] = anon_sym_dateTimeStamp,
  [anon_sym_dayTimeDuration] = anon_sym_dayTimeDuration,
  [anon_sym_yearMonthDuration] = anon_sym_yearMonthDuration,
  [anon_sym_integer] = anon_sym_integer,
  [anon_sym_long] = anon_sym_long,
  [anon_sym_int] = anon_sym_int,
  [anon_sym_short] = anon_sym_short,
  [anon_sym_byte] = anon_sym_byte,
  [anon_sym_nonNegativeInteger] = anon_sym_nonNegativeInteger,
  [anon_sym_positiveInteger] = anon_sym_positiveInteger,
  [anon_sym_unsignedLong] = anon_sym_unsignedLong,
  [anon_sym_unsignedInt] = anon_sym_unsignedInt,
  [anon_sym_unsigned] = anon_sym_unsigned,
  [anon_sym_unsignedShort] = anon_sym_unsignedShort,
  [anon_sym_unsignedByte] = anon_sym_unsignedByte,
  [anon_sym_nonPositiveInteger] = anon_sym_nonPositiveInteger,
  [anon_sym_negativeInteger] = anon_sym_negativeInteger,
  [anon_sym_normalizedString] = anon_sym_normalizedString,
  [anon_sym_token] = anon_sym_token,
  [anon_sym_language] = anon_sym_language,
  [anon_sym_binary] = anon_sym_binary,
  [anon_sym_iri] = anon_sym_iri,
  [anon_sym_without] = anon_sym_without,
  [anon_sym_datatype] = anon_sym_datatype,
  [sym_opaque] = sym_opaque,
  [anon_sym_length] = anon_sym_length,
  [anon_sym_maxLength] = anon_sym_maxLength,
  [anon_sym_minLength] = anon_sym_minLength,
  [anon_sym_fractionDigits] = anon_sym_fractionDigits,
  [anon_sym_totalDigits] = anon_sym_totalDigits,
  [anon_sym_maxExclusive] = anon_sym_maxExclusive,
  [anon_sym_minExclusive] = anon_sym_minExclusive,
  [anon_sym_maxInclusive] = anon_sym_maxInclusive,
  [anon_sym_minInclusive] = anon_sym_minInclusive,
  [anon_sym_explicitTimezone] = anon_sym_explicitTimezone,
  [anon_sym_required] = anon_sym_required,
  [anon_sym_prohibited] = anon_sym_prohibited,
  [anon_sym_optional] = anon_sym_optional,
  [anon_sym_pattern] = anon_sym_pattern,
  [sym_kw_is_fixed] = sym_kw_is_fixed,
  [anon_sym_dimension] = anon_sym_dimension,
  [anon_sym_parent] = anon_sym_parent,
  [anon_sym_source] = anon_sym_source,
  [anon_sym_entity] = anon_sym_entity,
  [anon_sym_enum] = anon_sym_enum,
  [anon_sym_of] = anon_sym_of,
  [anon_sym_metric] = anon_sym_metric,
  [anon_sym_group] = anon_sym_group,
  [anon_sym_on] = anon_sym_on,
  [anon_sym_ref] = anon_sym_ref,
  [anon_sym_property] = anon_sym_property,
  [anon_sym_rdf] = anon_sym_rdf,
  [anon_sym_a] = anon_sym_a,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_structure] = anon_sym_structure,
  [anon_sym_class] = anon_sym_class,
  [sym_wildcard] = sym_wildcard,
  [anon_sym_union] = anon_sym_union,
  [anon_sym_identity] = anon_sym_identity,
  [sym_unknown_type] = sym_unknown_type,
  [anon_sym_ordered] = anon_sym_ordered,
  [anon_sym_unordered] = anon_sym_unordered,
  [anon_sym_unique] = anon_sym_unique,
  [anon_sym_nonunique] = anon_sym_nonunique,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_u2260] = anon_sym_u2260,
  [sym_op_greater_than] = sym_op_greater_than,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_u2264] = anon_sym_u2264,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_u2265] = anon_sym_u2265,
  [anon_sym_not] = anon_sym_not,
  [anon_sym_u00ac] = anon_sym_u00ac,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_u2227] = anon_sym_u2227,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_u2228] = anon_sym_u2228,
  [anon_sym_xor] = anon_sym_xor,
  [anon_sym_u22bb] = anon_sym_u22bb,
  [anon_sym_implies] = anon_sym_implies,
  [anon_sym_EQ_EQ_GT] = anon_sym_EQ_EQ_GT,
  [anon_sym_u21d2] = anon_sym_u21d2,
  [anon_sym_iff] = anon_sym_iff,
  [anon_sym_LT_EQ_EQ_GT] = anon_sym_LT_EQ_EQ_GT,
  [anon_sym_u21d4] = anon_sym_u21d4,
  [anon_sym_forall] = anon_sym_forall,
  [anon_sym_u2200] = anon_sym_u2200,
  [anon_sym_exists] = anon_sym_exists,
  [anon_sym_u2203] = anon_sym_u2203,
  [anon_sym_exists_BANG] = anon_sym_exists_BANG,
  [anon_sym_u2203_BANG] = anon_sym_u2203_BANG,
  [anon_sym_u222a] = anon_sym_u222a,
  [anon_sym_intersects] = anon_sym_intersects,
  [anon_sym_u2229] = anon_sym_u2229,
  [anon_sym_subset] = anon_sym_subset,
  [anon_sym_u2282] = anon_sym_u2282,
  [anon_sym_subseteq] = anon_sym_subseteq,
  [anon_sym_u2286] = anon_sym_u2286,
  [anon_sym_supset] = anon_sym_supset,
  [anon_sym_u2283] = anon_sym_u2283,
  [anon_sym_supseteq] = anon_sym_supseteq,
  [anon_sym_u2287] = anon_sym_u2287,
  [anon_sym_complement] = anon_sym_complement,
  [anon_sym_u2216] = anon_sym_u2216,
  [anon_sym_product] = anon_sym_product,
  [anon_sym_u2a09] = anon_sym_u2a09,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_u2208] = anon_sym_u2208,
  [anon_sym_u2209] = anon_sym_u2209,
  [sym_seq_builder_separator] = sym_seq_builder_separator,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_u00d7] = anon_sym_u00d7,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_u00f7] = anon_sym_u00f7,
  [anon_sym_mod] = anon_sym_mod,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [sym_math_op_subtract] = sym_math_op_subtract,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_u2218] = anon_sym_u2218,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_u2254] = anon_sym_u2254,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_u2192] = anon_sym_u2192,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_u2190] = anon_sym_u2190,
  [sym_comment_aligned] = sym_comment_aligned,
  [sym_comment_local] = sym_comment_local,
  [sym_comment_page] = sym_comment_page,
  [sym_module] = sym_module,
  [sym_module_version] = sym_module_version,
  [sym_module_body] = sym_module_body,
  [sym_import_statement] = sym_import_statement,
  [sym_from_clause] = sym_from_clause,
  [sym_module_path_absolute] = sym_module_path_absolute,
  [sym_module_path_relative] = sym_module_path_relative,
  [sym__module_path_actual] = sym__module_path_actual,
  [sym_member_import] = sym_member_import,
  [sym_module_import] = sym_module_import,
  [sym_qualified_identifier] = sym_qualified_identifier,
  [sym_identifier_reference] = sym_identifier_reference,
  [sym_annotation] = sym_annotation,
  [sym_annotation_property] = sym_annotation_property,
  [sym_constraint] = sym_constraint,
  [sym_annotation_only_body] = sym_annotation_only_body,
  [sym_informal_constraint] = sym_informal_constraint,
  [sym_formal_constraint] = sym_formal_constraint,
  [sym_constraint_sentence] = sym_constraint_sentence,
  [sym_simple_sentence] = sym_simple_sentence,
  [sym_atomic_sentence] = sym_atomic_sentence,
  [sym_equation] = sym_equation,
  [sym_inequation] = sym_inequation,
  [sym_boolean_sentence] = sym_boolean_sentence,
  [sym_unary_boolean_sentence] = sym_unary_boolean_sentence,
  [sym_binary_boolean_sentence] = sym_binary_boolean_sentence,
  [sym_expression_sentence] = sym_expression_sentence,
  [sym_set_expression_sentence] = sym_set_expression_sentence,
  [sym_arithmetic_expression_sentence] = sym_arithmetic_expression_sentence,
  [sym_quantified_sentence] = sym_quantified_sentence,
  [sym_quantified_variable_binding] = sym_quantified_variable_binding,
  [sym_quantified_variable] = sym_quantified_variable,
  [sym_variable] = sym_variable,
  [sym_term] = sym_term,
  [sym_functional_term] = sym_functional_term,
  [sym_function_composition] = sym_function_composition,
  [sym_predicate_value] = sym_predicate_value,
  [sym_sequence_of_predicate_values] = sym_sequence_of_predicate_values,
  [sym_reserved_event] = sym_reserved_event,
  [sym_sentence_with_environment] = sym_sentence_with_environment,
  [sym_keyword_function_def] = sym_keyword_function_def,
  [sym_function_def] = sym_function_def,
  [sym_function_signature] = sym_function_signature,
  [sym_function_parameter] = sym_function_parameter,
  [sym_function_type_reference] = sym_function_type_reference,
  [sym_function_body] = sym_function_body,
  [sym_sequence_builder] = sym_sequence_builder,
  [sym_value] = sym_value,
  [sym_simple_value] = sym_simple_value,
  [sym_string] = sym_string,
  [sym_iri] = sym_iri,
  [sym_binary] = sym_binary,
  [sym_boolean] = sym_boolean,
  [sym_boolean_truth] = sym_boolean_truth,
  [sym_boolean_falsity] = sym_boolean_falsity,
  [sym_value_constructor] = sym_value_constructor,
  [sym_mapping_value] = sym_mapping_value,
  [sym_sequence_of_values] = sym_sequence_of_values,
  [sym_builtin_types] = sym_builtin_types,
  [sym_definition] = sym_definition,
  [sym_from_definition_clause] = sym_from_definition_clause,
  [sym_mixin_with_members] = sym_mixin_with_members,
  [sym_mixin_without_members] = sym_mixin_without_members,
  [sym_mixin_member] = sym_mixin_member,
  [sym_datatype_def] = sym_datatype_def,
  [sym_datatype_base_type_reference] = sym_datatype_base_type_reference,
  [sym_datatype_type_restrictions] = sym_datatype_type_restrictions,
  [sym_restriction_facet] = sym_restriction_facet,
  [sym_length_restriction_facet] = sym_length_restriction_facet,
  [sym_digit_restriction_facet] = sym_digit_restriction_facet,
  [sym_value_restriction_facet] = sym_value_restriction_facet,
  [sym_restriction_value] = sym_restriction_value,
  [sym_tz_restriction_facet] = sym_tz_restriction_facet,
  [sym_tz_restriction_value] = sym_tz_restriction_value,
  [sym_pattern_restriction_facet] = sym_pattern_restriction_facet,
  [sym_dimension_def] = sym_dimension_def,
  [sym_dimension_body] = sym_dimension_body,
  [sym_dimension_parent] = sym_dimension_parent,
  [sym_source_entity] = sym_source_entity,
  [sym_entity_def] = sym_entity_def,
  [sym_entity_body] = sym_entity_body,
  [sym_enum_def] = sym_enum_def,
  [sym_enum_body] = sym_enum_body,
  [sym_event_def] = sym_event_def,
  [sym_event_body] = sym_event_body,
  [sym_metric_def] = sym_metric_def,
  [sym_metric_group_def] = sym_metric_group_def,
  [sym_metric_event_binding] = sym_metric_event_binding,
  [sym_metric_group_body] = sym_metric_group_body,
  [sym_metric_ref] = sym_metric_ref,
  [sym_property_def] = sym_property_def,
  [sym_rdf_def] = sym_rdf_def,
  [sym_structure_def] = sym_structure_def,
  [sym_structure_body] = sym_structure_body,
  [sym_type_class_def] = sym_type_class_def,
  [sym_type_class_body] = sym_type_class_body,
  [sym_type_parameter] = sym_type_parameter,
  [sym_type_parameter_restriction] = sym_type_parameter_restriction,
  [sym_type_restriction_argument] = sym_type_restriction_argument,
  [sym_union_def] = sym_union_def,
  [sym_union_body] = sym_union_body,
  [sym_entity_identity] = sym_entity_identity,
  [sym_member] = sym_member,
  [sym_member_def] = sym_member_def,
  [sym_annotation_member_def] = sym_annotation_member_def,
  [sym_property_ref] = sym_property_ref,
  [sym_type_reference] = sym_type_reference,
  [sym_mapping_type] = sym_mapping_type,
  [sym_cardinality_expression] = sym_cardinality_expression,
  [sym_cardinality_reference_expression] = sym_cardinality_reference_expression,
  [sym_sequence_ordering] = sym_sequence_ordering,
  [sym_sequence_uniqueness] = sym_sequence_uniqueness,
  [sym_cardinality_range] = sym_cardinality_range,
  [sym_value_variant] = sym_value_variant,
  [sym_type_variant] = sym_type_variant,
  [sym_op_equality] = sym_op_equality,
  [sym_inequality_relation] = sym_inequality_relation,
  [sym_op_inequality] = sym_op_inequality,
  [sym_op_less_than] = sym_op_less_than,
  [sym_op_less_than_or_equal] = sym_op_less_than_or_equal,
  [sym_op_greater_than_or_equal] = sym_op_greater_than_or_equal,
  [sym_logical_op_negation] = sym_logical_op_negation,
  [sym_logical_connective] = sym_logical_connective,
  [sym_logical_conjunction] = sym_logical_conjunction,
  [sym_logical_disjunction] = sym_logical_disjunction,
  [sym_logical_exclusive_disjunction] = sym_logical_exclusive_disjunction,
  [sym_logical_implication] = sym_logical_implication,
  [sym_logical_biconditional] = sym_logical_biconditional,
  [sym_logical_quantifier] = sym_logical_quantifier,
  [sym_logical_quantifier_universal] = sym_logical_quantifier_universal,
  [sym_logical_quantifier_existential] = sym_logical_quantifier_existential,
  [sym_logical_quantifier_existential_one] = sym_logical_quantifier_existential_one,
  [sym_set_operator] = sym_set_operator,
  [sym_set_op_union] = sym_set_op_union,
  [sym_set_op_intersection] = sym_set_op_intersection,
  [sym_set_op_subset] = sym_set_op_subset,
  [sym_set_op_subset_or_equal] = sym_set_op_subset_or_equal,
  [sym_set_op_supset] = sym_set_op_supset,
  [sym_set_op_supset_or_equal] = sym_set_op_supset_or_equal,
  [sym_set_op_complement] = sym_set_op_complement,
  [sym_set_op_product] = sym_set_op_product,
  [sym_set_membership] = sym_set_membership,
  [sym_set_op_is_member] = sym_set_op_is_member,
  [sym_set_op_is_not_member] = sym_set_op_is_not_member,
  [sym_math_operator] = sym_math_operator,
  [sym_math_op_multiply] = sym_math_op_multiply,
  [sym_math_op_divide] = sym_math_op_divide,
  [sym_math_op_modulo] = sym_math_op_modulo,
  [sym_math_op_add] = sym_math_op_add,
  [sym_type_op_combiner] = sym_type_op_combiner,
  [sym_line_comment] = sym_line_comment,
  [aux_sym_module_body_repeat1] = aux_sym_module_body_repeat1,
  [aux_sym_module_body_repeat2] = aux_sym_module_body_repeat2,
  [aux_sym_module_body_repeat3] = aux_sym_module_body_repeat3,
  [aux_sym_import_statement_repeat1] = aux_sym_import_statement_repeat1,
  [aux_sym__module_path_actual_repeat1] = aux_sym__module_path_actual_repeat1,
  [aux_sym__actual_arguments_repeat1] = aux_sym__actual_arguments_repeat1,
  [aux_sym_function_composition_repeat1] = aux_sym_function_composition_repeat1,
  [aux_sym_sequence_of_predicate_values_repeat1] = aux_sym_sequence_of_predicate_values_repeat1,
  [aux_sym_sentence_with_environment_repeat1] = aux_sym_sentence_with_environment_repeat1,
  [aux_sym_function_signature_repeat1] = aux_sym_function_signature_repeat1,
  [aux_sym_sequence_builder_repeat1] = aux_sym_sequence_builder_repeat1,
  [aux_sym_binary_repeat1] = aux_sym_binary_repeat1,
  [aux_sym_sequence_of_values_repeat1] = aux_sym_sequence_of_values_repeat1,
  [aux_sym__mixin_member_or_sequence_repeat1] = aux_sym__mixin_member_or_sequence_repeat1,
  [aux_sym_datatype_type_restrictions_repeat1] = aux_sym_datatype_type_restrictions_repeat1,
  [aux_sym_pattern_restriction_facet_repeat1] = aux_sym_pattern_restriction_facet_repeat1,
  [aux_sym_dimension_body_repeat1] = aux_sym_dimension_body_repeat1,
  [aux_sym_dimension_body_repeat2] = aux_sym_dimension_body_repeat2,
  [aux_sym_enum_body_repeat1] = aux_sym_enum_body_repeat1,
  [aux_sym_metric_group_body_repeat1] = aux_sym_metric_group_body_repeat1,
  [aux_sym__rdf_types_repeat1] = aux_sym__rdf_types_repeat1,
  [aux_sym_type_class_body_repeat1] = aux_sym_type_class_body_repeat1,
  [aux_sym__class_type_parameters_repeat1] = aux_sym__class_type_parameters_repeat1,
  [aux_sym_type_parameter_repeat1] = aux_sym_type_parameter_repeat1,
  [aux_sym_type_parameter_restriction_repeat1] = aux_sym_type_parameter_restriction_repeat1,
  [aux_sym_union_body_repeat1] = aux_sym_union_body_repeat1,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_module] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_version] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_is] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_end] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_from] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [sym_controlled_language_tag] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [sym_reserved_self] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_event] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_def] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [sym_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym_language_tag] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_iri_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [sym_hex_byte] = {
    .visible = true,
    .named = true,
  },
  [sym_double] = {
    .visible = true,
    .named = true,
  },
  [sym_decimal] = {
    .visible = true,
    .named = true,
  },
  [sym_integer] = {
    .visible = true,
    .named = true,
  },
  [sym_rational] = {
    .visible = true,
    .named = true,
  },
  [sym_unsigned] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u22a4] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u22a5] = {
    .visible = true,
    .named = false,
  },
  [sym_value_empty_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_Thing] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_Nothing] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_real] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rational] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_anyURI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_base64Binary] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_boolean] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_date] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_dateTime] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_decimal] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_double] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_duration] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_float] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gDay] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gMonth] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gMonthDay] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gYearMonth] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gYear] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_hexBinary] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_string] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_time] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_dateTimeStamp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_dayTimeDuration] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_yearMonthDuration] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_integer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_long] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_int] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_short] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_byte] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonNegativeInteger] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_positiveInteger] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsignedLong] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsignedInt] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsigned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsignedShort] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsignedByte] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonPositiveInteger] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_negativeInteger] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_normalizedString] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_token] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_language] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_binary] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_iri] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_without] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_datatype] = {
    .visible = true,
    .named = false,
  },
  [sym_opaque] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_length] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_maxLength] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_minLength] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fractionDigits] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_totalDigits] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_maxExclusive] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_minExclusive] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_maxInclusive] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_minInclusive] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_explicitTimezone] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_required] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_prohibited] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_optional] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_pattern] = {
    .visible = true,
    .named = false,
  },
  [sym_kw_is_fixed] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_dimension] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_parent] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_source] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_entity] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_metric] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_group] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_on] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ref] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_property] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rdf] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_a] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_structure] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [sym_wildcard] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_union] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_identity] = {
    .visible = true,
    .named = false,
  },
  [sym_unknown_type] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_ordered] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unordered] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unique] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonunique] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2260] = {
    .visible = true,
    .named = false,
  },
  [sym_op_greater_than] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2264] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2265] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_not] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u00ac] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2227] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2228] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_xor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u22bb] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_implies] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u21d2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_iff] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u21d4] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_forall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2200] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exists] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2203] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exists_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2203_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u222a] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_intersects] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2229] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_subset] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2282] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_subseteq] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2286] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_supset] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2283] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_supseteq] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2287] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_complement] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2216] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_product] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2a09] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2208] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2209] = {
    .visible = true,
    .named = false,
  },
  [sym_seq_builder_separator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u00d7] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u00f7] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mod] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [sym_math_op_subtract] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2218] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2254] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2192] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u2190] = {
    .visible = true,
    .named = false,
  },
  [sym_comment_aligned] = {
    .visible = true,
    .named = true,
  },
  [sym_comment_local] = {
    .visible = true,
    .named = true,
  },
  [sym_comment_page] = {
    .visible = true,
    .named = true,
  },
  [sym_module] = {
    .visible = true,
    .named = true,
  },
  [sym_module_version] = {
    .visible = true,
    .named = true,
  },
  [sym_module_body] = {
    .visible = true,
    .named = true,
  },
  [sym_import_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_from_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_module_path_absolute] = {
    .visible = true,
    .named = true,
  },
  [sym_module_path_relative] = {
    .visible = true,
    .named = true,
  },
  [sym__module_path_actual] = {
    .visible = false,
    .named = true,
  },
  [sym_member_import] = {
    .visible = true,
    .named = true,
  },
  [sym_module_import] = {
    .visible = true,
    .named = true,
  },
  [sym_qualified_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_identifier_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_annotation] = {
    .visible = true,
    .named = true,
  },
  [sym_annotation_property] = {
    .visible = true,
    .named = true,
  },
  [sym_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_annotation_only_body] = {
    .visible = true,
    .named = true,
  },
  [sym_informal_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_formal_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_constraint_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_atomic_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_equation] = {
    .visible = true,
    .named = true,
  },
  [sym_inequation] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_boolean_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_boolean_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_set_expression_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_arithmetic_expression_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_quantified_sentence] = {
    .visible = true,
    .named = true,
  },
  [sym_quantified_variable_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_quantified_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_term] = {
    .visible = true,
    .named = true,
  },
  [sym_functional_term] = {
    .visible = true,
    .named = true,
  },
  [sym_function_composition] = {
    .visible = true,
    .named = true,
  },
  [sym_predicate_value] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_of_predicate_values] = {
    .visible = true,
    .named = true,
  },
  [sym_reserved_event] = {
    .visible = true,
    .named = true,
  },
  [sym_sentence_with_environment] = {
    .visible = true,
    .named = true,
  },
  [sym_keyword_function_def] = {
    .visible = true,
    .named = true,
  },
  [sym_function_def] = {
    .visible = true,
    .named = true,
  },
  [sym_function_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_function_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_function_body] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_builder] = {
    .visible = true,
    .named = true,
  },
  [sym_value] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_value] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_iri] = {
    .visible = true,
    .named = true,
  },
  [sym_binary] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_truth] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_falsity] = {
    .visible = true,
    .named = true,
  },
  [sym_value_constructor] = {
    .visible = true,
    .named = true,
  },
  [sym_mapping_value] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_of_values] = {
    .visible = true,
    .named = true,
  },
  [sym_builtin_types] = {
    .visible = true,
    .named = true,
  },
  [sym_definition] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_from_definition_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_mixin_with_members] = {
    .visible = true,
    .named = true,
  },
  [sym_mixin_without_members] = {
    .visible = true,
    .named = true,
  },
  [sym_mixin_member] = {
    .visible = true,
    .named = true,
  },
  [sym_datatype_def] = {
    .visible = true,
    .named = true,
  },
  [sym_datatype_base_type_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_datatype_type_restrictions] = {
    .visible = true,
    .named = true,
  },
  [sym_restriction_facet] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_length_restriction_facet] = {
    .visible = true,
    .named = true,
  },
  [sym_digit_restriction_facet] = {
    .visible = true,
    .named = true,
  },
  [sym_value_restriction_facet] = {
    .visible = true,
    .named = true,
  },
  [sym_restriction_value] = {
    .visible = true,
    .named = true,
  },
  [sym_tz_restriction_facet] = {
    .visible = true,
    .named = true,
  },
  [sym_tz_restriction_value] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern_restriction_facet] = {
    .visible = true,
    .named = true,
  },
  [sym_dimension_def] = {
    .visible = true,
    .named = true,
  },
  [sym_dimension_body] = {
    .visible = true,
    .named = true,
  },
  [sym_dimension_parent] = {
    .visible = true,
    .named = true,
  },
  [sym_source_entity] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_def] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_body] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_def] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_body] = {
    .visible = true,
    .named = true,
  },
  [sym_event_def] = {
    .visible = true,
    .named = true,
  },
  [sym_event_body] = {
    .visible = true,
    .named = true,
  },
  [sym_metric_def] = {
    .visible = true,
    .named = true,
  },
  [sym_metric_group_def] = {
    .visible = true,
    .named = true,
  },
  [sym_metric_event_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_metric_group_body] = {
    .visible = true,
    .named = true,
  },
  [sym_metric_ref] = {
    .visible = true,
    .named = true,
  },
  [sym_property_def] = {
    .visible = true,
    .named = true,
  },
  [sym_rdf_def] = {
    .visible = true,
    .named = true,
  },
  [sym_structure_def] = {
    .visible = true,
    .named = true,
  },
  [sym_structure_body] = {
    .visible = true,
    .named = true,
  },
  [sym_type_class_def] = {
    .visible = true,
    .named = true,
  },
  [sym_type_class_body] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter_restriction] = {
    .visible = true,
    .named = true,
  },
  [sym_type_restriction_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_union_def] = {
    .visible = true,
    .named = true,
  },
  [sym_union_body] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_identity] = {
    .visible = true,
    .named = true,
  },
  [sym_member] = {
    .visible = true,
    .named = true,
  },
  [sym_member_def] = {
    .visible = true,
    .named = true,
  },
  [sym_annotation_member_def] = {
    .visible = true,
    .named = true,
  },
  [sym_property_ref] = {
    .visible = true,
    .named = true,
  },
  [sym_type_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_mapping_type] = {
    .visible = true,
    .named = true,
  },
  [sym_cardinality_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cardinality_reference_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_ordering] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_uniqueness] = {
    .visible = true,
    .named = true,
  },
  [sym_cardinality_range] = {
    .visible = true,
    .named = true,
  },
  [sym_value_variant] = {
    .visible = true,
    .named = true,
  },
  [sym_type_variant] = {
    .visible = true,
    .named = true,
  },
  [sym_op_equality] = {
    .visible = true,
    .named = true,
  },
  [sym_inequality_relation] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_op_inequality] = {
    .visible = true,
    .named = true,
  },
  [sym_op_less_than] = {
    .visible = true,
    .named = true,
  },
  [sym_op_less_than_or_equal] = {
    .visible = true,
    .named = true,
  },
  [sym_op_greater_than_or_equal] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_op_negation] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_connective] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_logical_conjunction] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_disjunction] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_exclusive_disjunction] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_implication] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_biconditional] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_quantifier] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_logical_quantifier_universal] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_quantifier_existential] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_quantifier_existential_one] = {
    .visible = true,
    .named = true,
  },
  [sym_set_operator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_set_op_union] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_intersection] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_subset] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_subset_or_equal] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_supset] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_supset_or_equal] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_complement] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_product] = {
    .visible = true,
    .named = true,
  },
  [sym_set_membership] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_set_op_is_member] = {
    .visible = true,
    .named = true,
  },
  [sym_set_op_is_not_member] = {
    .visible = true,
    .named = true,
  },
  [sym_math_operator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_math_op_multiply] = {
    .visible = true,
    .named = true,
  },
  [sym_math_op_divide] = {
    .visible = true,
    .named = true,
  },
  [sym_math_op_modulo] = {
    .visible = true,
    .named = true,
  },
  [sym_math_op_add] = {
    .visible = true,
    .named = true,
  },
  [sym_type_op_combiner] = {
    .visible = true,
    .named = true,
  },
  [sym_line_comment] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_module_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_body_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_body_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_import_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__module_path_actual_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__actual_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_composition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequence_of_predicate_values_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sentence_with_environment_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_signature_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequence_builder_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_binary_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequence_of_values_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__mixin_member_or_sequence_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_datatype_type_restrictions_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pattern_restriction_facet_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dimension_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dimension_body_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_metric_group_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__rdf_types_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_class_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__class_type_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_parameter_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_parameter_restriction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_union_body_repeat1] = {
    .visible = false,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_argument = 1,
  field_base = 2,
  field_binding = 3,
  field_body = 4,
  field_cardinality = 5,
  field_class = 6,
  field_domain = 7,
  field_element = 8,
  field_empty = 9,
  field_facet = 10,
  field_from = 11,
  field_function = 12,
  field_identity = 13,
  field_is_fixed = 14,
  field_language = 15,
  field_lhs = 16,
  field_max = 17,
  field_member = 18,
  field_min = 19,
  field_module = 20,
  field_name = 21,
  field_opaque = 22,
  field_operator = 23,
  field_ordering = 24,
  field_parameter = 25,
  field_parent = 26,
  field_predicate = 27,
  field_property = 28,
  field_quantifier = 29,
  field_range = 30,
  field_referent = 31,
  field_relation = 32,
  field_rename = 33,
  field_restriction = 34,
  field_rhs = 35,
  field_segment = 36,
  field_signature = 37,
  field_source = 38,
  field_subject = 39,
  field_target = 40,
  field_type = 41,
  field_uniqueness = 42,
  field_value = 43,
  field_variable = 44,
  field_version_info = 45,
  field_version_uri = 46,
  field_wildcard = 47,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_argument] = "argument",
  [field_base] = "base",
  [field_binding] = "binding",
  [field_body] = "body",
  [field_cardinality] = "cardinality",
  [field_class] = "class",
  [field_domain] = "domain",
  [field_element] = "element",
  [field_empty] = "empty",
  [field_facet] = "facet",
  [field_from] = "from",
  [field_function] = "function",
  [field_identity] = "identity",
  [field_is_fixed] = "is_fixed",
  [field_language] = "language",
  [field_lhs] = "lhs",
  [field_max] = "max",
  [field_member] = "member",
  [field_min] = "min",
  [field_module] = "module",
  [field_name] = "name",
  [field_opaque] = "opaque",
  [field_operator] = "operator",
  [field_ordering] = "ordering",
  [field_parameter] = "parameter",
  [field_parent] = "parent",
  [field_predicate] = "predicate",
  [field_property] = "property",
  [field_quantifier] = "quantifier",
  [field_range] = "range",
  [field_referent] = "referent",
  [field_relation] = "relation",
  [field_rename] = "rename",
  [field_restriction] = "restriction",
  [field_rhs] = "rhs",
  [field_segment] = "segment",
  [field_signature] = "signature",
  [field_source] = "source",
  [field_subject] = "subject",
  [field_target] = "target",
  [field_type] = "type",
  [field_uniqueness] = "uniqueness",
  [field_value] = "value",
  [field_variable] = "variable",
  [field_version_info] = "version_info",
  [field_version_uri] = "version_uri",
  [field_wildcard] = "wildcard",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 2},
  [2] = {.index = 2, .length = 1},
  [3] = {.index = 3, .length = 2},
  [4] = {.index = 5, .length = 3},
  [5] = {.index = 8, .length = 2},
  [6] = {.index = 10, .length = 1},
  [7] = {.index = 11, .length = 1},
  [8] = {.index = 12, .length = 1},
  [9] = {.index = 13, .length = 1},
  [10] = {.index = 14, .length = 1},
  [11] = {.index = 15, .length = 1},
  [12] = {.index = 16, .length = 3},
  [13] = {.index = 19, .length = 2},
  [14] = {.index = 21, .length = 1},
  [15] = {.index = 22, .length = 2},
  [16] = {.index = 24, .length = 2},
  [17] = {.index = 26, .length = 2},
  [18] = {.index = 28, .length = 2},
  [19] = {.index = 30, .length = 2},
  [20] = {.index = 32, .length = 1},
  [21] = {.index = 33, .length = 2},
  [22] = {.index = 35, .length = 1},
  [23] = {.index = 36, .length = 2},
  [24] = {.index = 38, .length = 1},
  [25] = {.index = 39, .length = 3},
  [26] = {.index = 42, .length = 2},
  [27] = {.index = 44, .length = 1},
  [28] = {.index = 45, .length = 2},
  [29] = {.index = 47, .length = 1},
  [30] = {.index = 48, .length = 1},
  [31] = {.index = 49, .length = 2},
  [32] = {.index = 51, .length = 1},
  [33] = {.index = 52, .length = 3},
  [34] = {.index = 55, .length = 1},
  [35] = {.index = 56, .length = 1},
  [36] = {.index = 57, .length = 2},
  [37] = {.index = 59, .length = 1},
  [38] = {.index = 60, .length = 2},
  [39] = {.index = 62, .length = 2},
  [40] = {.index = 64, .length = 2},
  [41] = {.index = 66, .length = 1},
  [42] = {.index = 67, .length = 1},
  [43] = {.index = 68, .length = 3},
  [44] = {.index = 71, .length = 3},
  [45] = {.index = 74, .length = 3},
  [46] = {.index = 77, .length = 2},
  [47] = {.index = 79, .length = 2},
  [48] = {.index = 81, .length = 3},
  [49] = {.index = 84, .length = 2},
  [50] = {.index = 86, .length = 2},
  [51] = {.index = 88, .length = 3},
  [52] = {.index = 91, .length = 1},
  [53] = {.index = 92, .length = 3},
  [54] = {.index = 95, .length = 1},
  [55] = {.index = 96, .length = 1},
  [56] = {.index = 97, .length = 2},
  [57] = {.index = 99, .length = 2},
  [58] = {.index = 101, .length = 1},
  [59] = {.index = 102, .length = 2},
  [60] = {.index = 104, .length = 2},
  [61] = {.index = 106, .length = 2},
  [62] = {.index = 108, .length = 2},
  [63] = {.index = 110, .length = 3},
  [64] = {.index = 113, .length = 2},
  [65] = {.index = 115, .length = 1},
  [66] = {.index = 116, .length = 1},
  [67] = {.index = 117, .length = 2},
  [68] = {.index = 119, .length = 3},
  [69] = {.index = 122, .length = 1},
  [70] = {.index = 123, .length = 1},
  [71] = {.index = 124, .length = 1},
  [72] = {.index = 125, .length = 4},
  [73] = {.index = 129, .length = 4},
  [74] = {.index = 133, .length = 4},
  [75] = {.index = 137, .length = 1},
  [76] = {.index = 138, .length = 1},
  [77] = {.index = 139, .length = 1},
  [78] = {.index = 140, .length = 3},
  [79] = {.index = 143, .length = 4},
  [80] = {.index = 147, .length = 2},
  [81] = {.index = 149, .length = 2},
  [82] = {.index = 151, .length = 1},
  [83] = {.index = 152, .length = 1},
  [84] = {.index = 153, .length = 3},
  [85] = {.index = 156, .length = 3},
  [86] = {.index = 159, .length = 2},
  [87] = {.index = 161, .length = 2},
  [88] = {.index = 163, .length = 2},
  [89] = {.index = 165, .length = 2},
  [90] = {.index = 167, .length = 2},
  [91] = {.index = 169, .length = 2},
  [92] = {.index = 171, .length = 2},
  [93] = {.index = 173, .length = 2},
  [94] = {.index = 175, .length = 5},
  [95] = {.index = 180, .length = 3},
  [96] = {.index = 183, .length = 1},
  [97] = {.index = 184, .length = 2},
  [98] = {.index = 186, .length = 2},
  [99] = {.index = 188, .length = 2},
  [100] = {.index = 190, .length = 2},
  [101] = {.index = 192, .length = 1},
  [102] = {.index = 193, .length = 3},
  [103] = {.index = 196, .length = 2},
  [104] = {.index = 198, .length = 4},
  [105] = {.index = 202, .length = 3},
  [106] = {.index = 205, .length = 3},
  [107] = {.index = 208, .length = 1},
  [108] = {.index = 209, .length = 2},
  [109] = {.index = 211, .length = 1},
  [110] = {.index = 212, .length = 2},
  [111] = {.index = 214, .length = 2},
  [112] = {.index = 216, .length = 2},
  [113] = {.index = 218, .length = 1},
  [114] = {.index = 219, .length = 2},
  [115] = {.index = 221, .length = 4},
  [116] = {.index = 225, .length = 2},
  [117] = {.index = 227, .length = 3},
  [118] = {.index = 230, .length = 3},
  [119] = {.index = 233, .length = 3},
  [120] = {.index = 236, .length = 2},
  [121] = {.index = 238, .length = 2},
  [122] = {.index = 240, .length = 2},
  [123] = {.index = 242, .length = 3},
  [124] = {.index = 245, .length = 2},
  [125] = {.index = 247, .length = 3},
  [126] = {.index = 250, .length = 4},
  [127] = {.index = 254, .length = 2},
  [128] = {.index = 256, .length = 3},
  [129] = {.index = 259, .length = 3},
  [130] = {.index = 262, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_body, 2},
    {field_name, 1},
  [2] =
    {field_version_uri, 1},
  [3] =
    {field_body, 3},
    {field_name, 1},
  [5] =
    {field_base, 2},
    {field_body, 3},
    {field_name, 1},
  [8] =
    {field_version_info, 1},
    {field_version_uri, 2},
  [10] =
    {field_name, 0},
  [11] =
    {field_segment, 0},
  [12] =
    {field_segment, 0, .inherited = true},
  [13] =
    {field_name, 1},
  [14] =
    {field_signature, 0},
  [15] =
    {field_member, 1},
  [16] =
    {field_base, 2},
    {field_body, 4},
    {field_name, 1},
  [19] =
    {field_name, 0},
    {field_version_uri, 1},
  [21] =
    {field_segment, 1, .inherited = true},
  [22] =
    {field_segment, 0},
    {field_segment, 1, .inherited = true},
  [24] =
    {field_from, 2},
    {field_name, 1},
  [26] =
    {field_body, 1},
    {field_signature, 0},
  [28] =
    {field_name, 0},
    {field_rename, 2},
  [30] =
    {field_member, 2},
    {field_module, 0},
  [32] =
    {field_segment, 1},
  [33] =
    {field_segment, 0, .inherited = true},
    {field_segment, 1, .inherited = true},
  [35] =
    {field_empty, 0},
  [36] =
    {field_name, 1},
    {field_value, 3},
  [38] =
    {field_value, 1},
  [39] =
    {field_body, 3},
    {field_from, 2},
    {field_name, 1},
  [42] =
    {field_base, 3},
    {field_name, 1},
  [44] =
    {field_parameter, 0},
  [45] =
    {field_name, 0},
    {field_target, 2},
  [47] =
    {field_name, 2},
  [48] =
    {field_body, 1},
  [49] =
    {field_name, 1},
    {field_type, 3},
  [51] =
    {field_function, 0},
  [52] =
    {field_name, 0},
    {field_rename, 3},
    {field_version_uri, 1},
  [55] =
    {field_element, 0},
  [56] =
    {field_language, 1},
  [57] =
    {field_name, 1, .inherited = true},
    {field_subject, 0},
  [59] =
    {field_variable, 0},
  [60] =
    {field_operator, 0},
    {field_rhs, 1},
  [62] =
    {field_binding, 1},
    {field_quantifier, 0},
  [64] =
    {field_language, 2},
    {field_value, 1},
  [66] =
    {field_from, 1},
  [67] =
    {field_identity, 1},
  [68] =
    {field_base, 4},
    {field_name, 1},
    {field_opaque, 3},
  [71] =
    {field_base, 3},
    {field_body, 4},
    {field_name, 1},
  [74] =
    {field_base, 3},
    {field_name, 1},
    {field_restriction, 4},
  [77] =
    {field_body, 1},
    {field_name, 0},
  [79] =
    {field_parameter, 0, .inherited = true},
    {field_parameter, 1, .inherited = true},
  [81] =
    {field_cardinality, 2},
    {field_name, 0},
    {field_target, 3},
  [84] =
    {field_from, 4},
    {field_name, 2},
  [86] =
    {field_body, 4},
    {field_name, 2},
  [88] =
    {field_body, 3},
    {field_name, 0},
    {field_target, 2},
  [91] =
    {field_type, 0},
  [92] =
    {field_body, 4},
    {field_name, 1},
    {field_type, 3},
  [95] =
    {field_property, 1},
  [96] =
    {field_function, 1, .inherited = true},
  [97] =
    {field_function, 0, .inherited = true},
    {field_function, 1, .inherited = true},
  [99] =
    {field_name, 1},
    {field_parameter, 3, .inherited = true},
  [101] =
    {field_element, 1, .inherited = true},
  [102] =
    {field_element, 0, .inherited = true},
    {field_element, 1, .inherited = true},
  [104] =
    {field_domain, 0},
    {field_range, 2},
  [106] =
    {field_name, 0, .inherited = true},
    {field_name, 1, .inherited = true},
  [108] =
    {field_variable, 0, .inherited = true},
    {field_variable, 1, .inherited = true},
  [110] =
    {field_lhs, 0},
    {field_operator, 1},
    {field_rhs, 2},
  [113] =
    {field_binding, 0},
    {field_body, 2},
  [115] =
    {field_predicate, 0},
  [116] =
    {field_argument, 0},
  [117] =
    {field_lhs, 0},
    {field_rhs, 2},
  [119] =
    {field_lhs, 0},
    {field_relation, 1},
    {field_rhs, 2},
  [122] =
    {field_identity, 2},
  [123] =
    {field_member, 0},
  [124] =
    {field_wildcard, 1},
  [125] =
    {field_base, 4},
    {field_body, 5},
    {field_name, 1},
    {field_opaque, 3},
  [129] =
    {field_base, 4},
    {field_name, 1},
    {field_opaque, 3},
    {field_restriction, 5},
  [133] =
    {field_base, 3},
    {field_body, 5},
    {field_name, 1},
    {field_restriction, 4},
  [137] =
    {field_min, 1},
  [138] =
    {field_ordering, 1},
  [139] =
    {field_uniqueness, 1},
  [140] =
    {field_body, 5},
    {field_from, 4},
    {field_name, 2},
  [143] =
    {field_body, 4},
    {field_cardinality, 2},
    {field_name, 0},
    {field_target, 3},
  [147] =
    {field_name, 1},
    {field_type, 4, .inherited = true},
  [149] =
    {field_type, 0, .inherited = true},
    {field_type, 1, .inherited = true},
  [151] =
    {field_function, 2, .inherited = true},
  [152] =
    {field_class, 0},
  [153] =
    {field_from, 5},
    {field_name, 1},
    {field_parameter, 3, .inherited = true},
  [156] =
    {field_body, 5},
    {field_name, 1},
    {field_parameter, 3, .inherited = true},
  [159] =
    {field_empty, 3},
    {field_ordering, 1},
  [161] =
    {field_empty, 3},
    {field_uniqueness, 1},
  [163] =
    {field_name, 0},
    {field_value, 2},
  [165] =
    {field_name, 0},
    {field_range, 2},
  [167] =
    {field_argument, 2, .inherited = true},
    {field_predicate, 0},
  [169] =
    {field_argument, 2, .inherited = true},
    {field_function, 0},
  [171] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
  [173] =
    {field_source, 2},
    {field_variable, 0},
  [175] =
    {field_base, 4},
    {field_body, 6},
    {field_name, 1},
    {field_opaque, 3},
    {field_restriction, 5},
  [180] =
    {field_name, 0},
    {field_parameter, 2, .inherited = true},
    {field_target, 5},
  [183] =
    {field_max, 1},
  [184] =
    {field_min, 1},
    {field_range, 2},
  [186] =
    {field_min, 2},
    {field_ordering, 1},
  [188] =
    {field_ordering, 1},
    {field_uniqueness, 2},
  [190] =
    {field_min, 2},
    {field_uniqueness, 1},
  [192] =
    {field_referent, 1},
  [193] =
    {field_body, 6},
    {field_name, 1},
    {field_type, 4, .inherited = true},
  [196] =
    {field_cardinality, 0},
    {field_class, 1},
  [198] =
    {field_body, 6},
    {field_from, 5},
    {field_name, 1},
    {field_parameter, 3, .inherited = true},
  [202] =
    {field_body, 3},
    {field_name, 0},
    {field_rename, 2},
  [205] =
    {field_empty, 4},
    {field_ordering, 1},
    {field_uniqueness, 2},
  [208] =
    {field_body, 3},
  [209] =
    {field_body, 3},
    {field_variable, 1, .inherited = true},
  [211] =
    {field_member, 2, .inherited = true},
  [212] =
    {field_member, 0, .inherited = true},
    {field_member, 1, .inherited = true},
  [214] =
    {field_member, 0},
    {field_rename, 2},
  [216] =
    {field_facet, 0},
    {field_value, 2},
  [218] =
    {field_value, 2},
  [219] =
    {field_name, 1},
    {field_parent, 3},
  [221] =
    {field_cardinality, 5},
    {field_name, 0},
    {field_parameter, 2, .inherited = true},
    {field_target, 6},
  [225] =
    {field_domain, 1},
    {field_range, 3},
  [227] =
    {field_min, 2},
    {field_ordering, 1},
    {field_range, 3},
  [230] =
    {field_min, 3},
    {field_ordering, 1},
    {field_uniqueness, 2},
  [233] =
    {field_min, 2},
    {field_range, 3},
    {field_uniqueness, 1},
  [236] =
    {field_body, 2},
    {field_referent, 1},
  [238] =
    {field_element, 4, .inherited = true},
    {field_ordering, 1},
  [240] =
    {field_element, 4, .inherited = true},
    {field_uniqueness, 1},
  [242] =
    {field_facet, 0},
    {field_is_fixed, 2},
    {field_value, 3},
  [245] =
    {field_is_fixed, 2},
    {field_value, 3},
  [247] =
    {field_body, 4},
    {field_name, 1},
    {field_parent, 3},
  [250] =
    {field_min, 3},
    {field_ordering, 1},
    {field_range, 4},
    {field_uniqueness, 2},
  [254] =
    {field_argument, 2, .inherited = true},
    {field_class, 0},
  [256] =
    {field_element, 5, .inherited = true},
    {field_ordering, 1},
    {field_uniqueness, 2},
  [259] =
    {field_value, 2},
    {field_value, 3},
    {field_value, 4},
  [262] =
    {field_argument, 3, .inherited = true},
    {field_cardinality, 0},
    {field_class, 1},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 4,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 15,
  [18] = 16,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 20,
  [26] = 26,
  [27] = 22,
  [28] = 20,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 52,
  [54] = 52,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 59,
  [60] = 60,
  [61] = 61,
  [62] = 59,
  [63] = 59,
  [64] = 61,
  [65] = 60,
  [66] = 60,
  [67] = 60,
  [68] = 61,
  [69] = 61,
  [70] = 70,
  [71] = 71,
  [72] = 72,
  [73] = 73,
  [74] = 74,
  [75] = 75,
  [76] = 76,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 80,
  [81] = 81,
  [82] = 82,
  [83] = 83,
  [84] = 84,
  [85] = 85,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 90,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 103,
  [105] = 98,
  [106] = 12,
  [107] = 14,
  [108] = 99,
  [109] = 101,
  [110] = 110,
  [111] = 111,
  [112] = 30,
  [113] = 113,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 34,
  [118] = 118,
  [119] = 31,
  [120] = 33,
  [121] = 32,
  [122] = 122,
  [123] = 36,
  [124] = 35,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 3,
  [129] = 42,
  [130] = 130,
  [131] = 131,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 135,
  [136] = 49,
  [137] = 137,
  [138] = 138,
  [139] = 139,
  [140] = 44,
  [141] = 51,
  [142] = 37,
  [143] = 38,
  [144] = 50,
  [145] = 41,
  [146] = 40,
  [147] = 39,
  [148] = 43,
  [149] = 45,
  [150] = 46,
  [151] = 47,
  [152] = 48,
  [153] = 5,
  [154] = 6,
  [155] = 7,
  [156] = 8,
  [157] = 9,
  [158] = 10,
  [159] = 11,
  [160] = 160,
  [161] = 161,
  [162] = 161,
  [163] = 139,
  [164] = 131,
  [165] = 138,
  [166] = 134,
  [167] = 135,
  [168] = 137,
  [169] = 169,
  [170] = 169,
  [171] = 171,
  [172] = 161,
  [173] = 139,
  [174] = 131,
  [175] = 138,
  [176] = 134,
  [177] = 135,
  [178] = 137,
  [179] = 169,
  [180] = 161,
  [181] = 139,
  [182] = 131,
  [183] = 138,
  [184] = 134,
  [185] = 135,
  [186] = 137,
  [187] = 169,
  [188] = 188,
  [189] = 189,
  [190] = 2,
  [191] = 191,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 196,
  [197] = 197,
  [198] = 198,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 203,
  [205] = 202,
  [206] = 202,
  [207] = 203,
  [208] = 203,
  [209] = 209,
  [210] = 210,
  [211] = 211,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 224,
  [225] = 225,
  [226] = 226,
  [227] = 227,
  [228] = 218,
  [229] = 227,
  [230] = 219,
  [231] = 220,
  [232] = 216,
  [233] = 224,
  [234] = 225,
  [235] = 226,
  [236] = 218,
  [237] = 227,
  [238] = 219,
  [239] = 220,
  [240] = 216,
  [241] = 224,
  [242] = 225,
  [243] = 226,
  [244] = 218,
  [245] = 227,
  [246] = 219,
  [247] = 220,
  [248] = 216,
  [249] = 224,
  [250] = 225,
  [251] = 226,
  [252] = 252,
  [253] = 253,
  [254] = 30,
  [255] = 255,
  [256] = 32,
  [257] = 33,
  [258] = 31,
  [259] = 34,
  [260] = 260,
  [261] = 261,
  [262] = 35,
  [263] = 36,
  [264] = 3,
  [265] = 265,
  [266] = 132,
  [267] = 133,
  [268] = 45,
  [269] = 3,
  [270] = 270,
  [271] = 49,
  [272] = 272,
  [273] = 44,
  [274] = 5,
  [275] = 6,
  [276] = 7,
  [277] = 8,
  [278] = 51,
  [279] = 37,
  [280] = 9,
  [281] = 10,
  [282] = 38,
  [283] = 11,
  [284] = 50,
  [285] = 41,
  [286] = 40,
  [287] = 42,
  [288] = 39,
  [289] = 43,
  [290] = 46,
  [291] = 47,
  [292] = 48,
  [293] = 2,
  [294] = 14,
  [295] = 12,
  [296] = 296,
  [297] = 297,
  [298] = 7,
  [299] = 11,
  [300] = 300,
  [301] = 301,
  [302] = 302,
  [303] = 303,
  [304] = 127,
  [305] = 2,
  [306] = 302,
  [307] = 307,
  [308] = 308,
  [309] = 272,
  [310] = 8,
  [311] = 6,
  [312] = 312,
  [313] = 9,
  [314] = 314,
  [315] = 297,
  [316] = 316,
  [317] = 297,
  [318] = 10,
  [319] = 5,
  [320] = 302,
  [321] = 302,
  [322] = 322,
  [323] = 323,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 301,
  [329] = 14,
  [330] = 330,
  [331] = 331,
  [332] = 332,
  [333] = 333,
  [334] = 334,
  [335] = 335,
  [336] = 336,
  [337] = 337,
  [338] = 338,
  [339] = 339,
  [340] = 340,
  [341] = 341,
  [342] = 342,
  [343] = 343,
  [344] = 344,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 348,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 352,
  [353] = 353,
  [354] = 354,
  [355] = 12,
  [356] = 356,
  [357] = 357,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 361,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 365,
  [366] = 366,
  [367] = 367,
  [368] = 368,
  [369] = 369,
  [370] = 370,
  [371] = 371,
  [372] = 372,
  [373] = 373,
  [374] = 374,
  [375] = 375,
  [376] = 376,
  [377] = 377,
  [378] = 193,
  [379] = 341,
  [380] = 51,
  [381] = 37,
  [382] = 357,
  [383] = 38,
  [384] = 363,
  [385] = 358,
  [386] = 50,
  [387] = 361,
  [388] = 370,
  [389] = 41,
  [390] = 40,
  [391] = 377,
  [392] = 351,
  [393] = 353,
  [394] = 368,
  [395] = 39,
  [396] = 43,
  [397] = 366,
  [398] = 367,
  [399] = 369,
  [400] = 45,
  [401] = 46,
  [402] = 47,
  [403] = 371,
  [404] = 48,
  [405] = 405,
  [406] = 406,
  [407] = 407,
  [408] = 408,
  [409] = 409,
  [410] = 410,
  [411] = 42,
  [412] = 412,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 417,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 422,
  [423] = 423,
  [424] = 424,
  [425] = 425,
  [426] = 426,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 434,
  [435] = 435,
  [436] = 436,
  [437] = 437,
  [438] = 438,
  [439] = 439,
  [440] = 440,
  [441] = 441,
  [442] = 442,
  [443] = 443,
  [444] = 444,
  [445] = 445,
  [446] = 446,
  [447] = 447,
  [448] = 448,
  [449] = 449,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 490,
  [491] = 491,
  [492] = 492,
  [493] = 493,
  [494] = 494,
  [495] = 495,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 499,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 503,
  [504] = 504,
  [505] = 505,
  [506] = 506,
  [507] = 507,
  [508] = 508,
  [509] = 509,
  [510] = 510,
  [511] = 511,
  [512] = 512,
  [513] = 513,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 517,
  [518] = 518,
  [519] = 519,
  [520] = 520,
  [521] = 521,
  [522] = 522,
  [523] = 523,
  [524] = 524,
  [525] = 525,
  [526] = 526,
  [527] = 527,
  [528] = 528,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 534,
  [535] = 535,
  [536] = 536,
  [537] = 537,
  [538] = 538,
  [539] = 539,
  [540] = 512,
  [541] = 541,
  [542] = 542,
  [543] = 512,
  [544] = 544,
  [545] = 545,
  [546] = 546,
  [547] = 547,
  [548] = 548,
  [549] = 549,
  [550] = 550,
  [551] = 551,
  [552] = 552,
  [553] = 553,
  [554] = 554,
  [555] = 406,
  [556] = 556,
  [557] = 557,
  [558] = 558,
  [559] = 559,
  [560] = 261,
  [561] = 561,
  [562] = 562,
  [563] = 562,
  [564] = 562,
  [565] = 565,
  [566] = 566,
  [567] = 567,
  [568] = 261,
  [569] = 569,
  [570] = 570,
  [571] = 3,
  [572] = 572,
  [573] = 572,
  [574] = 574,
  [575] = 575,
  [576] = 576,
  [577] = 577,
  [578] = 578,
  [579] = 579,
  [580] = 580,
  [581] = 581,
  [582] = 582,
  [583] = 583,
  [584] = 584,
  [585] = 585,
  [586] = 3,
  [587] = 272,
  [588] = 127,
  [589] = 5,
  [590] = 590,
  [591] = 7,
  [592] = 8,
  [593] = 9,
  [594] = 10,
  [595] = 11,
  [596] = 127,
  [597] = 2,
  [598] = 127,
  [599] = 577,
  [600] = 600,
  [601] = 577,
  [602] = 602,
  [603] = 577,
  [604] = 604,
  [605] = 567,
  [606] = 606,
  [607] = 607,
  [608] = 6,
  [609] = 2,
  [610] = 301,
  [611] = 611,
  [612] = 8,
  [613] = 9,
  [614] = 10,
  [615] = 11,
  [616] = 616,
  [617] = 14,
  [618] = 618,
  [619] = 619,
  [620] = 618,
  [621] = 296,
  [622] = 272,
  [623] = 623,
  [624] = 624,
  [625] = 625,
  [626] = 312,
  [627] = 14,
  [628] = 451,
  [629] = 7,
  [630] = 12,
  [631] = 127,
  [632] = 127,
  [633] = 633,
  [634] = 14,
  [635] = 12,
  [636] = 636,
  [637] = 637,
  [638] = 458,
  [639] = 639,
  [640] = 640,
  [641] = 641,
  [642] = 642,
  [643] = 127,
  [644] = 644,
  [645] = 645,
  [646] = 127,
  [647] = 647,
  [648] = 644,
  [649] = 649,
  [650] = 644,
  [651] = 5,
  [652] = 6,
  [653] = 644,
  [654] = 618,
  [655] = 618,
  [656] = 12,
  [657] = 657,
  [658] = 127,
  [659] = 372,
  [660] = 362,
  [661] = 14,
  [662] = 662,
  [663] = 368,
  [664] = 664,
  [665] = 312,
  [666] = 341,
  [667] = 260,
  [668] = 357,
  [669] = 451,
  [670] = 363,
  [671] = 358,
  [672] = 434,
  [673] = 361,
  [674] = 370,
  [675] = 193,
  [676] = 377,
  [677] = 351,
  [678] = 353,
  [679] = 679,
  [680] = 265,
  [681] = 366,
  [682] = 367,
  [683] = 369,
  [684] = 684,
  [685] = 685,
  [686] = 12,
  [687] = 458,
  [688] = 127,
  [689] = 127,
  [690] = 690,
  [691] = 691,
  [692] = 14,
  [693] = 14,
  [694] = 12,
  [695] = 12,
  [696] = 696,
  [697] = 697,
  [698] = 698,
  [699] = 662,
  [700] = 700,
  [701] = 701,
  [702] = 356,
  [703] = 373,
  [704] = 347,
  [705] = 697,
  [706] = 301,
  [707] = 707,
  [708] = 335,
  [709] = 709,
  [710] = 710,
  [711] = 711,
  [712] = 712,
  [713] = 428,
  [714] = 371,
  [715] = 127,
  [716] = 443,
  [717] = 366,
  [718] = 367,
  [719] = 369,
  [720] = 371,
  [721] = 721,
  [722] = 722,
  [723] = 347,
  [724] = 724,
  [725] = 725,
  [726] = 726,
  [727] = 727,
  [728] = 728,
  [729] = 729,
  [730] = 730,
  [731] = 372,
  [732] = 362,
  [733] = 733,
  [734] = 734,
  [735] = 735,
  [736] = 736,
  [737] = 737,
  [738] = 738,
  [739] = 434,
  [740] = 740,
  [741] = 741,
  [742] = 742,
  [743] = 473,
  [744] = 744,
  [745] = 14,
  [746] = 335,
  [747] = 14,
  [748] = 12,
  [749] = 14,
  [750] = 750,
  [751] = 12,
  [752] = 752,
  [753] = 363,
  [754] = 358,
  [755] = 755,
  [756] = 449,
  [757] = 757,
  [758] = 758,
  [759] = 574,
  [760] = 460,
  [761] = 761,
  [762] = 439,
  [763] = 763,
  [764] = 764,
  [765] = 763,
  [766] = 764,
  [767] = 373,
  [768] = 480,
  [769] = 769,
  [770] = 770,
  [771] = 758,
  [772] = 772,
  [773] = 437,
  [774] = 761,
  [775] = 775,
  [776] = 763,
  [777] = 764,
  [778] = 361,
  [779] = 779,
  [780] = 12,
  [781] = 781,
  [782] = 370,
  [783] = 368,
  [784] = 784,
  [785] = 193,
  [786] = 786,
  [787] = 787,
  [788] = 377,
  [789] = 351,
  [790] = 341,
  [791] = 781,
  [792] = 353,
  [793] = 793,
  [794] = 442,
  [795] = 758,
  [796] = 444,
  [797] = 781,
  [798] = 798,
  [799] = 793,
  [800] = 800,
  [801] = 761,
  [802] = 260,
  [803] = 781,
  [804] = 793,
  [805] = 805,
  [806] = 793,
  [807] = 721,
  [808] = 356,
  [809] = 809,
  [810] = 721,
  [811] = 357,
  [812] = 812,
  [813] = 265,
  [814] = 814,
  [815] = 428,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 820,
  [821] = 814,
  [822] = 473,
  [823] = 823,
  [824] = 824,
  [825] = 735,
  [826] = 824,
  [827] = 827,
  [828] = 828,
  [829] = 449,
  [830] = 460,
  [831] = 439,
  [832] = 832,
  [833] = 480,
  [834] = 834,
  [835] = 835,
  [836] = 836,
  [837] = 837,
  [838] = 738,
  [839] = 839,
  [840] = 840,
  [841] = 841,
  [842] = 842,
  [843] = 442,
  [844] = 443,
  [845] = 444,
  [846] = 846,
  [847] = 847,
  [848] = 848,
  [849] = 849,
  [850] = 14,
  [851] = 851,
  [852] = 12,
  [853] = 408,
  [854] = 2,
  [855] = 855,
  [856] = 847,
  [857] = 857,
  [858] = 405,
  [859] = 859,
  [860] = 860,
  [861] = 861,
  [862] = 862,
  [863] = 842,
  [864] = 864,
  [865] = 865,
  [866] = 864,
  [867] = 867,
  [868] = 868,
  [869] = 12,
  [870] = 860,
  [871] = 871,
  [872] = 872,
  [873] = 847,
  [874] = 857,
  [875] = 875,
  [876] = 860,
  [877] = 862,
  [878] = 878,
  [879] = 879,
  [880] = 880,
  [881] = 868,
  [882] = 847,
  [883] = 883,
  [884] = 860,
  [885] = 885,
  [886] = 886,
  [887] = 859,
  [888] = 437,
  [889] = 889,
  [890] = 847,
  [891] = 860,
  [892] = 847,
  [893] = 860,
  [894] = 894,
  [895] = 779,
  [896] = 846,
  [897] = 851,
  [898] = 894,
  [899] = 434,
  [900] = 894,
  [901] = 846,
  [902] = 851,
  [903] = 862,
  [904] = 904,
  [905] = 905,
  [906] = 865,
  [907] = 907,
  [908] = 908,
  [909] = 909,
  [910] = 910,
  [911] = 911,
  [912] = 912,
  [913] = 913,
  [914] = 914,
  [915] = 915,
  [916] = 916,
  [917] = 917,
  [918] = 918,
  [919] = 919,
  [920] = 920,
  [921] = 909,
  [922] = 922,
  [923] = 923,
  [924] = 924,
  [925] = 925,
  [926] = 926,
  [927] = 927,
  [928] = 928,
  [929] = 913,
  [930] = 930,
  [931] = 418,
  [932] = 932,
  [933] = 933,
  [934] = 934,
  [935] = 412,
  [936] = 936,
  [937] = 924,
  [938] = 912,
  [939] = 939,
  [940] = 940,
  [941] = 925,
  [942] = 942,
  [943] = 943,
  [944] = 944,
  [945] = 945,
  [946] = 922,
  [947] = 923,
  [948] = 948,
  [949] = 925,
  [950] = 928,
  [951] = 913,
  [952] = 952,
  [953] = 953,
  [954] = 912,
  [955] = 955,
  [956] = 956,
  [957] = 957,
  [958] = 265,
  [959] = 959,
  [960] = 960,
  [961] = 922,
  [962] = 923,
  [963] = 924,
  [964] = 925,
  [965] = 928,
  [966] = 913,
  [967] = 967,
  [968] = 968,
  [969] = 969,
  [970] = 928,
  [971] = 922,
  [972] = 215,
  [973] = 437,
  [974] = 974,
  [975] = 923,
  [976] = 260,
  [977] = 322,
  [978] = 690,
  [979] = 979,
  [980] = 421,
  [981] = 916,
  [982] = 982,
  [983] = 926,
  [984] = 916,
  [985] = 916,
  [986] = 443,
  [987] = 924,
  [988] = 988,
  [989] = 989,
  [990] = 990,
  [991] = 991,
  [992] = 992,
  [993] = 993,
  [994] = 994,
  [995] = 995,
  [996] = 996,
  [997] = 997,
  [998] = 998,
  [999] = 999,
  [1000] = 1000,
  [1001] = 997,
  [1002] = 988,
  [1003] = 1003,
  [1004] = 1004,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 1011,
  [1012] = 1012,
  [1013] = 1006,
  [1014] = 1014,
  [1015] = 1015,
  [1016] = 1016,
  [1017] = 1017,
  [1018] = 991,
  [1019] = 1011,
  [1020] = 1020,
  [1021] = 1021,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 1025,
  [1026] = 1026,
  [1027] = 88,
  [1028] = 1005,
  [1029] = 991,
  [1030] = 1030,
  [1031] = 89,
  [1032] = 1032,
  [1033] = 992,
  [1034] = 92,
  [1035] = 91,
  [1036] = 1036,
  [1037] = 90,
  [1038] = 1017,
  [1039] = 1006,
  [1040] = 1040,
  [1041] = 1041,
  [1042] = 1042,
  [1043] = 1015,
  [1044] = 93,
  [1045] = 1045,
  [1046] = 1046,
  [1047] = 1047,
  [1048] = 1017,
  [1049] = 1011,
  [1050] = 1050,
  [1051] = 1051,
  [1052] = 1017,
  [1053] = 1053,
  [1054] = 1054,
  [1055] = 1055,
  [1056] = 1056,
  [1057] = 1057,
  [1058] = 1005,
  [1059] = 991,
  [1060] = 95,
  [1061] = 1061,
  [1062] = 1005,
  [1063] = 991,
  [1064] = 1064,
  [1065] = 1011,
  [1066] = 1005,
  [1067] = 991,
  [1068] = 85,
  [1069] = 991,
  [1070] = 991,
  [1071] = 991,
  [1072] = 991,
  [1073] = 991,
  [1074] = 991,
  [1075] = 991,
  [1076] = 996,
  [1077] = 1077,
  [1078] = 1078,
  [1079] = 1079,
  [1080] = 1021,
  [1081] = 1081,
  [1082] = 1014,
  [1083] = 86,
  [1084] = 999,
  [1085] = 991,
  [1086] = 997,
  [1087] = 988,
  [1088] = 1088,
  [1089] = 1089,
  [1090] = 1007,
  [1091] = 1091,
  [1092] = 1025,
  [1093] = 1015,
  [1094] = 996,
  [1095] = 1078,
  [1096] = 84,
  [1097] = 1088,
  [1098] = 1098,
  [1099] = 999,
  [1100] = 1100,
  [1101] = 1101,
  [1102] = 988,
  [1103] = 1088,
  [1104] = 1089,
  [1105] = 1078,
  [1106] = 992,
  [1107] = 87,
  [1108] = 1108,
  [1109] = 996,
  [1110] = 1110,
  [1111] = 1023,
  [1112] = 999,
  [1113] = 1113,
  [1114] = 1114,
  [1115] = 1115,
  [1116] = 1088,
  [1117] = 1089,
  [1118] = 94,
  [1119] = 996,
  [1120] = 996,
  [1121] = 996,
  [1122] = 1122,
  [1123] = 1123,
  [1124] = 1005,
  [1125] = 995,
  [1126] = 1040,
  [1127] = 1089,
  [1128] = 1030,
  [1129] = 1129,
  [1130] = 1130,
  [1131] = 1131,
  [1132] = 1006,
  [1133] = 1005,
  [1134] = 1007,
  [1135] = 991,
  [1136] = 1136,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym_definition,
  sym_inequality_relation,
  sym_logical_connective,
  sym_logical_quantifier,
  sym_math_operator,
  sym_restriction_facet,
  sym_set_membership,
  sym_set_operator,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym_definition] = {.index = 0, .length = 12},
  [sym_inequality_relation] = {.index = 12, .length = 5},
  [sym_logical_connective] = {.index = 17, .length = 5},
  [sym_logical_quantifier] = {.index = 22, .length = 3},
  [sym_math_operator] = {.index = 25, .length = 5},
  [sym_restriction_facet] = {.index = 30, .length = 5},
  [sym_set_membership] = {.index = 35, .length = 2},
  [sym_set_operator] = {.index = 37, .length = 9},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym_datatype_def,
    sym_dimension_def,
    sym_entity_def,
    sym_enum_def,
    sym_event_def,
    sym_metric_def,
    sym_metric_group_def,
    sym_property_def,
    sym_rdf_def,
    sym_structure_def,
    sym_type_class_def,
    sym_union_def,
  [12] =
    sym_op_greater_than,
    sym_op_greater_than_or_equal,
    sym_op_inequality,
    sym_op_less_than,
    sym_op_less_than_or_equal,
  [17] =
    sym_logical_biconditional,
    sym_logical_conjunction,
    sym_logical_disjunction,
    sym_logical_exclusive_disjunction,
    sym_logical_implication,
  [22] =
    sym_logical_quantifier_existential,
    sym_logical_quantifier_existential_one,
    sym_logical_quantifier_universal,
  [25] =
    sym_math_op_add,
    sym_math_op_divide,
    sym_math_op_modulo,
    sym_math_op_multiply,
    sym_math_op_subtract,
  [30] =
    sym_digit_restriction_facet,
    sym_length_restriction_facet,
    sym_pattern_restriction_facet,
    sym_tz_restriction_facet,
    sym_value_restriction_facet,
  [35] =
    sym_set_op_is_member,
    sym_set_op_is_not_member,
  [37] =
    sym_set_membership,
    sym_set_op_complement,
    sym_set_op_intersection,
    sym_set_op_product,
    sym_set_op_subset,
    sym_set_op_subset_or_equal,
    sym_set_op_supset,
    sym_set_op_supset_or_equal,
    sym_set_op_union,
};

static const TSCharacterRange sym_identifier_character_set_1[] = {
  {'A', 'Z'}, {'a', 'z'}, {0xb5, 0xb5}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x1ba}, {0x1bc, 0x1bf}, {0x1c4, 0x1c4},
  {0x1c6, 0x1c7}, {0x1c9, 0x1ca}, {0x1cc, 0x1f1}, {0x1f3, 0x293}, {0x295, 0x2af}, {0x370, 0x373}, {0x376, 0x377}, {0x37b, 0x37d},
  {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x48a, 0x52f},
  {0x531, 0x556}, {0x560, 0x588}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fd, 0x10ff}, {0x13a0, 0x13f5},
  {0x13f8, 0x13fd}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1d00, 0x1d2b}, {0x1d6b, 0x1d77}, {0x1d79, 0x1d9a}, {0x1e00, 0x1f15},
  {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d},
  {0x1f80, 0x1f87}, {0x1f90, 0x1f97}, {0x1fa0, 0x1fa7}, {0x1fb0, 0x1fb4}, {0x1fb6, 0x1fbb}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcb},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffb}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113},
  {0x2115, 0x2115}, {0x2119, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x212d}, {0x212f, 0x2134}, {0x2139, 0x2139},
  {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2183, 0x2184}, {0x2c00, 0x2c7b}, {0x2c7e, 0x2ce4}, {0x2ceb, 0x2cee}, {0x2cf2, 0x2cf3},
  {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0xa640, 0xa66d}, {0xa680, 0xa69b}, {0xa722, 0xa76f}, {0xa771, 0xa787}, {0xa78b, 0xa78e},
  {0xa790, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f5, 0xa7f6}, {0xa7fa, 0xa7fa}, {0xab30, 0xab5a}, {0xab60, 0xab68},
  {0xab70, 0xabbf}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0x10400, 0x1044f}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb},
  {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc},
  {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d50, 0x10d65}, {0x10d70, 0x10d85}, {0x118a0, 0x118df}, {0x16e40, 0x16e7f}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c},
  {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505},
  {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550},
  {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e},
  {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1df00, 0x1df09}, {0x1df0b, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e900, 0x1e943},
};

static const TSCharacterRange sym_identifier_character_set_2[] = {
  {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xb5, 0xb5}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x1ba},
  {0x1bc, 0x1bf}, {0x1c4, 0x1c4}, {0x1c6, 0x1c7}, {0x1c9, 0x1ca}, {0x1cc, 0x1f1}, {0x1f3, 0x293}, {0x295, 0x2af}, {0x370, 0x373},
  {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x48a, 0x52f}, {0x531, 0x556}, {0x560, 0x588}, {0x660, 0x669}, {0x6f0, 0x6f9}, {0x7c0, 0x7c9}, {0x966, 0x96f},
  {0x9e6, 0x9ef}, {0xa66, 0xa6f}, {0xae6, 0xaef}, {0xb66, 0xb6f}, {0xbe6, 0xbef}, {0xc66, 0xc6f}, {0xce6, 0xcef}, {0xd66, 0xd6f},
  {0xde6, 0xdef}, {0xe50, 0xe59}, {0xed0, 0xed9}, {0xf20, 0xf29}, {0x1040, 0x1049}, {0x1090, 0x1099}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7},
  {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fd, 0x10ff}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x17e0, 0x17e9}, {0x1810, 0x1819}, {0x1946, 0x194f},
  {0x19d0, 0x19d9}, {0x1a80, 0x1a89}, {0x1a90, 0x1a99}, {0x1b50, 0x1b59}, {0x1bb0, 0x1bb9}, {0x1c40, 0x1c49}, {0x1c50, 0x1c59}, {0x1c80, 0x1c8a},
  {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1d00, 0x1d2b}, {0x1d6b, 0x1d77}, {0x1d79, 0x1d9a}, {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45},
  {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1f87}, {0x1f90, 0x1f97},
  {0x1fa0, 0x1fa7}, {0x1fb0, 0x1fb4}, {0x1fb6, 0x1fbb}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcb}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb},
  {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffb}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2119, 0x211d},
  {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x212d}, {0x212f, 0x2134}, {0x2139, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149},
  {0x214e, 0x214e}, {0x2183, 0x2184}, {0x2c00, 0x2c7b}, {0x2c7e, 0x2ce4}, {0x2ceb, 0x2cee}, {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0xa620, 0xa629}, {0xa640, 0xa66d}, {0xa680, 0xa69b}, {0xa722, 0xa76f}, {0xa771, 0xa787}, {0xa78b, 0xa78e}, {0xa790, 0xa7cd},
  {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f5, 0xa7f6}, {0xa7fa, 0xa7fa}, {0xa8d0, 0xa8d9}, {0xa900, 0xa909}, {0xa9d0, 0xa9d9},
  {0xa9f0, 0xa9f9}, {0xaa50, 0xaa59}, {0xab30, 0xab5a}, {0xab60, 0xab68}, {0xab70, 0xabbf}, {0xabf0, 0xabf9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17},
  {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0x10400, 0x1044f}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10570, 0x1057a},
  {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x10c80, 0x10cb2},
  {0x10cc0, 0x10cf2}, {0x10d30, 0x10d39}, {0x10d40, 0x10d49}, {0x10d50, 0x10d65}, {0x10d70, 0x10d85}, {0x11066, 0x1106f}, {0x110f0, 0x110f9}, {0x11136, 0x1113f},
  {0x111d0, 0x111d9}, {0x112f0, 0x112f9}, {0x11450, 0x11459}, {0x114d0, 0x114d9}, {0x11650, 0x11659}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11730, 0x11739},
  {0x118a0, 0x118e9}, {0x11950, 0x11959}, {0x11bf0, 0x11bf9}, {0x11c50, 0x11c59}, {0x11d50, 0x11d59}, {0x11da0, 0x11da9}, {0x11f50, 0x11f59}, {0x16130, 0x16139},
  {0x16a60, 0x16a69}, {0x16ac0, 0x16ac9}, {0x16b50, 0x16b59}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x1ccf0, 0x1ccf9}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c},
  {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505},
  {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550},
  {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e},
  {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1df00, 0x1df09}, {0x1df0b, 0x1df1e}, {0x1df25, 0x1df2a},
  {0x1e140, 0x1e149}, {0x1e2f0, 0x1e2f9}, {0x1e4f0, 0x1e4f9}, {0x1e5f1, 0x1e5fa}, {0x1e900, 0x1e943}, {0x1e950, 0x1e959}, {0x1fbf0, 0x1fbf9},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(696);
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1333,
        '.', 1335,
        '/', 1327,
        '0', 1128,
        ':', 1059,
        ';', 74,
        '<', 1110,
        '=', 1065,
        '>', 1276,
        '@', 1062,
        'N', 436,
        'T', 286,
        '[', 705,
        ']', 706,
        '_', 1260,
        'a', 1256,
        'b', 101,
        'c', 344,
        'd', 102,
        'e', 380,
        'f', 103,
        'g', 82,
        'h', 204,
        'i', 175,
        'l', 110,
        'm', 112,
        'n', 179,
        'o', 253,
        'p', 105,
        'r', 114,
        's', 203,
        't', 317,
        'u', 383,
        'v', 212,
        'w', 299,
        'x', 450,
        'y', 210,
        '{', 1091,
        '|', 1324,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2190, 1342,
        0x2192, 1340,
        0x21d2, 1293,
        0x21d4, 1296,
        0x2200, 1299,
        0x2203, 1302,
        0x2205, 1136,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2218, 1336,
        0x2227, 1286,
        0x2228, 1288,
        0x2229, 1307,
        0x222a, 1305,
        0x2254, 1338,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x22a4, 1132,
        0x22a5, 1135,
        0x22bb, 1290,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(695);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1129);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '(', 1079,
        ')', 1080,
        ',', 1081,
        '.', 1334,
        '0', 1125,
        ':', 1058,
        ';', 74,
        '<', 1109,
        '@', 691,
        '[', 705,
        'e', 1041,
        'f', 732,
        's', 788,
        't', 956,
        '{', 1091,
        0x2205, 1136,
        0x2218, 1336,
        0x22a4, 1132,
        0x22a5, 1135,
        '+', 35,
        '-', 35,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(2);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '(', 1079,
        ')', 1080,
        ',', 1081,
        '0', 1125,
        ';', 74,
        '<', 1109,
        '[', 705,
        'e', 1041,
        'f', 732,
        's', 788,
        't', 956,
        '{', 1091,
        0x2205, 1136,
        0x22a4, 1132,
        0x22a5, 1135,
        '+', 35,
        '-', 35,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(2);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '(', 1079,
        '+', 35,
        '-', 34,
        '0', 1125,
        ':', 1058,
        ';', 74,
        '<', 1109,
        '@', 691,
        '[', 705,
        ']', 706,
        'f', 732,
        't', 956,
        '{', 1091,
        0x2192, 1340,
        0x2205, 1136,
        0x22a4, 1132,
        0x22a5, 1135,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(4);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '(', 1079,
        '+', 35,
        '-', 34,
        '0', 1125,
        ';', 74,
        '<', 1109,
        '[', 705,
        ']', 706,
        'f', 732,
        't', 956,
        '{', 1091,
        0x2192, 1340,
        0x2205, 1136,
        0x22a4, 1132,
        0x22a5, 1135,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(4);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '(', 1079,
        '0', 1125,
        ';', 74,
        '<', 1109,
        '[', 705,
        'e', 1040,
        'f', 731,
        'n', 929,
        's', 788,
        't', 956,
        'w', 848,
        '{', 1091,
        0xac, 1284,
        0x2200, 1299,
        0x2203, 1302,
        0x2205, 1136,
        0x22a4, 1132,
        0x22a5, 1135,
        '+', 35,
        '-', 35,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(5);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '0', 1125,
        ':', 1058,
        ';', 74,
        '<', 1109,
        '@', 1061,
        'a', 532,
        'c', 343,
        'd', 133,
        'e', 381,
        'f', 104,
        'i', 371,
        'm', 208,
        'p', 505,
        'r', 168,
        's', 587,
        't', 499,
        'u', 431,
        'v', 212,
        0x2200, 1299,
        0x2203, 1302,
        0x22a4, 1132,
        0x22a5, 1135,
        '+', 35,
        '-', 35,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '0', 1125,
        ';', 74,
        '<', 1109,
        '@', 1061,
        'a', 532,
        'c', 343,
        'd', 133,
        'e', 381,
        'f', 104,
        'i', 371,
        'm', 208,
        'p', 505,
        'r', 168,
        's', 587,
        't', 499,
        'u', 431,
        'v', 212,
        0x2200, 1299,
        0x2203, 1302,
        0x22a4, 1132,
        0x22a5, 1135,
        '+', 35,
        '-', 35,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '0', 1125,
        ';', 74,
        '<', 1109,
        'f', 730,
        't', 956,
        0x22a4, 1132,
        0x22a5, 1135,
        '+', 35,
        '-', 35,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 9:
      if (lookahead == '"') ADVANCE(1093);
      if (lookahead == '\\') ADVANCE(609);
      if (lookahead > 0x08 &&
          (lookahead < 0x0b || 0x1f < lookahead) &&
          lookahead != 0x7f) ADVANCE(9);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1333,
        '.', 33,
        '/', 1327,
        '0', 1128,
        ':', 41,
        ';', 74,
        '<', 1111,
        '=', 1064,
        '>', 1276,
        '@', 1061,
        'a', 532,
        'c', 344,
        'd', 132,
        'e', 382,
        'f', 444,
        'i', 177,
        'l', 211,
        'm', 112,
        'n', 439,
        'o', 521,
        'p', 135,
        'r', 169,
        's', 442,
        't', 463,
        'u', 395,
        'v', 212,
        'w', 299,
        '{', 1091,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2192, 1340,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2229, 1307,
        0x222a, 1305,
        0x2254, 1338,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(10);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1129);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1333,
        '/', 1327,
        ';', 74,
        '<', 1111,
        '=', 1064,
        '>', 1276,
        '@', 1061,
        'a', 546,
        'c', 344,
        'd', 132,
        'e', 382,
        'f', 445,
        'i', 178,
        'l', 211,
        'm', 112,
        'n', 462,
        'p', 135,
        'r', 169,
        's', 442,
        't', 463,
        'u', 431,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2192, 1340,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2229, 1307,
        0x222a, 1305,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(11);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1333,
        '/', 1327,
        ';', 74,
        '<', 1111,
        '=', 1064,
        '>', 1276,
        '@', 1063,
        'a', 546,
        'c', 344,
        'd', 132,
        'e', 382,
        'f', 445,
        'i', 178,
        'l', 211,
        'm', 112,
        'n', 462,
        'p', 135,
        'r', 169,
        's', 442,
        't', 463,
        'u', 431,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2192, 1340,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2229, 1307,
        0x222a, 1305,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(11);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        '-', 1332,
        '/', 1327,
        ';', 74,
        '<', 1110,
        '=', 1065,
        '>', 1276,
        '@', 691,
        'a', 404,
        'c', 344,
        'd', 132,
        'e', 379,
        'f', 443,
        'i', 252,
        'm', 209,
        'n', 462,
        'o', 254,
        'p', 507,
        'r', 169,
        's', 588,
        'u', 431,
        'x', 450,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x21d2, 1293,
        0x21d4, 1296,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2227, 1286,
        0x2228, 1288,
        0x2229, 1307,
        0x222a, 1305,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x22bb, 1290,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        '-', 1332,
        '/', 1327,
        ';', 74,
        '<', 1110,
        '=', 1065,
        '>', 1276,
        'a', 404,
        'c', 344,
        'd', 132,
        'e', 379,
        'f', 443,
        'i', 252,
        'm', 209,
        'n', 462,
        'o', 254,
        'p', 507,
        'r', 169,
        's', 588,
        'u', 431,
        'x', 450,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x21d2, 1293,
        0x21d4, 1296,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2227, 1286,
        0x2228, 1288,
        0x2229, 1307,
        0x222a, 1305,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x22bb, 1290,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1332,
        '.', 1334,
        '/', 1327,
        ':', 1058,
        ';', 74,
        '<', 1111,
        '=', 1064,
        '>', 1276,
        '@', 1062,
        'a', 546,
        'c', 344,
        'd', 132,
        'e', 379,
        'f', 443,
        'i', 178,
        'm', 209,
        'n', 462,
        'p', 507,
        'r', 169,
        's', 442,
        'u', 431,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2218, 1336,
        0x2229, 1307,
        0x222a, 1305,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '%', 1330,
        '(', 1079,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1332,
        '/', 1327,
        ';', 74,
        '<', 1111,
        '=', 1064,
        '>', 1276,
        '@', 1061,
        'a', 546,
        'c', 344,
        'd', 132,
        'e', 379,
        'f', 443,
        'i', 178,
        'm', 209,
        'n', 462,
        'p', 507,
        'r', 169,
        's', 442,
        'u', 431,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2229, 1307,
        0x222a, 1305,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '(', 1079,
        ')', 1080,
        '+', 1331,
        '-', 77,
        ':', 1058,
        ';', 74,
        '<', 27,
        '@', 1061,
        '[', 705,
        ']', 706,
        '_', 1260,
        '{', 1091,
        '|', 1324,
        0x2190, 1342,
        0x2192, 1340,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(18);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '(', 1079,
        ')', 1080,
        '+', 1331,
        '-', 77,
        ';', 74,
        '<', 27,
        '@', 1061,
        '[', 705,
        ']', 706,
        '_', 1260,
        '{', 1091,
        '|', 1324,
        0x2190, 1342,
        0x2192, 1340,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(18);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '(', 1079,
        ')', 1080,
        '-', 77,
        '.', 1334,
        ':', 1060,
        ';', 74,
        '<', 28,
        '=', 1065,
        '>', 1119,
        '@', 1061,
        'a', 403,
        'c', 343,
        'd', 132,
        'e', 382,
        'f', 444,
        'i', 176,
        'l', 211,
        'm', 113,
        'n', 462,
        'o', 487,
        'p', 134,
        'r', 169,
        's', 441,
        't', 463,
        'u', 431,
        'w', 299,
        'x', 450,
        '{', 1091,
        '}', 1092,
        0xac, 1284,
        0x2190, 1342,
        0x2192, 1340,
        0x21d2, 1293,
        0x21d4, 1296,
        0x2208, 1322,
        0x2209, 1323,
        0x2218, 1336,
        0x2227, 1286,
        0x2228, 1288,
        0x2254, 1338,
        0x22bb, 1290,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      END_STATE();
    case 20:
      ADVANCE_MAP(
        '(', 1079,
        ')', 1080,
        '-', 77,
        ':', 75,
        ';', 74,
        '<', 28,
        '=', 1065,
        '@', 1061,
        'a', 403,
        'c', 343,
        'd', 132,
        'e', 382,
        'f', 444,
        'i', 176,
        'l', 211,
        'm', 113,
        'n', 462,
        'o', 487,
        'p', 134,
        'r', 169,
        's', 441,
        't', 463,
        'u', 431,
        'w', 299,
        'x', 450,
        '{', 1091,
        '}', 1092,
        0xac, 1284,
        0x2190, 1342,
        0x2192, 1340,
        0x21d2, 1293,
        0x21d4, 1296,
        0x2208, 1322,
        0x2209, 1323,
        0x2227, 1286,
        0x2228, 1288,
        0x2254, 1338,
        0x22bb, 1290,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '(', 1079,
        '-', 77,
        ':', 1058,
        ';', 74,
        '@', 1061,
        'a', 992,
        'e', 905,
        'r', 789,
        0x2192, 1340,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 22:
      ADVANCE_MAP(
        '(', 1079,
        '-', 77,
        ':', 1058,
        ';', 74,
        '@', 1061,
        'a', 992,
        'e', 905,
        0x2192, 1340,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '(', 1079,
        '-', 77,
        ';', 74,
        '@', 1061,
        'a', 992,
        'e', 905,
        'r', 789,
        0x2192, 1340,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 24:
      if (lookahead == '(') ADVANCE(1079);
      if (lookahead == '-') ADVANCE(77);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1061);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 0x2192) ADVANCE(1340);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '(', 1079,
        ';', 74,
        'N', 933,
        'T', 838,
        'a', 891,
        'b', 754,
        'd', 733,
        'f', 879,
        'g', 716,
        'h', 773,
        'i', 908,
        'l', 739,
        'n', 774,
        'p', 931,
        'r', 745,
        's', 839,
        't', 844,
        'u', 892,
        'y', 792,
        '{', 1091,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(25);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '(', 1079,
        ';', 74,
        'N', 933,
        'T', 838,
        'a', 891,
        'b', 754,
        'd', 733,
        'f', 879,
        'g', 716,
        'h', 773,
        'i', 908,
        'l', 739,
        'n', 774,
        'p', 931,
        'r', 745,
        's', 839,
        't', 844,
        'u', 907,
        'y', 792,
        '{', 1091,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(26);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 27:
      if (lookahead == '-') ADVANCE(1341);
      END_STATE();
    case 28:
      if (lookahead == '-') ADVANCE(1341);
      if (lookahead == '=') ADVANCE(76);
      END_STATE();
    case 29:
      if (lookahead == '-') ADVANCE(77);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1061);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (lookahead == 0x2192) ADVANCE(1340);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 30:
      if (lookahead == '-') ADVANCE(77);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1061);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 0x2192) ADVANCE(1340);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 31:
      if (lookahead == '-') ADVANCE(77);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1063);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (lookahead == 0x2192) ADVANCE(1340);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 32:
      if (lookahead == '-') ADVANCE(77);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1063);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 0x2192) ADVANCE(1340);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 33:
      if (lookahead == '.') ADVANCE(1273);
      END_STATE();
    case 34:
      if (lookahead == '0') ADVANCE(1125);
      if (lookahead == '>') ADVANCE(1339);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      END_STATE();
    case 35:
      if (lookahead == '0') ADVANCE(1125);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      END_STATE();
    case 36:
      if (lookahead == '0') ADVANCE(1122);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(37);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1123);
      END_STATE();
    case 37:
      if (lookahead == '0') ADVANCE(1122);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1123);
      END_STATE();
    case 38:
      if (lookahead == '4') ADVANCE(81);
      END_STATE();
    case 39:
      if (lookahead == '6') ADVANCE(38);
      END_STATE();
    case 40:
      if (lookahead == ':') ADVANCE(708);
      END_STATE();
    case 41:
      if (lookahead == ':') ADVANCE(708);
      if (lookahead == '=') ADVANCE(1337);
      END_STATE();
    case 42:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '<') ADVANCE(1109);
      if (lookahead == ']') ADVANCE(706);
      if (lookahead == 'a') ADVANCE(984);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 43:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'a') ADVANCE(984);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(58);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 44:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(61);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 45:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(62);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 46:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (lookahead == 'w') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(64);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 47:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(65);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 48:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (lookahead == 'w') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 49:
      if (lookahead == ':') ADVANCE(1058);
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(67);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 50:
      if (lookahead == ':') ADVANCE(40);
      if (lookahead == ';') ADVANCE(74);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(50);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 51:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '<') ADVANCE(1109);
      if (lookahead == ']') ADVANCE(706);
      if (lookahead == 'a') ADVANCE(984);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 52:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1062);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 53:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1062);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 54:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1061);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 55:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == '@') ADVANCE(1061);
      if (lookahead == 'a') ADVANCE(992);
      if (lookahead == 'e') ADVANCE(905);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 56:
      ADVANCE_MAP(
        ';', 74,
        'N', 933,
        'T', 838,
        'a', 891,
        'b', 754,
        'd', 733,
        'f', 879,
        'g', 716,
        'h', 773,
        'i', 908,
        'l', 739,
        'n', 774,
        'o', 951,
        'p', 931,
        'r', 745,
        's', 839,
        't', 844,
        'u', 907,
        'y', 792,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 57:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == ']') ADVANCE(706);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(57);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(673);
      END_STATE();
    case 58:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'a') ADVANCE(984);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(58);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 59:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'a') ADVANCE(984);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(59);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 60:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'a') ADVANCE(984);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(60);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 61:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(61);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 62:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(62);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 63:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'i') ADVANCE(985);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(63);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 64:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (lookahead == 'w') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(64);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 65:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'p') ADVANCE(753);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(65);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 66:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (lookahead == 'w') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 67:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(67);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 68:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'e') ADVANCE(905);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(68);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 69:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'g') ADVANCE(970);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(69);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 70:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'i') ADVANCE(885);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(70);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 71:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'n') ADVANCE(941);
      if (lookahead == 'o') ADVANCE(968);
      if (lookahead == 'u') ADVANCE(900);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(71);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 72:
      if (lookahead == ';') ADVANCE(74);
      if (lookahead == 'r') ADVANCE(789);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(72);
      if (set_contains(sym_identifier_character_set_1, 152, lookahead)) ADVANCE(1057);
      END_STATE();
    case 73:
      if (lookahead == ';') ADVANCE(1372);
      if (lookahead != 0) ADVANCE(1362);
      END_STATE();
    case 74:
      if (lookahead == ';') ADVANCE(73);
      if (lookahead != 0) ADVANCE(1352);
      END_STATE();
    case 75:
      if (lookahead == '=') ADVANCE(1337);
      END_STATE();
    case 76:
      if (lookahead == '=') ADVANCE(79);
      END_STATE();
    case 77:
      if (lookahead == '>') ADVANCE(1339);
      END_STATE();
    case 78:
      if (lookahead == '>') ADVANCE(1292);
      END_STATE();
    case 79:
      if (lookahead == '>') ADVANCE(1295);
      END_STATE();
    case 80:
      if (lookahead == 'B') ADVANCE(333);
      END_STATE();
    case 81:
      if (lookahead == 'B') ADVANCE(335);
      END_STATE();
    case 82:
      if (lookahead == 'D') ADVANCE(108);
      if (lookahead == 'M') ADVANCE(466);
      if (lookahead == 'Y') ADVANCE(235);
      if (lookahead == 'r') ADVANCE(448);
      END_STATE();
    case 83:
      if (lookahead == 'D') ADVANCE(629);
      END_STATE();
    case 84:
      if (lookahead == 'D') ADVANCE(305);
      END_STATE();
    case 85:
      if (lookahead == 'D') ADVANCE(334);
      END_STATE();
    case 86:
      if (lookahead == 'D') ADVANCE(630);
      END_STATE();
    case 87:
      if (lookahead == 'E') ADVANCE(642);
      if (lookahead == 'I') ADVANCE(434);
      if (lookahead == 'L') ADVANCE(233);
      END_STATE();
    case 88:
      if (lookahead == 'E') ADVANCE(644);
      if (lookahead == 'I') ADVANCE(435);
      if (lookahead == 'L') ADVANCE(236);
      END_STATE();
    case 89:
      if (lookahead == 'I') ADVANCE(421);
      END_STATE();
    case 90:
      if (lookahead == 'I') ADVANCE(428);
      END_STATE();
    case 91:
      if (lookahead == 'I') ADVANCE(429);
      END_STATE();
    case 92:
      if (lookahead == 'I') ADVANCE(430);
      END_STATE();
    case 93:
      if (lookahead == 'M') ADVANCE(469);
      END_STATE();
    case 94:
      if (lookahead == 'N') ADVANCE(251);
      if (lookahead == 'P') ADVANCE(472);
      if (lookahead == 'u') ADVANCE(411);
      END_STATE();
    case 95:
      if (lookahead == 'S') ADVANCE(586);
      END_STATE();
    case 96:
      if (lookahead == 'T') ADVANCE(321);
      END_STATE();
    case 97:
      if (lookahead == 'T') ADVANCE(327);
      END_STATE();
    case 98:
      if (lookahead == 'U') ADVANCE(682);
      if (lookahead == 'u') ADVANCE(678);
      END_STATE();
    case 99:
      if (lookahead == '[') ADVANCE(1120);
      END_STATE();
    case 100:
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 101:
      if (lookahead == 'a') ADVANCE(533);
      if (lookahead == 'i') ADVANCE(419);
      if (lookahead == 'o') ADVANCE(446);
      if (lookahead == 'y') ADVANCE(573);
      END_STATE();
    case 102:
      if (lookahead == 'a') ADVANCE(550);
      if (lookahead == 'e') ADVANCE(147);
      if (lookahead == 'i') ADVANCE(376);
      if (lookahead == 'o') ADVANCE(610);
      if (lookahead == 'u') ADVANCE(501);
      END_STATE();
    case 103:
      if (lookahead == 'a') ADVANCE(361);
      if (lookahead == 'i') ADVANCE(643);
      if (lookahead == 'l') ADVANCE(449);
      if (lookahead == 'o') ADVANCE(488);
      if (lookahead == 'r') ADVANCE(107);
      END_STATE();
    case 104:
      if (lookahead == 'a') ADVANCE(361);
      if (lookahead == 'o') ADVANCE(520);
      if (lookahead == 'r') ADVANCE(447);
      END_STATE();
    case 105:
      if (lookahead == 'a') ADVANCE(523);
      if (lookahead == 'o') ADVANCE(542);
      if (lookahead == 'r') ADVANCE(437);
      END_STATE();
    case 106:
      if (lookahead == 'a') ADVANCE(155);
      END_STATE();
    case 107:
      if (lookahead == 'a') ADVANCE(155);
      if (lookahead == 'o') ADVANCE(367);
      END_STATE();
    case 108:
      if (lookahead == 'a') ADVANCE(645);
      END_STATE();
    case 109:
      if (lookahead == 'a') ADVANCE(485);
      if (lookahead == 't') ADVANCE(304);
      END_STATE();
    case 110:
      if (lookahead == 'a') ADVANCE(396);
      if (lookahead == 'e') ADVANCE(397);
      if (lookahead == 'o') ADVANCE(398);
      END_STATE();
    case 111:
      if (lookahead == 'a') ADVANCE(534);
      END_STATE();
    case 112:
      if (lookahead == 'a') ADVANCE(641);
      if (lookahead == 'e') ADVANCE(576);
      if (lookahead == 'i') ADVANCE(427);
      if (lookahead == 'o') ADVANCE(159);
      END_STATE();
    case 113:
      if (lookahead == 'a') ADVANCE(641);
      if (lookahead == 'e') ADVANCE(576);
      if (lookahead == 'i') ADVANCE(427);
      if (lookahead == 'o') ADVANCE(172);
      END_STATE();
    case 114:
      if (lookahead == 'a') ADVANCE(605);
      if (lookahead == 'd') ADVANCE(257);
      if (lookahead == 'e') ADVANCE(115);
      END_STATE();
    case 115:
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'f') ADVANCE(1252);
      if (lookahead == 'q') ADVANCE(616);
      END_STATE();
    case 116:
      if (lookahead == 'a') ADVANCE(567);
      END_STATE();
    case 117:
      if (lookahead == 'a') ADVANCE(567);
      if (lookahead == 'e') ADVANCE(1150);
      END_STATE();
    case 118:
      if (lookahead == 'a') ADVANCE(346);
      END_STATE();
    case 119:
      if (lookahead == 'a') ADVANCE(650);
      END_STATE();
    case 120:
      if (lookahead == 'a') ADVANCE(347);
      END_STATE();
    case 121:
      if (lookahead == 'a') ADVANCE(269);
      END_STATE();
    case 122:
      if (lookahead == 'a') ADVANCE(491);
      END_STATE();
    case 123:
      if (lookahead == 'a') ADVANCE(348);
      END_STATE();
    case 124:
      if (lookahead == 'a') ADVANCE(553);
      END_STATE();
    case 125:
      if (lookahead == 'a') ADVANCE(498);
      END_STATE();
    case 126:
      if (lookahead == 'a') ADVANCE(349);
      END_STATE();
    case 127:
      if (lookahead == 'a') ADVANCE(492);
      END_STATE();
    case 128:
      if (lookahead == 'a') ADVANCE(357);
      END_STATE();
    case 129:
      if (lookahead == 'a') ADVANCE(503);
      END_STATE();
    case 130:
      if (lookahead == 'a') ADVANCE(504);
      END_STATE();
    case 131:
      if (lookahead == 'a') ADVANCE(387);
      END_STATE();
    case 132:
      if (lookahead == 'a') ADVANCE(593);
      if (lookahead == 'e') ADVANCE(255);
      if (lookahead == 'i') ADVANCE(376);
      END_STATE();
    case 133:
      if (lookahead == 'a') ADVANCE(593);
      if (lookahead == 'i') ADVANCE(376);
      END_STATE();
    case 134:
      if (lookahead == 'a') ADVANCE(574);
      if (lookahead == 'r') ADVANCE(451);
      END_STATE();
    case 135:
      if (lookahead == 'a') ADVANCE(574);
      if (lookahead == 'r') ADVANCE(440);
      END_STATE();
    case 136:
      if (lookahead == 'a') ADVANCE(369);
      END_STATE();
    case 137:
      if (lookahead == 'a') ADVANCE(597);
      END_STATE();
    case 138:
      if (lookahead == 'a') ADVANCE(359);
      END_STATE();
    case 139:
      if (lookahead == 'a') ADVANCE(585);
      END_STATE();
    case 140:
      if (lookahead == 'a') ADVANCE(602);
      END_STATE();
    case 141:
      if (lookahead == 'a') ADVANCE(603);
      END_STATE();
    case 142:
      if (lookahead == 'a') ADVANCE(607);
      END_STATE();
    case 143:
      if (lookahead == 'b') ADVANCE(358);
      END_STATE();
    case 144:
      if (lookahead == 'b') ADVANCE(537);
      if (lookahead == 'p') ADVANCE(540);
      END_STATE();
    case 145:
      if (lookahead == 'b') ADVANCE(316);
      END_STATE();
    case 146:
      if (lookahead == 'c') ADVANCE(1248);
      END_STATE();
    case 147:
      if (lookahead == 'c') ADVANCE(319);
      if (lookahead == 'f') ADVANCE(1090);
      END_STATE();
    case 148:
      if (lookahead == 'c') ADVANCE(353);
      END_STATE();
    case 149:
      if (lookahead == 'c') ADVANCE(188);
      END_STATE();
    case 150:
      if (lookahead == 'c') ADVANCE(592);
      END_STATE();
    case 151:
      if (lookahead == 'c') ADVANCE(559);
      END_STATE();
    case 152:
      if (lookahead == 'c') ADVANCE(581);
      END_STATE();
    case 153:
      if (lookahead == 'c') ADVANCE(311);
      END_STATE();
    case 154:
      if (lookahead == 'c') ADVANCE(363);
      END_STATE();
    case 155:
      if (lookahead == 'c') ADVANCE(599);
      END_STATE();
    case 156:
      if (lookahead == 'c') ADVANCE(364);
      END_STATE();
    case 157:
      if (lookahead == 'c') ADVANCE(365);
      END_STATE();
    case 158:
      if (lookahead == 'd') ADVANCE(701);
      if (lookahead == 't') ADVANCE(303);
      if (lookahead == 'u') ADVANCE(366);
      END_STATE();
    case 159:
      if (lookahead == 'd') ADVANCE(1329);
      END_STATE();
    case 160:
      if (lookahead == 'd') ADVANCE(1239);
      END_STATE();
    case 161:
      if (lookahead == 'd') ADVANCE(1235);
      END_STATE();
    case 162:
      if (lookahead == 'd') ADVANCE(1201);
      END_STATE();
    case 163:
      if (lookahead == 'd') ADVANCE(1267);
      END_STATE();
    case 164:
      if (lookahead == 'd') ADVANCE(95);
      END_STATE();
    case 165:
      if (lookahead == 'd') ADVANCE(1236);
      END_STATE();
    case 166:
      if (lookahead == 'd') ADVANCE(1285);
      END_STATE();
    case 167:
      if (lookahead == 'd') ADVANCE(1265);
      END_STATE();
    case 168:
      if (lookahead == 'd') ADVANCE(257);
      END_STATE();
    case 169:
      if (lookahead == 'd') ADVANCE(257);
      if (lookahead == 'e') ADVANCE(258);
      END_STATE();
    case 170:
      if (lookahead == 'd') ADVANCE(623);
      if (lookahead == 'h') ADVANCE(295);
      if (lookahead == 'p') ADVANCE(237);
      END_STATE();
    case 171:
      if (lookahead == 'd') ADVANCE(623);
      if (lookahead == 'p') ADVANCE(237);
      END_STATE();
    case 172:
      if (lookahead == 'd') ADVANCE(624);
      END_STATE();
    case 173:
      if (lookahead == 'd') ADVANCE(234);
      END_STATE();
    case 174:
      if (lookahead == 'd') ADVANCE(241);
      END_STATE();
    case 175:
      if (lookahead == 'd') ADVANCE(238);
      if (lookahead == 'f') ADVANCE(256);
      if (lookahead == 'm') ADVANCE(475);
      if (lookahead == 'n') ADVANCE(1321);
      if (lookahead == 'r') ADVANCE(290);
      if (lookahead == 's') ADVANCE(699);
      END_STATE();
    case 176:
      if (lookahead == 'd') ADVANCE(238);
      if (lookahead == 'f') ADVANCE(256);
      if (lookahead == 'm') ADVANCE(475);
      if (lookahead == 'n') ADVANCE(1320);
      if (lookahead == 's') ADVANCE(699);
      END_STATE();
    case 177:
      if (lookahead == 'd') ADVANCE(238);
      if (lookahead == 'm') ADVANCE(481);
      if (lookahead == 'n') ADVANCE(1321);
      if (lookahead == 's') ADVANCE(699);
      END_STATE();
    case 178:
      if (lookahead == 'd') ADVANCE(238);
      if (lookahead == 'n') ADVANCE(1321);
      if (lookahead == 's') ADVANCE(699);
      END_STATE();
    case 179:
      if (lookahead == 'e') ADVANCE(278);
      if (lookahead == 'o') ADVANCE(384);
      END_STATE();
    case 180:
      if (lookahead == 'e') ADVANCE(39);
      END_STATE();
    case 181:
      if (lookahead == 'e') ADVANCE(1190);
      END_STATE();
    case 182:
      if (lookahead == 'e') ADVANCE(1176);
      END_STATE();
    case 183:
      if (lookahead == 'e') ADVANCE(1130);
      END_STATE();
    case 184:
      if (lookahead == 'e') ADVANCE(1257);
      END_STATE();
    case 185:
      if (lookahead == 'e') ADVANCE(1133);
      END_STATE();
    case 186:
      if (lookahead == 'e') ADVANCE(1156);
      END_STATE();
    case 187:
      if (lookahead == 'e') ADVANCE(1223);
      END_STATE();
    case 188:
      if (lookahead == 'e') ADVANCE(1244);
      END_STATE();
    case 189:
      if (lookahead == 'e') ADVANCE(1269);
      END_STATE();
    case 190:
      if (lookahead == 'e') ADVANCE(1222);
      END_STATE();
    case 191:
      if (lookahead == 'e') ADVANCE(1153);
      END_STATE();
    case 192:
      if (lookahead == 'e') ADVANCE(1214);
      END_STATE();
    case 193:
      if (lookahead == 'e') ADVANCE(89);
      END_STATE();
    case 194:
      if (lookahead == 'e') ADVANCE(1271);
      END_STATE();
    case 195:
      if (lookahead == 'e') ADVANCE(1258);
      END_STATE();
    case 196:
      if (lookahead == 'e') ADVANCE(1230);
      END_STATE();
    case 197:
      if (lookahead == 'e') ADVANCE(1232);
      END_STATE();
    case 198:
      if (lookahead == 'e') ADVANCE(1231);
      END_STATE();
    case 199:
      if (lookahead == 'e') ADVANCE(1233);
      END_STATE();
    case 200:
      if (lookahead == 'e') ADVANCE(1204);
      END_STATE();
    case 201:
      if (lookahead == 'e') ADVANCE(1234);
      END_STATE();
    case 202:
      if (lookahead == 'e') ADVANCE(697);
      END_STATE();
    case 203:
      if (lookahead == 'e') ADVANCE(351);
      if (lookahead == 'h') ADVANCE(452);
      if (lookahead == 'o') ADVANCE(612);
      if (lookahead == 't') ADVANCE(489);
      if (lookahead == 'u') ADVANCE(144);
      END_STATE();
    case 204:
      if (lookahead == 'e') ADVANCE(640);
      END_STATE();
    case 205:
      if (lookahead == 'e') ADVANCE(160);
      END_STATE();
    case 206:
      if (lookahead == 'e') ADVANCE(83);
      END_STATE();
    case 207:
      if (lookahead == 'e') ADVANCE(656);
      END_STATE();
    case 208:
      if (lookahead == 'e') ADVANCE(576);
      END_STATE();
    case 209:
      if (lookahead == 'e') ADVANCE(576);
      if (lookahead == 'o') ADVANCE(159);
      END_STATE();
    case 210:
      if (lookahead == 'e') ADVANCE(122);
      END_STATE();
    case 211:
      if (lookahead == 'e') ADVANCE(397);
      END_STATE();
    case 212:
      if (lookahead == 'e') ADVANCE(502);
      END_STATE();
    case 213:
      if (lookahead == 'e') ADVANCE(161);
      END_STATE();
    case 214:
      if (lookahead == 'e') ADVANCE(162);
      END_STATE();
    case 215:
      if (lookahead == 'e') ADVANCE(163);
      END_STATE();
    case 216:
      if (lookahead == 'e') ADVANCE(164);
      END_STATE();
    case 217:
      if (lookahead == 'e') ADVANCE(402);
      END_STATE();
    case 218:
      if (lookahead == 'e') ADVANCE(528);
      END_STATE();
    case 219:
      if (lookahead == 'e') ADVANCE(131);
      END_STATE();
    case 220:
      if (lookahead == 'e') ADVANCE(165);
      END_STATE();
    case 221:
      if (lookahead == 'e') ADVANCE(433);
      END_STATE();
    case 222:
      if (lookahead == 'e') ADVANCE(167);
      END_STATE();
    case 223:
      if (lookahead == 'e') ADVANCE(272);
      END_STATE();
    case 224:
      if (lookahead == 'e') ADVANCE(519);
      END_STATE();
    case 225:
      if (lookahead == 'e') ADVANCE(385);
      END_STATE();
    case 226:
      if (lookahead == 'e') ADVANCE(510);
      END_STATE();
    case 227:
      if (lookahead == 'e') ADVANCE(557);
      END_STATE();
    case 228:
      if (lookahead == 'e') ADVANCE(493);
      END_STATE();
    case 229:
      if (lookahead == 'e') ADVANCE(558);
      END_STATE();
    case 230:
      if (lookahead == 'e') ADVANCE(494);
      END_STATE();
    case 231:
      if (lookahead == 'e') ADVANCE(495);
      END_STATE();
    case 232:
      if (lookahead == 'e') ADVANCE(496);
      END_STATE();
    case 233:
      if (lookahead == 'e') ADVANCE(420);
      END_STATE();
    case 234:
      if (lookahead == 'e') ADVANCE(513);
      END_STATE();
    case 235:
      if (lookahead == 'e') ADVANCE(127);
      END_STATE();
    case 236:
      if (lookahead == 'e') ADVANCE(422);
      END_STATE();
    case 237:
      if (lookahead == 'e') ADVANCE(512);
      END_STATE();
    case 238:
      if (lookahead == 'e') ADVANCE(409);
      END_STATE();
    case 239:
      if (lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 240:
      if (lookahead == 'e') ADVANCE(412);
      END_STATE();
    case 241:
      if (lookahead == 'e') ADVANCE(517);
      END_STATE();
    case 242:
      if (lookahead == 'e') ADVANCE(516);
      END_STATE();
    case 243:
      if (lookahead == 'e') ADVANCE(416);
      END_STATE();
    case 244:
      if (lookahead == 'e') ADVANCE(378);
      END_STATE();
    case 245:
      if (lookahead == 'e') ADVANCE(274);
      END_STATE();
    case 246:
      if (lookahead == 'e') ADVANCE(276);
      END_STATE();
    case 247:
      if (lookahead == 'e') ADVANCE(277);
      END_STATE();
    case 248:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 249:
      if (lookahead == 'e') ADVANCE(91);
      END_STATE();
    case 250:
      if (lookahead == 'e') ADVANCE(92);
      END_STATE();
    case 251:
      if (lookahead == 'e') ADVANCE(279);
      END_STATE();
    case 252:
      if (lookahead == 'f') ADVANCE(256);
      if (lookahead == 'm') ADVANCE(480);
      if (lookahead == 'n') ADVANCE(1321);
      if (lookahead == 's') ADVANCE(699);
      END_STATE();
    case 253:
      if (lookahead == 'f') ADVANCE(1247);
      if (lookahead == 'n') ADVANCE(1251);
      if (lookahead == 'p') ADVANCE(109);
      if (lookahead == 'r') ADVANCE(1287);
      END_STATE();
    case 254:
      if (lookahead == 'f') ADVANCE(1247);
      if (lookahead == 'r') ADVANCE(1287);
      END_STATE();
    case 255:
      if (lookahead == 'f') ADVANCE(1090);
      END_STATE();
    case 256:
      if (lookahead == 'f') ADVANCE(1294);
      END_STATE();
    case 257:
      if (lookahead == 'f') ADVANCE(1255);
      END_STATE();
    case 258:
      if (lookahead == 'f') ADVANCE(1252);
      END_STATE();
    case 259:
      if (lookahead == 'f') ADVANCE(1082);
      END_STATE();
    case 260:
      if (lookahead == 'g') ADVANCE(1185);
      END_STATE();
    case 261:
      if (lookahead == 'g') ADVANCE(1137);
      END_STATE();
    case 262:
      if (lookahead == 'g') ADVANCE(1174);
      END_STATE();
    case 263:
      if (lookahead == 'g') ADVANCE(1139);
      END_STATE();
    case 264:
      if (lookahead == 'g') ADVANCE(1196);
      END_STATE();
    case 265:
      if (lookahead == 'g') ADVANCE(1210);
      END_STATE();
    case 266:
      if (lookahead == 'g') ADVANCE(614);
      END_STATE();
    case 267:
      if (lookahead == 'g') ADVANCE(568);
      END_STATE();
    case 268:
      if (lookahead == 'g') ADVANCE(413);
      END_STATE();
    case 269:
      if (lookahead == 'g') ADVANCE(192);
      END_STATE();
    case 270:
      if (lookahead == 'g') ADVANCE(570);
      END_STATE();
    case 271:
      if (lookahead == 'g') ADVANCE(572);
      END_STATE();
    case 272:
      if (lookahead == 'g') ADVANCE(228);
      END_STATE();
    case 273:
      if (lookahead == 'g') ADVANCE(313);
      END_STATE();
    case 274:
      if (lookahead == 'g') ADVANCE(230);
      END_STATE();
    case 275:
      if (lookahead == 'g') ADVANCE(314);
      END_STATE();
    case 276:
      if (lookahead == 'g') ADVANCE(231);
      END_STATE();
    case 277:
      if (lookahead == 'g') ADVANCE(232);
      END_STATE();
    case 278:
      if (lookahead == 'g') ADVANCE(139);
      END_STATE();
    case 279:
      if (lookahead == 'g') ADVANCE(142);
      END_STATE();
    case 280:
      if (lookahead == 'h') ADVANCE(1088);
      END_STATE();
    case 281:
      if (lookahead == 'h') ADVANCE(1164);
      END_STATE();
    case 282:
      if (lookahead == 'h') ADVANCE(1225);
      END_STATE();
    case 283:
      if (lookahead == 'h') ADVANCE(1226);
      END_STATE();
    case 284:
      if (lookahead == 'h') ADVANCE(1227);
      END_STATE();
    case 285:
      if (lookahead == 'h') ADVANCE(1168);
      END_STATE();
    case 286:
      if (lookahead == 'h') ADVANCE(302);
      END_STATE();
    case 287:
      if (lookahead == 'h') ADVANCE(309);
      END_STATE();
    case 288:
      if (lookahead == 'h') ADVANCE(468);
      END_STATE();
    case 289:
      if (lookahead == 'h') ADVANCE(86);
      END_STATE();
    case 290:
      if (lookahead == 'i') ADVANCE(1218);
      END_STATE();
    case 291:
      if (lookahead == 'i') ADVANCE(454);
      if (lookahead == 'k') ADVANCE(405);
      if (lookahead == 'o') ADVANCE(500);
      if (lookahead == 's') ADVANCE(301);
      END_STATE();
    case 292:
      if (lookahead == 'i') ADVANCE(454);
      if (lookahead == 'o') ADVANCE(500);
      END_STATE();
    case 293:
      if (lookahead == 'i') ADVANCE(631);
      END_STATE();
    case 294:
      if (lookahead == 'i') ADVANCE(655);
      END_STATE();
    case 295:
      if (lookahead == 'i') ADVANCE(145);
      END_STATE();
    case 296:
      if (lookahead == 'i') ADVANCE(153);
      END_STATE();
    case 297:
      if (lookahead == 'i') ADVANCE(535);
      END_STATE();
    case 298:
      if (lookahead == 'i') ADVANCE(535);
      if (lookahead == 'p') ADVANCE(352);
      END_STATE();
    case 299:
      if (lookahead == 'i') ADVANCE(565);
      END_STATE();
    case 300:
      if (lookahead == 'i') ADVANCE(146);
      END_STATE();
    case 301:
      if (lookahead == 'i') ADVANCE(268);
      END_STATE();
    case 302:
      if (lookahead == 'i') ADVANCE(399);
      END_STATE();
    case 303:
      if (lookahead == 'i') ADVANCE(569);
      END_STATE();
    case 304:
      if (lookahead == 'i') ADVANCE(464);
      END_STATE();
    case 305:
      if (lookahead == 'i') ADVANCE(273);
      END_STATE();
    case 306:
      if (lookahead == 'i') ADVANCE(400);
      if (lookahead == 'u') ADVANCE(150);
      END_STATE();
    case 307:
      if (lookahead == 'i') ADVANCE(511);
      END_STATE();
    case 308:
      if (lookahead == 'i') ADVANCE(218);
      END_STATE();
    case 309:
      if (lookahead == 'i') ADVANCE(401);
      END_STATE();
    case 310:
      if (lookahead == 'i') ADVANCE(571);
      END_STATE();
    case 311:
      if (lookahead == 'i') ADVANCE(604);
      END_STATE();
    case 312:
      if (lookahead == 'i') ADVANCE(410);
      END_STATE();
    case 313:
      if (lookahead == 'i') ADVANCE(582);
      END_STATE();
    case 314:
      if (lookahead == 'i') ADVANCE(583);
      END_STATE();
    case 315:
      if (lookahead == 'i') ADVANCE(632);
      END_STATE();
    case 316:
      if (lookahead == 'i') ADVANCE(589);
      END_STATE();
    case 317:
      if (lookahead == 'i') ADVANCE(372);
      if (lookahead == 'o') ADVANCE(342);
      if (lookahead == 'r') ADVANCE(613);
      if (lookahead == 'y') ADVANCE(478);
      END_STATE();
    case 318:
      if (lookahead == 'i') ADVANCE(455);
      END_STATE();
    case 319:
      if (lookahead == 'i') ADVANCE(377);
      END_STATE();
    case 320:
      if (lookahead == 'i') ADVANCE(633);
      END_STATE();
    case 321:
      if (lookahead == 'i') ADVANCE(373);
      END_STATE();
    case 322:
      if (lookahead == 'i') ADVANCE(456);
      END_STATE();
    case 323:
      if (lookahead == 'i') ADVANCE(634);
      END_STATE();
    case 324:
      if (lookahead == 'i') ADVANCE(374);
      END_STATE();
    case 325:
      if (lookahead == 'i') ADVANCE(457);
      END_STATE();
    case 326:
      if (lookahead == 'i') ADVANCE(635);
      END_STATE();
    case 327:
      if (lookahead == 'i') ADVANCE(375);
      END_STATE();
    case 328:
      if (lookahead == 'i') ADVANCE(458);
      END_STATE();
    case 329:
      if (lookahead == 'i') ADVANCE(460);
      END_STATE();
    case 330:
      if (lookahead == 'i') ADVANCE(461);
      END_STATE();
    case 331:
      if (lookahead == 'i') ADVANCE(453);
      END_STATE();
    case 332:
      if (lookahead == 'i') ADVANCE(486);
      END_STATE();
    case 333:
      if (lookahead == 'i') ADVANCE(424);
      END_STATE();
    case 334:
      if (lookahead == 'i') ADVANCE(275);
      END_STATE();
    case 335:
      if (lookahead == 'i') ADVANCE(426);
      END_STATE();
    case 336:
      if (lookahead == 'i') ADVANCE(471);
      END_STATE();
    case 337:
      if (lookahead == 'i') ADVANCE(636);
      END_STATE();
    case 338:
      if (lookahead == 'i') ADVANCE(606);
      END_STATE();
    case 339:
      if (lookahead == 'i') ADVANCE(637);
      END_STATE();
    case 340:
      if (lookahead == 'i') ADVANCE(638);
      END_STATE();
    case 341:
      if (lookahead == 'i') ADVANCE(608);
      END_STATE();
    case 342:
      if (lookahead == 'k') ADVANCE(225);
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 343:
      if (lookahead == 'l') ADVANCE(111);
      END_STATE();
    case 344:
      if (lookahead == 'l') ADVANCE(111);
      if (lookahead == 'o') ADVANCE(368);
      END_STATE();
    case 345:
      if (lookahead == 'l') ADVANCE(1141);
      END_STATE();
    case 346:
      if (lookahead == 'l') ADVANCE(84);
      END_STATE();
    case 347:
      if (lookahead == 'l') ADVANCE(1154);
      END_STATE();
    case 348:
      if (lookahead == 'l') ADVANCE(1237);
      END_STATE();
    case 349:
      if (lookahead == 'l') ADVANCE(1143);
      END_STATE();
    case 350:
      if (lookahead == 'l') ADVANCE(1297);
      END_STATE();
    case 351:
      if (lookahead == 'l') ADVANCE(259);
      END_STATE();
    case 352:
      if (lookahead == 'l') ADVANCE(296);
      END_STATE();
    case 353:
      if (lookahead == 'l') ADVANCE(622);
      END_STATE();
    case 354:
      if (lookahead == 'l') ADVANCE(308);
      END_STATE();
    case 355:
      if (lookahead == 'l') ADVANCE(308);
      if (lookahead == 'o') ADVANCE(509);
      END_STATE();
    case 356:
      if (lookahead == 'l') ADVANCE(244);
      END_STATE();
    case 357:
      if (lookahead == 'l') ADVANCE(350);
      END_STATE();
    case 358:
      if (lookahead == 'l') ADVANCE(186);
      END_STATE();
    case 359:
      if (lookahead == 'l') ADVANCE(294);
      END_STATE();
    case 360:
      if (lookahead == 'l') ADVANCE(202);
      END_STATE();
    case 361:
      if (lookahead == 'l') ADVANCE(536);
      END_STATE();
    case 362:
      if (lookahead == 'l') ADVANCE(219);
      END_STATE();
    case 363:
      if (lookahead == 'l') ADVANCE(626);
      END_STATE();
    case 364:
      if (lookahead == 'l') ADVANCE(627);
      END_STATE();
    case 365:
      if (lookahead == 'l') ADVANCE(628);
      END_STATE();
    case 366:
      if (lookahead == 'm') ADVANCE(1246);
      END_STATE();
    case 367:
      if (lookahead == 'm') ADVANCE(707);
      END_STATE();
    case 368:
      if (lookahead == 'm') ADVANCE(477);
      END_STATE();
    case 369:
      if (lookahead == 'm') ADVANCE(474);
      END_STATE();
    case 370:
      if (lookahead == 'm') ADVANCE(138);
      END_STATE();
    case 371:
      if (lookahead == 'm') ADVANCE(481);
      if (lookahead == 's') ADVANCE(699);
      END_STATE();
    case 372:
      if (lookahead == 'm') ADVANCE(182);
      END_STATE();
    case 373:
      if (lookahead == 'm') ADVANCE(206);
      END_STATE();
    case 374:
      if (lookahead == 'm') ADVANCE(191);
      END_STATE();
    case 375:
      if (lookahead == 'm') ADVANCE(207);
      END_STATE();
    case 376:
      if (lookahead == 'm') ADVANCE(221);
      END_STATE();
    case 377:
      if (lookahead == 'm') ADVANCE(120);
      END_STATE();
    case 378:
      if (lookahead == 'm') ADVANCE(243);
      END_STATE();
    case 379:
      if (lookahead == 'n') ADVANCE(158);
      if (lookahead == 'v') ADVANCE(217);
      END_STATE();
    case 380:
      if (lookahead == 'n') ADVANCE(158);
      if (lookahead == 'v') ADVANCE(217);
      if (lookahead == 'x') ADVANCE(298);
      END_STATE();
    case 381:
      if (lookahead == 'n') ADVANCE(158);
      if (lookahead == 'v') ADVANCE(217);
      if (lookahead == 'x') ADVANCE(297);
      END_STATE();
    case 382:
      if (lookahead == 'n') ADVANCE(158);
      if (lookahead == 'v') ADVANCE(217);
      if (lookahead == 'x') ADVANCE(476);
      END_STATE();
    case 383:
      if (lookahead == 'n') ADVANCE(291);
      END_STATE();
    case 384:
      if (lookahead == 'n') ADVANCE(94);
      if (lookahead == 'r') ADVANCE(370);
      if (lookahead == 't') ADVANCE(1282);
      END_STATE();
    case 385:
      if (lookahead == 'n') ADVANCE(1212);
      END_STATE();
    case 386:
      if (lookahead == 'n') ADVANCE(1261);
      END_STATE();
    case 387:
      if (lookahead == 'n') ADVANCE(1148);
      END_STATE();
    case 388:
      if (lookahead == 'n') ADVANCE(1238);
      END_STATE();
    case 389:
      if (lookahead == 'n') ADVANCE(1263);
      END_STATE();
    case 390:
      if (lookahead == 'n') ADVANCE(698);
      END_STATE();
    case 391:
      if (lookahead == 'n') ADVANCE(1158);
      END_STATE();
    case 392:
      if (lookahead == 'n') ADVANCE(1241);
      END_STATE();
    case 393:
      if (lookahead == 'n') ADVANCE(1180);
      END_STATE();
    case 394:
      if (lookahead == 'n') ADVANCE(1182);
      END_STATE();
    case 395:
      if (lookahead == 'n') ADVANCE(292);
      END_STATE();
    case 396:
      if (lookahead == 'n') ADVANCE(266);
      END_STATE();
    case 397:
      if (lookahead == 'n') ADVANCE(267);
      END_STATE();
    case 398:
      if (lookahead == 'n') ADVANCE(260);
      END_STATE();
    case 399:
      if (lookahead == 'n') ADVANCE(261);
      END_STATE();
    case 400:
      if (lookahead == 'n') ADVANCE(262);
      END_STATE();
    case 401:
      if (lookahead == 'n') ADVANCE(263);
      END_STATE();
    case 402:
      if (lookahead == 'n') ADVANCE(552);
      END_STATE();
    case 403:
      if (lookahead == 'n') ADVANCE(166);
      if (lookahead == 's') ADVANCE(711);
      END_STATE();
    case 404:
      if (lookahead == 'n') ADVANCE(166);
      if (lookahead == 's') ADVANCE(709);
      END_STATE();
    case 405:
      if (lookahead == 'n') ADVANCE(438);
      END_STATE();
    case 406:
      if (lookahead == 'n') ADVANCE(264);
      END_STATE();
    case 407:
      if (lookahead == 'n') ADVANCE(566);
      END_STATE();
    case 408:
      if (lookahead == 'n') ADVANCE(615);
      if (lookahead == 't') ADVANCE(1282);
      END_STATE();
    case 409:
      if (lookahead == 'n') ADVANCE(595);
      END_STATE();
    case 410:
      if (lookahead == 'n') ADVANCE(265);
      END_STATE();
    case 411:
      if (lookahead == 'n') ADVANCE(332);
      END_STATE();
    case 412:
      if (lookahead == 'n') ADVANCE(556);
      END_STATE();
    case 413:
      if (lookahead == 'n') ADVANCE(214);
      END_STATE();
    case 414:
      if (lookahead == 'n') ADVANCE(578);
      END_STATE();
    case 415:
      if (lookahead == 'n') ADVANCE(580);
      END_STATE();
    case 416:
      if (lookahead == 'n') ADVANCE(561);
      END_STATE();
    case 417:
      if (lookahead == 'n') ADVANCE(562);
      END_STATE();
    case 418:
      if (lookahead == 'n') ADVANCE(201);
      END_STATE();
    case 419:
      if (lookahead == 'n') ADVANCE(125);
      END_STATE();
    case 420:
      if (lookahead == 'n') ADVANCE(270);
      END_STATE();
    case 421:
      if (lookahead == 'n') ADVANCE(590);
      END_STATE();
    case 422:
      if (lookahead == 'n') ADVANCE(271);
      END_STATE();
    case 423:
      if (lookahead == 'n') ADVANCE(123);
      END_STATE();
    case 424:
      if (lookahead == 'n') ADVANCE(129);
      END_STATE();
    case 425:
      if (lookahead == 'n') ADVANCE(126);
      END_STATE();
    case 426:
      if (lookahead == 'n') ADVANCE(130);
      END_STATE();
    case 427:
      if (lookahead == 'n') ADVANCE(88);
      END_STATE();
    case 428:
      if (lookahead == 'n') ADVANCE(598);
      END_STATE();
    case 429:
      if (lookahead == 'n') ADVANCE(600);
      END_STATE();
    case 430:
      if (lookahead == 'n') ADVANCE(601);
      END_STATE();
    case 431:
      if (lookahead == 'n') ADVANCE(331);
      END_STATE();
    case 432:
      if (lookahead == 'n') ADVANCE(85);
      END_STATE();
    case 433:
      if (lookahead == 'n') ADVANCE(548);
      END_STATE();
    case 434:
      if (lookahead == 'n') ADVANCE(154);
      END_STATE();
    case 435:
      if (lookahead == 'n') ADVANCE(157);
      END_STATE();
    case 436:
      if (lookahead == 'o') ADVANCE(596);
      END_STATE();
    case 437:
      if (lookahead == 'o') ADVANCE(170);
      END_STATE();
    case 438:
      if (lookahead == 'o') ADVANCE(639);
      END_STATE();
    case 439:
      if (lookahead == 'o') ADVANCE(408);
      END_STATE();
    case 440:
      if (lookahead == 'o') ADVANCE(171);
      END_STATE();
    case 441:
      if (lookahead == 'o') ADVANCE(612);
      if (lookahead == 't') ADVANCE(522);
      END_STATE();
    case 442:
      if (lookahead == 'o') ADVANCE(612);
      if (lookahead == 't') ADVANCE(522);
      if (lookahead == 'u') ADVANCE(144);
      END_STATE();
    case 443:
      if (lookahead == 'o') ADVANCE(488);
      END_STATE();
    case 444:
      if (lookahead == 'o') ADVANCE(488);
      if (lookahead == 'r') ADVANCE(107);
      END_STATE();
    case 445:
      if (lookahead == 'o') ADVANCE(488);
      if (lookahead == 'r') ADVANCE(106);
      END_STATE();
    case 446:
      if (lookahead == 'o') ADVANCE(362);
      END_STATE();
    case 447:
      if (lookahead == 'o') ADVANCE(367);
      END_STATE();
    case 448:
      if (lookahead == 'o') ADVANCE(611);
      END_STATE();
    case 449:
      if (lookahead == 'o') ADVANCE(124);
      END_STATE();
    case 450:
      if (lookahead == 'o') ADVANCE(490);
      END_STATE();
    case 451:
      if (lookahead == 'o') ADVANCE(482);
      END_STATE();
    case 452:
      if (lookahead == 'o') ADVANCE(508);
      END_STATE();
    case 453:
      if (lookahead == 'o') ADVANCE(386);
      END_STATE();
    case 454:
      if (lookahead == 'o') ADVANCE(386);
      if (lookahead == 'q') ADVANCE(618);
      END_STATE();
    case 455:
      if (lookahead == 'o') ADVANCE(390);
      END_STATE();
    case 456:
      if (lookahead == 'o') ADVANCE(391);
      END_STATE();
    case 457:
      if (lookahead == 'o') ADVANCE(432);
      END_STATE();
    case 458:
      if (lookahead == 'o') ADVANCE(392);
      END_STATE();
    case 459:
      if (lookahead == 'o') ADVANCE(406);
      END_STATE();
    case 460:
      if (lookahead == 'o') ADVANCE(393);
      END_STATE();
    case 461:
      if (lookahead == 'o') ADVANCE(394);
      END_STATE();
    case 462:
      if (lookahead == 'o') ADVANCE(551);
      END_STATE();
    case 463:
      if (lookahead == 'o') ADVANCE(575);
      END_STATE();
    case 464:
      if (lookahead == 'o') ADVANCE(423);
      END_STATE();
    case 465:
      if (lookahead == 'o') ADVANCE(509);
      END_STATE();
    case 466:
      if (lookahead == 'o') ADVANCE(407);
      END_STATE();
    case 467:
      if (lookahead == 'o') ADVANCE(418);
      END_STATE();
    case 468:
      if (lookahead == 'o') ADVANCE(515);
      END_STATE();
    case 469:
      if (lookahead == 'o') ADVANCE(414);
      END_STATE();
    case 470:
      if (lookahead == 'o') ADVANCE(415);
      END_STATE();
    case 471:
      if (lookahead == 'o') ADVANCE(425);
      END_STATE();
    case 472:
      if (lookahead == 'o') ADVANCE(549);
      END_STATE();
    case 473:
      if (lookahead == 'p') ADVANCE(1249);
      END_STATE();
    case 474:
      if (lookahead == 'p') ADVANCE(1178);
      END_STATE();
    case 475:
      if (lookahead == 'p') ADVANCE(355);
      END_STATE();
    case 476:
      if (lookahead == 'p') ADVANCE(352);
      END_STATE();
    case 477:
      if (lookahead == 'p') ADVANCE(356);
      END_STATE();
    case 478:
      if (lookahead == 'p') ADVANCE(184);
      END_STATE();
    case 479:
      if (lookahead == 'p') ADVANCE(190);
      END_STATE();
    case 480:
      if (lookahead == 'p') ADVANCE(354);
      END_STATE();
    case 481:
      if (lookahead == 'p') ADVANCE(465);
      END_STATE();
    case 482:
      if (lookahead == 'p') ADVANCE(237);
      END_STATE();
    case 483:
      if (lookahead == 'q') ADVANCE(1310);
      END_STATE();
    case 484:
      if (lookahead == 'q') ADVANCE(1314);
      END_STATE();
    case 485:
      if (lookahead == 'q') ADVANCE(617);
      END_STATE();
    case 486:
      if (lookahead == 'q') ADVANCE(620);
      END_STATE();
    case 487:
      if (lookahead == 'r') ADVANCE(1287);
      END_STATE();
    case 488:
      if (lookahead == 'r') ADVANCE(1089);
      END_STATE();
    case 489:
      if (lookahead == 'r') ADVANCE(306);
      END_STATE();
    case 490:
      if (lookahead == 'r') ADVANCE(1289);
      END_STATE();
    case 491:
      if (lookahead == 'r') ADVANCE(93);
      END_STATE();
    case 492:
      if (lookahead == 'r') ADVANCE(1170);
      END_STATE();
    case 493:
      if (lookahead == 'r') ADVANCE(1208);
      END_STATE();
    case 494:
      if (lookahead == 'r') ADVANCE(1194);
      END_STATE();
    case 495:
      if (lookahead == 'r') ADVANCE(1192);
      END_STATE();
    case 496:
      if (lookahead == 'r') ADVANCE(1206);
      END_STATE();
    case 497:
      if (lookahead == 'r') ADVANCE(149);
      END_STATE();
    case 498:
      if (lookahead == 'r') ADVANCE(646);
      END_STATE();
    case 499:
      if (lookahead == 'r') ADVANCE(613);
      END_STATE();
    case 500:
      if (lookahead == 'r') ADVANCE(173);
      END_STATE();
    case 501:
      if (lookahead == 'r') ADVANCE(137);
      END_STATE();
    case 502:
      if (lookahead == 'r') ADVANCE(538);
      END_STATE();
    case 503:
      if (lookahead == 'r') ADVANCE(651);
      END_STATE();
    case 504:
      if (lookahead == 'r') ADVANCE(652);
      END_STATE();
    case 505:
      if (lookahead == 'r') ADVANCE(451);
      END_STATE();
    case 506:
      if (lookahead == 'r') ADVANCE(300);
      END_STATE();
    case 507:
      if (lookahead == 'r') ADVANCE(440);
      END_STATE();
    case 508:
      if (lookahead == 'r') ADVANCE(554);
      END_STATE();
    case 509:
      if (lookahead == 'r') ADVANCE(555);
      END_STATE();
    case 510:
      if (lookahead == 'r') ADVANCE(388);
      END_STATE();
    case 511:
      if (lookahead == 'r') ADVANCE(213);
      END_STATE();
    case 512:
      if (lookahead == 'r') ADVANCE(577);
      END_STATE();
    case 513:
      if (lookahead == 'r') ADVANCE(215);
      END_STATE();
    case 514:
      if (lookahead == 'r') ADVANCE(195);
      END_STATE();
    case 515:
      if (lookahead == 'r') ADVANCE(563);
      END_STATE();
    case 516:
      if (lookahead == 'r') ADVANCE(564);
      END_STATE();
    case 517:
      if (lookahead == 'r') ADVANCE(222);
      END_STATE();
    case 518:
      if (lookahead == 'r') ADVANCE(312);
      END_STATE();
    case 519:
      if (lookahead == 'r') ADVANCE(539);
      END_STATE();
    case 520:
      if (lookahead == 'r') ADVANCE(128);
      END_STATE();
    case 521:
      if (lookahead == 'r') ADVANCE(174);
      END_STATE();
    case 522:
      if (lookahead == 'r') ADVANCE(621);
      END_STATE();
    case 523:
      if (lookahead == 'r') ADVANCE(240);
      if (lookahead == 't') ADVANCE(594);
      END_STATE();
    case 524:
      if (lookahead == 'r') ADVANCE(140);
      END_STATE();
    case 525:
      if (lookahead == 'r') ADVANCE(141);
      END_STATE();
    case 526:
      if (lookahead == 's') ADVANCE(1259);
      END_STATE();
    case 527:
      if (lookahead == 's') ADVANCE(1300);
      END_STATE();
    case 528:
      if (lookahead == 's') ADVANCE(1291);
      END_STATE();
    case 529:
      if (lookahead == 's') ADVANCE(1306);
      END_STATE();
    case 530:
      if (lookahead == 's') ADVANCE(1229);
      END_STATE();
    case 531:
      if (lookahead == 's') ADVANCE(1228);
      END_STATE();
    case 532:
      if (lookahead == 's') ADVANCE(711);
      END_STATE();
    case 533:
      if (lookahead == 's') ADVANCE(180);
      END_STATE();
    case 534:
      if (lookahead == 's') ADVANCE(526);
      END_STATE();
    case 535:
      if (lookahead == 's') ADVANCE(579);
      END_STATE();
    case 536:
      if (lookahead == 's') ADVANCE(185);
      END_STATE();
    case 537:
      if (lookahead == 's') ADVANCE(227);
      END_STATE();
    case 538:
      if (lookahead == 's') ADVANCE(318);
      END_STATE();
    case 539:
      if (lookahead == 's') ADVANCE(239);
      END_STATE();
    case 540:
      if (lookahead == 's') ADVANCE(229);
      END_STATE();
    case 541:
      if (lookahead == 's') ADVANCE(315);
      END_STATE();
    case 542:
      if (lookahead == 's') ADVANCE(338);
      END_STATE();
    case 543:
      if (lookahead == 's') ADVANCE(320);
      END_STATE();
    case 544:
      if (lookahead == 's') ADVANCE(323);
      END_STATE();
    case 545:
      if (lookahead == 's') ADVANCE(326);
      END_STATE();
    case 546:
      if (lookahead == 's') ADVANCE(547);
      END_STATE();
    case 547:
      if (lookahead == 's') ADVANCE(242);
      END_STATE();
    case 548:
      if (lookahead == 's') ADVANCE(328);
      END_STATE();
    case 549:
      if (lookahead == 's') ADVANCE(341);
      END_STATE();
    case 550:
      if (lookahead == 't') ADVANCE(117);
      if (lookahead == 'y') ADVANCE(96);
      END_STATE();
    case 551:
      if (lookahead == 't') ADVANCE(1282);
      END_STATE();
    case 552:
      if (lookahead == 't') ADVANCE(1084);
      END_STATE();
    case 553:
      if (lookahead == 't') ADVANCE(1160);
      END_STATE();
    case 554:
      if (lookahead == 't') ADVANCE(1188);
      END_STATE();
    case 555:
      if (lookahead == 't') ADVANCE(703);
      END_STATE();
    case 556:
      if (lookahead == 't') ADVANCE(1242);
      END_STATE();
    case 557:
      if (lookahead == 't') ADVANCE(1308);
      END_STATE();
    case 558:
      if (lookahead == 't') ADVANCE(1312);
      END_STATE();
    case 559:
      if (lookahead == 't') ADVANCE(1318);
      END_STATE();
    case 560:
      if (lookahead == 't') ADVANCE(1220);
      END_STATE();
    case 561:
      if (lookahead == 't') ADVANCE(1316);
      END_STATE();
    case 562:
      if (lookahead == 't') ADVANCE(1198);
      END_STATE();
    case 563:
      if (lookahead == 't') ADVANCE(1202);
      END_STATE();
    case 564:
      if (lookahead == 't') ADVANCE(1066);
      END_STATE();
    case 565:
      if (lookahead == 't') ADVANCE(280);
      END_STATE();
    case 566:
      if (lookahead == 't') ADVANCE(281);
      END_STATE();
    case 567:
      if (lookahead == 't') ADVANCE(653);
      END_STATE();
    case 568:
      if (lookahead == 't') ADVANCE(282);
      END_STATE();
    case 569:
      if (lookahead == 't') ADVANCE(647);
      END_STATE();
    case 570:
      if (lookahead == 't') ADVANCE(283);
      END_STATE();
    case 571:
      if (lookahead == 't') ADVANCE(648);
      END_STATE();
    case 572:
      if (lookahead == 't') ADVANCE(284);
      END_STATE();
    case 573:
      if (lookahead == 't') ADVANCE(181);
      END_STATE();
    case 574:
      if (lookahead == 't') ADVANCE(594);
      END_STATE();
    case 575:
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 576:
      if (lookahead == 't') ADVANCE(506);
      END_STATE();
    case 577:
      if (lookahead == 't') ADVANCE(649);
      END_STATE();
    case 578:
      if (lookahead == 't') ADVANCE(289);
      END_STATE();
    case 579:
      if (lookahead == 't') ADVANCE(527);
      END_STATE();
    case 580:
      if (lookahead == 't') ADVANCE(285);
      END_STATE();
    case 581:
      if (lookahead == 't') ADVANCE(529);
      END_STATE();
    case 582:
      if (lookahead == 't') ADVANCE(530);
      END_STATE();
    case 583:
      if (lookahead == 't') ADVANCE(531);
      END_STATE();
    case 584:
      if (lookahead == 't') ADVANCE(136);
      END_STATE();
    case 585:
      if (lookahead == 't') ADVANCE(293);
      END_STATE();
    case 586:
      if (lookahead == 't') ADVANCE(518);
      END_STATE();
    case 587:
      if (lookahead == 't') ADVANCE(522);
      END_STATE();
    case 588:
      if (lookahead == 't') ADVANCE(522);
      if (lookahead == 'u') ADVANCE(144);
      END_STATE();
    case 589:
      if (lookahead == 't') ADVANCE(220);
      END_STATE();
    case 590:
      if (lookahead == 't') ADVANCE(223);
      END_STATE();
    case 591:
      if (lookahead == 't') ADVANCE(200);
      END_STATE();
    case 592:
      if (lookahead == 't') ADVANCE(625);
      END_STATE();
    case 593:
      if (lookahead == 't') ADVANCE(116);
      END_STATE();
    case 594:
      if (lookahead == 't') ADVANCE(226);
      END_STATE();
    case 595:
      if (lookahead == 't') ADVANCE(310);
      END_STATE();
    case 596:
      if (lookahead == 't') ADVANCE(287);
      END_STATE();
    case 597:
      if (lookahead == 't') ADVANCE(322);
      END_STATE();
    case 598:
      if (lookahead == 't') ADVANCE(245);
      END_STATE();
    case 599:
      if (lookahead == 't') ADVANCE(325);
      END_STATE();
    case 600:
      if (lookahead == 't') ADVANCE(246);
      END_STATE();
    case 601:
      if (lookahead == 't') ADVANCE(247);
      END_STATE();
    case 602:
      if (lookahead == 't') ADVANCE(329);
      END_STATE();
    case 603:
      if (lookahead == 't') ADVANCE(330);
      END_STATE();
    case 604:
      if (lookahead == 't') ADVANCE(97);
      END_STATE();
    case 605:
      if (lookahead == 't') ADVANCE(336);
      END_STATE();
    case 606:
      if (lookahead == 't') ADVANCE(337);
      END_STATE();
    case 607:
      if (lookahead == 't') ADVANCE(339);
      END_STATE();
    case 608:
      if (lookahead == 't') ADVANCE(340);
      END_STATE();
    case 609:
      ADVANCE_MAP(
        'u', 657,
        '"', 9,
        '/', 9,
        '\\', 9,
        'a', 9,
        'b', 9,
        'e', 9,
        'f', 9,
        'n', 9,
        'r', 9,
        't', 9,
        'v', 9,
      );
      END_STATE();
    case 610:
      if (lookahead == 'u') ADVANCE(143);
      END_STATE();
    case 611:
      if (lookahead == 'u') ADVANCE(473);
      END_STATE();
    case 612:
      if (lookahead == 'u') ADVANCE(497);
      END_STATE();
    case 613:
      if (lookahead == 'u') ADVANCE(183);
      END_STATE();
    case 614:
      if (lookahead == 'u') ADVANCE(121);
      END_STATE();
    case 615:
      if (lookahead == 'u') ADVANCE(411);
      END_STATE();
    case 616:
      if (lookahead == 'u') ADVANCE(307);
      END_STATE();
    case 617:
      if (lookahead == 'u') ADVANCE(187);
      END_STATE();
    case 618:
      if (lookahead == 'u') ADVANCE(189);
      END_STATE();
    case 619:
      if (lookahead == 'u') ADVANCE(560);
      END_STATE();
    case 620:
      if (lookahead == 'u') ADVANCE(194);
      END_STATE();
    case 621:
      if (lookahead == 'u') ADVANCE(150);
      END_STATE();
    case 622:
      if (lookahead == 'u') ADVANCE(541);
      END_STATE();
    case 623:
      if (lookahead == 'u') ADVANCE(151);
      END_STATE();
    case 624:
      if (lookahead == 'u') ADVANCE(360);
      END_STATE();
    case 625:
      if (lookahead == 'u') ADVANCE(514);
      END_STATE();
    case 626:
      if (lookahead == 'u') ADVANCE(543);
      END_STATE();
    case 627:
      if (lookahead == 'u') ADVANCE(544);
      END_STATE();
    case 628:
      if (lookahead == 'u') ADVANCE(545);
      END_STATE();
    case 629:
      if (lookahead == 'u') ADVANCE(524);
      END_STATE();
    case 630:
      if (lookahead == 'u') ADVANCE(525);
      END_STATE();
    case 631:
      if (lookahead == 'v') ADVANCE(193);
      END_STATE();
    case 632:
      if (lookahead == 'v') ADVANCE(196);
      END_STATE();
    case 633:
      if (lookahead == 'v') ADVANCE(197);
      END_STATE();
    case 634:
      if (lookahead == 'v') ADVANCE(198);
      END_STATE();
    case 635:
      if (lookahead == 'v') ADVANCE(199);
      END_STATE();
    case 636:
      if (lookahead == 'v') ADVANCE(248);
      END_STATE();
    case 637:
      if (lookahead == 'v') ADVANCE(249);
      END_STATE();
    case 638:
      if (lookahead == 'v') ADVANCE(250);
      END_STATE();
    case 639:
      if (lookahead == 'w') ADVANCE(389);
      END_STATE();
    case 640:
      if (lookahead == 'x') ADVANCE(80);
      END_STATE();
    case 641:
      if (lookahead == 'x') ADVANCE(87);
      END_STATE();
    case 642:
      if (lookahead == 'x') ADVANCE(148);
      END_STATE();
    case 643:
      if (lookahead == 'x') ADVANCE(205);
      END_STATE();
    case 644:
      if (lookahead == 'x') ADVANCE(156);
      END_STATE();
    case 645:
      if (lookahead == 'y') ADVANCE(1162);
      END_STATE();
    case 646:
      if (lookahead == 'y') ADVANCE(1216);
      END_STATE();
    case 647:
      if (lookahead == 'y') ADVANCE(1245);
      END_STATE();
    case 648:
      if (lookahead == 'y') ADVANCE(1262);
      END_STATE();
    case 649:
      if (lookahead == 'y') ADVANCE(1254);
      END_STATE();
    case 650:
      if (lookahead == 'y') ADVANCE(1166);
      END_STATE();
    case 651:
      if (lookahead == 'y') ADVANCE(1172);
      END_STATE();
    case 652:
      if (lookahead == 'y') ADVANCE(1146);
      END_STATE();
    case 653:
      if (lookahead == 'y') ADVANCE(479);
      END_STATE();
    case 654:
      if (lookahead == 'y') ADVANCE(591);
      END_STATE();
    case 655:
      if (lookahead == 'z') ADVANCE(216);
      END_STATE();
    case 656:
      if (lookahead == 'z') ADVANCE(467);
      END_STATE();
    case 657:
      if (lookahead == '{') ADVANCE(675);
      END_STATE();
    case 658:
      if (lookahead == '}') ADVANCE(9);
      END_STATE();
    case 659:
      if (lookahead == '}') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(658);
      END_STATE();
    case 660:
      if (lookahead == '}') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(659);
      END_STATE();
    case 661:
      if (lookahead == '}') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(660);
      END_STATE();
    case 662:
      if (lookahead == '}') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(661);
      END_STATE();
    case 663:
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1127);
      END_STATE();
    case 664:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(671);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(684);
      END_STATE();
    case 665:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(671);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(685);
      END_STATE();
    case 666:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(671);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(686);
      END_STATE();
    case 667:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(671);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(687);
      END_STATE();
    case 668:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(671);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(688);
      END_STATE();
    case 669:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1094);
      END_STATE();
    case 670:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1124);
      END_STATE();
    case 671:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(669);
      END_STATE();
    case 672:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(662);
      END_STATE();
    case 673:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1121);
      END_STATE();
    case 674:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1115);
      END_STATE();
    case 675:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(672);
      END_STATE();
    case 676:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(674);
      END_STATE();
    case 677:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(676);
      END_STATE();
    case 678:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(677);
      END_STATE();
    case 679:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(678);
      END_STATE();
    case 680:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(679);
      END_STATE();
    case 681:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(680);
      END_STATE();
    case 682:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(681);
      END_STATE();
    case 683:
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(694);
      END_STATE();
    case 684:
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1106);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      END_STATE();
    case 685:
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1094);
      END_STATE();
    case 686:
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1094);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      END_STATE();
    case 687:
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1107);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      END_STATE();
    case 688:
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1108);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(693);
      END_STATE();
    case 689:
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1070);
      END_STATE();
    case 690:
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1100);
      END_STATE();
    case 691:
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(690);
      END_STATE();
    case 692:
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1102);
      END_STATE();
    case 693:
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(692);
      END_STATE();
    case 694:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1078);
      END_STATE();
    case 695:
      if (eof) ADVANCE(696);
      ADVANCE_MAP(
        '"', 9,
        '#', 99,
        '%', 1330,
        '(', 1079,
        ')', 1080,
        '*', 1325,
        '+', 1331,
        ',', 1081,
        '-', 1333,
        '.', 33,
        '/', 1327,
        '0', 1128,
        ':', 41,
        ';', 74,
        '<', 1110,
        '=', 1065,
        '>', 1276,
        '@', 1061,
        'N', 436,
        'T', 286,
        '[', 705,
        ']', 706,
        '_', 1260,
        'a', 1256,
        'b', 101,
        'c', 344,
        'd', 102,
        'e', 380,
        'f', 103,
        'g', 82,
        'h', 204,
        'i', 175,
        'l', 110,
        'm', 112,
        'n', 179,
        'o', 253,
        'p', 105,
        'r', 114,
        's', 203,
        't', 317,
        'u', 383,
        'v', 212,
        'w', 299,
        'x', 450,
        'y', 210,
        '{', 1091,
        '|', 1324,
        '}', 1092,
        0xac, 1284,
        0xd7, 1326,
        0xf7, 1328,
        0x2190, 1342,
        0x2192, 1340,
        0x21d2, 1293,
        0x21d4, 1296,
        0x2200, 1299,
        0x2203, 1302,
        0x2205, 1136,
        0x2208, 1322,
        0x2209, 1323,
        0x2216, 1317,
        0x2227, 1286,
        0x2228, 1288,
        0x2229, 1307,
        0x222a, 1305,
        0x2254, 1338,
        0x2260, 1275,
        0x2264, 1279,
        0x2265, 1281,
        0x2282, 1309,
        0x2283, 1313,
        0x2286, 1311,
        0x2287, 1315,
        0x22a4, 1132,
        0x22a5, 1135,
        0x22bb, 1290,
        0x2a09, 1319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(695);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(1129);
      END_STATE();
    case 696:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 697:
      ACCEPT_TOKEN(anon_sym_module);
      END_STATE();
    case 698:
      ACCEPT_TOKEN(anon_sym_version);
      END_STATE();
    case 699:
      ACCEPT_TOKEN(anon_sym_is);
      END_STATE();
    case 700:
      ACCEPT_TOKEN(anon_sym_is);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 701:
      ACCEPT_TOKEN(anon_sym_end);
      END_STATE();
    case 702:
      ACCEPT_TOKEN(anon_sym_end);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 703:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 704:
      ACCEPT_TOKEN(anon_sym_import);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 705:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 706:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 707:
      ACCEPT_TOKEN(anon_sym_from);
      END_STATE();
    case 708:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 709:
      ACCEPT_TOKEN(anon_sym_as);
      END_STATE();
    case 710:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 711:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 's') ADVANCE(242);
      END_STATE();
    case 712:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '4') ADVANCE(715);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 713:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '6') ADVANCE(712);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 714:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'B') ADVANCE(863);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 715:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'B') ADVANCE(865);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 716:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'D') ADVANCE(735);
      if (lookahead == 'M') ADVANCE(942);
      if (lookahead == 'Y') ADVANCE(805);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 717:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'D') ADVANCE(1038);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 718:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'D') ADVANCE(1039);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 719:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'I') ADVANCE(1145);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 720:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'I') ADVANCE(922);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 721:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'I') ADVANCE(926);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 722:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'I') ADVANCE(927);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 723:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'I') ADVANCE(928);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 724:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'M') ADVANCE(945);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 725:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'N') ADVANCE(815);
      if (lookahead == 'P') ADVANCE(948);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 726:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'R') ADVANCE(719);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 727:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'S') ADVANCE(1019);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 728:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'T') ADVANCE(856);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 729:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'U') ADVANCE(726);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 730:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(878);
      if (lookahead == 'i') ADVANCE(1048);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 731:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(878);
      if (lookahead == 'o') ADVANCE(957);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 732:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(878);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 733:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1014);
      if (lookahead == 'e') ADVANCE(761);
      if (lookahead == 'o') ADVANCE(1029);
      if (lookahead == 'u') ADVANCE(966);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 734:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(953);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 735:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1050);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 736:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(880);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 737:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(884);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 738:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1052);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 739:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(901);
      if (lookahead == 'o') ADVANCE(903);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 740:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(958);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 741:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(965);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 742:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(875);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 743:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(959);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 744:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(881);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 745:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1010);
      if (lookahead == 'e') ADVANCE(742);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 746:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(876);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 747:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(998);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 748:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(967);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 749:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(877);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 750:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(969);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 751:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1015);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 752:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(895);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 753:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(977);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 754:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(988);
      if (lookahead == 'i') ADVANCE(921);
      if (lookahead == 'o') ADVANCE(934);
      if (lookahead == 'y') ADVANCE(1013);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 755:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(827);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 756:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1020);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 757:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1022);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 758:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1024);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 759:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'a') ADVANCE(1027);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 760:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'b') ADVANCE(883);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 761:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'c') ADVANCE(862);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 762:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(1200);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 763:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(727);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 764:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(1240);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 765:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(702);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 766:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(1266);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 767:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(1268);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 768:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(803);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 769:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(804);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 770:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(902);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 771:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1131);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 772:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1134);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 773:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1047);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 774:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(832);
      if (lookahead == 'o') ADVANCE(893);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 775:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(713);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 776:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1191);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 777:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1151);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 778:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1177);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 779:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1157);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 780:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(717);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 781:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(762);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 782:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1152);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 783:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1215);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 784:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1205);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 785:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1224);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 786:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1270);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 787:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(1272);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 788:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(873);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 789:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(817);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 790:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(720);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 791:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(763);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 792:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(740);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 793:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(764);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 794:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(752);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 795:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(766);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 796:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(767);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 797:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(960);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 798:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(894);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 799:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(961);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 800:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(962);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 801:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(963);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 802:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(964);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 803:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(975);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 804:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(976);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 805:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(743);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 806:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(973);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 807:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(828);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 808:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(829);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 809:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(830);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 810:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(831);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 811:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(920);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 812:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(721);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 813:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(722);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 814:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(723);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 815:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(833);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 816:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'f') ADVANCE(1083);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 817:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'f') ADVANCE(1253);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 818:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1186);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 819:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1138);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 820:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1175);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 821:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1140);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 822:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1197);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 823:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1211);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 824:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(1032);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 825:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(919);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 826:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(797);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 827:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(783);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 828:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(799);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 829:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(800);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 830:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(801);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 831:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(802);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 832:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(751);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 833:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'g') ADVANCE(759);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 834:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(1087);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 835:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(1165);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 836:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(1169);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 837:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(1086);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 838:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(851);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 839:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(932);
      if (lookahead == 't') ADVANCE(978);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 840:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(944);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 841:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(853);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 842:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'h') ADVANCE(718);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 843:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(987);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 844:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(886);
      if (lookahead == 'o') ADVANCE(872);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 845:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1219);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 846:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1042);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 847:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1056);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 848:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(995);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 849:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(825);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 850:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(936);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 851:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(904);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 852:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(909);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 853:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(911);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 854:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(914);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 855:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1012);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 856:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(887);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 857:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(937);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 858:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(888);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 859:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(939);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 860:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(940);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 861:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(954);
      if (lookahead == 'o') ADVANCE(980);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 862:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(890);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 863:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(923);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 864:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(955);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 865:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(925);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 866:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1043);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 867:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1026);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 868:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1044);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 869:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1045);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 870:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'i') ADVANCE(1028);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 871:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'k') ADVANCE(910);
      if (lookahead == 's') ADVANCE(849);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 872:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'k') ADVANCE(798);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 873:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(816);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 874:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(1298);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 875:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(1142);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 876:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(1155);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 877:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(1144);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 878:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(986);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 879:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(943);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 880:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(874);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 881:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(847);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 882:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(794);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 883:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(779);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 884:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(949);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 885:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(952);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 886:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(778);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 887:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(780);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 888:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(782);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 889:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(744);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 890:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'm') ADVANCE(746);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 891:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1049);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 892:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(871);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 893:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(725);
      if (lookahead == 'r') ADVANCE(889);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 894:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1213);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 895:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1149);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 896:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1264);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 897:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1159);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 898:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1181);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 899:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1183);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 900:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(861);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 901:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(824);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 902:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(996);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 903:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(818);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 904:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(819);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 905:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(765);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 906:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1034);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 907:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(989);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 908:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(997);
      if (lookahead == 'r') ADVANCE(845);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 909:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(820);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 910:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(930);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 911:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(821);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 912:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(822);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 913:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1007);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 914:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(823);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 915:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(864);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 916:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1008);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 917:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1009);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 918:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1000);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 919:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(781);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 920:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1003);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 921:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(741);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 922:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1016);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 923:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(748);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 924:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(749);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 925:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(750);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 926:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1021);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 927:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1023);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 928:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'n') ADVANCE(1025);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 929:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(994);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 930:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(1046);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 931:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(991);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 932:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(971);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 933:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(1018);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 934:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(882);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 935:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(1031);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 936:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(924);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 937:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(897);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 938:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(912);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 939:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(898);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 940:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(899);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 941:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(906);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 942:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(913);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 943:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(747);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 944:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(972);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 945:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(916);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 946:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(917);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 947:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(974);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 948:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(993);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 949:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'p') ADVANCE(1179);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 950:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'p') ADVANCE(1250);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 951:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'p') ADVANCE(734);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 952:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'p') ADVANCE(947);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 953:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'q') ADVANCE(1035);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 954:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'q') ADVANCE(1036);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 955:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'q') ADVANCE(1037);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 956:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1030);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 957:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(736);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 958:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(724);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 959:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1171);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 960:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1184);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 961:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1209);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 962:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1195);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 963:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1193);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 964:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1207);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 965:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1051);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 966:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(756);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 967:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1053);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 968:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(768);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 969:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1054);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 970:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(935);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 971:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(999);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 972:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1001);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 973:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1002);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 974:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(1005);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 975:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(795);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 976:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(796);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 977:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(811);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 978:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(852);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 979:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(854);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 980:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(769);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 981:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(757);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 982:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'r') ADVANCE(758);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 983:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(1301);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 984:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(710);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 985:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(700);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 986:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(772);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 987:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(1006);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 988:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(775);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 989:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(849);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 990:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(806);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 991:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(867);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 992:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(990);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 993:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 's') ADVANCE(870);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 994:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1283);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 995:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(834);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 996:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1085);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 997:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1187);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 998:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1161);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 999:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1189);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1000:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1199);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1001:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1203);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1002:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1067);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1003:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1243);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1004:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(1221);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1005:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(704);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1006:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(983);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1007:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(835);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1008:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(842);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1009:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(836);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1010:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(850);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1011:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(737);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1012:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(837);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1013:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(776);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1014:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(777);
      if (lookahead == 'y') ADVANCE(728);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1015:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(846);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1016:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(807);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1017:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(784);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1018:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(841);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1019:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(979);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1020:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(857);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1021:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(808);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1022:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(859);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1023:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(809);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1024:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(860);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1025:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(810);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1026:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(866);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1027:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(868);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1028:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 't') ADVANCE(869);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1029:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(760);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1030:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(771);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1031:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(950);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1032:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(755);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1033:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(1004);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1034:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(915);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1035:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(785);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1036:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(786);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1037:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(787);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1038:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(981);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1039:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'u') ADVANCE(982);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1040:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'v') ADVANCE(770);
      if (lookahead == 'x') ADVANCE(843);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1041:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'v') ADVANCE(770);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1042:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'v') ADVANCE(790);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1043:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'v') ADVANCE(812);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1044:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'v') ADVANCE(813);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1045:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'v') ADVANCE(814);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1046:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'w') ADVANCE(896);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1047:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'x') ADVANCE(714);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1048:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'x') ADVANCE(793);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1049:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(729);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1050:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(1163);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1051:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(1217);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1052:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(1167);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1053:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(1173);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1054:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(1147);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1055:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'y') ADVANCE(1017);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1056:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'z') ADVANCE(791);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1057:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1058:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 1059:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(708);
      if (lookahead == '=') ADVANCE(1337);
      END_STATE();
    case 1060:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '=') ADVANCE(1337);
      END_STATE();
    case 1061:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 1062:
      ACCEPT_TOKEN(anon_sym_AT);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(689);
      END_STATE();
    case 1063:
      ACCEPT_TOKEN(anon_sym_AT);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(690);
      END_STATE();
    case 1064:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 1065:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(78);
      END_STATE();
    case 1066:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 1067:
      ACCEPT_TOKEN(anon_sym_assert);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1068:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      END_STATE();
    case 1069:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (lookahead == '-') ADVANCE(683);
      END_STATE();
    case 1070:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (lookahead == '-') ADVANCE(683);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1069);
      END_STATE();
    case 1071:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1068);
      END_STATE();
    case 1072:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1071);
      END_STATE();
    case 1073:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1072);
      END_STATE();
    case 1074:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1073);
      END_STATE();
    case 1075:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1074);
      END_STATE();
    case 1076:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1075);
      END_STATE();
    case 1077:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1076);
      END_STATE();
    case 1078:
      ACCEPT_TOKEN(sym_controlled_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1077);
      END_STATE();
    case 1079:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 1080:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 1081:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 1082:
      ACCEPT_TOKEN(sym_reserved_self);
      END_STATE();
    case 1083:
      ACCEPT_TOKEN(sym_reserved_self);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1084:
      ACCEPT_TOKEN(anon_sym_event);
      END_STATE();
    case 1085:
      ACCEPT_TOKEN(anon_sym_event);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1086:
      ACCEPT_TOKEN(anon_sym_with);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'o') ADVANCE(1033);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1087:
      ACCEPT_TOKEN(anon_sym_with);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1088:
      ACCEPT_TOKEN(anon_sym_with);
      if (lookahead == 'o') ADVANCE(619);
      END_STATE();
    case 1089:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 1090:
      ACCEPT_TOKEN(anon_sym_def);
      END_STATE();
    case 1091:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 1092:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 1093:
      ACCEPT_TOKEN(sym_quoted_string);
      END_STATE();
    case 1094:
      ACCEPT_TOKEN(sym_language_tag);
      END_STATE();
    case 1095:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      END_STATE();
    case 1096:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1101);
      END_STATE();
    case 1097:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1096);
      END_STATE();
    case 1098:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1097);
      END_STATE();
    case 1099:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1098);
      END_STATE();
    case 1100:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1099);
      END_STATE();
    case 1101:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(664);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(1095);
      END_STATE();
    case 1102:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(665);
      END_STATE();
    case 1103:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(666);
      END_STATE();
    case 1104:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(667);
      END_STATE();
    case 1105:
      ACCEPT_TOKEN(sym_language_tag);
      if (lookahead == '-') ADVANCE(668);
      END_STATE();
    case 1106:
      ACCEPT_TOKEN(sym_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1105);
      END_STATE();
    case 1107:
      ACCEPT_TOKEN(sym_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1103);
      END_STATE();
    case 1108:
      ACCEPT_TOKEN(sym_language_tag);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(1104);
      END_STATE();
    case 1109:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 1110:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '=') ADVANCE(1278);
      END_STATE();
    case 1111:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '=') ADVANCE(1277);
      END_STATE();
    case 1112:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == ';') ADVANCE(1113);
      if (lookahead == '\\') ADVANCE(98);
      if (lookahead > ' ' &&
          lookahead != '"' &&
          lookahead != ';' &&
          lookahead != '<' &&
          lookahead != '>' &&
          lookahead != '^' &&
          lookahead != '`' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1115);
      END_STATE();
    case 1113:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == ';') ADVANCE(1114);
      if (lookahead == '\\') ADVANCE(1343);
      if ((!eof && lookahead <= ' ') ||
          lookahead == '"' ||
          lookahead == '<' ||
          lookahead == '>' ||
          lookahead == '^' ||
          lookahead == '`' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1352);
      if (lookahead > ' ') ADVANCE(1116);
      END_STATE();
    case 1114:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == ';') ADVANCE(1118);
      if (lookahead == '\\') ADVANCE(1353);
      if ((!eof && lookahead <= ' ') ||
          lookahead == '"' ||
          lookahead == '<' ||
          lookahead == '>' ||
          lookahead == '^' ||
          lookahead == '`' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1362);
      if (lookahead > ' ') ADVANCE(1117);
      END_STATE();
    case 1115:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == '\\') ADVANCE(98);
      if (lookahead > ' ' &&
          lookahead != '"' &&
          lookahead != '<' &&
          lookahead != '>' &&
          lookahead != '^' &&
          lookahead != '`' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(1115);
      END_STATE();
    case 1116:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == '\\') ADVANCE(1343);
      if ((!eof && lookahead <= '\t') ||
          (0x0b <= lookahead && lookahead <= ' ') ||
          lookahead == '"' ||
          lookahead == '<' ||
          lookahead == '>' ||
          lookahead == '^' ||
          lookahead == '`' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1352);
      if (lookahead > ' ') ADVANCE(1116);
      END_STATE();
    case 1117:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == '\\') ADVANCE(1353);
      if ((!eof && lookahead <= '\t') ||
          (0x0b <= lookahead && lookahead <= ' ') ||
          lookahead == '"' ||
          lookahead == '<' ||
          lookahead == '>' ||
          lookahead == '^' ||
          lookahead == '`' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1362);
      if (lookahead > ' ') ADVANCE(1117);
      END_STATE();
    case 1118:
      ACCEPT_TOKEN(aux_sym_iri_token1);
      if (lookahead == '\\') ADVANCE(1363);
      if ((!eof && lookahead <= '\t') ||
          (0x0b <= lookahead && lookahead <= ' ') ||
          lookahead == '"' ||
          lookahead == '<' ||
          lookahead == '>' ||
          lookahead == '^' ||
          lookahead == '`' ||
          ('{' <= lookahead && lookahead <= '}')) ADVANCE(1372);
      if (lookahead > ' ') ADVANCE(1118);
      END_STATE();
    case 1119:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 1120:
      ACCEPT_TOKEN(anon_sym_POUND_LBRACK);
      END_STATE();
    case 1121:
      ACCEPT_TOKEN(sym_hex_byte);
      END_STATE();
    case 1122:
      ACCEPT_TOKEN(sym_double);
      END_STATE();
    case 1123:
      ACCEPT_TOKEN(sym_double);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1123);
      END_STATE();
    case 1124:
      ACCEPT_TOKEN(sym_decimal);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(36);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1124);
      END_STATE();
    case 1125:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(670);
      if (lookahead == '/' ||
          lookahead == 0x2044) ADVANCE(663);
      END_STATE();
    case 1126:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(670);
      if (lookahead == '/' ||
          lookahead == 0x2044) ADVANCE(663);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1126);
      END_STATE();
    case 1127:
      ACCEPT_TOKEN(sym_rational);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1127);
      END_STATE();
    case 1128:
      ACCEPT_TOKEN(sym_unsigned);
      END_STATE();
    case 1129:
      ACCEPT_TOKEN(sym_unsigned);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1129);
      END_STATE();
    case 1130:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 1131:
      ACCEPT_TOKEN(anon_sym_true);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1132:
      ACCEPT_TOKEN(anon_sym_u22a4);
      END_STATE();
    case 1133:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 1134:
      ACCEPT_TOKEN(anon_sym_false);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1135:
      ACCEPT_TOKEN(anon_sym_u22a5);
      END_STATE();
    case 1136:
      ACCEPT_TOKEN(sym_value_empty_sequence);
      END_STATE();
    case 1137:
      ACCEPT_TOKEN(anon_sym_Thing);
      END_STATE();
    case 1138:
      ACCEPT_TOKEN(anon_sym_Thing);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1139:
      ACCEPT_TOKEN(anon_sym_Nothing);
      END_STATE();
    case 1140:
      ACCEPT_TOKEN(anon_sym_Nothing);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1141:
      ACCEPT_TOKEN(anon_sym_real);
      END_STATE();
    case 1142:
      ACCEPT_TOKEN(anon_sym_real);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1143:
      ACCEPT_TOKEN(anon_sym_rational);
      END_STATE();
    case 1144:
      ACCEPT_TOKEN(anon_sym_rational);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1145:
      ACCEPT_TOKEN(anon_sym_anyURI);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1146:
      ACCEPT_TOKEN(anon_sym_base64Binary);
      END_STATE();
    case 1147:
      ACCEPT_TOKEN(anon_sym_base64Binary);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1148:
      ACCEPT_TOKEN(anon_sym_boolean);
      END_STATE();
    case 1149:
      ACCEPT_TOKEN(anon_sym_boolean);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1150:
      ACCEPT_TOKEN(anon_sym_date);
      if (lookahead == 'T') ADVANCE(324);
      END_STATE();
    case 1151:
      ACCEPT_TOKEN(anon_sym_date);
      if (lookahead == 'T') ADVANCE(858);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1152:
      ACCEPT_TOKEN(anon_sym_dateTime);
      if (lookahead == 'S') ADVANCE(1011);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1153:
      ACCEPT_TOKEN(anon_sym_dateTime);
      if (lookahead == 'S') ADVANCE(584);
      END_STATE();
    case 1154:
      ACCEPT_TOKEN(anon_sym_decimal);
      END_STATE();
    case 1155:
      ACCEPT_TOKEN(anon_sym_decimal);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1156:
      ACCEPT_TOKEN(anon_sym_double);
      END_STATE();
    case 1157:
      ACCEPT_TOKEN(anon_sym_double);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1158:
      ACCEPT_TOKEN(anon_sym_duration);
      END_STATE();
    case 1159:
      ACCEPT_TOKEN(anon_sym_duration);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1160:
      ACCEPT_TOKEN(anon_sym_float);
      END_STATE();
    case 1161:
      ACCEPT_TOKEN(anon_sym_float);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1162:
      ACCEPT_TOKEN(anon_sym_gDay);
      END_STATE();
    case 1163:
      ACCEPT_TOKEN(anon_sym_gDay);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1164:
      ACCEPT_TOKEN(anon_sym_gMonth);
      if (lookahead == 'D') ADVANCE(119);
      END_STATE();
    case 1165:
      ACCEPT_TOKEN(anon_sym_gMonth);
      if (lookahead == 'D') ADVANCE(738);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1166:
      ACCEPT_TOKEN(anon_sym_gMonthDay);
      END_STATE();
    case 1167:
      ACCEPT_TOKEN(anon_sym_gMonthDay);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1168:
      ACCEPT_TOKEN(anon_sym_gYearMonth);
      END_STATE();
    case 1169:
      ACCEPT_TOKEN(anon_sym_gYearMonth);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1170:
      ACCEPT_TOKEN(anon_sym_gYear);
      if (lookahead == 'M') ADVANCE(470);
      END_STATE();
    case 1171:
      ACCEPT_TOKEN(anon_sym_gYear);
      if (lookahead == 'M') ADVANCE(946);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1172:
      ACCEPT_TOKEN(anon_sym_hexBinary);
      END_STATE();
    case 1173:
      ACCEPT_TOKEN(anon_sym_hexBinary);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1174:
      ACCEPT_TOKEN(anon_sym_string);
      END_STATE();
    case 1175:
      ACCEPT_TOKEN(anon_sym_string);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1176:
      ACCEPT_TOKEN(anon_sym_time);
      END_STATE();
    case 1177:
      ACCEPT_TOKEN(anon_sym_time);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1178:
      ACCEPT_TOKEN(anon_sym_dateTimeStamp);
      END_STATE();
    case 1179:
      ACCEPT_TOKEN(anon_sym_dateTimeStamp);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1180:
      ACCEPT_TOKEN(anon_sym_dayTimeDuration);
      END_STATE();
    case 1181:
      ACCEPT_TOKEN(anon_sym_dayTimeDuration);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1182:
      ACCEPT_TOKEN(anon_sym_yearMonthDuration);
      END_STATE();
    case 1183:
      ACCEPT_TOKEN(anon_sym_yearMonthDuration);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1184:
      ACCEPT_TOKEN(anon_sym_integer);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1185:
      ACCEPT_TOKEN(anon_sym_long);
      END_STATE();
    case 1186:
      ACCEPT_TOKEN(anon_sym_long);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1187:
      ACCEPT_TOKEN(anon_sym_int);
      if (lookahead == '_') ADVANCE(100);
      if (lookahead == 'e') ADVANCE(826);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1188:
      ACCEPT_TOKEN(anon_sym_short);
      END_STATE();
    case 1189:
      ACCEPT_TOKEN(anon_sym_short);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1190:
      ACCEPT_TOKEN(anon_sym_byte);
      END_STATE();
    case 1191:
      ACCEPT_TOKEN(anon_sym_byte);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1192:
      ACCEPT_TOKEN(anon_sym_nonNegativeInteger);
      END_STATE();
    case 1193:
      ACCEPT_TOKEN(anon_sym_nonNegativeInteger);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1194:
      ACCEPT_TOKEN(anon_sym_positiveInteger);
      END_STATE();
    case 1195:
      ACCEPT_TOKEN(anon_sym_positiveInteger);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1196:
      ACCEPT_TOKEN(anon_sym_unsignedLong);
      END_STATE();
    case 1197:
      ACCEPT_TOKEN(anon_sym_unsignedLong);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1198:
      ACCEPT_TOKEN(anon_sym_unsignedInt);
      END_STATE();
    case 1199:
      ACCEPT_TOKEN(anon_sym_unsignedInt);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1200:
      ACCEPT_TOKEN(anon_sym_unsigned);
      if (lookahead == 'B') ADVANCE(1055);
      if (lookahead == 'I') ADVANCE(918);
      if (lookahead == 'L') ADVANCE(938);
      if (lookahead == 'S') ADVANCE(840);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1201:
      ACCEPT_TOKEN(anon_sym_unsigned);
      if (lookahead == 'B') ADVANCE(654);
      if (lookahead == 'I') ADVANCE(417);
      if (lookahead == 'L') ADVANCE(459);
      if (lookahead == 'S') ADVANCE(288);
      END_STATE();
    case 1202:
      ACCEPT_TOKEN(anon_sym_unsignedShort);
      END_STATE();
    case 1203:
      ACCEPT_TOKEN(anon_sym_unsignedShort);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1204:
      ACCEPT_TOKEN(anon_sym_unsignedByte);
      END_STATE();
    case 1205:
      ACCEPT_TOKEN(anon_sym_unsignedByte);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1206:
      ACCEPT_TOKEN(anon_sym_nonPositiveInteger);
      END_STATE();
    case 1207:
      ACCEPT_TOKEN(anon_sym_nonPositiveInteger);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1208:
      ACCEPT_TOKEN(anon_sym_negativeInteger);
      END_STATE();
    case 1209:
      ACCEPT_TOKEN(anon_sym_negativeInteger);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1210:
      ACCEPT_TOKEN(anon_sym_normalizedString);
      END_STATE();
    case 1211:
      ACCEPT_TOKEN(anon_sym_normalizedString);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1212:
      ACCEPT_TOKEN(anon_sym_token);
      END_STATE();
    case 1213:
      ACCEPT_TOKEN(anon_sym_token);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1214:
      ACCEPT_TOKEN(anon_sym_language);
      END_STATE();
    case 1215:
      ACCEPT_TOKEN(anon_sym_language);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1216:
      ACCEPT_TOKEN(anon_sym_binary);
      END_STATE();
    case 1217:
      ACCEPT_TOKEN(anon_sym_binary);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1218:
      ACCEPT_TOKEN(anon_sym_iri);
      END_STATE();
    case 1219:
      ACCEPT_TOKEN(anon_sym_iri);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1220:
      ACCEPT_TOKEN(anon_sym_without);
      END_STATE();
    case 1221:
      ACCEPT_TOKEN(anon_sym_without);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1222:
      ACCEPT_TOKEN(anon_sym_datatype);
      END_STATE();
    case 1223:
      ACCEPT_TOKEN(sym_opaque);
      END_STATE();
    case 1224:
      ACCEPT_TOKEN(sym_opaque);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1225:
      ACCEPT_TOKEN(anon_sym_length);
      END_STATE();
    case 1226:
      ACCEPT_TOKEN(anon_sym_maxLength);
      END_STATE();
    case 1227:
      ACCEPT_TOKEN(anon_sym_minLength);
      END_STATE();
    case 1228:
      ACCEPT_TOKEN(anon_sym_fractionDigits);
      END_STATE();
    case 1229:
      ACCEPT_TOKEN(anon_sym_totalDigits);
      END_STATE();
    case 1230:
      ACCEPT_TOKEN(anon_sym_maxExclusive);
      END_STATE();
    case 1231:
      ACCEPT_TOKEN(anon_sym_minExclusive);
      END_STATE();
    case 1232:
      ACCEPT_TOKEN(anon_sym_maxInclusive);
      END_STATE();
    case 1233:
      ACCEPT_TOKEN(anon_sym_minInclusive);
      END_STATE();
    case 1234:
      ACCEPT_TOKEN(anon_sym_explicitTimezone);
      END_STATE();
    case 1235:
      ACCEPT_TOKEN(anon_sym_required);
      END_STATE();
    case 1236:
      ACCEPT_TOKEN(anon_sym_prohibited);
      END_STATE();
    case 1237:
      ACCEPT_TOKEN(anon_sym_optional);
      END_STATE();
    case 1238:
      ACCEPT_TOKEN(anon_sym_pattern);
      END_STATE();
    case 1239:
      ACCEPT_TOKEN(sym_kw_is_fixed);
      END_STATE();
    case 1240:
      ACCEPT_TOKEN(sym_kw_is_fixed);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1241:
      ACCEPT_TOKEN(anon_sym_dimension);
      END_STATE();
    case 1242:
      ACCEPT_TOKEN(anon_sym_parent);
      END_STATE();
    case 1243:
      ACCEPT_TOKEN(anon_sym_parent);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1244:
      ACCEPT_TOKEN(anon_sym_source);
      END_STATE();
    case 1245:
      ACCEPT_TOKEN(anon_sym_entity);
      END_STATE();
    case 1246:
      ACCEPT_TOKEN(anon_sym_enum);
      END_STATE();
    case 1247:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 1248:
      ACCEPT_TOKEN(anon_sym_metric);
      END_STATE();
    case 1249:
      ACCEPT_TOKEN(anon_sym_group);
      END_STATE();
    case 1250:
      ACCEPT_TOKEN(anon_sym_group);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1251:
      ACCEPT_TOKEN(anon_sym_on);
      END_STATE();
    case 1252:
      ACCEPT_TOKEN(anon_sym_ref);
      END_STATE();
    case 1253:
      ACCEPT_TOKEN(anon_sym_ref);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1254:
      ACCEPT_TOKEN(anon_sym_property);
      END_STATE();
    case 1255:
      ACCEPT_TOKEN(anon_sym_rdf);
      END_STATE();
    case 1256:
      ACCEPT_TOKEN(anon_sym_a);
      END_STATE();
    case 1257:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 1258:
      ACCEPT_TOKEN(anon_sym_structure);
      END_STATE();
    case 1259:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 1260:
      ACCEPT_TOKEN(sym_wildcard);
      END_STATE();
    case 1261:
      ACCEPT_TOKEN(anon_sym_union);
      END_STATE();
    case 1262:
      ACCEPT_TOKEN(anon_sym_identity);
      END_STATE();
    case 1263:
      ACCEPT_TOKEN(sym_unknown_type);
      END_STATE();
    case 1264:
      ACCEPT_TOKEN(sym_unknown_type);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1265:
      ACCEPT_TOKEN(anon_sym_ordered);
      END_STATE();
    case 1266:
      ACCEPT_TOKEN(anon_sym_ordered);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1267:
      ACCEPT_TOKEN(anon_sym_unordered);
      END_STATE();
    case 1268:
      ACCEPT_TOKEN(anon_sym_unordered);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1269:
      ACCEPT_TOKEN(anon_sym_unique);
      END_STATE();
    case 1270:
      ACCEPT_TOKEN(anon_sym_unique);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1271:
      ACCEPT_TOKEN(anon_sym_nonunique);
      END_STATE();
    case 1272:
      ACCEPT_TOKEN(anon_sym_nonunique);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1273:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 1274:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 1275:
      ACCEPT_TOKEN(anon_sym_u2260);
      END_STATE();
    case 1276:
      ACCEPT_TOKEN(sym_op_greater_than);
      if (lookahead == '=') ADVANCE(1280);
      END_STATE();
    case 1277:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 1278:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == '=') ADVANCE(79);
      END_STATE();
    case 1279:
      ACCEPT_TOKEN(anon_sym_u2264);
      END_STATE();
    case 1280:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 1281:
      ACCEPT_TOKEN(anon_sym_u2265);
      END_STATE();
    case 1282:
      ACCEPT_TOKEN(anon_sym_not);
      END_STATE();
    case 1283:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1284:
      ACCEPT_TOKEN(anon_sym_u00ac);
      END_STATE();
    case 1285:
      ACCEPT_TOKEN(anon_sym_and);
      END_STATE();
    case 1286:
      ACCEPT_TOKEN(anon_sym_u2227);
      END_STATE();
    case 1287:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 1288:
      ACCEPT_TOKEN(anon_sym_u2228);
      END_STATE();
    case 1289:
      ACCEPT_TOKEN(anon_sym_xor);
      END_STATE();
    case 1290:
      ACCEPT_TOKEN(anon_sym_u22bb);
      END_STATE();
    case 1291:
      ACCEPT_TOKEN(anon_sym_implies);
      END_STATE();
    case 1292:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_GT);
      END_STATE();
    case 1293:
      ACCEPT_TOKEN(anon_sym_u21d2);
      END_STATE();
    case 1294:
      ACCEPT_TOKEN(anon_sym_iff);
      END_STATE();
    case 1295:
      ACCEPT_TOKEN(anon_sym_LT_EQ_EQ_GT);
      END_STATE();
    case 1296:
      ACCEPT_TOKEN(anon_sym_u21d4);
      END_STATE();
    case 1297:
      ACCEPT_TOKEN(anon_sym_forall);
      END_STATE();
    case 1298:
      ACCEPT_TOKEN(anon_sym_forall);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1299:
      ACCEPT_TOKEN(anon_sym_u2200);
      END_STATE();
    case 1300:
      ACCEPT_TOKEN(anon_sym_exists);
      if (lookahead == '!') ADVANCE(1303);
      END_STATE();
    case 1301:
      ACCEPT_TOKEN(anon_sym_exists);
      if (lookahead == '!') ADVANCE(1303);
      if (lookahead == '_') ADVANCE(100);
      if (set_contains(sym_identifier_character_set_2, 223, lookahead)) ADVANCE(1057);
      END_STATE();
    case 1302:
      ACCEPT_TOKEN(anon_sym_u2203);
      if (lookahead == '!') ADVANCE(1304);
      END_STATE();
    case 1303:
      ACCEPT_TOKEN(anon_sym_exists_BANG);
      END_STATE();
    case 1304:
      ACCEPT_TOKEN(anon_sym_u2203_BANG);
      END_STATE();
    case 1305:
      ACCEPT_TOKEN(anon_sym_u222a);
      END_STATE();
    case 1306:
      ACCEPT_TOKEN(anon_sym_intersects);
      END_STATE();
    case 1307:
      ACCEPT_TOKEN(anon_sym_u2229);
      END_STATE();
    case 1308:
      ACCEPT_TOKEN(anon_sym_subset);
      if (lookahead == 'e') ADVANCE(483);
      END_STATE();
    case 1309:
      ACCEPT_TOKEN(anon_sym_u2282);
      END_STATE();
    case 1310:
      ACCEPT_TOKEN(anon_sym_subseteq);
      END_STATE();
    case 1311:
      ACCEPT_TOKEN(anon_sym_u2286);
      END_STATE();
    case 1312:
      ACCEPT_TOKEN(anon_sym_supset);
      if (lookahead == 'e') ADVANCE(484);
      END_STATE();
    case 1313:
      ACCEPT_TOKEN(anon_sym_u2283);
      END_STATE();
    case 1314:
      ACCEPT_TOKEN(anon_sym_supseteq);
      END_STATE();
    case 1315:
      ACCEPT_TOKEN(anon_sym_u2287);
      END_STATE();
    case 1316:
      ACCEPT_TOKEN(anon_sym_complement);
      END_STATE();
    case 1317:
      ACCEPT_TOKEN(anon_sym_u2216);
      END_STATE();
    case 1318:
      ACCEPT_TOKEN(anon_sym_product);
      END_STATE();
    case 1319:
      ACCEPT_TOKEN(anon_sym_u2a09);
      END_STATE();
    case 1320:
      ACCEPT_TOKEN(anon_sym_in);
      END_STATE();
    case 1321:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 't') ADVANCE(224);
      END_STATE();
    case 1322:
      ACCEPT_TOKEN(anon_sym_u2208);
      END_STATE();
    case 1323:
      ACCEPT_TOKEN(anon_sym_u2209);
      END_STATE();
    case 1324:
      ACCEPT_TOKEN(sym_seq_builder_separator);
      END_STATE();
    case 1325:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 1326:
      ACCEPT_TOKEN(anon_sym_u00d7);
      END_STATE();
    case 1327:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '=') ADVANCE(1274);
      END_STATE();
    case 1328:
      ACCEPT_TOKEN(anon_sym_u00f7);
      END_STATE();
    case 1329:
      ACCEPT_TOKEN(anon_sym_mod);
      END_STATE();
    case 1330:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 1331:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 1332:
      ACCEPT_TOKEN(sym_math_op_subtract);
      END_STATE();
    case 1333:
      ACCEPT_TOKEN(sym_math_op_subtract);
      if (lookahead == '>') ADVANCE(1339);
      END_STATE();
    case 1334:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 1335:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(1273);
      END_STATE();
    case 1336:
      ACCEPT_TOKEN(anon_sym_u2218);
      END_STATE();
    case 1337:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 1338:
      ACCEPT_TOKEN(anon_sym_u2254);
      END_STATE();
    case 1339:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 1340:
      ACCEPT_TOKEN(anon_sym_u2192);
      END_STATE();
    case 1341:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 1342:
      ACCEPT_TOKEN(anon_sym_u2190);
      END_STATE();
    case 1343:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (lookahead == 'U') ADVANCE(1351);
      if (lookahead == 'u') ADVANCE(1347);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1344:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1116);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1345:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1344);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1346:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1345);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1347:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1346);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1348:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1347);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1349:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1348);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1350:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1349);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1351:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1350);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1352:
      ACCEPT_TOKEN(sym_comment_aligned);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1352);
      END_STATE();
    case 1353:
      ACCEPT_TOKEN(sym_comment_local);
      if (lookahead == 'U') ADVANCE(1361);
      if (lookahead == 'u') ADVANCE(1357);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1354:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1355);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1355:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1117);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1356:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1354);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1357:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1356);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1358:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1357);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1359:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1358);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1360:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1359);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1361:
      ACCEPT_TOKEN(sym_comment_local);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1360);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1362:
      ACCEPT_TOKEN(sym_comment_local);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1362);
      END_STATE();
    case 1363:
      ACCEPT_TOKEN(sym_comment_page);
      if (lookahead == 'U') ADVANCE(1371);
      if (lookahead == 'u') ADVANCE(1367);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1364:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1118);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1365:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1364);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1366:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1365);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1367:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1366);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1368:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1367);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1369:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1368);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1370:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1369);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1371:
      ACCEPT_TOKEN(sym_comment_page);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1370);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    case 1372:
      ACCEPT_TOKEN(sym_comment_page);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(1372);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0},
  [1] = {.lex_state = 19},
  [2] = {.lex_state = 10},
  [3] = {.lex_state = 12},
  [4] = {.lex_state = 15},
  [5] = {.lex_state = 10},
  [6] = {.lex_state = 10},
  [7] = {.lex_state = 10},
  [8] = {.lex_state = 10},
  [9] = {.lex_state = 10},
  [10] = {.lex_state = 10},
  [11] = {.lex_state = 10},
  [12] = {.lex_state = 10},
  [13] = {.lex_state = 15},
  [14] = {.lex_state = 10},
  [15] = {.lex_state = 13},
  [16] = {.lex_state = 13},
  [17] = {.lex_state = 13},
  [18] = {.lex_state = 13},
  [19] = {.lex_state = 5},
  [20] = {.lex_state = 5},
  [21] = {.lex_state = 5},
  [22] = {.lex_state = 5},
  [23] = {.lex_state = 5},
  [24] = {.lex_state = 5},
  [25] = {.lex_state = 5},
  [26] = {.lex_state = 5},
  [27] = {.lex_state = 5},
  [28] = {.lex_state = 5},
  [29] = {.lex_state = 15},
  [30] = {.lex_state = 15},
  [31] = {.lex_state = 15},
  [32] = {.lex_state = 15},
  [33] = {.lex_state = 15},
  [34] = {.lex_state = 15},
  [35] = {.lex_state = 15},
  [36] = {.lex_state = 15},
  [37] = {.lex_state = 15},
  [38] = {.lex_state = 15},
  [39] = {.lex_state = 15},
  [40] = {.lex_state = 15},
  [41] = {.lex_state = 15},
  [42] = {.lex_state = 15},
  [43] = {.lex_state = 15},
  [44] = {.lex_state = 15},
  [45] = {.lex_state = 15},
  [46] = {.lex_state = 15},
  [47] = {.lex_state = 15},
  [48] = {.lex_state = 15},
  [49] = {.lex_state = 15},
  [50] = {.lex_state = 15},
  [51] = {.lex_state = 15},
  [52] = {.lex_state = 25},
  [53] = {.lex_state = 25},
  [54] = {.lex_state = 25},
  [55] = {.lex_state = 26},
  [56] = {.lex_state = 26},
  [57] = {.lex_state = 26},
  [58] = {.lex_state = 26},
  [59] = {.lex_state = 25},
  [60] = {.lex_state = 25},
  [61] = {.lex_state = 25},
  [62] = {.lex_state = 25},
  [63] = {.lex_state = 25},
  [64] = {.lex_state = 25},
  [65] = {.lex_state = 25},
  [66] = {.lex_state = 25},
  [67] = {.lex_state = 25},
  [68] = {.lex_state = 25},
  [69] = {.lex_state = 25},
  [70] = {.lex_state = 26},
  [71] = {.lex_state = 26},
  [72] = {.lex_state = 26},
  [73] = {.lex_state = 26},
  [74] = {.lex_state = 56},
  [75] = {.lex_state = 26},
  [76] = {.lex_state = 25},
  [77] = {.lex_state = 25},
  [78] = {.lex_state = 25},
  [79] = {.lex_state = 25},
  [80] = {.lex_state = 25},
  [81] = {.lex_state = 25},
  [82] = {.lex_state = 25},
  [83] = {.lex_state = 25},
  [84] = {.lex_state = 26},
  [85] = {.lex_state = 26},
  [86] = {.lex_state = 26},
  [87] = {.lex_state = 26},
  [88] = {.lex_state = 26},
  [89] = {.lex_state = 26},
  [90] = {.lex_state = 26},
  [91] = {.lex_state = 26},
  [92] = {.lex_state = 26},
  [93] = {.lex_state = 26},
  [94] = {.lex_state = 26},
  [95] = {.lex_state = 26},
  [96] = {.lex_state = 19},
  [97] = {.lex_state = 19},
  [98] = {.lex_state = 1},
  [99] = {.lex_state = 1},
  [100] = {.lex_state = 1},
  [101] = {.lex_state = 1},
  [102] = {.lex_state = 19},
  [103] = {.lex_state = 1},
  [104] = {.lex_state = 1},
  [105] = {.lex_state = 1},
  [106] = {.lex_state = 19},
  [107] = {.lex_state = 19},
  [108] = {.lex_state = 1},
  [109] = {.lex_state = 1},
  [110] = {.lex_state = 19},
  [111] = {.lex_state = 19},
  [112] = {.lex_state = 19},
  [113] = {.lex_state = 1},
  [114] = {.lex_state = 1},
  [115] = {.lex_state = 1},
  [116] = {.lex_state = 1},
  [117] = {.lex_state = 19},
  [118] = {.lex_state = 1},
  [119] = {.lex_state = 19},
  [120] = {.lex_state = 19},
  [121] = {.lex_state = 19},
  [122] = {.lex_state = 19},
  [123] = {.lex_state = 19},
  [124] = {.lex_state = 19},
  [125] = {.lex_state = 19},
  [126] = {.lex_state = 19},
  [127] = {.lex_state = 19},
  [128] = {.lex_state = 13},
  [129] = {.lex_state = 19},
  [130] = {.lex_state = 19},
  [131] = {.lex_state = 3},
  [132] = {.lex_state = 19},
  [133] = {.lex_state = 19},
  [134] = {.lex_state = 3},
  [135] = {.lex_state = 3},
  [136] = {.lex_state = 19},
  [137] = {.lex_state = 3},
  [138] = {.lex_state = 3},
  [139] = {.lex_state = 3},
  [140] = {.lex_state = 19},
  [141] = {.lex_state = 19},
  [142] = {.lex_state = 19},
  [143] = {.lex_state = 19},
  [144] = {.lex_state = 19},
  [145] = {.lex_state = 19},
  [146] = {.lex_state = 19},
  [147] = {.lex_state = 19},
  [148] = {.lex_state = 19},
  [149] = {.lex_state = 19},
  [150] = {.lex_state = 19},
  [151] = {.lex_state = 19},
  [152] = {.lex_state = 19},
  [153] = {.lex_state = 19},
  [154] = {.lex_state = 19},
  [155] = {.lex_state = 19},
  [156] = {.lex_state = 19},
  [157] = {.lex_state = 19},
  [158] = {.lex_state = 19},
  [159] = {.lex_state = 19},
  [160] = {.lex_state = 19},
  [161] = {.lex_state = 3},
  [162] = {.lex_state = 3},
  [163] = {.lex_state = 3},
  [164] = {.lex_state = 3},
  [165] = {.lex_state = 3},
  [166] = {.lex_state = 3},
  [167] = {.lex_state = 3},
  [168] = {.lex_state = 3},
  [169] = {.lex_state = 3},
  [170] = {.lex_state = 3},
  [171] = {.lex_state = 3},
  [172] = {.lex_state = 3},
  [173] = {.lex_state = 3},
  [174] = {.lex_state = 3},
  [175] = {.lex_state = 3},
  [176] = {.lex_state = 3},
  [177] = {.lex_state = 3},
  [178] = {.lex_state = 3},
  [179] = {.lex_state = 3},
  [180] = {.lex_state = 3},
  [181] = {.lex_state = 3},
  [182] = {.lex_state = 3},
  [183] = {.lex_state = 3},
  [184] = {.lex_state = 3},
  [185] = {.lex_state = 3},
  [186] = {.lex_state = 3},
  [187] = {.lex_state = 3},
  [188] = {.lex_state = 19},
  [189] = {.lex_state = 19},
  [190] = {.lex_state = 19},
  [191] = {.lex_state = 19},
  [192] = {.lex_state = 19},
  [193] = {.lex_state = 19},
  [194] = {.lex_state = 19},
  [195] = {.lex_state = 19},
  [196] = {.lex_state = 19},
  [197] = {.lex_state = 19},
  [198] = {.lex_state = 19},
  [199] = {.lex_state = 19},
  [200] = {.lex_state = 19},
  [201] = {.lex_state = 19},
  [202] = {.lex_state = 3},
  [203] = {.lex_state = 3},
  [204] = {.lex_state = 3},
  [205] = {.lex_state = 3},
  [206] = {.lex_state = 3},
  [207] = {.lex_state = 3},
  [208] = {.lex_state = 3},
  [209] = {.lex_state = 5},
  [210] = {.lex_state = 5},
  [211] = {.lex_state = 5},
  [212] = {.lex_state = 5},
  [213] = {.lex_state = 5},
  [214] = {.lex_state = 5},
  [215] = {.lex_state = 5},
  [216] = {.lex_state = 3},
  [217] = {.lex_state = 3},
  [218] = {.lex_state = 3},
  [219] = {.lex_state = 3},
  [220] = {.lex_state = 3},
  [221] = {.lex_state = 0},
  [222] = {.lex_state = 0},
  [223] = {.lex_state = 0},
  [224] = {.lex_state = 3},
  [225] = {.lex_state = 3},
  [226] = {.lex_state = 3},
  [227] = {.lex_state = 3},
  [228] = {.lex_state = 3},
  [229] = {.lex_state = 3},
  [230] = {.lex_state = 3},
  [231] = {.lex_state = 3},
  [232] = {.lex_state = 3},
  [233] = {.lex_state = 3},
  [234] = {.lex_state = 3},
  [235] = {.lex_state = 3},
  [236] = {.lex_state = 3},
  [237] = {.lex_state = 3},
  [238] = {.lex_state = 3},
  [239] = {.lex_state = 3},
  [240] = {.lex_state = 3},
  [241] = {.lex_state = 3},
  [242] = {.lex_state = 3},
  [243] = {.lex_state = 3},
  [244] = {.lex_state = 3},
  [245] = {.lex_state = 3},
  [246] = {.lex_state = 3},
  [247] = {.lex_state = 3},
  [248] = {.lex_state = 3},
  [249] = {.lex_state = 3},
  [250] = {.lex_state = 3},
  [251] = {.lex_state = 3},
  [252] = {.lex_state = 0},
  [253] = {.lex_state = 8},
  [254] = {.lex_state = 1},
  [255] = {.lex_state = 3},
  [256] = {.lex_state = 1},
  [257] = {.lex_state = 1},
  [258] = {.lex_state = 1},
  [259] = {.lex_state = 1},
  [260] = {.lex_state = 19},
  [261] = {.lex_state = 19},
  [262] = {.lex_state = 1},
  [263] = {.lex_state = 1},
  [264] = {.lex_state = 1},
  [265] = {.lex_state = 19},
  [266] = {.lex_state = 1},
  [267] = {.lex_state = 1},
  [268] = {.lex_state = 1},
  [269] = {.lex_state = 3},
  [270] = {.lex_state = 1},
  [271] = {.lex_state = 1},
  [272] = {.lex_state = 19},
  [273] = {.lex_state = 1},
  [274] = {.lex_state = 1},
  [275] = {.lex_state = 1},
  [276] = {.lex_state = 1},
  [277] = {.lex_state = 1},
  [278] = {.lex_state = 1},
  [279] = {.lex_state = 1},
  [280] = {.lex_state = 1},
  [281] = {.lex_state = 1},
  [282] = {.lex_state = 1},
  [283] = {.lex_state = 1},
  [284] = {.lex_state = 1},
  [285] = {.lex_state = 1},
  [286] = {.lex_state = 1},
  [287] = {.lex_state = 1},
  [288] = {.lex_state = 1},
  [289] = {.lex_state = 1},
  [290] = {.lex_state = 1},
  [291] = {.lex_state = 1},
  [292] = {.lex_state = 1},
  [293] = {.lex_state = 1},
  [294] = {.lex_state = 1},
  [295] = {.lex_state = 1},
  [296] = {.lex_state = 6},
  [297] = {.lex_state = 19},
  [298] = {.lex_state = 3},
  [299] = {.lex_state = 3},
  [300] = {.lex_state = 0},
  [301] = {.lex_state = 19},
  [302] = {.lex_state = 6},
  [303] = {.lex_state = 19},
  [304] = {.lex_state = 3},
  [305] = {.lex_state = 3},
  [306] = {.lex_state = 6},
  [307] = {.lex_state = 3},
  [308] = {.lex_state = 19},
  [309] = {.lex_state = 3},
  [310] = {.lex_state = 3},
  [311] = {.lex_state = 3},
  [312] = {.lex_state = 15},
  [313] = {.lex_state = 3},
  [314] = {.lex_state = 0},
  [315] = {.lex_state = 19},
  [316] = {.lex_state = 19},
  [317] = {.lex_state = 19},
  [318] = {.lex_state = 3},
  [319] = {.lex_state = 3},
  [320] = {.lex_state = 6},
  [321] = {.lex_state = 6},
  [322] = {.lex_state = 0},
  [323] = {.lex_state = 1},
  [324] = {.lex_state = 1},
  [325] = {.lex_state = 1},
  [326] = {.lex_state = 1},
  [327] = {.lex_state = 1},
  [328] = {.lex_state = 3},
  [329] = {.lex_state = 3},
  [330] = {.lex_state = 1},
  [331] = {.lex_state = 1},
  [332] = {.lex_state = 1},
  [333] = {.lex_state = 0},
  [334] = {.lex_state = 1},
  [335] = {.lex_state = 19},
  [336] = {.lex_state = 0},
  [337] = {.lex_state = 1},
  [338] = {.lex_state = 1},
  [339] = {.lex_state = 1},
  [340] = {.lex_state = 3},
  [341] = {.lex_state = 19},
  [342] = {.lex_state = 1},
  [343] = {.lex_state = 1},
  [344] = {.lex_state = 1},
  [345] = {.lex_state = 1},
  [346] = {.lex_state = 1},
  [347] = {.lex_state = 19},
  [348] = {.lex_state = 1},
  [349] = {.lex_state = 1},
  [350] = {.lex_state = 1},
  [351] = {.lex_state = 19},
  [352] = {.lex_state = 1},
  [353] = {.lex_state = 19},
  [354] = {.lex_state = 1},
  [355] = {.lex_state = 3},
  [356] = {.lex_state = 19},
  [357] = {.lex_state = 19},
  [358] = {.lex_state = 19},
  [359] = {.lex_state = 0},
  [360] = {.lex_state = 0},
  [361] = {.lex_state = 19},
  [362] = {.lex_state = 19},
  [363] = {.lex_state = 19},
  [364] = {.lex_state = 0},
  [365] = {.lex_state = 0},
  [366] = {.lex_state = 19},
  [367] = {.lex_state = 19},
  [368] = {.lex_state = 19},
  [369] = {.lex_state = 19},
  [370] = {.lex_state = 19},
  [371] = {.lex_state = 19},
  [372] = {.lex_state = 19},
  [373] = {.lex_state = 19},
  [374] = {.lex_state = 1},
  [375] = {.lex_state = 0},
  [376] = {.lex_state = 1},
  [377] = {.lex_state = 19},
  [378] = {.lex_state = 3},
  [379] = {.lex_state = 3},
  [380] = {.lex_state = 3},
  [381] = {.lex_state = 3},
  [382] = {.lex_state = 3},
  [383] = {.lex_state = 3},
  [384] = {.lex_state = 3},
  [385] = {.lex_state = 3},
  [386] = {.lex_state = 3},
  [387] = {.lex_state = 3},
  [388] = {.lex_state = 3},
  [389] = {.lex_state = 3},
  [390] = {.lex_state = 3},
  [391] = {.lex_state = 3},
  [392] = {.lex_state = 3},
  [393] = {.lex_state = 3},
  [394] = {.lex_state = 3},
  [395] = {.lex_state = 3},
  [396] = {.lex_state = 3},
  [397] = {.lex_state = 3},
  [398] = {.lex_state = 3},
  [399] = {.lex_state = 3},
  [400] = {.lex_state = 3},
  [401] = {.lex_state = 3},
  [402] = {.lex_state = 3},
  [403] = {.lex_state = 3},
  [404] = {.lex_state = 3},
  [405] = {.lex_state = 19},
  [406] = {.lex_state = 0},
  [407] = {.lex_state = 3},
  [408] = {.lex_state = 19},
  [409] = {.lex_state = 0},
  [410] = {.lex_state = 0},
  [411] = {.lex_state = 3},
  [412] = {.lex_state = 19},
  [413] = {.lex_state = 0},
  [414] = {.lex_state = 19},
  [415] = {.lex_state = 19},
  [416] = {.lex_state = 0},
  [417] = {.lex_state = 3},
  [418] = {.lex_state = 19},
  [419] = {.lex_state = 0},
  [420] = {.lex_state = 19},
  [421] = {.lex_state = 19},
  [422] = {.lex_state = 0},
  [423] = {.lex_state = 0},
  [424] = {.lex_state = 0},
  [425] = {.lex_state = 0},
  [426] = {.lex_state = 19},
  [427] = {.lex_state = 0},
  [428] = {.lex_state = 19},
  [429] = {.lex_state = 0},
  [430] = {.lex_state = 0},
  [431] = {.lex_state = 0},
  [432] = {.lex_state = 19},
  [433] = {.lex_state = 3},
  [434] = {.lex_state = 13},
  [435] = {.lex_state = 0},
  [436] = {.lex_state = 0},
  [437] = {.lex_state = 0},
  [438] = {.lex_state = 0},
  [439] = {.lex_state = 0},
  [440] = {.lex_state = 0},
  [441] = {.lex_state = 0},
  [442] = {.lex_state = 0},
  [443] = {.lex_state = 0},
  [444] = {.lex_state = 0},
  [445] = {.lex_state = 0},
  [446] = {.lex_state = 0},
  [447] = {.lex_state = 0},
  [448] = {.lex_state = 0},
  [449] = {.lex_state = 0},
  [450] = {.lex_state = 0},
  [451] = {.lex_state = 0},
  [452] = {.lex_state = 0},
  [453] = {.lex_state = 3},
  [454] = {.lex_state = 0},
  [455] = {.lex_state = 0},
  [456] = {.lex_state = 0},
  [457] = {.lex_state = 0},
  [458] = {.lex_state = 0},
  [459] = {.lex_state = 0},
  [460] = {.lex_state = 0},
  [461] = {.lex_state = 0},
  [462] = {.lex_state = 0},
  [463] = {.lex_state = 0},
  [464] = {.lex_state = 21},
  [465] = {.lex_state = 0},
  [466] = {.lex_state = 21},
  [467] = {.lex_state = 19},
  [468] = {.lex_state = 22},
  [469] = {.lex_state = 0},
  [470] = {.lex_state = 0},
  [471] = {.lex_state = 0},
  [472] = {.lex_state = 0},
  [473] = {.lex_state = 0},
  [474] = {.lex_state = 0},
  [475] = {.lex_state = 0},
  [476] = {.lex_state = 0},
  [477] = {.lex_state = 0},
  [478] = {.lex_state = 0},
  [479] = {.lex_state = 0},
  [480] = {.lex_state = 0},
  [481] = {.lex_state = 0},
  [482] = {.lex_state = 0},
  [483] = {.lex_state = 0},
  [484] = {.lex_state = 0},
  [485] = {.lex_state = 0},
  [486] = {.lex_state = 0},
  [487] = {.lex_state = 0},
  [488] = {.lex_state = 0},
  [489] = {.lex_state = 0},
  [490] = {.lex_state = 0},
  [491] = {.lex_state = 0},
  [492] = {.lex_state = 0},
  [493] = {.lex_state = 0},
  [494] = {.lex_state = 0},
  [495] = {.lex_state = 0},
  [496] = {.lex_state = 0},
  [497] = {.lex_state = 0},
  [498] = {.lex_state = 0},
  [499] = {.lex_state = 0},
  [500] = {.lex_state = 0},
  [501] = {.lex_state = 0},
  [502] = {.lex_state = 0},
  [503] = {.lex_state = 0},
  [504] = {.lex_state = 0},
  [505] = {.lex_state = 0},
  [506] = {.lex_state = 0},
  [507] = {.lex_state = 0},
  [508] = {.lex_state = 0},
  [509] = {.lex_state = 22},
  [510] = {.lex_state = 0},
  [511] = {.lex_state = 0},
  [512] = {.lex_state = 6},
  [513] = {.lex_state = 0},
  [514] = {.lex_state = 0},
  [515] = {.lex_state = 0},
  [516] = {.lex_state = 0},
  [517] = {.lex_state = 0},
  [518] = {.lex_state = 0},
  [519] = {.lex_state = 19},
  [520] = {.lex_state = 0},
  [521] = {.lex_state = 0},
  [522] = {.lex_state = 0},
  [523] = {.lex_state = 0},
  [524] = {.lex_state = 0},
  [525] = {.lex_state = 0},
  [526] = {.lex_state = 0},
  [527] = {.lex_state = 0},
  [528] = {.lex_state = 0},
  [529] = {.lex_state = 0},
  [530] = {.lex_state = 0},
  [531] = {.lex_state = 0},
  [532] = {.lex_state = 0},
  [533] = {.lex_state = 0},
  [534] = {.lex_state = 0},
  [535] = {.lex_state = 0},
  [536] = {.lex_state = 0},
  [537] = {.lex_state = 0},
  [538] = {.lex_state = 0},
  [539] = {.lex_state = 0},
  [540] = {.lex_state = 6},
  [541] = {.lex_state = 0},
  [542] = {.lex_state = 0},
  [543] = {.lex_state = 6},
  [544] = {.lex_state = 0},
  [545] = {.lex_state = 0},
  [546] = {.lex_state = 0},
  [547] = {.lex_state = 0},
  [548] = {.lex_state = 0},
  [549] = {.lex_state = 22},
  [550] = {.lex_state = 19},
  [551] = {.lex_state = 19},
  [552] = {.lex_state = 19},
  [553] = {.lex_state = 47},
  [554] = {.lex_state = 47},
  [555] = {.lex_state = 0},
  [556] = {.lex_state = 47},
  [557] = {.lex_state = 47},
  [558] = {.lex_state = 19},
  [559] = {.lex_state = 22},
  [560] = {.lex_state = 21},
  [561] = {.lex_state = 19},
  [562] = {.lex_state = 71},
  [563] = {.lex_state = 71},
  [564] = {.lex_state = 71},
  [565] = {.lex_state = 19},
  [566] = {.lex_state = 19},
  [567] = {.lex_state = 46},
  [568] = {.lex_state = 22},
  [569] = {.lex_state = 19},
  [570] = {.lex_state = 19},
  [571] = {.lex_state = 31},
  [572] = {.lex_state = 10},
  [573] = {.lex_state = 10},
  [574] = {.lex_state = 19},
  [575] = {.lex_state = 49},
  [576] = {.lex_state = 49},
  [577] = {.lex_state = 19},
  [578] = {.lex_state = 10},
  [579] = {.lex_state = 49},
  [580] = {.lex_state = 49},
  [581] = {.lex_state = 49},
  [582] = {.lex_state = 49},
  [583] = {.lex_state = 49},
  [584] = {.lex_state = 49},
  [585] = {.lex_state = 49},
  [586] = {.lex_state = 32},
  [587] = {.lex_state = 21},
  [588] = {.lex_state = 46},
  [589] = {.lex_state = 21},
  [590] = {.lex_state = 49},
  [591] = {.lex_state = 21},
  [592] = {.lex_state = 21},
  [593] = {.lex_state = 21},
  [594] = {.lex_state = 21},
  [595] = {.lex_state = 21},
  [596] = {.lex_state = 17},
  [597] = {.lex_state = 21},
  [598] = {.lex_state = 21},
  [599] = {.lex_state = 19},
  [600] = {.lex_state = 49},
  [601] = {.lex_state = 19},
  [602] = {.lex_state = 49},
  [603] = {.lex_state = 19},
  [604] = {.lex_state = 49},
  [605] = {.lex_state = 48},
  [606] = {.lex_state = 49},
  [607] = {.lex_state = 49},
  [608] = {.lex_state = 21},
  [609] = {.lex_state = 22},
  [610] = {.lex_state = 21},
  [611] = {.lex_state = 17},
  [612] = {.lex_state = 22},
  [613] = {.lex_state = 22},
  [614] = {.lex_state = 22},
  [615] = {.lex_state = 22},
  [616] = {.lex_state = 68},
  [617] = {.lex_state = 46},
  [618] = {.lex_state = 10},
  [619] = {.lex_state = 17},
  [620] = {.lex_state = 10},
  [621] = {.lex_state = 42},
  [622] = {.lex_state = 22},
  [623] = {.lex_state = 47},
  [624] = {.lex_state = 17},
  [625] = {.lex_state = 68},
  [626] = {.lex_state = 52},
  [627] = {.lex_state = 17},
  [628] = {.lex_state = 44},
  [629] = {.lex_state = 22},
  [630] = {.lex_state = 17},
  [631] = {.lex_state = 44},
  [632] = {.lex_state = 22},
  [633] = {.lex_state = 68},
  [634] = {.lex_state = 21},
  [635] = {.lex_state = 21},
  [636] = {.lex_state = 10},
  [637] = {.lex_state = 17},
  [638] = {.lex_state = 44},
  [639] = {.lex_state = 6},
  [640] = {.lex_state = 0},
  [641] = {.lex_state = 0},
  [642] = {.lex_state = 0},
  [643] = {.lex_state = 48},
  [644] = {.lex_state = 19},
  [645] = {.lex_state = 17},
  [646] = {.lex_state = 19},
  [647] = {.lex_state = 0},
  [648] = {.lex_state = 19},
  [649] = {.lex_state = 44},
  [650] = {.lex_state = 19},
  [651] = {.lex_state = 22},
  [652] = {.lex_state = 22},
  [653] = {.lex_state = 19},
  [654] = {.lex_state = 10},
  [655] = {.lex_state = 10},
  [656] = {.lex_state = 46},
  [657] = {.lex_state = 17},
  [658] = {.lex_state = 43},
  [659] = {.lex_state = 21},
  [660] = {.lex_state = 21},
  [661] = {.lex_state = 48},
  [662] = {.lex_state = 0},
  [663] = {.lex_state = 21},
  [664] = {.lex_state = 17},
  [665] = {.lex_state = 53},
  [666] = {.lex_state = 21},
  [667] = {.lex_state = 44},
  [668] = {.lex_state = 21},
  [669] = {.lex_state = 45},
  [670] = {.lex_state = 21},
  [671] = {.lex_state = 21},
  [672] = {.lex_state = 59},
  [673] = {.lex_state = 21},
  [674] = {.lex_state = 21},
  [675] = {.lex_state = 21},
  [676] = {.lex_state = 21},
  [677] = {.lex_state = 21},
  [678] = {.lex_state = 21},
  [679] = {.lex_state = 17},
  [680] = {.lex_state = 44},
  [681] = {.lex_state = 21},
  [682] = {.lex_state = 21},
  [683] = {.lex_state = 21},
  [684] = {.lex_state = 17},
  [685] = {.lex_state = 50},
  [686] = {.lex_state = 48},
  [687] = {.lex_state = 45},
  [688] = {.lex_state = 47},
  [689] = {.lex_state = 45},
  [690] = {.lex_state = 19},
  [691] = {.lex_state = 43},
  [692] = {.lex_state = 44},
  [693] = {.lex_state = 22},
  [694] = {.lex_state = 44},
  [695] = {.lex_state = 22},
  [696] = {.lex_state = 17},
  [697] = {.lex_state = 72},
  [698] = {.lex_state = 17},
  [699] = {.lex_state = 0},
  [700] = {.lex_state = 0},
  [701] = {.lex_state = 17},
  [702] = {.lex_state = 21},
  [703] = {.lex_state = 21},
  [704] = {.lex_state = 21},
  [705] = {.lex_state = 72},
  [706] = {.lex_state = 22},
  [707] = {.lex_state = 17},
  [708] = {.lex_state = 21},
  [709] = {.lex_state = 17},
  [710] = {.lex_state = 17},
  [711] = {.lex_state = 17},
  [712] = {.lex_state = 17},
  [713] = {.lex_state = 44},
  [714] = {.lex_state = 21},
  [715] = {.lex_state = 49},
  [716] = {.lex_state = 47},
  [717] = {.lex_state = 22},
  [718] = {.lex_state = 22},
  [719] = {.lex_state = 22},
  [720] = {.lex_state = 22},
  [721] = {.lex_state = 17},
  [722] = {.lex_state = 17},
  [723] = {.lex_state = 22},
  [724] = {.lex_state = 17},
  [725] = {.lex_state = 17},
  [726] = {.lex_state = 0},
  [727] = {.lex_state = 69},
  [728] = {.lex_state = 17},
  [729] = {.lex_state = 17},
  [730] = {.lex_state = 0},
  [731] = {.lex_state = 22},
  [732] = {.lex_state = 22},
  [733] = {.lex_state = 0},
  [734] = {.lex_state = 0},
  [735] = {.lex_state = 47},
  [736] = {.lex_state = 17},
  [737] = {.lex_state = 17},
  [738] = {.lex_state = 47},
  [739] = {.lex_state = 60},
  [740] = {.lex_state = 63},
  [741] = {.lex_state = 47},
  [742] = {.lex_state = 17},
  [743] = {.lex_state = 47},
  [744] = {.lex_state = 17},
  [745] = {.lex_state = 47},
  [746] = {.lex_state = 22},
  [747] = {.lex_state = 45},
  [748] = {.lex_state = 47},
  [749] = {.lex_state = 43},
  [750] = {.lex_state = 17},
  [751] = {.lex_state = 45},
  [752] = {.lex_state = 0},
  [753] = {.lex_state = 22},
  [754] = {.lex_state = 22},
  [755] = {.lex_state = 17},
  [756] = {.lex_state = 47},
  [757] = {.lex_state = 0},
  [758] = {.lex_state = 0},
  [759] = {.lex_state = 17},
  [760] = {.lex_state = 47},
  [761] = {.lex_state = 17},
  [762] = {.lex_state = 47},
  [763] = {.lex_state = 17},
  [764] = {.lex_state = 17},
  [765] = {.lex_state = 17},
  [766] = {.lex_state = 17},
  [767] = {.lex_state = 22},
  [768] = {.lex_state = 47},
  [769] = {.lex_state = 0},
  [770] = {.lex_state = 47},
  [771] = {.lex_state = 0},
  [772] = {.lex_state = 0},
  [773] = {.lex_state = 47},
  [774] = {.lex_state = 17},
  [775] = {.lex_state = 0},
  [776] = {.lex_state = 17},
  [777] = {.lex_state = 17},
  [778] = {.lex_state = 22},
  [779] = {.lex_state = 47},
  [780] = {.lex_state = 43},
  [781] = {.lex_state = 0},
  [782] = {.lex_state = 22},
  [783] = {.lex_state = 22},
  [784] = {.lex_state = 17},
  [785] = {.lex_state = 22},
  [786] = {.lex_state = 0},
  [787] = {.lex_state = 17},
  [788] = {.lex_state = 22},
  [789] = {.lex_state = 22},
  [790] = {.lex_state = 22},
  [791] = {.lex_state = 0},
  [792] = {.lex_state = 22},
  [793] = {.lex_state = 0},
  [794] = {.lex_state = 47},
  [795] = {.lex_state = 0},
  [796] = {.lex_state = 47},
  [797] = {.lex_state = 0},
  [798] = {.lex_state = 68},
  [799] = {.lex_state = 0},
  [800] = {.lex_state = 68},
  [801] = {.lex_state = 17},
  [802] = {.lex_state = 45},
  [803] = {.lex_state = 0},
  [804] = {.lex_state = 0},
  [805] = {.lex_state = 63},
  [806] = {.lex_state = 0},
  [807] = {.lex_state = 17},
  [808] = {.lex_state = 22},
  [809] = {.lex_state = 17},
  [810] = {.lex_state = 17},
  [811] = {.lex_state = 22},
  [812] = {.lex_state = 68},
  [813] = {.lex_state = 45},
  [814] = {.lex_state = 47},
  [815] = {.lex_state = 45},
  [816] = {.lex_state = 19},
  [817] = {.lex_state = 17},
  [818] = {.lex_state = 17},
  [819] = {.lex_state = 17},
  [820] = {.lex_state = 0},
  [821] = {.lex_state = 49},
  [822] = {.lex_state = 49},
  [823] = {.lex_state = 0},
  [824] = {.lex_state = 17},
  [825] = {.lex_state = 49},
  [826] = {.lex_state = 17},
  [827] = {.lex_state = 10},
  [828] = {.lex_state = 17},
  [829] = {.lex_state = 49},
  [830] = {.lex_state = 49},
  [831] = {.lex_state = 49},
  [832] = {.lex_state = 10},
  [833] = {.lex_state = 49},
  [834] = {.lex_state = 0},
  [835] = {.lex_state = 10},
  [836] = {.lex_state = 17},
  [837] = {.lex_state = 0},
  [838] = {.lex_state = 49},
  [839] = {.lex_state = 17},
  [840] = {.lex_state = 0},
  [841] = {.lex_state = 0},
  [842] = {.lex_state = 10},
  [843] = {.lex_state = 49},
  [844] = {.lex_state = 49},
  [845] = {.lex_state = 49},
  [846] = {.lex_state = 17},
  [847] = {.lex_state = 57},
  [848] = {.lex_state = 10},
  [849] = {.lex_state = 5},
  [850] = {.lex_state = 49},
  [851] = {.lex_state = 17},
  [852] = {.lex_state = 49},
  [853] = {.lex_state = 42},
  [854] = {.lex_state = 42},
  [855] = {.lex_state = 17},
  [856] = {.lex_state = 57},
  [857] = {.lex_state = 17},
  [858] = {.lex_state = 42},
  [859] = {.lex_state = 17},
  [860] = {.lex_state = 57},
  [861] = {.lex_state = 17},
  [862] = {.lex_state = 17},
  [863] = {.lex_state = 10},
  [864] = {.lex_state = 10},
  [865] = {.lex_state = 10},
  [866] = {.lex_state = 10},
  [867] = {.lex_state = 17},
  [868] = {.lex_state = 10},
  [869] = {.lex_state = 42},
  [870] = {.lex_state = 57},
  [871] = {.lex_state = 17},
  [872] = {.lex_state = 17},
  [873] = {.lex_state = 57},
  [874] = {.lex_state = 17},
  [875] = {.lex_state = 17},
  [876] = {.lex_state = 57},
  [877] = {.lex_state = 17},
  [878] = {.lex_state = 70},
  [879] = {.lex_state = 10},
  [880] = {.lex_state = 17},
  [881] = {.lex_state = 10},
  [882] = {.lex_state = 57},
  [883] = {.lex_state = 19},
  [884] = {.lex_state = 57},
  [885] = {.lex_state = 10},
  [886] = {.lex_state = 17},
  [887] = {.lex_state = 17},
  [888] = {.lex_state = 49},
  [889] = {.lex_state = 49},
  [890] = {.lex_state = 57},
  [891] = {.lex_state = 57},
  [892] = {.lex_state = 57},
  [893] = {.lex_state = 57},
  [894] = {.lex_state = 17},
  [895] = {.lex_state = 49},
  [896] = {.lex_state = 17},
  [897] = {.lex_state = 17},
  [898] = {.lex_state = 17},
  [899] = {.lex_state = 42},
  [900] = {.lex_state = 17},
  [901] = {.lex_state = 17},
  [902] = {.lex_state = 17},
  [903] = {.lex_state = 17},
  [904] = {.lex_state = 57},
  [905] = {.lex_state = 10},
  [906] = {.lex_state = 10},
  [907] = {.lex_state = 17},
  [908] = {.lex_state = 0},
  [909] = {.lex_state = 0},
  [910] = {.lex_state = 17},
  [911] = {.lex_state = 0},
  [912] = {.lex_state = 19},
  [913] = {.lex_state = 0},
  [914] = {.lex_state = 17},
  [915] = {.lex_state = 0},
  [916] = {.lex_state = 19},
  [917] = {.lex_state = 0},
  [918] = {.lex_state = 0},
  [919] = {.lex_state = 68},
  [920] = {.lex_state = 10},
  [921] = {.lex_state = 0},
  [922] = {.lex_state = 0},
  [923] = {.lex_state = 0},
  [924] = {.lex_state = 0},
  [925] = {.lex_state = 0},
  [926] = {.lex_state = 0},
  [927] = {.lex_state = 17},
  [928] = {.lex_state = 0},
  [929] = {.lex_state = 0},
  [930] = {.lex_state = 19},
  [931] = {.lex_state = 17},
  [932] = {.lex_state = 57},
  [933] = {.lex_state = 0},
  [934] = {.lex_state = 0},
  [935] = {.lex_state = 17},
  [936] = {.lex_state = 17},
  [937] = {.lex_state = 0},
  [938] = {.lex_state = 19},
  [939] = {.lex_state = 5},
  [940] = {.lex_state = 17},
  [941] = {.lex_state = 0},
  [942] = {.lex_state = 19},
  [943] = {.lex_state = 19},
  [944] = {.lex_state = 68},
  [945] = {.lex_state = 0},
  [946] = {.lex_state = 0},
  [947] = {.lex_state = 0},
  [948] = {.lex_state = 17},
  [949] = {.lex_state = 0},
  [950] = {.lex_state = 0},
  [951] = {.lex_state = 0},
  [952] = {.lex_state = 17},
  [953] = {.lex_state = 68},
  [954] = {.lex_state = 19},
  [955] = {.lex_state = 0},
  [956] = {.lex_state = 0},
  [957] = {.lex_state = 0},
  [958] = {.lex_state = 17},
  [959] = {.lex_state = 17},
  [960] = {.lex_state = 17},
  [961] = {.lex_state = 0},
  [962] = {.lex_state = 0},
  [963] = {.lex_state = 0},
  [964] = {.lex_state = 0},
  [965] = {.lex_state = 0},
  [966] = {.lex_state = 0},
  [967] = {.lex_state = 19},
  [968] = {.lex_state = 17},
  [969] = {.lex_state = 68},
  [970] = {.lex_state = 0},
  [971] = {.lex_state = 0},
  [972] = {.lex_state = 19},
  [973] = {.lex_state = 68},
  [974] = {.lex_state = 0},
  [975] = {.lex_state = 0},
  [976] = {.lex_state = 17},
  [977] = {.lex_state = 17},
  [978] = {.lex_state = 17},
  [979] = {.lex_state = 68},
  [980] = {.lex_state = 17},
  [981] = {.lex_state = 19},
  [982] = {.lex_state = 0},
  [983] = {.lex_state = 0},
  [984] = {.lex_state = 19},
  [985] = {.lex_state = 19},
  [986] = {.lex_state = 17},
  [987] = {.lex_state = 0},
  [988] = {.lex_state = 0},
  [989] = {.lex_state = 0},
  [990] = {.lex_state = 17},
  [991] = {.lex_state = 17},
  [992] = {.lex_state = 0},
  [993] = {.lex_state = 0},
  [994] = {.lex_state = 0},
  [995] = {.lex_state = 0},
  [996] = {.lex_state = 1112},
  [997] = {.lex_state = 17},
  [998] = {.lex_state = 0},
  [999] = {.lex_state = 0},
  [1000] = {.lex_state = 0},
  [1001] = {.lex_state = 17},
  [1002] = {.lex_state = 0},
  [1003] = {.lex_state = 17},
  [1004] = {.lex_state = 0},
  [1005] = {.lex_state = 19},
  [1006] = {.lex_state = 0},
  [1007] = {.lex_state = 0},
  [1008] = {.lex_state = 17},
  [1009] = {.lex_state = 17},
  [1010] = {.lex_state = 0},
  [1011] = {.lex_state = 0},
  [1012] = {.lex_state = 0},
  [1013] = {.lex_state = 0},
  [1014] = {.lex_state = 17},
  [1015] = {.lex_state = 0},
  [1016] = {.lex_state = 0},
  [1017] = {.lex_state = 17},
  [1018] = {.lex_state = 17},
  [1019] = {.lex_state = 0},
  [1020] = {.lex_state = 0},
  [1021] = {.lex_state = 0},
  [1022] = {.lex_state = 0},
  [1023] = {.lex_state = 0},
  [1024] = {.lex_state = 0},
  [1025] = {.lex_state = 0},
  [1026] = {.lex_state = 0},
  [1027] = {.lex_state = 17},
  [1028] = {.lex_state = 19},
  [1029] = {.lex_state = 17},
  [1030] = {.lex_state = 17},
  [1031] = {.lex_state = 17},
  [1032] = {.lex_state = 17},
  [1033] = {.lex_state = 0},
  [1034] = {.lex_state = 17},
  [1035] = {.lex_state = 17},
  [1036] = {.lex_state = 0},
  [1037] = {.lex_state = 17},
  [1038] = {.lex_state = 17},
  [1039] = {.lex_state = 0},
  [1040] = {.lex_state = 17},
  [1041] = {.lex_state = 0},
  [1042] = {.lex_state = 0},
  [1043] = {.lex_state = 0},
  [1044] = {.lex_state = 17},
  [1045] = {.lex_state = 0},
  [1046] = {.lex_state = 0},
  [1047] = {.lex_state = 0},
  [1048] = {.lex_state = 17},
  [1049] = {.lex_state = 0},
  [1050] = {.lex_state = 0},
  [1051] = {.lex_state = 0},
  [1052] = {.lex_state = 17},
  [1053] = {.lex_state = 0},
  [1054] = {.lex_state = 17},
  [1055] = {.lex_state = 0},
  [1056] = {.lex_state = 0},
  [1057] = {.lex_state = 0},
  [1058] = {.lex_state = 19},
  [1059] = {.lex_state = 17},
  [1060] = {.lex_state = 17},
  [1061] = {.lex_state = 0},
  [1062] = {.lex_state = 19},
  [1063] = {.lex_state = 17},
  [1064] = {.lex_state = 17},
  [1065] = {.lex_state = 0},
  [1066] = {.lex_state = 19},
  [1067] = {.lex_state = 17},
  [1068] = {.lex_state = 17},
  [1069] = {.lex_state = 17},
  [1070] = {.lex_state = 17},
  [1071] = {.lex_state = 17},
  [1072] = {.lex_state = 17},
  [1073] = {.lex_state = 17},
  [1074] = {.lex_state = 17},
  [1075] = {.lex_state = 17},
  [1076] = {.lex_state = 1112},
  [1077] = {.lex_state = 17},
  [1078] = {.lex_state = 17},
  [1079] = {.lex_state = 17},
  [1080] = {.lex_state = 0},
  [1081] = {.lex_state = 0},
  [1082] = {.lex_state = 17},
  [1083] = {.lex_state = 17},
  [1084] = {.lex_state = 0},
  [1085] = {.lex_state = 17},
  [1086] = {.lex_state = 17},
  [1087] = {.lex_state = 0},
  [1088] = {.lex_state = 0},
  [1089] = {.lex_state = 0},
  [1090] = {.lex_state = 0},
  [1091] = {.lex_state = 17},
  [1092] = {.lex_state = 0},
  [1093] = {.lex_state = 0},
  [1094] = {.lex_state = 1112},
  [1095] = {.lex_state = 17},
  [1096] = {.lex_state = 17},
  [1097] = {.lex_state = 0},
  [1098] = {.lex_state = 0},
  [1099] = {.lex_state = 0},
  [1100] = {.lex_state = 17},
  [1101] = {.lex_state = 0},
  [1102] = {.lex_state = 0},
  [1103] = {.lex_state = 0},
  [1104] = {.lex_state = 0},
  [1105] = {.lex_state = 17},
  [1106] = {.lex_state = 0},
  [1107] = {.lex_state = 17},
  [1108] = {.lex_state = 17},
  [1109] = {.lex_state = 1112},
  [1110] = {.lex_state = 17},
  [1111] = {.lex_state = 0},
  [1112] = {.lex_state = 0},
  [1113] = {.lex_state = 17},
  [1114] = {.lex_state = 17},
  [1115] = {.lex_state = 17},
  [1116] = {.lex_state = 0},
  [1117] = {.lex_state = 0},
  [1118] = {.lex_state = 17},
  [1119] = {.lex_state = 1112},
  [1120] = {.lex_state = 1112},
  [1121] = {.lex_state = 1112},
  [1122] = {.lex_state = 0},
  [1123] = {.lex_state = 17},
  [1124] = {.lex_state = 19},
  [1125] = {.lex_state = 0},
  [1126] = {.lex_state = 17},
  [1127] = {.lex_state = 0},
  [1128] = {.lex_state = 17},
  [1129] = {.lex_state = 17},
  [1130] = {.lex_state = 0},
  [1131] = {.lex_state = 0},
  [1132] = {.lex_state = 0},
  [1133] = {.lex_state = 19},
  [1134] = {.lex_state = 0},
  [1135] = {.lex_state = 17},
  [1136] = {(TSStateId)(-1),},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_line_comment] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [anon_sym_version] = ACTIONS(1),
    [anon_sym_is] = ACTIONS(1),
    [anon_sym_end] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_from] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [sym_controlled_language_tag] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [sym_reserved_self] = ACTIONS(1),
    [anon_sym_event] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_def] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [sym_quoted_string] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_POUND_LBRACK] = ACTIONS(1),
    [sym_unsigned] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_u22a4] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_u22a5] = ACTIONS(1),
    [sym_value_empty_sequence] = ACTIONS(1),
    [anon_sym_Thing] = ACTIONS(1),
    [anon_sym_Nothing] = ACTIONS(1),
    [anon_sym_real] = ACTIONS(1),
    [anon_sym_rational] = ACTIONS(1),
    [anon_sym_base64Binary] = ACTIONS(1),
    [anon_sym_boolean] = ACTIONS(1),
    [anon_sym_date] = ACTIONS(1),
    [anon_sym_dateTime] = ACTIONS(1),
    [anon_sym_decimal] = ACTIONS(1),
    [anon_sym_double] = ACTIONS(1),
    [anon_sym_duration] = ACTIONS(1),
    [anon_sym_float] = ACTIONS(1),
    [anon_sym_gDay] = ACTIONS(1),
    [anon_sym_gMonth] = ACTIONS(1),
    [anon_sym_gMonthDay] = ACTIONS(1),
    [anon_sym_gYearMonth] = ACTIONS(1),
    [anon_sym_gYear] = ACTIONS(1),
    [anon_sym_hexBinary] = ACTIONS(1),
    [anon_sym_string] = ACTIONS(1),
    [anon_sym_time] = ACTIONS(1),
    [anon_sym_dateTimeStamp] = ACTIONS(1),
    [anon_sym_dayTimeDuration] = ACTIONS(1),
    [anon_sym_yearMonthDuration] = ACTIONS(1),
    [anon_sym_long] = ACTIONS(1),
    [anon_sym_short] = ACTIONS(1),
    [anon_sym_byte] = ACTIONS(1),
    [anon_sym_nonNegativeInteger] = ACTIONS(1),
    [anon_sym_positiveInteger] = ACTIONS(1),
    [anon_sym_unsignedLong] = ACTIONS(1),
    [anon_sym_unsignedInt] = ACTIONS(1),
    [anon_sym_unsigned] = ACTIONS(1),
    [anon_sym_unsignedShort] = ACTIONS(1),
    [anon_sym_unsignedByte] = ACTIONS(1),
    [anon_sym_nonPositiveInteger] = ACTIONS(1),
    [anon_sym_negativeInteger] = ACTIONS(1),
    [anon_sym_normalizedString] = ACTIONS(1),
    [anon_sym_token] = ACTIONS(1),
    [anon_sym_language] = ACTIONS(1),
    [anon_sym_binary] = ACTIONS(1),
    [anon_sym_iri] = ACTIONS(1),
    [anon_sym_without] = ACTIONS(1),
    [anon_sym_datatype] = ACTIONS(1),
    [sym_opaque] = ACTIONS(1),
    [anon_sym_length] = ACTIONS(1),
    [anon_sym_maxLength] = ACTIONS(1),
    [anon_sym_minLength] = ACTIONS(1),
    [anon_sym_fractionDigits] = ACTIONS(1),
    [anon_sym_totalDigits] = ACTIONS(1),
    [anon_sym_maxExclusive] = ACTIONS(1),
    [anon_sym_minExclusive] = ACTIONS(1),
    [anon_sym_maxInclusive] = ACTIONS(1),
    [anon_sym_minInclusive] = ACTIONS(1),
    [anon_sym_explicitTimezone] = ACTIONS(1),
    [anon_sym_required] = ACTIONS(1),
    [anon_sym_prohibited] = ACTIONS(1),
    [anon_sym_optional] = ACTIONS(1),
    [anon_sym_pattern] = ACTIONS(1),
    [sym_kw_is_fixed] = ACTIONS(1),
    [anon_sym_dimension] = ACTIONS(1),
    [anon_sym_parent] = ACTIONS(1),
    [anon_sym_source] = ACTIONS(1),
    [anon_sym_entity] = ACTIONS(1),
    [anon_sym_enum] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_metric] = ACTIONS(1),
    [anon_sym_group] = ACTIONS(1),
    [anon_sym_on] = ACTIONS(1),
    [anon_sym_ref] = ACTIONS(1),
    [anon_sym_property] = ACTIONS(1),
    [anon_sym_rdf] = ACTIONS(1),
    [anon_sym_a] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_structure] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [sym_wildcard] = ACTIONS(1),
    [anon_sym_union] = ACTIONS(1),
    [anon_sym_identity] = ACTIONS(1),
    [sym_unknown_type] = ACTIONS(1),
    [anon_sym_unordered] = ACTIONS(1),
    [anon_sym_unique] = ACTIONS(1),
    [anon_sym_nonunique] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_u2260] = ACTIONS(1),
    [sym_op_greater_than] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_u2264] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_u2265] = ACTIONS(1),
    [anon_sym_not] = ACTIONS(1),
    [anon_sym_u00ac] = ACTIONS(1),
    [anon_sym_u2227] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_u2228] = ACTIONS(1),
    [anon_sym_xor] = ACTIONS(1),
    [anon_sym_u22bb] = ACTIONS(1),
    [anon_sym_implies] = ACTIONS(1),
    [anon_sym_EQ_EQ_GT] = ACTIONS(1),
    [anon_sym_u21d2] = ACTIONS(1),
    [anon_sym_iff] = ACTIONS(1),
    [anon_sym_LT_EQ_EQ_GT] = ACTIONS(1),
    [anon_sym_u21d4] = ACTIONS(1),
    [anon_sym_u2200] = ACTIONS(1),
    [anon_sym_exists] = ACTIONS(1),
    [anon_sym_u2203] = ACTIONS(1),
    [anon_sym_exists_BANG] = ACTIONS(1),
    [anon_sym_u2203_BANG] = ACTIONS(1),
    [anon_sym_u222a] = ACTIONS(1),
    [anon_sym_intersects] = ACTIONS(1),
    [anon_sym_u2229] = ACTIONS(1),
    [anon_sym_subset] = ACTIONS(1),
    [anon_sym_u2282] = ACTIONS(1),
    [anon_sym_subseteq] = ACTIONS(1),
    [anon_sym_u2286] = ACTIONS(1),
    [anon_sym_supset] = ACTIONS(1),
    [anon_sym_u2283] = ACTIONS(1),
    [anon_sym_supseteq] = ACTIONS(1),
    [anon_sym_u2287] = ACTIONS(1),
    [anon_sym_complement] = ACTIONS(1),
    [anon_sym_u2216] = ACTIONS(1),
    [anon_sym_product] = ACTIONS(1),
    [anon_sym_u2a09] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_u2208] = ACTIONS(1),
    [anon_sym_u2209] = ACTIONS(1),
    [sym_seq_builder_separator] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_u00d7] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_u00f7] = ACTIONS(1),
    [anon_sym_mod] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [sym_math_op_subtract] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_u2218] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_u2254] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_u2192] = ACTIONS(1),
    [anon_sym_u2190] = ACTIONS(1),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(1)] = {
    [sym_module] = STATE(1016),
    [sym_line_comment] = STATE(1),
    [anon_sym_module] = ACTIONS(5),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(2)] = {
    [sym_line_comment] = STATE(2),
    [anon_sym_version] = ACTIONS(7),
    [anon_sym_is] = ACTIONS(7),
    [anon_sym_end] = ACTIONS(7),
    [anon_sym_import] = ACTIONS(7),
    [anon_sym_from] = ACTIONS(7),
    [anon_sym_as] = ACTIONS(9),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_assert] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_event] = ACTIONS(7),
    [anon_sym_for] = ACTIONS(7),
    [anon_sym_def] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(9),
    [anon_sym_datatype] = ACTIONS(7),
    [anon_sym_length] = ACTIONS(7),
    [anon_sym_maxLength] = ACTIONS(7),
    [anon_sym_minLength] = ACTIONS(7),
    [anon_sym_fractionDigits] = ACTIONS(7),
    [anon_sym_totalDigits] = ACTIONS(7),
    [anon_sym_maxExclusive] = ACTIONS(7),
    [anon_sym_minExclusive] = ACTIONS(7),
    [anon_sym_maxInclusive] = ACTIONS(7),
    [anon_sym_minInclusive] = ACTIONS(7),
    [anon_sym_explicitTimezone] = ACTIONS(7),
    [anon_sym_pattern] = ACTIONS(7),
    [anon_sym_dimension] = ACTIONS(7),
    [anon_sym_source] = ACTIONS(7),
    [anon_sym_entity] = ACTIONS(7),
    [anon_sym_enum] = ACTIONS(7),
    [anon_sym_metric] = ACTIONS(7),
    [anon_sym_ref] = ACTIONS(7),
    [anon_sym_property] = ACTIONS(7),
    [anon_sym_rdf] = ACTIONS(7),
    [anon_sym_structure] = ACTIONS(7),
    [anon_sym_class] = ACTIONS(7),
    [anon_sym_union] = ACTIONS(7),
    [anon_sym_identity] = ACTIONS(7),
    [anon_sym_SLASH_EQ] = ACTIONS(7),
    [anon_sym_u2260] = ACTIONS(7),
    [sym_op_greater_than] = ACTIONS(9),
    [anon_sym_LT_EQ] = ACTIONS(7),
    [anon_sym_u2264] = ACTIONS(7),
    [anon_sym_GT_EQ] = ACTIONS(7),
    [anon_sym_u2265] = ACTIONS(7),
    [anon_sym_not] = ACTIONS(7),
    [anon_sym_u00ac] = ACTIONS(7),
    [anon_sym_u222a] = ACTIONS(7),
    [anon_sym_intersects] = ACTIONS(7),
    [anon_sym_u2229] = ACTIONS(7),
    [anon_sym_subset] = ACTIONS(9),
    [anon_sym_u2282] = ACTIONS(7),
    [anon_sym_subseteq] = ACTIONS(7),
    [anon_sym_u2286] = ACTIONS(7),
    [anon_sym_supset] = ACTIONS(9),
    [anon_sym_u2283] = ACTIONS(7),
    [anon_sym_supseteq] = ACTIONS(7),
    [anon_sym_u2287] = ACTIONS(7),
    [anon_sym_complement] = ACTIONS(7),
    [anon_sym_u2216] = ACTIONS(7),
    [anon_sym_product] = ACTIONS(7),
    [anon_sym_u2a09] = ACTIONS(7),
    [anon_sym_in] = ACTIONS(9),
    [anon_sym_u2208] = ACTIONS(7),
    [anon_sym_u2209] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_u00d7] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(9),
    [anon_sym_u00f7] = ACTIONS(7),
    [anon_sym_mod] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [sym_math_op_subtract] = ACTIONS(9),
    [anon_sym_DASH_GT] = ACTIONS(7),
    [anon_sym_u2192] = ACTIONS(7),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(3)] = {
    [sym_line_comment] = STATE(3),
    [anon_sym_is] = ACTIONS(11),
    [anon_sym_end] = ACTIONS(11),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_EQ] = ACTIONS(11),
    [anon_sym_assert] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(11),
    [anon_sym_RPAREN] = ACTIONS(11),
    [anon_sym_COMMA] = ACTIONS(11),
    [anon_sym_event] = ACTIONS(11),
    [anon_sym_for] = ACTIONS(11),
    [anon_sym_def] = ACTIONS(11),
    [anon_sym_RBRACE] = ACTIONS(11),
    [sym_language_tag] = ACTIONS(15),
    [anon_sym_LT] = ACTIONS(13),
    [anon_sym_datatype] = ACTIONS(11),
    [anon_sym_length] = ACTIONS(11),
    [anon_sym_maxLength] = ACTIONS(11),
    [anon_sym_minLength] = ACTIONS(11),
    [anon_sym_fractionDigits] = ACTIONS(11),
    [anon_sym_totalDigits] = ACTIONS(11),
    [anon_sym_maxExclusive] = ACTIONS(11),
    [anon_sym_minExclusive] = ACTIONS(11),
    [anon_sym_maxInclusive] = ACTIONS(11),
    [anon_sym_minInclusive] = ACTIONS(11),
    [anon_sym_explicitTimezone] = ACTIONS(11),
    [anon_sym_pattern] = ACTIONS(11),
    [anon_sym_dimension] = ACTIONS(11),
    [anon_sym_source] = ACTIONS(11),
    [anon_sym_entity] = ACTIONS(11),
    [anon_sym_enum] = ACTIONS(11),
    [anon_sym_metric] = ACTIONS(11),
    [anon_sym_ref] = ACTIONS(11),
    [anon_sym_property] = ACTIONS(11),
    [anon_sym_rdf] = ACTIONS(11),
    [anon_sym_structure] = ACTIONS(11),
    [anon_sym_class] = ACTIONS(11),
    [anon_sym_union] = ACTIONS(11),
    [anon_sym_identity] = ACTIONS(11),
    [anon_sym_SLASH_EQ] = ACTIONS(11),
    [anon_sym_u2260] = ACTIONS(11),
    [sym_op_greater_than] = ACTIONS(13),
    [anon_sym_LT_EQ] = ACTIONS(11),
    [anon_sym_u2264] = ACTIONS(11),
    [anon_sym_GT_EQ] = ACTIONS(11),
    [anon_sym_u2265] = ACTIONS(11),
    [anon_sym_not] = ACTIONS(11),
    [anon_sym_u00ac] = ACTIONS(11),
    [anon_sym_u222a] = ACTIONS(11),
    [anon_sym_intersects] = ACTIONS(11),
    [anon_sym_u2229] = ACTIONS(11),
    [anon_sym_subset] = ACTIONS(13),
    [anon_sym_u2282] = ACTIONS(11),
    [anon_sym_subseteq] = ACTIONS(11),
    [anon_sym_u2286] = ACTIONS(11),
    [anon_sym_supset] = ACTIONS(13),
    [anon_sym_u2283] = ACTIONS(11),
    [anon_sym_supseteq] = ACTIONS(11),
    [anon_sym_u2287] = ACTIONS(11),
    [anon_sym_complement] = ACTIONS(11),
    [anon_sym_u2216] = ACTIONS(11),
    [anon_sym_product] = ACTIONS(11),
    [anon_sym_u2a09] = ACTIONS(11),
    [anon_sym_in] = ACTIONS(13),
    [anon_sym_u2208] = ACTIONS(11),
    [anon_sym_u2209] = ACTIONS(11),
    [anon_sym_STAR] = ACTIONS(11),
    [anon_sym_u00d7] = ACTIONS(11),
    [anon_sym_SLASH] = ACTIONS(13),
    [anon_sym_u00f7] = ACTIONS(11),
    [anon_sym_mod] = ACTIONS(11),
    [anon_sym_PERCENT] = ACTIONS(11),
    [anon_sym_PLUS] = ACTIONS(11),
    [sym_math_op_subtract] = ACTIONS(13),
    [anon_sym_DASH_GT] = ACTIONS(11),
    [anon_sym_u2192] = ACTIONS(11),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(4)] = {
    [sym_op_equality] = STATE(118),
    [sym_inequality_relation] = STATE(114),
    [sym_op_inequality] = STATE(324),
    [sym_op_less_than] = STATE(324),
    [sym_op_less_than_or_equal] = STATE(324),
    [sym_op_greater_than_or_equal] = STATE(324),
    [sym_logical_op_negation] = STATE(883),
    [sym_set_operator] = STATE(113),
    [sym_set_op_union] = STATE(352),
    [sym_set_op_intersection] = STATE(352),
    [sym_set_op_subset] = STATE(352),
    [sym_set_op_subset_or_equal] = STATE(352),
    [sym_set_op_supset] = STATE(352),
    [sym_set_op_supset_or_equal] = STATE(352),
    [sym_set_op_complement] = STATE(352),
    [sym_set_op_product] = STATE(352),
    [sym_set_membership] = STATE(352),
    [sym_set_op_is_member] = STATE(354),
    [sym_set_op_is_not_member] = STATE(354),
    [sym_math_operator] = STATE(116),
    [sym_math_op_multiply] = STATE(350),
    [sym_math_op_divide] = STATE(350),
    [sym_math_op_modulo] = STATE(350),
    [sym_math_op_add] = STATE(350),
    [sym_line_comment] = STATE(4),
    [anon_sym_is] = ACTIONS(17),
    [anon_sym_end] = ACTIONS(17),
    [anon_sym_EQ] = ACTIONS(19),
    [anon_sym_LPAREN] = ACTIONS(21),
    [anon_sym_event] = ACTIONS(17),
    [anon_sym_LT] = ACTIONS(23),
    [anon_sym_datatype] = ACTIONS(17),
    [anon_sym_dimension] = ACTIONS(17),
    [anon_sym_entity] = ACTIONS(17),
    [anon_sym_enum] = ACTIONS(17),
    [anon_sym_metric] = ACTIONS(17),
    [anon_sym_property] = ACTIONS(17),
    [anon_sym_rdf] = ACTIONS(17),
    [anon_sym_structure] = ACTIONS(17),
    [anon_sym_class] = ACTIONS(17),
    [anon_sym_union] = ACTIONS(17),
    [anon_sym_SLASH_EQ] = ACTIONS(25),
    [anon_sym_u2260] = ACTIONS(25),
    [sym_op_greater_than] = ACTIONS(27),
    [anon_sym_LT_EQ] = ACTIONS(29),
    [anon_sym_u2264] = ACTIONS(29),
    [anon_sym_GT_EQ] = ACTIONS(31),
    [anon_sym_u2265] = ACTIONS(31),
    [anon_sym_not] = ACTIONS(33),
    [anon_sym_u00ac] = ACTIONS(33),
    [anon_sym_u222a] = ACTIONS(35),
    [anon_sym_intersects] = ACTIONS(37),
    [anon_sym_u2229] = ACTIONS(37),
    [anon_sym_subset] = ACTIONS(39),
    [anon_sym_u2282] = ACTIONS(41),
    [anon_sym_subseteq] = ACTIONS(43),
    [anon_sym_u2286] = ACTIONS(43),
    [anon_sym_supset] = ACTIONS(45),
    [anon_sym_u2283] = ACTIONS(47),
    [anon_sym_supseteq] = ACTIONS(49),
    [anon_sym_u2287] = ACTIONS(49),
    [anon_sym_complement] = ACTIONS(51),
    [anon_sym_u2216] = ACTIONS(51),
    [anon_sym_product] = ACTIONS(53),
    [anon_sym_u2a09] = ACTIONS(53),
    [anon_sym_in] = ACTIONS(55),
    [anon_sym_u2208] = ACTIONS(57),
    [anon_sym_u2209] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_u00d7] = ACTIONS(61),
    [anon_sym_SLASH] = ACTIONS(63),
    [anon_sym_u00f7] = ACTIONS(65),
    [anon_sym_mod] = ACTIONS(67),
    [anon_sym_PERCENT] = ACTIONS(67),
    [anon_sym_PLUS] = ACTIONS(69),
    [sym_math_op_subtract] = ACTIONS(71),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(5)] = {
    [sym_line_comment] = STATE(5),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_end] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_EQ] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(73),
    [anon_sym_RPAREN] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_event] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_def] = ACTIONS(73),
    [anon_sym_RBRACE] = ACTIONS(73),
    [anon_sym_LT] = ACTIONS(75),
    [anon_sym_datatype] = ACTIONS(73),
    [anon_sym_length] = ACTIONS(73),
    [anon_sym_maxLength] = ACTIONS(73),
    [anon_sym_minLength] = ACTIONS(73),
    [anon_sym_fractionDigits] = ACTIONS(73),
    [anon_sym_totalDigits] = ACTIONS(73),
    [anon_sym_maxExclusive] = ACTIONS(73),
    [anon_sym_minExclusive] = ACTIONS(73),
    [anon_sym_maxInclusive] = ACTIONS(73),
    [anon_sym_minInclusive] = ACTIONS(73),
    [anon_sym_explicitTimezone] = ACTIONS(73),
    [anon_sym_pattern] = ACTIONS(73),
    [anon_sym_dimension] = ACTIONS(73),
    [anon_sym_source] = ACTIONS(73),
    [anon_sym_entity] = ACTIONS(73),
    [anon_sym_enum] = ACTIONS(73),
    [anon_sym_metric] = ACTIONS(73),
    [anon_sym_ref] = ACTIONS(73),
    [anon_sym_property] = ACTIONS(73),
    [anon_sym_rdf] = ACTIONS(73),
    [anon_sym_structure] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_union] = ACTIONS(73),
    [anon_sym_identity] = ACTIONS(73),
    [anon_sym_SLASH_EQ] = ACTIONS(73),
    [anon_sym_u2260] = ACTIONS(73),
    [sym_op_greater_than] = ACTIONS(75),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_u2264] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_u2265] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(73),
    [anon_sym_u00ac] = ACTIONS(73),
    [anon_sym_u222a] = ACTIONS(73),
    [anon_sym_intersects] = ACTIONS(73),
    [anon_sym_u2229] = ACTIONS(73),
    [anon_sym_subset] = ACTIONS(75),
    [anon_sym_u2282] = ACTIONS(73),
    [anon_sym_subseteq] = ACTIONS(73),
    [anon_sym_u2286] = ACTIONS(73),
    [anon_sym_supset] = ACTIONS(75),
    [anon_sym_u2283] = ACTIONS(73),
    [anon_sym_supseteq] = ACTIONS(73),
    [anon_sym_u2287] = ACTIONS(73),
    [anon_sym_complement] = ACTIONS(73),
    [anon_sym_u2216] = ACTIONS(73),
    [anon_sym_product] = ACTIONS(73),
    [anon_sym_u2a09] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(75),
    [anon_sym_u2208] = ACTIONS(73),
    [anon_sym_u2209] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_u00d7] = ACTIONS(73),
    [anon_sym_SLASH] = ACTIONS(75),
    [anon_sym_u00f7] = ACTIONS(73),
    [anon_sym_mod] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(73),
    [sym_math_op_subtract] = ACTIONS(75),
    [anon_sym_DASH_GT] = ACTIONS(73),
    [anon_sym_u2192] = ACTIONS(73),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(6)] = {
    [sym_line_comment] = STATE(6),
    [anon_sym_is] = ACTIONS(77),
    [anon_sym_end] = ACTIONS(77),
    [anon_sym_AT] = ACTIONS(77),
    [anon_sym_EQ] = ACTIONS(77),
    [anon_sym_assert] = ACTIONS(77),
    [anon_sym_LPAREN] = ACTIONS(77),
    [anon_sym_RPAREN] = ACTIONS(77),
    [anon_sym_COMMA] = ACTIONS(77),
    [anon_sym_event] = ACTIONS(77),
    [anon_sym_for] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(77),
    [anon_sym_RBRACE] = ACTIONS(77),
    [anon_sym_LT] = ACTIONS(79),
    [anon_sym_datatype] = ACTIONS(77),
    [anon_sym_length] = ACTIONS(77),
    [anon_sym_maxLength] = ACTIONS(77),
    [anon_sym_minLength] = ACTIONS(77),
    [anon_sym_fractionDigits] = ACTIONS(77),
    [anon_sym_totalDigits] = ACTIONS(77),
    [anon_sym_maxExclusive] = ACTIONS(77),
    [anon_sym_minExclusive] = ACTIONS(77),
    [anon_sym_maxInclusive] = ACTIONS(77),
    [anon_sym_minInclusive] = ACTIONS(77),
    [anon_sym_explicitTimezone] = ACTIONS(77),
    [anon_sym_pattern] = ACTIONS(77),
    [anon_sym_dimension] = ACTIONS(77),
    [anon_sym_source] = ACTIONS(77),
    [anon_sym_entity] = ACTIONS(77),
    [anon_sym_enum] = ACTIONS(77),
    [anon_sym_metric] = ACTIONS(77),
    [anon_sym_ref] = ACTIONS(77),
    [anon_sym_property] = ACTIONS(77),
    [anon_sym_rdf] = ACTIONS(77),
    [anon_sym_structure] = ACTIONS(77),
    [anon_sym_class] = ACTIONS(77),
    [anon_sym_union] = ACTIONS(77),
    [anon_sym_identity] = ACTIONS(77),
    [anon_sym_SLASH_EQ] = ACTIONS(77),
    [anon_sym_u2260] = ACTIONS(77),
    [sym_op_greater_than] = ACTIONS(79),
    [anon_sym_LT_EQ] = ACTIONS(77),
    [anon_sym_u2264] = ACTIONS(77),
    [anon_sym_GT_EQ] = ACTIONS(77),
    [anon_sym_u2265] = ACTIONS(77),
    [anon_sym_not] = ACTIONS(77),
    [anon_sym_u00ac] = ACTIONS(77),
    [anon_sym_u222a] = ACTIONS(77),
    [anon_sym_intersects] = ACTIONS(77),
    [anon_sym_u2229] = ACTIONS(77),
    [anon_sym_subset] = ACTIONS(79),
    [anon_sym_u2282] = ACTIONS(77),
    [anon_sym_subseteq] = ACTIONS(77),
    [anon_sym_u2286] = ACTIONS(77),
    [anon_sym_supset] = ACTIONS(79),
    [anon_sym_u2283] = ACTIONS(77),
    [anon_sym_supseteq] = ACTIONS(77),
    [anon_sym_u2287] = ACTIONS(77),
    [anon_sym_complement] = ACTIONS(77),
    [anon_sym_u2216] = ACTIONS(77),
    [anon_sym_product] = ACTIONS(77),
    [anon_sym_u2a09] = ACTIONS(77),
    [anon_sym_in] = ACTIONS(79),
    [anon_sym_u2208] = ACTIONS(77),
    [anon_sym_u2209] = ACTIONS(77),
    [anon_sym_STAR] = ACTIONS(77),
    [anon_sym_u00d7] = ACTIONS(77),
    [anon_sym_SLASH] = ACTIONS(79),
    [anon_sym_u00f7] = ACTIONS(77),
    [anon_sym_mod] = ACTIONS(77),
    [anon_sym_PERCENT] = ACTIONS(77),
    [anon_sym_PLUS] = ACTIONS(77),
    [sym_math_op_subtract] = ACTIONS(79),
    [anon_sym_DASH_GT] = ACTIONS(77),
    [anon_sym_u2192] = ACTIONS(77),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(7)] = {
    [sym_line_comment] = STATE(7),
    [anon_sym_is] = ACTIONS(81),
    [anon_sym_end] = ACTIONS(81),
    [anon_sym_AT] = ACTIONS(81),
    [anon_sym_EQ] = ACTIONS(81),
    [anon_sym_assert] = ACTIONS(81),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_RPAREN] = ACTIONS(81),
    [anon_sym_COMMA] = ACTIONS(81),
    [anon_sym_event] = ACTIONS(81),
    [anon_sym_for] = ACTIONS(81),
    [anon_sym_def] = ACTIONS(81),
    [anon_sym_RBRACE] = ACTIONS(81),
    [anon_sym_LT] = ACTIONS(83),
    [anon_sym_datatype] = ACTIONS(81),
    [anon_sym_length] = ACTIONS(81),
    [anon_sym_maxLength] = ACTIONS(81),
    [anon_sym_minLength] = ACTIONS(81),
    [anon_sym_fractionDigits] = ACTIONS(81),
    [anon_sym_totalDigits] = ACTIONS(81),
    [anon_sym_maxExclusive] = ACTIONS(81),
    [anon_sym_minExclusive] = ACTIONS(81),
    [anon_sym_maxInclusive] = ACTIONS(81),
    [anon_sym_minInclusive] = ACTIONS(81),
    [anon_sym_explicitTimezone] = ACTIONS(81),
    [anon_sym_pattern] = ACTIONS(81),
    [anon_sym_dimension] = ACTIONS(81),
    [anon_sym_source] = ACTIONS(81),
    [anon_sym_entity] = ACTIONS(81),
    [anon_sym_enum] = ACTIONS(81),
    [anon_sym_metric] = ACTIONS(81),
    [anon_sym_ref] = ACTIONS(81),
    [anon_sym_property] = ACTIONS(81),
    [anon_sym_rdf] = ACTIONS(81),
    [anon_sym_structure] = ACTIONS(81),
    [anon_sym_class] = ACTIONS(81),
    [anon_sym_union] = ACTIONS(81),
    [anon_sym_identity] = ACTIONS(81),
    [anon_sym_SLASH_EQ] = ACTIONS(81),
    [anon_sym_u2260] = ACTIONS(81),
    [sym_op_greater_than] = ACTIONS(83),
    [anon_sym_LT_EQ] = ACTIONS(81),
    [anon_sym_u2264] = ACTIONS(81),
    [anon_sym_GT_EQ] = ACTIONS(81),
    [anon_sym_u2265] = ACTIONS(81),
    [anon_sym_not] = ACTIONS(81),
    [anon_sym_u00ac] = ACTIONS(81),
    [anon_sym_u222a] = ACTIONS(81),
    [anon_sym_intersects] = ACTIONS(81),
    [anon_sym_u2229] = ACTIONS(81),
    [anon_sym_subset] = ACTIONS(83),
    [anon_sym_u2282] = ACTIONS(81),
    [anon_sym_subseteq] = ACTIONS(81),
    [anon_sym_u2286] = ACTIONS(81),
    [anon_sym_supset] = ACTIONS(83),
    [anon_sym_u2283] = ACTIONS(81),
    [anon_sym_supseteq] = ACTIONS(81),
    [anon_sym_u2287] = ACTIONS(81),
    [anon_sym_complement] = ACTIONS(81),
    [anon_sym_u2216] = ACTIONS(81),
    [anon_sym_product] = ACTIONS(81),
    [anon_sym_u2a09] = ACTIONS(81),
    [anon_sym_in] = ACTIONS(83),
    [anon_sym_u2208] = ACTIONS(81),
    [anon_sym_u2209] = ACTIONS(81),
    [anon_sym_STAR] = ACTIONS(81),
    [anon_sym_u00d7] = ACTIONS(81),
    [anon_sym_SLASH] = ACTIONS(83),
    [anon_sym_u00f7] = ACTIONS(81),
    [anon_sym_mod] = ACTIONS(81),
    [anon_sym_PERCENT] = ACTIONS(81),
    [anon_sym_PLUS] = ACTIONS(81),
    [sym_math_op_subtract] = ACTIONS(83),
    [anon_sym_DASH_GT] = ACTIONS(81),
    [anon_sym_u2192] = ACTIONS(81),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(8)] = {
    [sym_line_comment] = STATE(8),
    [anon_sym_is] = ACTIONS(85),
    [anon_sym_end] = ACTIONS(85),
    [anon_sym_AT] = ACTIONS(85),
    [anon_sym_EQ] = ACTIONS(85),
    [anon_sym_assert] = ACTIONS(85),
    [anon_sym_LPAREN] = ACTIONS(85),
    [anon_sym_RPAREN] = ACTIONS(85),
    [anon_sym_COMMA] = ACTIONS(85),
    [anon_sym_event] = ACTIONS(85),
    [anon_sym_for] = ACTIONS(85),
    [anon_sym_def] = ACTIONS(85),
    [anon_sym_RBRACE] = ACTIONS(85),
    [anon_sym_LT] = ACTIONS(87),
    [anon_sym_datatype] = ACTIONS(85),
    [anon_sym_length] = ACTIONS(85),
    [anon_sym_maxLength] = ACTIONS(85),
    [anon_sym_minLength] = ACTIONS(85),
    [anon_sym_fractionDigits] = ACTIONS(85),
    [anon_sym_totalDigits] = ACTIONS(85),
    [anon_sym_maxExclusive] = ACTIONS(85),
    [anon_sym_minExclusive] = ACTIONS(85),
    [anon_sym_maxInclusive] = ACTIONS(85),
    [anon_sym_minInclusive] = ACTIONS(85),
    [anon_sym_explicitTimezone] = ACTIONS(85),
    [anon_sym_pattern] = ACTIONS(85),
    [anon_sym_dimension] = ACTIONS(85),
    [anon_sym_source] = ACTIONS(85),
    [anon_sym_entity] = ACTIONS(85),
    [anon_sym_enum] = ACTIONS(85),
    [anon_sym_metric] = ACTIONS(85),
    [anon_sym_ref] = ACTIONS(85),
    [anon_sym_property] = ACTIONS(85),
    [anon_sym_rdf] = ACTIONS(85),
    [anon_sym_structure] = ACTIONS(85),
    [anon_sym_class] = ACTIONS(85),
    [anon_sym_union] = ACTIONS(85),
    [anon_sym_identity] = ACTIONS(85),
    [anon_sym_SLASH_EQ] = ACTIONS(85),
    [anon_sym_u2260] = ACTIONS(85),
    [sym_op_greater_than] = ACTIONS(87),
    [anon_sym_LT_EQ] = ACTIONS(85),
    [anon_sym_u2264] = ACTIONS(85),
    [anon_sym_GT_EQ] = ACTIONS(85),
    [anon_sym_u2265] = ACTIONS(85),
    [anon_sym_not] = ACTIONS(85),
    [anon_sym_u00ac] = ACTIONS(85),
    [anon_sym_u222a] = ACTIONS(85),
    [anon_sym_intersects] = ACTIONS(85),
    [anon_sym_u2229] = ACTIONS(85),
    [anon_sym_subset] = ACTIONS(87),
    [anon_sym_u2282] = ACTIONS(85),
    [anon_sym_subseteq] = ACTIONS(85),
    [anon_sym_u2286] = ACTIONS(85),
    [anon_sym_supset] = ACTIONS(87),
    [anon_sym_u2283] = ACTIONS(85),
    [anon_sym_supseteq] = ACTIONS(85),
    [anon_sym_u2287] = ACTIONS(85),
    [anon_sym_complement] = ACTIONS(85),
    [anon_sym_u2216] = ACTIONS(85),
    [anon_sym_product] = ACTIONS(85),
    [anon_sym_u2a09] = ACTIONS(85),
    [anon_sym_in] = ACTIONS(87),
    [anon_sym_u2208] = ACTIONS(85),
    [anon_sym_u2209] = ACTIONS(85),
    [anon_sym_STAR] = ACTIONS(85),
    [anon_sym_u00d7] = ACTIONS(85),
    [anon_sym_SLASH] = ACTIONS(87),
    [anon_sym_u00f7] = ACTIONS(85),
    [anon_sym_mod] = ACTIONS(85),
    [anon_sym_PERCENT] = ACTIONS(85),
    [anon_sym_PLUS] = ACTIONS(85),
    [sym_math_op_subtract] = ACTIONS(87),
    [anon_sym_DASH_GT] = ACTIONS(85),
    [anon_sym_u2192] = ACTIONS(85),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(9)] = {
    [sym_line_comment] = STATE(9),
    [anon_sym_is] = ACTIONS(89),
    [anon_sym_end] = ACTIONS(89),
    [anon_sym_AT] = ACTIONS(89),
    [anon_sym_EQ] = ACTIONS(89),
    [anon_sym_assert] = ACTIONS(89),
    [anon_sym_LPAREN] = ACTIONS(89),
    [anon_sym_RPAREN] = ACTIONS(89),
    [anon_sym_COMMA] = ACTIONS(89),
    [anon_sym_event] = ACTIONS(89),
    [anon_sym_for] = ACTIONS(89),
    [anon_sym_def] = ACTIONS(89),
    [anon_sym_RBRACE] = ACTIONS(89),
    [anon_sym_LT] = ACTIONS(91),
    [anon_sym_datatype] = ACTIONS(89),
    [anon_sym_length] = ACTIONS(89),
    [anon_sym_maxLength] = ACTIONS(89),
    [anon_sym_minLength] = ACTIONS(89),
    [anon_sym_fractionDigits] = ACTIONS(89),
    [anon_sym_totalDigits] = ACTIONS(89),
    [anon_sym_maxExclusive] = ACTIONS(89),
    [anon_sym_minExclusive] = ACTIONS(89),
    [anon_sym_maxInclusive] = ACTIONS(89),
    [anon_sym_minInclusive] = ACTIONS(89),
    [anon_sym_explicitTimezone] = ACTIONS(89),
    [anon_sym_pattern] = ACTIONS(89),
    [anon_sym_dimension] = ACTIONS(89),
    [anon_sym_source] = ACTIONS(89),
    [anon_sym_entity] = ACTIONS(89),
    [anon_sym_enum] = ACTIONS(89),
    [anon_sym_metric] = ACTIONS(89),
    [anon_sym_ref] = ACTIONS(89),
    [anon_sym_property] = ACTIONS(89),
    [anon_sym_rdf] = ACTIONS(89),
    [anon_sym_structure] = ACTIONS(89),
    [anon_sym_class] = ACTIONS(89),
    [anon_sym_union] = ACTIONS(89),
    [anon_sym_identity] = ACTIONS(89),
    [anon_sym_SLASH_EQ] = ACTIONS(89),
    [anon_sym_u2260] = ACTIONS(89),
    [sym_op_greater_than] = ACTIONS(91),
    [anon_sym_LT_EQ] = ACTIONS(89),
    [anon_sym_u2264] = ACTIONS(89),
    [anon_sym_GT_EQ] = ACTIONS(89),
    [anon_sym_u2265] = ACTIONS(89),
    [anon_sym_not] = ACTIONS(89),
    [anon_sym_u00ac] = ACTIONS(89),
    [anon_sym_u222a] = ACTIONS(89),
    [anon_sym_intersects] = ACTIONS(89),
    [anon_sym_u2229] = ACTIONS(89),
    [anon_sym_subset] = ACTIONS(91),
    [anon_sym_u2282] = ACTIONS(89),
    [anon_sym_subseteq] = ACTIONS(89),
    [anon_sym_u2286] = ACTIONS(89),
    [anon_sym_supset] = ACTIONS(91),
    [anon_sym_u2283] = ACTIONS(89),
    [anon_sym_supseteq] = ACTIONS(89),
    [anon_sym_u2287] = ACTIONS(89),
    [anon_sym_complement] = ACTIONS(89),
    [anon_sym_u2216] = ACTIONS(89),
    [anon_sym_product] = ACTIONS(89),
    [anon_sym_u2a09] = ACTIONS(89),
    [anon_sym_in] = ACTIONS(91),
    [anon_sym_u2208] = ACTIONS(89),
    [anon_sym_u2209] = ACTIONS(89),
    [anon_sym_STAR] = ACTIONS(89),
    [anon_sym_u00d7] = ACTIONS(89),
    [anon_sym_SLASH] = ACTIONS(91),
    [anon_sym_u00f7] = ACTIONS(89),
    [anon_sym_mod] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(89),
    [sym_math_op_subtract] = ACTIONS(91),
    [anon_sym_DASH_GT] = ACTIONS(89),
    [anon_sym_u2192] = ACTIONS(89),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(10)] = {
    [sym_line_comment] = STATE(10),
    [anon_sym_is] = ACTIONS(93),
    [anon_sym_end] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_EQ] = ACTIONS(93),
    [anon_sym_assert] = ACTIONS(93),
    [anon_sym_LPAREN] = ACTIONS(93),
    [anon_sym_RPAREN] = ACTIONS(93),
    [anon_sym_COMMA] = ACTIONS(93),
    [anon_sym_event] = ACTIONS(93),
    [anon_sym_for] = ACTIONS(93),
    [anon_sym_def] = ACTIONS(93),
    [anon_sym_RBRACE] = ACTIONS(93),
    [anon_sym_LT] = ACTIONS(95),
    [anon_sym_datatype] = ACTIONS(93),
    [anon_sym_length] = ACTIONS(93),
    [anon_sym_maxLength] = ACTIONS(93),
    [anon_sym_minLength] = ACTIONS(93),
    [anon_sym_fractionDigits] = ACTIONS(93),
    [anon_sym_totalDigits] = ACTIONS(93),
    [anon_sym_maxExclusive] = ACTIONS(93),
    [anon_sym_minExclusive] = ACTIONS(93),
    [anon_sym_maxInclusive] = ACTIONS(93),
    [anon_sym_minInclusive] = ACTIONS(93),
    [anon_sym_explicitTimezone] = ACTIONS(93),
    [anon_sym_pattern] = ACTIONS(93),
    [anon_sym_dimension] = ACTIONS(93),
    [anon_sym_source] = ACTIONS(93),
    [anon_sym_entity] = ACTIONS(93),
    [anon_sym_enum] = ACTIONS(93),
    [anon_sym_metric] = ACTIONS(93),
    [anon_sym_ref] = ACTIONS(93),
    [anon_sym_property] = ACTIONS(93),
    [anon_sym_rdf] = ACTIONS(93),
    [anon_sym_structure] = ACTIONS(93),
    [anon_sym_class] = ACTIONS(93),
    [anon_sym_union] = ACTIONS(93),
    [anon_sym_identity] = ACTIONS(93),
    [anon_sym_SLASH_EQ] = ACTIONS(93),
    [anon_sym_u2260] = ACTIONS(93),
    [sym_op_greater_than] = ACTIONS(95),
    [anon_sym_LT_EQ] = ACTIONS(93),
    [anon_sym_u2264] = ACTIONS(93),
    [anon_sym_GT_EQ] = ACTIONS(93),
    [anon_sym_u2265] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(93),
    [anon_sym_u00ac] = ACTIONS(93),
    [anon_sym_u222a] = ACTIONS(93),
    [anon_sym_intersects] = ACTIONS(93),
    [anon_sym_u2229] = ACTIONS(93),
    [anon_sym_subset] = ACTIONS(95),
    [anon_sym_u2282] = ACTIONS(93),
    [anon_sym_subseteq] = ACTIONS(93),
    [anon_sym_u2286] = ACTIONS(93),
    [anon_sym_supset] = ACTIONS(95),
    [anon_sym_u2283] = ACTIONS(93),
    [anon_sym_supseteq] = ACTIONS(93),
    [anon_sym_u2287] = ACTIONS(93),
    [anon_sym_complement] = ACTIONS(93),
    [anon_sym_u2216] = ACTIONS(93),
    [anon_sym_product] = ACTIONS(93),
    [anon_sym_u2a09] = ACTIONS(93),
    [anon_sym_in] = ACTIONS(95),
    [anon_sym_u2208] = ACTIONS(93),
    [anon_sym_u2209] = ACTIONS(93),
    [anon_sym_STAR] = ACTIONS(93),
    [anon_sym_u00d7] = ACTIONS(93),
    [anon_sym_SLASH] = ACTIONS(95),
    [anon_sym_u00f7] = ACTIONS(93),
    [anon_sym_mod] = ACTIONS(93),
    [anon_sym_PERCENT] = ACTIONS(93),
    [anon_sym_PLUS] = ACTIONS(93),
    [sym_math_op_subtract] = ACTIONS(95),
    [anon_sym_DASH_GT] = ACTIONS(93),
    [anon_sym_u2192] = ACTIONS(93),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(11)] = {
    [sym_line_comment] = STATE(11),
    [anon_sym_is] = ACTIONS(97),
    [anon_sym_end] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(97),
    [anon_sym_EQ] = ACTIONS(97),
    [anon_sym_assert] = ACTIONS(97),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_RPAREN] = ACTIONS(97),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_event] = ACTIONS(97),
    [anon_sym_for] = ACTIONS(97),
    [anon_sym_def] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(99),
    [anon_sym_datatype] = ACTIONS(97),
    [anon_sym_length] = ACTIONS(97),
    [anon_sym_maxLength] = ACTIONS(97),
    [anon_sym_minLength] = ACTIONS(97),
    [anon_sym_fractionDigits] = ACTIONS(97),
    [anon_sym_totalDigits] = ACTIONS(97),
    [anon_sym_maxExclusive] = ACTIONS(97),
    [anon_sym_minExclusive] = ACTIONS(97),
    [anon_sym_maxInclusive] = ACTIONS(97),
    [anon_sym_minInclusive] = ACTIONS(97),
    [anon_sym_explicitTimezone] = ACTIONS(97),
    [anon_sym_pattern] = ACTIONS(97),
    [anon_sym_dimension] = ACTIONS(97),
    [anon_sym_source] = ACTIONS(97),
    [anon_sym_entity] = ACTIONS(97),
    [anon_sym_enum] = ACTIONS(97),
    [anon_sym_metric] = ACTIONS(97),
    [anon_sym_ref] = ACTIONS(97),
    [anon_sym_property] = ACTIONS(97),
    [anon_sym_rdf] = ACTIONS(97),
    [anon_sym_structure] = ACTIONS(97),
    [anon_sym_class] = ACTIONS(97),
    [anon_sym_union] = ACTIONS(97),
    [anon_sym_identity] = ACTIONS(97),
    [anon_sym_SLASH_EQ] = ACTIONS(97),
    [anon_sym_u2260] = ACTIONS(97),
    [sym_op_greater_than] = ACTIONS(99),
    [anon_sym_LT_EQ] = ACTIONS(97),
    [anon_sym_u2264] = ACTIONS(97),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_u2265] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_u00ac] = ACTIONS(97),
    [anon_sym_u222a] = ACTIONS(97),
    [anon_sym_intersects] = ACTIONS(97),
    [anon_sym_u2229] = ACTIONS(97),
    [anon_sym_subset] = ACTIONS(99),
    [anon_sym_u2282] = ACTIONS(97),
    [anon_sym_subseteq] = ACTIONS(97),
    [anon_sym_u2286] = ACTIONS(97),
    [anon_sym_supset] = ACTIONS(99),
    [anon_sym_u2283] = ACTIONS(97),
    [anon_sym_supseteq] = ACTIONS(97),
    [anon_sym_u2287] = ACTIONS(97),
    [anon_sym_complement] = ACTIONS(97),
    [anon_sym_u2216] = ACTIONS(97),
    [anon_sym_product] = ACTIONS(97),
    [anon_sym_u2a09] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(99),
    [anon_sym_u2208] = ACTIONS(97),
    [anon_sym_u2209] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(97),
    [anon_sym_u00d7] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(99),
    [anon_sym_u00f7] = ACTIONS(97),
    [anon_sym_mod] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_PLUS] = ACTIONS(97),
    [sym_math_op_subtract] = ACTIONS(99),
    [anon_sym_DASH_GT] = ACTIONS(97),
    [anon_sym_u2192] = ACTIONS(97),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(12)] = {
    [sym_line_comment] = STATE(12),
    [anon_sym_is] = ACTIONS(101),
    [anon_sym_end] = ACTIONS(101),
    [anon_sym_import] = ACTIONS(101),
    [anon_sym_from] = ACTIONS(101),
    [anon_sym_as] = ACTIONS(103),
    [anon_sym_AT] = ACTIONS(101),
    [anon_sym_EQ] = ACTIONS(101),
    [anon_sym_assert] = ACTIONS(101),
    [anon_sym_LPAREN] = ACTIONS(101),
    [anon_sym_RPAREN] = ACTIONS(101),
    [anon_sym_COMMA] = ACTIONS(101),
    [anon_sym_event] = ACTIONS(101),
    [anon_sym_with] = ACTIONS(103),
    [anon_sym_for] = ACTIONS(101),
    [anon_sym_def] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(101),
    [anon_sym_LT] = ACTIONS(103),
    [anon_sym_without] = ACTIONS(101),
    [anon_sym_datatype] = ACTIONS(101),
    [anon_sym_dimension] = ACTIONS(101),
    [anon_sym_source] = ACTIONS(101),
    [anon_sym_entity] = ACTIONS(101),
    [anon_sym_enum] = ACTIONS(101),
    [anon_sym_metric] = ACTIONS(101),
    [anon_sym_ref] = ACTIONS(101),
    [anon_sym_property] = ACTIONS(101),
    [anon_sym_rdf] = ACTIONS(101),
    [anon_sym_structure] = ACTIONS(101),
    [anon_sym_class] = ACTIONS(101),
    [anon_sym_union] = ACTIONS(101),
    [anon_sym_identity] = ACTIONS(101),
    [anon_sym_SLASH_EQ] = ACTIONS(101),
    [anon_sym_u2260] = ACTIONS(101),
    [sym_op_greater_than] = ACTIONS(103),
    [anon_sym_LT_EQ] = ACTIONS(101),
    [anon_sym_u2264] = ACTIONS(101),
    [anon_sym_GT_EQ] = ACTIONS(101),
    [anon_sym_u2265] = ACTIONS(101),
    [anon_sym_not] = ACTIONS(101),
    [anon_sym_u00ac] = ACTIONS(101),
    [anon_sym_u222a] = ACTIONS(101),
    [anon_sym_intersects] = ACTIONS(101),
    [anon_sym_u2229] = ACTIONS(101),
    [anon_sym_subset] = ACTIONS(103),
    [anon_sym_u2282] = ACTIONS(101),
    [anon_sym_subseteq] = ACTIONS(101),
    [anon_sym_u2286] = ACTIONS(101),
    [anon_sym_supset] = ACTIONS(103),
    [anon_sym_u2283] = ACTIONS(101),
    [anon_sym_supseteq] = ACTIONS(101),
    [anon_sym_u2287] = ACTIONS(101),
    [anon_sym_complement] = ACTIONS(101),
    [anon_sym_u2216] = ACTIONS(101),
    [anon_sym_product] = ACTIONS(101),
    [anon_sym_u2a09] = ACTIONS(101),
    [anon_sym_in] = ACTIONS(103),
    [anon_sym_u2208] = ACTIONS(101),
    [anon_sym_u2209] = ACTIONS(101),
    [anon_sym_STAR] = ACTIONS(101),
    [anon_sym_u00d7] = ACTIONS(101),
    [anon_sym_SLASH] = ACTIONS(103),
    [anon_sym_u00f7] = ACTIONS(101),
    [anon_sym_mod] = ACTIONS(101),
    [anon_sym_PERCENT] = ACTIONS(101),
    [anon_sym_PLUS] = ACTIONS(101),
    [sym_math_op_subtract] = ACTIONS(103),
    [anon_sym_COLON_EQ] = ACTIONS(101),
    [anon_sym_u2254] = ACTIONS(101),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_u2192] = ACTIONS(101),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(13)] = {
    [sym_op_equality] = STATE(118),
    [sym_inequality_relation] = STATE(114),
    [sym_op_inequality] = STATE(324),
    [sym_op_less_than] = STATE(324),
    [sym_op_less_than_or_equal] = STATE(324),
    [sym_op_greater_than_or_equal] = STATE(324),
    [sym_logical_op_negation] = STATE(883),
    [sym_set_operator] = STATE(113),
    [sym_set_op_union] = STATE(352),
    [sym_set_op_intersection] = STATE(352),
    [sym_set_op_subset] = STATE(352),
    [sym_set_op_subset_or_equal] = STATE(352),
    [sym_set_op_supset] = STATE(352),
    [sym_set_op_supset_or_equal] = STATE(352),
    [sym_set_op_complement] = STATE(352),
    [sym_set_op_product] = STATE(352),
    [sym_set_membership] = STATE(352),
    [sym_set_op_is_member] = STATE(354),
    [sym_set_op_is_not_member] = STATE(354),
    [sym_math_operator] = STATE(116),
    [sym_math_op_multiply] = STATE(350),
    [sym_math_op_divide] = STATE(350),
    [sym_math_op_modulo] = STATE(350),
    [sym_math_op_add] = STATE(350),
    [sym_line_comment] = STATE(13),
    [anon_sym_is] = ACTIONS(17),
    [anon_sym_end] = ACTIONS(17),
    [anon_sym_EQ] = ACTIONS(19),
    [anon_sym_LPAREN] = ACTIONS(21),
    [anon_sym_for] = ACTIONS(17),
    [anon_sym_def] = ACTIONS(17),
    [anon_sym_LT] = ACTIONS(23),
    [anon_sym_metric] = ACTIONS(17),
    [anon_sym_ref] = ACTIONS(17),
    [anon_sym_union] = ACTIONS(35),
    [anon_sym_SLASH_EQ] = ACTIONS(25),
    [anon_sym_u2260] = ACTIONS(25),
    [sym_op_greater_than] = ACTIONS(27),
    [anon_sym_LT_EQ] = ACTIONS(29),
    [anon_sym_u2264] = ACTIONS(29),
    [anon_sym_GT_EQ] = ACTIONS(31),
    [anon_sym_u2265] = ACTIONS(31),
    [anon_sym_not] = ACTIONS(33),
    [anon_sym_u00ac] = ACTIONS(33),
    [anon_sym_u222a] = ACTIONS(35),
    [anon_sym_intersects] = ACTIONS(37),
    [anon_sym_u2229] = ACTIONS(37),
    [anon_sym_subset] = ACTIONS(39),
    [anon_sym_u2282] = ACTIONS(41),
    [anon_sym_subseteq] = ACTIONS(43),
    [anon_sym_u2286] = ACTIONS(43),
    [anon_sym_supset] = ACTIONS(45),
    [anon_sym_u2283] = ACTIONS(47),
    [anon_sym_supseteq] = ACTIONS(49),
    [anon_sym_u2287] = ACTIONS(49),
    [anon_sym_complement] = ACTIONS(51),
    [anon_sym_u2216] = ACTIONS(51),
    [anon_sym_product] = ACTIONS(53),
    [anon_sym_u2a09] = ACTIONS(53),
    [anon_sym_in] = ACTIONS(55),
    [anon_sym_u2208] = ACTIONS(57),
    [anon_sym_u2209] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_u00d7] = ACTIONS(61),
    [anon_sym_SLASH] = ACTIONS(63),
    [anon_sym_u00f7] = ACTIONS(65),
    [anon_sym_mod] = ACTIONS(67),
    [anon_sym_PERCENT] = ACTIONS(67),
    [anon_sym_PLUS] = ACTIONS(69),
    [sym_math_op_subtract] = ACTIONS(71),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(14)] = {
    [sym_line_comment] = STATE(14),
    [anon_sym_is] = ACTIONS(105),
    [anon_sym_end] = ACTIONS(105),
    [anon_sym_from] = ACTIONS(105),
    [anon_sym_AT] = ACTIONS(105),
    [anon_sym_EQ] = ACTIONS(105),
    [anon_sym_assert] = ACTIONS(105),
    [anon_sym_LPAREN] = ACTIONS(105),
    [anon_sym_RPAREN] = ACTIONS(105),
    [anon_sym_COMMA] = ACTIONS(105),
    [anon_sym_event] = ACTIONS(105),
    [anon_sym_with] = ACTIONS(107),
    [anon_sym_for] = ACTIONS(105),
    [anon_sym_def] = ACTIONS(105),
    [anon_sym_LBRACE] = ACTIONS(105),
    [anon_sym_LT] = ACTIONS(107),
    [anon_sym_without] = ACTIONS(105),
    [anon_sym_datatype] = ACTIONS(105),
    [anon_sym_dimension] = ACTIONS(105),
    [anon_sym_source] = ACTIONS(105),
    [anon_sym_entity] = ACTIONS(105),
    [anon_sym_enum] = ACTIONS(105),
    [anon_sym_metric] = ACTIONS(105),
    [anon_sym_ref] = ACTIONS(105),
    [anon_sym_property] = ACTIONS(105),
    [anon_sym_rdf] = ACTIONS(105),
    [anon_sym_structure] = ACTIONS(105),
    [anon_sym_class] = ACTIONS(105),
    [anon_sym_union] = ACTIONS(105),
    [anon_sym_identity] = ACTIONS(105),
    [anon_sym_SLASH_EQ] = ACTIONS(105),
    [anon_sym_u2260] = ACTIONS(105),
    [sym_op_greater_than] = ACTIONS(107),
    [anon_sym_LT_EQ] = ACTIONS(105),
    [anon_sym_u2264] = ACTIONS(105),
    [anon_sym_GT_EQ] = ACTIONS(105),
    [anon_sym_u2265] = ACTIONS(105),
    [anon_sym_not] = ACTIONS(105),
    [anon_sym_u00ac] = ACTIONS(105),
    [anon_sym_u222a] = ACTIONS(105),
    [anon_sym_intersects] = ACTIONS(105),
    [anon_sym_u2229] = ACTIONS(105),
    [anon_sym_subset] = ACTIONS(107),
    [anon_sym_u2282] = ACTIONS(105),
    [anon_sym_subseteq] = ACTIONS(105),
    [anon_sym_u2286] = ACTIONS(105),
    [anon_sym_supset] = ACTIONS(107),
    [anon_sym_u2283] = ACTIONS(105),
    [anon_sym_supseteq] = ACTIONS(105),
    [anon_sym_u2287] = ACTIONS(105),
    [anon_sym_complement] = ACTIONS(105),
    [anon_sym_u2216] = ACTIONS(105),
    [anon_sym_product] = ACTIONS(105),
    [anon_sym_u2a09] = ACTIONS(105),
    [anon_sym_in] = ACTIONS(107),
    [anon_sym_u2208] = ACTIONS(105),
    [anon_sym_u2209] = ACTIONS(105),
    [anon_sym_STAR] = ACTIONS(105),
    [anon_sym_u00d7] = ACTIONS(105),
    [anon_sym_SLASH] = ACTIONS(107),
    [anon_sym_u00f7] = ACTIONS(105),
    [anon_sym_mod] = ACTIONS(105),
    [anon_sym_PERCENT] = ACTIONS(105),
    [anon_sym_PLUS] = ACTIONS(105),
    [sym_math_op_subtract] = ACTIONS(107),
    [anon_sym_COLON_EQ] = ACTIONS(105),
    [anon_sym_u2254] = ACTIONS(105),
    [anon_sym_DASH_GT] = ACTIONS(105),
    [anon_sym_u2192] = ACTIONS(105),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(15)] = {
    [sym_line_comment] = STATE(15),
    [anon_sym_is] = ACTIONS(109),
    [anon_sym_end] = ACTIONS(109),
    [anon_sym_EQ] = ACTIONS(111),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_RPAREN] = ACTIONS(109),
    [anon_sym_event] = ACTIONS(109),
    [anon_sym_for] = ACTIONS(109),
    [anon_sym_def] = ACTIONS(109),
    [anon_sym_RBRACE] = ACTIONS(109),
    [anon_sym_LT] = ACTIONS(111),
    [anon_sym_datatype] = ACTIONS(109),
    [anon_sym_dimension] = ACTIONS(109),
    [anon_sym_entity] = ACTIONS(109),
    [anon_sym_enum] = ACTIONS(109),
    [anon_sym_metric] = ACTIONS(109),
    [anon_sym_ref] = ACTIONS(109),
    [anon_sym_property] = ACTIONS(109),
    [anon_sym_rdf] = ACTIONS(109),
    [anon_sym_structure] = ACTIONS(109),
    [anon_sym_class] = ACTIONS(109),
    [anon_sym_union] = ACTIONS(113),
    [anon_sym_SLASH_EQ] = ACTIONS(113),
    [anon_sym_u2260] = ACTIONS(113),
    [sym_op_greater_than] = ACTIONS(111),
    [anon_sym_LT_EQ] = ACTIONS(111),
    [anon_sym_u2264] = ACTIONS(113),
    [anon_sym_GT_EQ] = ACTIONS(113),
    [anon_sym_u2265] = ACTIONS(113),
    [anon_sym_not] = ACTIONS(113),
    [anon_sym_u00ac] = ACTIONS(113),
    [anon_sym_and] = ACTIONS(109),
    [anon_sym_u2227] = ACTIONS(109),
    [anon_sym_or] = ACTIONS(109),
    [anon_sym_u2228] = ACTIONS(109),
    [anon_sym_xor] = ACTIONS(109),
    [anon_sym_u22bb] = ACTIONS(109),
    [anon_sym_implies] = ACTIONS(109),
    [anon_sym_EQ_EQ_GT] = ACTIONS(109),
    [anon_sym_u21d2] = ACTIONS(109),
    [anon_sym_iff] = ACTIONS(109),
    [anon_sym_LT_EQ_EQ_GT] = ACTIONS(109),
    [anon_sym_u21d4] = ACTIONS(109),
    [anon_sym_u222a] = ACTIONS(113),
    [anon_sym_intersects] = ACTIONS(113),
    [anon_sym_u2229] = ACTIONS(113),
    [anon_sym_subset] = ACTIONS(111),
    [anon_sym_u2282] = ACTIONS(113),
    [anon_sym_subseteq] = ACTIONS(113),
    [anon_sym_u2286] = ACTIONS(113),
    [anon_sym_supset] = ACTIONS(111),
    [anon_sym_u2283] = ACTIONS(113),
    [anon_sym_supseteq] = ACTIONS(113),
    [anon_sym_u2287] = ACTIONS(113),
    [anon_sym_complement] = ACTIONS(113),
    [anon_sym_u2216] = ACTIONS(113),
    [anon_sym_product] = ACTIONS(113),
    [anon_sym_u2a09] = ACTIONS(113),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_u2208] = ACTIONS(113),
    [anon_sym_u2209] = ACTIONS(113),
    [anon_sym_STAR] = ACTIONS(113),
    [anon_sym_u00d7] = ACTIONS(113),
    [anon_sym_SLASH] = ACTIONS(111),
    [anon_sym_u00f7] = ACTIONS(113),
    [anon_sym_mod] = ACTIONS(113),
    [anon_sym_PERCENT] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(113),
    [sym_math_op_subtract] = ACTIONS(113),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(16)] = {
    [sym_line_comment] = STATE(16),
    [anon_sym_is] = ACTIONS(115),
    [anon_sym_end] = ACTIONS(115),
    [anon_sym_EQ] = ACTIONS(117),
    [anon_sym_LPAREN] = ACTIONS(119),
    [anon_sym_RPAREN] = ACTIONS(115),
    [anon_sym_event] = ACTIONS(115),
    [anon_sym_for] = ACTIONS(115),
    [anon_sym_def] = ACTIONS(115),
    [anon_sym_RBRACE] = ACTIONS(115),
    [anon_sym_LT] = ACTIONS(117),
    [anon_sym_datatype] = ACTIONS(115),
    [anon_sym_dimension] = ACTIONS(115),
    [anon_sym_entity] = ACTIONS(115),
    [anon_sym_enum] = ACTIONS(115),
    [anon_sym_metric] = ACTIONS(115),
    [anon_sym_ref] = ACTIONS(115),
    [anon_sym_property] = ACTIONS(115),
    [anon_sym_rdf] = ACTIONS(115),
    [anon_sym_structure] = ACTIONS(115),
    [anon_sym_class] = ACTIONS(115),
    [anon_sym_union] = ACTIONS(119),
    [anon_sym_SLASH_EQ] = ACTIONS(119),
    [anon_sym_u2260] = ACTIONS(119),
    [sym_op_greater_than] = ACTIONS(117),
    [anon_sym_LT_EQ] = ACTIONS(117),
    [anon_sym_u2264] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_u2265] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(119),
    [anon_sym_u00ac] = ACTIONS(119),
    [anon_sym_and] = ACTIONS(115),
    [anon_sym_u2227] = ACTIONS(115),
    [anon_sym_or] = ACTIONS(115),
    [anon_sym_u2228] = ACTIONS(115),
    [anon_sym_xor] = ACTIONS(115),
    [anon_sym_u22bb] = ACTIONS(115),
    [anon_sym_implies] = ACTIONS(115),
    [anon_sym_EQ_EQ_GT] = ACTIONS(115),
    [anon_sym_u21d2] = ACTIONS(115),
    [anon_sym_iff] = ACTIONS(115),
    [anon_sym_LT_EQ_EQ_GT] = ACTIONS(115),
    [anon_sym_u21d4] = ACTIONS(115),
    [anon_sym_u222a] = ACTIONS(119),
    [anon_sym_intersects] = ACTIONS(119),
    [anon_sym_u2229] = ACTIONS(119),
    [anon_sym_subset] = ACTIONS(117),
    [anon_sym_u2282] = ACTIONS(119),
    [anon_sym_subseteq] = ACTIONS(119),
    [anon_sym_u2286] = ACTIONS(119),
    [anon_sym_supset] = ACTIONS(117),
    [anon_sym_u2283] = ACTIONS(119),
    [anon_sym_supseteq] = ACTIONS(119),
    [anon_sym_u2287] = ACTIONS(119),
    [anon_sym_complement] = ACTIONS(119),
    [anon_sym_u2216] = ACTIONS(119),
    [anon_sym_product] = ACTIONS(119),
    [anon_sym_u2a09] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(117),
    [anon_sym_u2208] = ACTIONS(119),
    [anon_sym_u2209] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_u00d7] = ACTIONS(119),
    [anon_sym_SLASH] = ACTIONS(117),
    [anon_sym_u00f7] = ACTIONS(119),
    [anon_sym_mod] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(119),
    [sym_math_op_subtract] = ACTIONS(119),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(17)] = {
    [sym_line_comment] = STATE(17),
    [anon_sym_is] = ACTIONS(113),
    [anon_sym_end] = ACTIONS(113),
    [anon_sym_EQ] = ACTIONS(111),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_event] = ACTIONS(113),
    [anon_sym_for] = ACTIONS(113),
    [anon_sym_def] = ACTIONS(113),
    [anon_sym_LT] = ACTIONS(111),
    [anon_sym_datatype] = ACTIONS(113),
    [anon_sym_dimension] = ACTIONS(113),
    [anon_sym_entity] = ACTIONS(113),
    [anon_sym_enum] = ACTIONS(113),
    [anon_sym_metric] = ACTIONS(113),
    [anon_sym_ref] = ACTIONS(113),
    [anon_sym_property] = ACTIONS(113),
    [anon_sym_rdf] = ACTIONS(113),
    [anon_sym_structure] = ACTIONS(113),
    [anon_sym_class] = ACTIONS(113),
    [anon_sym_union] = ACTIONS(113),
    [anon_sym_SLASH_EQ] = ACTIONS(113),
    [anon_sym_u2260] = ACTIONS(113),
    [sym_op_greater_than] = ACTIONS(111),
    [anon_sym_LT_EQ] = ACTIONS(111),
    [anon_sym_u2264] = ACTIONS(113),
    [anon_sym_GT_EQ] = ACTIONS(113),
    [anon_sym_u2265] = ACTIONS(113),
    [anon_sym_not] = ACTIONS(113),
    [anon_sym_u00ac] = ACTIONS(113),
    [anon_sym_and] = ACTIONS(109),
    [anon_sym_u2227] = ACTIONS(109),
    [anon_sym_or] = ACTIONS(109),
    [anon_sym_u2228] = ACTIONS(109),
    [anon_sym_xor] = ACTIONS(109),
    [anon_sym_u22bb] = ACTIONS(109),
    [anon_sym_implies] = ACTIONS(109),
    [anon_sym_EQ_EQ_GT] = ACTIONS(109),
    [anon_sym_u21d2] = ACTIONS(109),
    [anon_sym_iff] = ACTIONS(109),
    [anon_sym_LT_EQ_EQ_GT] = ACTIONS(109),
    [anon_sym_u21d4] = ACTIONS(109),
    [anon_sym_u222a] = ACTIONS(113),
    [anon_sym_intersects] = ACTIONS(113),
    [anon_sym_u2229] = ACTIONS(113),
    [anon_sym_subset] = ACTIONS(111),
    [anon_sym_u2282] = ACTIONS(113),
    [anon_sym_subseteq] = ACTIONS(113),
    [anon_sym_u2286] = ACTIONS(113),
    [anon_sym_supset] = ACTIONS(111),
    [anon_sym_u2283] = ACTIONS(113),
    [anon_sym_supseteq] = ACTIONS(113),
    [anon_sym_u2287] = ACTIONS(113),
    [anon_sym_complement] = ACTIONS(113),
    [anon_sym_u2216] = ACTIONS(113),
    [anon_sym_product] = ACTIONS(113),
    [anon_sym_u2a09] = ACTIONS(113),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_u2208] = ACTIONS(113),
    [anon_sym_u2209] = ACTIONS(113),
    [anon_sym_STAR] = ACTIONS(113),
    [anon_sym_u00d7] = ACTIONS(113),
    [anon_sym_SLASH] = ACTIONS(111),
    [anon_sym_u00f7] = ACTIONS(113),
    [anon_sym_mod] = ACTIONS(113),
    [anon_sym_PERCENT] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(113),
    [sym_math_op_subtract] = ACTIONS(113),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(18)] = {
    [sym_line_comment] = STATE(18),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_end] = ACTIONS(119),
    [anon_sym_EQ] = ACTIONS(117),
    [anon_sym_LPAREN] = ACTIONS(119),
    [anon_sym_event] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_def] = ACTIONS(119),
    [anon_sym_LT] = ACTIONS(117),
    [anon_sym_datatype] = ACTIONS(119),
    [anon_sym_dimension] = ACTIONS(119),
    [anon_sym_entity] = ACTIONS(119),
    [anon_sym_enum] = ACTIONS(119),
    [anon_sym_metric] = ACTIONS(119),
    [anon_sym_ref] = ACTIONS(119),
    [anon_sym_property] = ACTIONS(119),
    [anon_sym_rdf] = ACTIONS(119),
    [anon_sym_structure] = ACTIONS(119),
    [anon_sym_class] = ACTIONS(119),
    [anon_sym_union] = ACTIONS(119),
    [anon_sym_SLASH_EQ] = ACTIONS(119),
    [anon_sym_u2260] = ACTIONS(119),
    [sym_op_greater_than] = ACTIONS(117),
    [anon_sym_LT_EQ] = ACTIONS(117),
    [anon_sym_u2264] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_u2265] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(119),
    [anon_sym_u00ac] = ACTIONS(119),
    [anon_sym_and] = ACTIONS(115),
    [anon_sym_u2227] = ACTIONS(115),
    [anon_sym_or] = ACTIONS(115),
    [anon_sym_u2228] = ACTIONS(115),
    [anon_sym_xor] = ACTIONS(115),
    [anon_sym_u22bb] = ACTIONS(115),
    [anon_sym_implies] = ACTIONS(115),
    [anon_sym_EQ_EQ_GT] = ACTIONS(115),
    [anon_sym_u21d2] = ACTIONS(115),
    [anon_sym_iff] = ACTIONS(115),
    [anon_sym_LT_EQ_EQ_GT] = ACTIONS(115),
    [anon_sym_u21d4] = ACTIONS(115),
    [anon_sym_u222a] = ACTIONS(119),
    [anon_sym_intersects] = ACTIONS(119),
    [anon_sym_u2229] = ACTIONS(119),
    [anon_sym_subset] = ACTIONS(117),
    [anon_sym_u2282] = ACTIONS(119),
    [anon_sym_subseteq] = ACTIONS(119),
    [anon_sym_u2286] = ACTIONS(119),
    [anon_sym_supset] = ACTIONS(117),
    [anon_sym_u2283] = ACTIONS(119),
    [anon_sym_supseteq] = ACTIONS(119),
    [anon_sym_u2287] = ACTIONS(119),
    [anon_sym_complement] = ACTIONS(119),
    [anon_sym_u2216] = ACTIONS(119),
    [anon_sym_product] = ACTIONS(119),
    [anon_sym_u2a09] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(117),
    [anon_sym_u2208] = ACTIONS(119),
    [anon_sym_u2209] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_u00d7] = ACTIONS(119),
    [anon_sym_SLASH] = ACTIONS(117),
    [anon_sym_u00f7] = ACTIONS(119),
    [anon_sym_mod] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(119),
    [sym_math_op_subtract] = ACTIONS(119),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(19)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(303),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(19),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(20)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(315),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(20),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(21)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(102),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(21),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(22)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(122),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(4),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(22),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(23)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(110),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(23),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(24)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(111),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(24),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(25)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(317),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(25),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(26)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(308),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(26),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(27)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(122),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(13),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(27),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(28)] = {
    [sym_qualified_identifier] = STATE(14),
    [sym_identifier_reference] = STATE(49),
    [sym_constraint_sentence] = STATE(297),
    [sym_simple_sentence] = STATE(201),
    [sym_atomic_sentence] = STATE(197),
    [sym_equation] = STATE(197),
    [sym_inequation] = STATE(197),
    [sym_boolean_sentence] = STATE(198),
    [sym_unary_boolean_sentence] = STATE(191),
    [sym_binary_boolean_sentence] = STATE(191),
    [sym_expression_sentence] = STATE(192),
    [sym_set_expression_sentence] = STATE(196),
    [sym_arithmetic_expression_sentence] = STATE(196),
    [sym_quantified_sentence] = STATE(195),
    [sym_quantified_variable_binding] = STATE(1022),
    [sym_term] = STATE(29),
    [sym_functional_term] = STATE(49),
    [sym_function_composition] = STATE(49),
    [sym_predicate_value] = STATE(49),
    [sym_sequence_of_predicate_values] = STATE(37),
    [sym_reserved_event] = STATE(34),
    [sym_sentence_with_environment] = STATE(199),
    [sym_sequence_builder] = STATE(49),
    [sym_simple_value] = STATE(37),
    [sym_string] = STATE(5),
    [sym_iri] = STATE(5),
    [sym_binary] = STATE(5),
    [sym_boolean] = STATE(5),
    [sym_boolean_truth] = STATE(8),
    [sym_boolean_falsity] = STATE(8),
    [sym_logical_op_negation] = STATE(21),
    [sym_logical_quantifier] = STATE(872),
    [sym_logical_quantifier_universal] = STATE(1129),
    [sym_logical_quantifier_existential] = STATE(1129),
    [sym_logical_quantifier_existential_one] = STATE(1129),
    [sym_line_comment] = STATE(28),
    [anon_sym_LBRACK] = ACTIONS(121),
    [sym_identifier] = ACTIONS(123),
    [anon_sym_LPAREN] = ACTIONS(125),
    [sym_reserved_self] = ACTIONS(127),
    [anon_sym_event] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(131),
    [anon_sym_LBRACE] = ACTIONS(133),
    [sym_quoted_string] = ACTIONS(135),
    [anon_sym_LT] = ACTIONS(137),
    [anon_sym_POUND_LBRACK] = ACTIONS(139),
    [sym_double] = ACTIONS(141),
    [sym_decimal] = ACTIONS(143),
    [sym_integer] = ACTIONS(143),
    [sym_rational] = ACTIONS(141),
    [sym_unsigned] = ACTIONS(143),
    [anon_sym_true] = ACTIONS(145),
    [anon_sym_u22a4] = ACTIONS(147),
    [anon_sym_false] = ACTIONS(149),
    [anon_sym_u22a5] = ACTIONS(151),
    [sym_value_empty_sequence] = ACTIONS(153),
    [anon_sym_not] = ACTIONS(155),
    [anon_sym_u00ac] = ACTIONS(157),
    [anon_sym_forall] = ACTIONS(159),
    [anon_sym_u2200] = ACTIONS(161),
    [anon_sym_exists] = ACTIONS(163),
    [anon_sym_u2203] = ACTIONS(163),
    [anon_sym_exists_BANG] = ACTIONS(165),
    [anon_sym_u2203_BANG] = ACTIONS(165),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
  [STATE(29)] = {
    [sym_op_equality] = STATE(118),
    [sym_inequality_relation] = STATE(114),
    [sym_op_inequality] = STATE(324),
    [sym_op_less_than] = STATE(324),
    [sym_op_less_than_or_equal] = STATE(324),
    [sym_op_greater_than_or_equal] = STATE(324),
    [sym_logical_op_negation] = STATE(883),
    [sym_set_operator] = STATE(113),
    [sym_set_op_union] = STATE(352),
    [sym_set_op_intersection] = STATE(352),
    [sym_set_op_subset] = STATE(352),
    [sym_set_op_subset_or_equal] = STATE(352),
    [sym_set_op_supset] = STATE(352),
    [sym_set_op_supset_or_equal] = STATE(352),
    [sym_set_op_complement] = STATE(352),
    [sym_set_op_product] = STATE(352),
    [sym_set_membership] = STATE(352),
    [sym_set_op_is_member] = STATE(354),
    [sym_set_op_is_not_member] = STATE(354),
    [sym_math_operator] = STATE(116),
    [sym_math_op_multiply] = STATE(350),
    [sym_math_op_divide] = STATE(350),
    [sym_math_op_modulo] = STATE(350),
    [sym_math_op_add] = STATE(350),
    [sym_line_comment] = STATE(29),
    [anon_sym_EQ] = ACTIONS(19),
    [anon_sym_LPAREN] = ACTIONS(167),
    [anon_sym_LT] = ACTIONS(23),
    [anon_sym_union] = ACTIONS(35),
    [anon_sym_SLASH_EQ] = ACTIONS(25),
    [anon_sym_u2260] = ACTIONS(25),
    [sym_op_greater_than] = ACTIONS(27),
    [anon_sym_LT_EQ] = ACTIONS(29),
    [anon_sym_u2264] = ACTIONS(29),
    [anon_sym_GT_EQ] = ACTIONS(31),
    [anon_sym_u2265] = ACTIONS(31),
    [anon_sym_not] = ACTIONS(33),
    [anon_sym_u00ac] = ACTIONS(33),
    [anon_sym_u222a] = ACTIONS(35),
    [anon_sym_intersects] = ACTIONS(37),
    [anon_sym_u2229] = ACTIONS(37),
    [anon_sym_subset] = ACTIONS(39),
    [anon_sym_u2282] = ACTIONS(41),
    [anon_sym_subseteq] = ACTIONS(43),
    [anon_sym_u2286] = ACTIONS(43),
    [anon_sym_supset] = ACTIONS(45),
    [anon_sym_u2283] = ACTIONS(47),
    [anon_sym_supseteq] = ACTIONS(49),
    [anon_sym_u2287] = ACTIONS(49),
    [anon_sym_complement] = ACTIONS(51),
    [anon_sym_u2216] = ACTIONS(51),
    [anon_sym_product] = ACTIONS(53),
    [anon_sym_u2a09] = ACTIONS(53),
    [anon_sym_in] = ACTIONS(55),
    [anon_sym_u2208] = ACTIONS(57),
    [anon_sym_u2209] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_u00d7] = ACTIONS(61),
    [anon_sym_SLASH] = ACTIONS(63),
    [anon_sym_u00f7] = ACTIONS(65),
    [anon_sym_mod] = ACTIONS(67),
    [anon_sym_PERCENT] = ACTIONS(67),
    [anon_sym_PLUS] = ACTIONS(69),
    [sym_math_op_subtract] = ACTIONS(71),
    [sym_comment_aligned] = ACTIONS(3),
    [sym_comment_local] = ACTIONS(3),
    [sym_comment_page] = ACTIONS(3),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 7,
    ACTIONS(169), 1,
      anon_sym_COLON,
    STATE(30), 1,
      sym_line_comment,
    STATE(31), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(171), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(105), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [78] = 6,
    STATE(31), 1,
      sym_line_comment,
    STATE(33), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(171), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(175), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(173), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [153] = 6,
    STATE(32), 1,
      sym_line_comment,
    STATE(33), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(171), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(175), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(173), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [228] = 5,
    ACTIONS(181), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    STATE(33), 2,
      sym_line_comment,
      aux_sym_function_composition_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(179), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(177), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [301] = 6,
    STATE(32), 1,
      aux_sym_function_composition_repeat1,
    STATE(34), 1,
      sym_line_comment,
    ACTIONS(171), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(186), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(184), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [376] = 4,
    STATE(35), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(190), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(188), 51,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
      anon_sym_DOT,
      anon_sym_u2218,
  [446] = 4,
    STATE(36), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(194), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(192), 51,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
      anon_sym_DOT,
      anon_sym_u2218,
  [516] = 4,
    STATE(37), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(198), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(196), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [584] = 4,
    STATE(38), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(202), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(200), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [652] = 4,
    STATE(39), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(206), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(204), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [720] = 4,
    STATE(40), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(210), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(208), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [788] = 4,
    STATE(41), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(214), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(212), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [856] = 4,
    STATE(42), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(218), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(216), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [924] = 4,
    STATE(43), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(222), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(220), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [992] = 4,
    STATE(44), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(226), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(224), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1060] = 4,
    STATE(45), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(230), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(228), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1128] = 4,
    STATE(46), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(234), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(232), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1196] = 4,
    STATE(47), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(238), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(236), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1264] = 4,
    STATE(48), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(242), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(240), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1332] = 4,
    STATE(49), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(186), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(184), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1400] = 4,
    STATE(50), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(246), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(244), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1468] = 4,
    STATE(51), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(250), 6,
      anon_sym_LT,
      sym_op_greater_than,
      anon_sym_subset,
      anon_sym_supset,
      anon_sym_in,
      anon_sym_SLASH,
    ACTIONS(248), 49,
      anon_sym_is,
      anon_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_SLASH_EQ,
      anon_sym_u2260,
      anon_sym_LT_EQ,
      anon_sym_u2264,
      anon_sym_GT_EQ,
      anon_sym_u2265,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_u222a,
      anon_sym_intersects,
      anon_sym_u2229,
      anon_sym_u2282,
      anon_sym_subseteq,
      anon_sym_u2286,
      anon_sym_u2283,
      anon_sym_supseteq,
      anon_sym_u2287,
      anon_sym_complement,
      anon_sym_u2216,
      anon_sym_product,
      anon_sym_u2a09,
      anon_sym_u2208,
      anon_sym_u2209,
      anon_sym_STAR,
      anon_sym_u00d7,
      anon_sym_u00f7,
      anon_sym_mod,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      sym_math_op_subtract,
  [1536] = 11,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(256), 1,
      anon_sym_LBRACE,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(52), 1,
      sym_line_comment,
    STATE(63), 1,
      sym_cardinality_expression,
    STATE(451), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [1616] = 11,
    ACTIONS(256), 1,
      anon_sym_LBRACE,
    ACTIONS(262), 1,
      sym_identifier,
    ACTIONS(264), 1,
      anon_sym_LPAREN,
    ACTIONS(268), 1,
      sym_unknown_type,
    STATE(53), 1,
      sym_line_comment,
    STATE(62), 1,
      sym_cardinality_expression,
    STATE(669), 1,
      sym_type_reference,
    STATE(747), 1,
      sym_qualified_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(815), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(266), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [1696] = 11,
    ACTIONS(256), 1,
      anon_sym_LBRACE,
    ACTIONS(270), 1,
      sym_identifier,
    ACTIONS(272), 1,
      anon_sym_LPAREN,
    ACTIONS(276), 1,
      sym_unknown_type,
    STATE(54), 1,
      sym_line_comment,
    STATE(59), 1,
      sym_cardinality_expression,
    STATE(628), 1,
      sym_type_reference,
    STATE(692), 1,
      sym_qualified_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(713), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(274), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [1776] = 10,
    ACTIONS(278), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_LPAREN,
    ACTIONS(282), 1,
      anon_sym_LBRACE,
    STATE(55), 1,
      sym_line_comment,
    STATE(71), 1,
      sym_cardinality_reference_expression,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(959), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(977), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(284), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [1853] = 10,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(282), 1,
      anon_sym_LBRACE,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(56), 1,
      sym_line_comment,
    STATE(72), 1,
      sym_cardinality_reference_expression,
    STATE(336), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(322), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [1930] = 10,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(282), 1,
      anon_sym_LBRACE,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(57), 1,
      sym_line_comment,
    STATE(70), 1,
      sym_cardinality_reference_expression,
    STATE(333), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(322), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2007] = 10,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(282), 1,
      anon_sym_LBRACE,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(58), 1,
      sym_line_comment,
    STATE(73), 1,
      sym_cardinality_reference_expression,
    STATE(364), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(322), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2084] = 9,
    ACTIONS(270), 1,
      sym_identifier,
    ACTIONS(272), 1,
      anon_sym_LPAREN,
    ACTIONS(276), 1,
      sym_unknown_type,
    STATE(59), 1,
      sym_line_comment,
    STATE(638), 1,
      sym_type_reference,
    STATE(692), 1,
      sym_qualified_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(713), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(274), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2158] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(60), 1,
      sym_line_comment,
    STATE(1019), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2232] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(61), 1,
      sym_line_comment,
    STATE(985), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2306] = 9,
    ACTIONS(262), 1,
      sym_identifier,
    ACTIONS(264), 1,
      anon_sym_LPAREN,
    ACTIONS(268), 1,
      sym_unknown_type,
    STATE(62), 1,
      sym_line_comment,
    STATE(687), 1,
      sym_type_reference,
    STATE(747), 1,
      sym_qualified_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(815), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(266), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2380] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(63), 1,
      sym_line_comment,
    STATE(458), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2454] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(64), 1,
      sym_line_comment,
    STATE(916), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2528] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(65), 1,
      sym_line_comment,
    STATE(1065), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2602] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(66), 1,
      sym_line_comment,
    STATE(1011), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2676] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(67), 1,
      sym_line_comment,
    STATE(1049), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2750] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(68), 1,
      sym_line_comment,
    STATE(981), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2824] = 9,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    ACTIONS(260), 1,
      sym_unknown_type,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(69), 1,
      sym_line_comment,
    STATE(984), 1,
      sym_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(428), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2898] = 8,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(70), 1,
      sym_line_comment,
    STATE(359), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(322), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [2969] = 8,
    ACTIONS(278), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_LPAREN,
    STATE(71), 1,
      sym_line_comment,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(940), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(977), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(284), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3040] = 8,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(72), 1,
      sym_line_comment,
    STATE(360), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(322), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3111] = 8,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(254), 1,
      anon_sym_LPAREN,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(73), 1,
      sym_line_comment,
    STATE(375), 1,
      sym_function_type_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(322), 3,
      sym_identifier_reference,
      sym_builtin_types,
      sym_mapping_type,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3182] = 8,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(286), 1,
      sym_opaque,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(74), 1,
      sym_line_comment,
    STATE(422), 1,
      sym_datatype_base_type_reference,
    STATE(448), 2,
      sym_identifier_reference,
      sym_builtin_types,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3252] = 7,
    ACTIONS(252), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(75), 1,
      sym_line_comment,
    STATE(430), 1,
      sym_datatype_base_type_reference,
    STATE(448), 2,
      sym_identifier_reference,
      sym_builtin_types,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(258), 43,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3319] = 4,
    ACTIONS(290), 1,
      anon_sym_LPAREN,
    STATE(76), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(288), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3378] = 4,
    ACTIONS(294), 1,
      anon_sym_LPAREN,
    STATE(77), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(292), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3437] = 4,
    ACTIONS(298), 1,
      anon_sym_LPAREN,
    STATE(78), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(296), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3496] = 4,
    ACTIONS(302), 1,
      anon_sym_LPAREN,
    STATE(79), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(300), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3555] = 4,
    ACTIONS(306), 1,
      anon_sym_LPAREN,
    STATE(80), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(304), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3614] = 4,
    ACTIONS(310), 1,
      anon_sym_LPAREN,
    STATE(81), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(308), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3673] = 4,
    ACTIONS(314), 1,
      anon_sym_LPAREN,
    STATE(82), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(312), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3732] = 4,
    ACTIONS(318), 1,
      anon_sym_LPAREN,
    STATE(83), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(316), 45,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
      sym_unknown_type,
  [3791] = 4,
    ACTIONS(322), 1,
      anon_sym_LPAREN,
    STATE(84), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(320), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3849] = 4,
    ACTIONS(326), 1,
      anon_sym_LPAREN,
    STATE(85), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(324), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3907] = 4,
    ACTIONS(330), 1,
      anon_sym_LPAREN,
    STATE(86), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(328), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [3965] = 4,
    ACTIONS(334), 1,
      anon_sym_LPAREN,
    STATE(87), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(332), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4023] = 4,
    ACTIONS(338), 1,
      anon_sym_LPAREN,
    STATE(88), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(336), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4081] = 4,
    ACTIONS(342), 1,
      anon_sym_LPAREN,
    STATE(89), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(340), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4139] = 4,
    ACTIONS(346), 1,
      anon_sym_LPAREN,
    STATE(90), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(344), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4197] = 4,
    ACTIONS(350), 1,
      anon_sym_LPAREN,
    STATE(91), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(348), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4255] = 4,
    ACTIONS(354), 1,
      anon_sym_LPAREN,
    STATE(92), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(352), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4313] = 4,
    ACTIONS(358), 1,
      anon_sym_LPAREN,
    STATE(93), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(356), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4371] = 4,
    ACTIONS(362), 1,
      anon_sym_LPAREN,
    STATE(94), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(360), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4429] = 4,
    ACTIONS(366), 1,
      anon_sym_LPAREN,
    STATE(95), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(364), 44,
      sym_identifier,
      anon_sym_Thing,
      anon_sym_Nothing,
      anon_sym_real,
      anon_sym_rational,
      anon_sym_anyURI,
      anon_sym_base64Binary,
      anon_sym_boolean,
      anon_sym_date,
      anon_sym_dateTime,
      anon_sym_decimal,
      anon_sym_double,
      anon_sym_duration,
      anon_sym_float,
      anon_sym_gDay,
      anon_sym_gMonth,
      anon_sym_gMonthDay,
      anon_sym_gYearMonth,
      anon_sym_gYear,
      anon_sym_hexBinary,
      anon_sym_string,
      anon_sym_time,
      anon_sym_dateTimeStamp,
      anon_sym_dayTimeDuration,
      anon_sym_yearMonthDuration,
      anon_sym_integer,
      anon_sym_long,
      anon_sym_int,
      anon_sym_short,
      anon_sym_byte,
      anon_sym_nonNegativeInteger,
      anon_sym_positiveInteger,
      anon_sym_unsignedLong,
      anon_sym_unsignedInt,
      anon_sym_unsigned,
      anon_sym_unsignedShort,
      anon_sym_unsignedByte,
      anon_sym_nonPositiveInteger,
      anon_sym_negativeInteger,
      anon_sym_normalizedString,
      anon_sym_token,
      anon_sym_language,
      anon_sym_binary,
      anon_sym_iri,
  [4487] = 27,
    ACTIONS(368), 1,
      anon_sym_end,
    ACTIONS(370), 1,
      anon_sym_import,
    ACTIONS(372), 1,
      anon_sym_from,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    STATE(96), 1,
      sym_line_comment,
    STATE(126), 1,
      aux_sym_module_body_repeat2,
    STATE(221), 1,
      aux_sym_module_body_repeat3,
    STATE(316), 1,
      aux_sym_module_body_repeat1,
    STATE(356), 1,
      sym_annotation,
    STATE(415), 1,
      sym_import_statement,
    STATE(503), 1,
      sym_definition,
    STATE(1131), 1,
      sym_from_clause,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [4583] = 27,
    ACTIONS(370), 1,
      anon_sym_import,
    ACTIONS(372), 1,
      anon_sym_from,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    ACTIONS(400), 1,
      anon_sym_end,
    STATE(96), 1,
      aux_sym_module_body_repeat1,
    STATE(97), 1,
      sym_line_comment,
    STATE(125), 1,
      aux_sym_module_body_repeat2,
    STATE(223), 1,
      aux_sym_module_body_repeat3,
    STATE(356), 1,
      sym_annotation,
    STATE(415), 1,
      sym_import_statement,
    STATE(503), 1,
      sym_definition,
    STATE(1131), 1,
      sym_from_clause,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [4679] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_RPAREN,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    STATE(98), 1,
      sym_line_comment,
    STATE(100), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [4772] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(434), 1,
      anon_sym_RPAREN,
    STATE(99), 1,
      sym_line_comment,
    STATE(100), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [4865] = 25,
    ACTIONS(436), 1,
      anon_sym_LBRACK,
    ACTIONS(439), 1,
      sym_identifier,
    ACTIONS(442), 1,
      anon_sym_RPAREN,
    ACTIONS(444), 1,
      sym_reserved_self,
    ACTIONS(447), 1,
      anon_sym_event,
    ACTIONS(450), 1,
      anon_sym_LBRACE,
    ACTIONS(453), 1,
      sym_quoted_string,
    ACTIONS(456), 1,
      anon_sym_LT,
    ACTIONS(459), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(468), 1,
      anon_sym_true,
    ACTIONS(471), 1,
      anon_sym_u22a4,
    ACTIONS(474), 1,
      anon_sym_false,
    ACTIONS(477), 1,
      anon_sym_u22a5,
    ACTIONS(480), 1,
      sym_value_empty_sequence,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(462), 2,
      sym_double,
      sym_rational,
    STATE(100), 2,
      sym_line_comment,
      aux_sym__actual_arguments_repeat1,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(465), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [4956] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(483), 1,
      anon_sym_RPAREN,
    STATE(98), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(101), 1,
      sym_line_comment,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5049] = 5,
    STATE(23), 1,
      sym_logical_connective,
    STATE(102), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
    ACTIONS(485), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [5100] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(487), 1,
      anon_sym_RPAREN,
    STATE(99), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(103), 1,
      sym_line_comment,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5193] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(489), 1,
      anon_sym_RPAREN,
    STATE(104), 1,
      sym_line_comment,
    STATE(108), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5286] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(491), 1,
      anon_sym_RPAREN,
    STATE(100), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(105), 1,
      sym_line_comment,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5379] = 3,
    STATE(106), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(101), 36,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
      anon_sym_in,
      anon_sym_u2208,
      anon_sym_u2209,
  [5426] = 3,
    STATE(107), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(105), 36,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
      anon_sym_in,
      anon_sym_u2208,
      anon_sym_u2209,
  [5473] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(493), 1,
      anon_sym_RPAREN,
    STATE(100), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(108), 1,
      sym_line_comment,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5566] = 26,
    ACTIONS(402), 1,
      anon_sym_LBRACK,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(408), 1,
      sym_reserved_self,
    ACTIONS(410), 1,
      anon_sym_event,
    ACTIONS(412), 1,
      anon_sym_LBRACE,
    ACTIONS(414), 1,
      sym_quoted_string,
    ACTIONS(416), 1,
      anon_sym_LT,
    ACTIONS(418), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(424), 1,
      anon_sym_true,
    ACTIONS(426), 1,
      anon_sym_u22a4,
    ACTIONS(428), 1,
      anon_sym_false,
    ACTIONS(430), 1,
      anon_sym_u22a5,
    ACTIONS(432), 1,
      sym_value_empty_sequence,
    ACTIONS(495), 1,
      anon_sym_RPAREN,
    STATE(105), 1,
      aux_sym__actual_arguments_repeat1,
    STATE(109), 1,
      sym_line_comment,
    STATE(259), 1,
      sym_reserved_event,
    STATE(270), 1,
      sym_term,
    STATE(294), 1,
      sym_qualified_identifier,
    ACTIONS(420), 2,
      sym_double,
      sym_rational,
    STATE(277), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(279), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(422), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(274), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(271), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5659] = 5,
    STATE(23), 1,
      sym_logical_connective,
    STATE(110), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
    ACTIONS(497), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [5710] = 10,
    STATE(23), 1,
      sym_logical_connective,
    STATE(111), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
    ACTIONS(499), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [5771] = 6,
    ACTIONS(511), 1,
      anon_sym_COLON,
    STATE(112), 1,
      sym_line_comment,
    STATE(119), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(513), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(105), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [5823] = 24,
    ACTIONS(515), 1,
      anon_sym_LBRACK,
    ACTIONS(517), 1,
      sym_identifier,
    ACTIONS(519), 1,
      sym_reserved_self,
    ACTIONS(521), 1,
      anon_sym_event,
    ACTIONS(523), 1,
      anon_sym_LBRACE,
    ACTIONS(525), 1,
      sym_quoted_string,
    ACTIONS(527), 1,
      anon_sym_LT,
    ACTIONS(529), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(535), 1,
      anon_sym_true,
    ACTIONS(537), 1,
      anon_sym_u22a4,
    ACTIONS(539), 1,
      anon_sym_false,
    ACTIONS(541), 1,
      anon_sym_u22a5,
    ACTIONS(543), 1,
      sym_value_empty_sequence,
    STATE(107), 1,
      sym_qualified_identifier,
    STATE(113), 1,
      sym_line_comment,
    STATE(117), 1,
      sym_reserved_event,
    STATE(188), 1,
      sym_term,
    ACTIONS(531), 2,
      sym_double,
      sym_rational,
    STATE(142), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    STATE(156), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(533), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(153), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(136), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5910] = 24,
    ACTIONS(515), 1,
      anon_sym_LBRACK,
    ACTIONS(517), 1,
      sym_identifier,
    ACTIONS(519), 1,
      sym_reserved_self,
    ACTIONS(521), 1,
      anon_sym_event,
    ACTIONS(523), 1,
      anon_sym_LBRACE,
    ACTIONS(525), 1,
      sym_quoted_string,
    ACTIONS(527), 1,
      anon_sym_LT,
    ACTIONS(529), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(535), 1,
      anon_sym_true,
    ACTIONS(537), 1,
      anon_sym_u22a4,
    ACTIONS(539), 1,
      anon_sym_false,
    ACTIONS(541), 1,
      anon_sym_u22a5,
    ACTIONS(543), 1,
      sym_value_empty_sequence,
    STATE(107), 1,
      sym_qualified_identifier,
    STATE(114), 1,
      sym_line_comment,
    STATE(117), 1,
      sym_reserved_event,
    STATE(130), 1,
      sym_term,
    ACTIONS(531), 2,
      sym_double,
      sym_rational,
    STATE(142), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    STATE(156), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(533), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(153), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(136), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [5997] = 24,
    ACTIONS(121), 1,
      anon_sym_LBRACK,
    ACTIONS(123), 1,
      sym_identifier,
    ACTIONS(127), 1,
      sym_reserved_self,
    ACTIONS(129), 1,
      anon_sym_event,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(145), 1,
      anon_sym_true,
    ACTIONS(147), 1,
      anon_sym_u22a4,
    ACTIONS(149), 1,
      anon_sym_false,
    ACTIONS(151), 1,
      anon_sym_u22a5,
    ACTIONS(153), 1,
      sym_value_empty_sequence,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(34), 1,
      sym_reserved_event,
    STATE(115), 1,
      sym_line_comment,
    STATE(918), 1,
      sym_term,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(37), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(49), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [6084] = 24,
    ACTIONS(515), 1,
      anon_sym_LBRACK,
    ACTIONS(517), 1,
      sym_identifier,
    ACTIONS(519), 1,
      sym_reserved_self,
    ACTIONS(521), 1,
      anon_sym_event,
    ACTIONS(523), 1,
      anon_sym_LBRACE,
    ACTIONS(525), 1,
      sym_quoted_string,
    ACTIONS(527), 1,
      anon_sym_LT,
    ACTIONS(529), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(535), 1,
      anon_sym_true,
    ACTIONS(537), 1,
      anon_sym_u22a4,
    ACTIONS(539), 1,
      anon_sym_false,
    ACTIONS(541), 1,
      anon_sym_u22a5,
    ACTIONS(543), 1,
      sym_value_empty_sequence,
    STATE(107), 1,
      sym_qualified_identifier,
    STATE(116), 1,
      sym_line_comment,
    STATE(117), 1,
      sym_reserved_event,
    STATE(189), 1,
      sym_term,
    ACTIONS(531), 2,
      sym_double,
      sym_rational,
    STATE(142), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    STATE(156), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(533), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(153), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(136), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [6171] = 5,
    STATE(117), 1,
      sym_line_comment,
    STATE(121), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(513), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(184), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6220] = 24,
    ACTIONS(515), 1,
      anon_sym_LBRACK,
    ACTIONS(517), 1,
      sym_identifier,
    ACTIONS(519), 1,
      sym_reserved_self,
    ACTIONS(521), 1,
      anon_sym_event,
    ACTIONS(523), 1,
      anon_sym_LBRACE,
    ACTIONS(525), 1,
      sym_quoted_string,
    ACTIONS(527), 1,
      anon_sym_LT,
    ACTIONS(529), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(535), 1,
      anon_sym_true,
    ACTIONS(537), 1,
      anon_sym_u22a4,
    ACTIONS(539), 1,
      anon_sym_false,
    ACTIONS(541), 1,
      anon_sym_u22a5,
    ACTIONS(543), 1,
      sym_value_empty_sequence,
    STATE(107), 1,
      sym_qualified_identifier,
    STATE(117), 1,
      sym_reserved_event,
    STATE(118), 1,
      sym_line_comment,
    STATE(160), 1,
      sym_term,
    ACTIONS(531), 2,
      sym_double,
      sym_rational,
    STATE(142), 2,
      sym_sequence_of_predicate_values,
      sym_simple_value,
    STATE(156), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(533), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(153), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
    STATE(136), 5,
      sym_identifier_reference,
      sym_functional_term,
      sym_function_composition,
      sym_predicate_value,
      sym_sequence_builder,
  [6307] = 5,
    STATE(119), 1,
      sym_line_comment,
    STATE(120), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(513), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(173), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6356] = 4,
    ACTIONS(545), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    STATE(120), 2,
      sym_line_comment,
      aux_sym_function_composition_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(177), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6403] = 5,
    STATE(120), 1,
      aux_sym_function_composition_repeat1,
    STATE(121), 1,
      sym_line_comment,
    ACTIONS(513), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(173), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6452] = 10,
    STATE(23), 1,
      sym_logical_connective,
    STATE(122), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
    ACTIONS(17), 16,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [6511] = 3,
    STATE(123), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(192), 33,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
      anon_sym_DOT,
      anon_sym_u2218,
  [6555] = 3,
    STATE(124), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(188), 33,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
      anon_sym_DOT,
      anon_sym_u2218,
  [6599] = 22,
    ACTIONS(368), 1,
      anon_sym_end,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    STATE(125), 1,
      sym_line_comment,
    STATE(221), 1,
      aux_sym_module_body_repeat3,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(503), 1,
      sym_definition,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [6680] = 22,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    ACTIONS(548), 1,
      anon_sym_end,
    STATE(126), 1,
      sym_line_comment,
    STATE(252), 1,
      aux_sym_module_body_repeat3,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(503), 1,
      sym_definition,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [6761] = 5,
    ACTIONS(107), 1,
      anon_sym_with,
    ACTIONS(550), 1,
      anon_sym_COLON,
    STATE(127), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(105), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_assert,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_LBRACE,
      anon_sym_without,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [6808] = 4,
    ACTIONS(552), 1,
      sym_language_tag,
    STATE(128), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(11), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6853] = 3,
    STATE(129), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(216), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6895] = 4,
    ACTIONS(556), 1,
      anon_sym_LPAREN,
    STATE(130), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(554), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [6939] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(560), 1,
      anon_sym_RBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    STATE(131), 1,
      sym_line_comment,
    STATE(134), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [7023] = 3,
    STATE(132), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(113), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7065] = 3,
    STATE(133), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(119), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7107] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(586), 1,
      anon_sym_RBRACK,
    STATE(134), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [7191] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(588), 1,
      anon_sym_RBRACK,
    STATE(135), 1,
      sym_line_comment,
    STATE(169), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [7275] = 3,
    STATE(136), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(184), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7317] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(590), 1,
      anon_sym_RBRACK,
    STATE(137), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [7401] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(592), 1,
      anon_sym_RBRACK,
    STATE(137), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(138), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [7485] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(594), 1,
      anon_sym_RBRACK,
    STATE(139), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [7569] = 3,
    STATE(140), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(224), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7611] = 3,
    STATE(141), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(248), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7653] = 3,
    STATE(142), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(196), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7695] = 3,
    STATE(143), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(200), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7737] = 3,
    STATE(144), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(244), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7779] = 3,
    STATE(145), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(212), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7821] = 3,
    STATE(146), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(208), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7863] = 3,
    STATE(147), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(204), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7905] = 3,
    STATE(148), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(220), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7947] = 3,
    STATE(149), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(228), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [7989] = 3,
    STATE(150), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(232), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8031] = 3,
    STATE(151), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(236), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8073] = 3,
    STATE(152), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(240), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8115] = 3,
    STATE(153), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(73), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8157] = 3,
    STATE(154), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(77), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8199] = 3,
    STATE(155), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(81), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8241] = 3,
    STATE(156), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(85), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8283] = 3,
    STATE(157), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(89), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8325] = 3,
    STATE(158), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(93), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8367] = 3,
    STATE(159), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(97), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8409] = 4,
    ACTIONS(556), 1,
      anon_sym_LPAREN,
    STATE(160), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(596), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [8453] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(598), 1,
      anon_sym_RBRACK,
    STATE(139), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(161), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [8537] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(600), 1,
      anon_sym_RBRACK,
    STATE(162), 1,
      sym_line_comment,
    STATE(163), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [8621] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(602), 1,
      anon_sym_RBRACK,
    STATE(163), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [8705] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(604), 1,
      anon_sym_RBRACK,
    STATE(164), 1,
      sym_line_comment,
    STATE(166), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [8789] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(606), 1,
      anon_sym_RBRACK,
    STATE(165), 1,
      sym_line_comment,
    STATE(168), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [8873] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(608), 1,
      anon_sym_RBRACK,
    STATE(166), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [8957] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(610), 1,
      anon_sym_RBRACK,
    STATE(167), 1,
      sym_line_comment,
    STATE(170), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9041] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(612), 1,
      anon_sym_RBRACK,
    STATE(168), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9125] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(614), 1,
      anon_sym_RBRACK,
    STATE(169), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9209] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(616), 1,
      anon_sym_RBRACK,
    STATE(170), 1,
      sym_line_comment,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9293] = 23,
    ACTIONS(618), 1,
      anon_sym_LBRACK,
    ACTIONS(621), 1,
      anon_sym_RBRACK,
    ACTIONS(623), 1,
      sym_identifier,
    ACTIONS(626), 1,
      anon_sym_LBRACE,
    ACTIONS(629), 1,
      sym_quoted_string,
    ACTIONS(632), 1,
      anon_sym_LT,
    ACTIONS(635), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(644), 1,
      anon_sym_true,
    ACTIONS(647), 1,
      anon_sym_u22a4,
    ACTIONS(650), 1,
      anon_sym_false,
    ACTIONS(653), 1,
      anon_sym_u22a5,
    ACTIONS(656), 1,
      sym_value_empty_sequence,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(638), 2,
      sym_double,
      sym_rational,
    STATE(171), 2,
      sym_line_comment,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(641), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9375] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(659), 1,
      anon_sym_RBRACK,
    STATE(172), 1,
      sym_line_comment,
    STATE(173), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9459] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(661), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(173), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9543] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(663), 1,
      anon_sym_RBRACK,
    STATE(174), 1,
      sym_line_comment,
    STATE(176), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9627] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(665), 1,
      anon_sym_RBRACK,
    STATE(175), 1,
      sym_line_comment,
    STATE(178), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9711] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(667), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(176), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9795] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(669), 1,
      anon_sym_RBRACK,
    STATE(177), 1,
      sym_line_comment,
    STATE(179), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9879] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(671), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(178), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [9963] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(673), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(179), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10047] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(675), 1,
      anon_sym_RBRACK,
    STATE(180), 1,
      sym_line_comment,
    STATE(181), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10131] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(677), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(181), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10215] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(679), 1,
      anon_sym_RBRACK,
    STATE(182), 1,
      sym_line_comment,
    STATE(184), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10299] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(681), 1,
      anon_sym_RBRACK,
    STATE(183), 1,
      sym_line_comment,
    STATE(186), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10383] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(683), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(184), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10467] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(685), 1,
      anon_sym_RBRACK,
    STATE(185), 1,
      sym_line_comment,
    STATE(187), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10551] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(687), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(186), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10635] = 24,
    ACTIONS(558), 1,
      anon_sym_LBRACK,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(564), 1,
      anon_sym_LBRACE,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(584), 1,
      sym_value_empty_sequence,
    ACTIONS(689), 1,
      anon_sym_RBRACK,
    STATE(171), 1,
      aux_sym_sequence_of_predicate_values_repeat1,
    STATE(187), 1,
      sym_line_comment,
    STATE(307), 1,
      sym_simple_value,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(340), 1,
      sym_identifier_reference,
    STATE(381), 1,
      sym_sequence_of_predicate_values,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(407), 3,
      sym_predicate_value,
      sym_value_constructor,
      sym_mapping_value,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [10719] = 4,
    ACTIONS(556), 1,
      anon_sym_LPAREN,
    STATE(188), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(691), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [10763] = 4,
    ACTIONS(556), 1,
      anon_sym_LPAREN,
    STATE(189), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(693), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [10807] = 3,
    STATE(190), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(7), 31,
      anon_sym_is,
      anon_sym_end,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [10849] = 3,
    STATE(191), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(695), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [10890] = 3,
    STATE(192), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(697), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [10931] = 3,
    STATE(193), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(699), 30,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [10972] = 3,
    STATE(194), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(701), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11013] = 3,
    STATE(195), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(697), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11054] = 3,
    STATE(196), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(703), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11095] = 3,
    STATE(197), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(705), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11136] = 3,
    STATE(198), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(697), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11177] = 3,
    STATE(199), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(697), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11218] = 3,
    STATE(200), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(707), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11259] = 3,
    STATE(201), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(697), 30,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_RBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_and,
      anon_sym_u2227,
      anon_sym_or,
      anon_sym_u2228,
      anon_sym_xor,
      anon_sym_u22bb,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
  [11300] = 22,
    ACTIONS(709), 1,
      anon_sym_LBRACK,
    ACTIONS(711), 1,
      sym_identifier,
    ACTIONS(713), 1,
      anon_sym_LBRACE,
    ACTIONS(715), 1,
      sym_quoted_string,
    ACTIONS(717), 1,
      anon_sym_LT,
    ACTIONS(719), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(725), 1,
      anon_sym_true,
    ACTIONS(727), 1,
      anon_sym_u22a4,
    ACTIONS(729), 1,
      anon_sym_false,
    ACTIONS(731), 1,
      anon_sym_u22a5,
    ACTIONS(733), 1,
      sym_value_empty_sequence,
    STATE(202), 1,
      sym_line_comment,
    STATE(622), 1,
      sym_simple_value,
    STATE(693), 1,
      sym_qualified_identifier,
    STATE(706), 1,
      sym_identifier_reference,
    STATE(746), 1,
      sym_value,
    ACTIONS(721), 2,
      sym_double,
      sym_rational,
    STATE(612), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(723), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(790), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(651), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11378] = 22,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(735), 1,
      anon_sym_LBRACK,
    ACTIONS(737), 1,
      anon_sym_LBRACE,
    ACTIONS(739), 1,
      sym_value_empty_sequence,
    STATE(203), 1,
      sym_line_comment,
    STATE(309), 1,
      sym_simple_value,
    STATE(328), 1,
      sym_identifier_reference,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(385), 1,
      sym_value,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(379), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11456] = 22,
    ACTIONS(709), 1,
      anon_sym_LBRACK,
    ACTIONS(711), 1,
      sym_identifier,
    ACTIONS(713), 1,
      anon_sym_LBRACE,
    ACTIONS(715), 1,
      sym_quoted_string,
    ACTIONS(717), 1,
      anon_sym_LT,
    ACTIONS(719), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(725), 1,
      anon_sym_true,
    ACTIONS(727), 1,
      anon_sym_u22a4,
    ACTIONS(729), 1,
      anon_sym_false,
    ACTIONS(731), 1,
      anon_sym_u22a5,
    ACTIONS(733), 1,
      sym_value_empty_sequence,
    STATE(204), 1,
      sym_line_comment,
    STATE(622), 1,
      sym_simple_value,
    STATE(693), 1,
      sym_qualified_identifier,
    STATE(706), 1,
      sym_identifier_reference,
    STATE(754), 1,
      sym_value,
    ACTIONS(721), 2,
      sym_double,
      sym_rational,
    STATE(612), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(723), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(790), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(651), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11534] = 22,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(145), 1,
      anon_sym_true,
    ACTIONS(147), 1,
      anon_sym_u22a4,
    ACTIONS(149), 1,
      anon_sym_false,
    ACTIONS(151), 1,
      anon_sym_u22a5,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(741), 1,
      anon_sym_LBRACK,
    ACTIONS(743), 1,
      anon_sym_LBRACE,
    ACTIONS(745), 1,
      sym_value_empty_sequence,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(205), 1,
      sym_line_comment,
    STATE(272), 1,
      sym_simple_value,
    STATE(301), 1,
      sym_identifier_reference,
    STATE(335), 1,
      sym_value,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(341), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11612] = 22,
    ACTIONS(747), 1,
      anon_sym_LBRACK,
    ACTIONS(749), 1,
      sym_identifier,
    ACTIONS(751), 1,
      anon_sym_LBRACE,
    ACTIONS(753), 1,
      sym_quoted_string,
    ACTIONS(755), 1,
      anon_sym_LT,
    ACTIONS(757), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(763), 1,
      anon_sym_true,
    ACTIONS(765), 1,
      anon_sym_u22a4,
    ACTIONS(767), 1,
      anon_sym_false,
    ACTIONS(769), 1,
      anon_sym_u22a5,
    ACTIONS(771), 1,
      sym_value_empty_sequence,
    STATE(206), 1,
      sym_line_comment,
    STATE(587), 1,
      sym_simple_value,
    STATE(610), 1,
      sym_identifier_reference,
    STATE(634), 1,
      sym_qualified_identifier,
    STATE(708), 1,
      sym_value,
    ACTIONS(759), 2,
      sym_double,
      sym_rational,
    STATE(592), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(761), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(666), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(589), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11690] = 22,
    ACTIONS(747), 1,
      anon_sym_LBRACK,
    ACTIONS(749), 1,
      sym_identifier,
    ACTIONS(751), 1,
      anon_sym_LBRACE,
    ACTIONS(753), 1,
      sym_quoted_string,
    ACTIONS(755), 1,
      anon_sym_LT,
    ACTIONS(757), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(763), 1,
      anon_sym_true,
    ACTIONS(765), 1,
      anon_sym_u22a4,
    ACTIONS(767), 1,
      anon_sym_false,
    ACTIONS(769), 1,
      anon_sym_u22a5,
    ACTIONS(771), 1,
      sym_value_empty_sequence,
    STATE(207), 1,
      sym_line_comment,
    STATE(587), 1,
      sym_simple_value,
    STATE(610), 1,
      sym_identifier_reference,
    STATE(634), 1,
      sym_qualified_identifier,
    STATE(671), 1,
      sym_value,
    ACTIONS(759), 2,
      sym_double,
      sym_rational,
    STATE(592), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(761), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(666), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(589), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11768] = 22,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(145), 1,
      anon_sym_true,
    ACTIONS(147), 1,
      anon_sym_u22a4,
    ACTIONS(149), 1,
      anon_sym_false,
    ACTIONS(151), 1,
      anon_sym_u22a5,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(741), 1,
      anon_sym_LBRACK,
    ACTIONS(743), 1,
      anon_sym_LBRACE,
    ACTIONS(745), 1,
      sym_value_empty_sequence,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(208), 1,
      sym_line_comment,
    STATE(272), 1,
      sym_simple_value,
    STATE(301), 1,
      sym_identifier_reference,
    STATE(358), 1,
      sym_value,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(341), 3,
      sym_value_constructor,
      sym_mapping_value,
      sym_sequence_of_values,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [11846] = 4,
    STATE(209), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(775), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(773), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [11887] = 4,
    STATE(210), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(779), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(777), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [11928] = 4,
    STATE(211), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(783), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(781), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [11969] = 4,
    STATE(212), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(787), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(785), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [12010] = 4,
    STATE(213), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(791), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(789), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [12051] = 4,
    STATE(214), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(795), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(793), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [12092] = 4,
    STATE(215), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(799), 13,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      anon_sym_with,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
      anon_sym_not,
      anon_sym_forall,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(797), 15,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_u00ac,
      anon_sym_u2200,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
  [12133] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(801), 1,
      anon_sym_RBRACK,
    STATE(216), 1,
      sym_line_comment,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12204] = 19,
    ACTIONS(803), 1,
      anon_sym_RBRACK,
    ACTIONS(805), 1,
      sym_identifier,
    ACTIONS(808), 1,
      sym_quoted_string,
    ACTIONS(811), 1,
      anon_sym_LT,
    ACTIONS(814), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(823), 1,
      anon_sym_true,
    ACTIONS(826), 1,
      anon_sym_u22a4,
    ACTIONS(829), 1,
      anon_sym_false,
    ACTIONS(832), 1,
      anon_sym_u22a5,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(817), 2,
      sym_double,
      sym_rational,
    STATE(217), 2,
      sym_line_comment,
      aux_sym_sequence_of_values_repeat1,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(820), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12273] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(835), 1,
      anon_sym_RBRACK,
    STATE(218), 1,
      sym_line_comment,
    STATE(227), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12344] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(837), 1,
      anon_sym_RBRACK,
    STATE(216), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(219), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12415] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(839), 1,
      anon_sym_RBRACK,
    STATE(220), 1,
      sym_line_comment,
    STATE(225), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12486] = 17,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    ACTIONS(548), 1,
      anon_sym_end,
    STATE(221), 1,
      sym_line_comment,
    STATE(222), 1,
      aux_sym_module_body_repeat3,
    STATE(503), 1,
      sym_definition,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [12551] = 16,
    ACTIONS(841), 1,
      anon_sym_end,
    ACTIONS(843), 1,
      anon_sym_event,
    ACTIONS(846), 1,
      anon_sym_datatype,
    ACTIONS(849), 1,
      anon_sym_dimension,
    ACTIONS(852), 1,
      anon_sym_entity,
    ACTIONS(855), 1,
      anon_sym_enum,
    ACTIONS(858), 1,
      anon_sym_metric,
    ACTIONS(861), 1,
      anon_sym_property,
    ACTIONS(864), 1,
      anon_sym_rdf,
    ACTIONS(867), 1,
      anon_sym_structure,
    ACTIONS(870), 1,
      anon_sym_class,
    ACTIONS(873), 1,
      anon_sym_union,
    STATE(503), 1,
      sym_definition,
    STATE(222), 2,
      sym_line_comment,
      aux_sym_module_body_repeat3,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [12614] = 17,
    ACTIONS(368), 1,
      anon_sym_end,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    STATE(222), 1,
      aux_sym_module_body_repeat3,
    STATE(223), 1,
      sym_line_comment,
    STATE(503), 1,
      sym_definition,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [12679] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(876), 1,
      anon_sym_RBRACK,
    STATE(224), 1,
      sym_line_comment,
    STATE(226), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12750] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(878), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(225), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12821] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(880), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(226), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12892] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(882), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(227), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [12963] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(884), 1,
      anon_sym_RBRACK,
    STATE(228), 1,
      sym_line_comment,
    STATE(229), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13034] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(886), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(229), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13105] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(888), 1,
      anon_sym_RBRACK,
    STATE(230), 1,
      sym_line_comment,
    STATE(232), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13176] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(890), 1,
      anon_sym_RBRACK,
    STATE(231), 1,
      sym_line_comment,
    STATE(234), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13247] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(892), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(232), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13318] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(894), 1,
      anon_sym_RBRACK,
    STATE(233), 1,
      sym_line_comment,
    STATE(235), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13389] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(896), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(234), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13460] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(898), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(235), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13531] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(900), 1,
      anon_sym_RBRACK,
    STATE(236), 1,
      sym_line_comment,
    STATE(237), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13602] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(902), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(237), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13673] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(904), 1,
      anon_sym_RBRACK,
    STATE(238), 1,
      sym_line_comment,
    STATE(240), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13744] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(906), 1,
      anon_sym_RBRACK,
    STATE(239), 1,
      sym_line_comment,
    STATE(242), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13815] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(908), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(240), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13886] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(910), 1,
      anon_sym_RBRACK,
    STATE(241), 1,
      sym_line_comment,
    STATE(243), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [13957] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(912), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(242), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14028] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(914), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(243), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14099] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(916), 1,
      anon_sym_RBRACK,
    STATE(244), 1,
      sym_line_comment,
    STATE(245), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14170] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(918), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(245), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14241] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(920), 1,
      anon_sym_RBRACK,
    STATE(246), 1,
      sym_line_comment,
    STATE(248), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14312] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(922), 1,
      anon_sym_RBRACK,
    STATE(247), 1,
      sym_line_comment,
    STATE(250), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14383] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(924), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(248), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14454] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(926), 1,
      anon_sym_RBRACK,
    STATE(249), 1,
      sym_line_comment,
    STATE(251), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14525] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(928), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(250), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14596] = 20,
    ACTIONS(562), 1,
      sym_identifier,
    ACTIONS(566), 1,
      sym_quoted_string,
    ACTIONS(568), 1,
      anon_sym_LT,
    ACTIONS(570), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(576), 1,
      anon_sym_true,
    ACTIONS(578), 1,
      anon_sym_u22a4,
    ACTIONS(580), 1,
      anon_sym_false,
    ACTIONS(582), 1,
      anon_sym_u22a5,
    ACTIONS(930), 1,
      anon_sym_RBRACK,
    STATE(217), 1,
      aux_sym_sequence_of_values_repeat1,
    STATE(251), 1,
      sym_line_comment,
    STATE(329), 1,
      sym_qualified_identifier,
    STATE(417), 1,
      sym_simple_value,
    STATE(433), 1,
      sym_identifier_reference,
    ACTIONS(572), 2,
      sym_double,
      sym_rational,
    STATE(310), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(453), 2,
      sym_value_constructor,
      sym_mapping_value,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(574), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(319), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14667] = 17,
    ACTIONS(378), 1,
      anon_sym_event,
    ACTIONS(380), 1,
      anon_sym_datatype,
    ACTIONS(382), 1,
      anon_sym_dimension,
    ACTIONS(384), 1,
      anon_sym_entity,
    ACTIONS(386), 1,
      anon_sym_enum,
    ACTIONS(388), 1,
      anon_sym_metric,
    ACTIONS(390), 1,
      anon_sym_property,
    ACTIONS(392), 1,
      anon_sym_rdf,
    ACTIONS(394), 1,
      anon_sym_structure,
    ACTIONS(396), 1,
      anon_sym_class,
    ACTIONS(398), 1,
      anon_sym_union,
    ACTIONS(932), 1,
      anon_sym_end,
    STATE(222), 1,
      aux_sym_module_body_repeat3,
    STATE(252), 1,
      sym_line_comment,
    STATE(503), 1,
      sym_definition,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(469), 12,
      sym_datatype_def,
      sym_dimension_def,
      sym_entity_def,
      sym_enum_def,
      sym_event_def,
      sym_metric_def,
      sym_metric_group_def,
      sym_property_def,
      sym_rdf_def,
      sym_structure_def,
      sym_type_class_def,
      sym_union_def,
  [14732] = 19,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(145), 1,
      anon_sym_true,
    ACTIONS(147), 1,
      anon_sym_u22a4,
    ACTIONS(149), 1,
      anon_sym_false,
    ACTIONS(151), 1,
      anon_sym_u22a5,
    ACTIONS(252), 1,
      sym_identifier,
    ACTIONS(934), 1,
      sym_kw_is_fixed,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(253), 1,
      sym_line_comment,
    STATE(498), 1,
      sym_restriction_value,
    STATE(1122), 1,
      sym_identifier_reference,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(497), 2,
      sym_simple_value,
      sym_value_constructor,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14800] = 7,
    ACTIONS(936), 1,
      anon_sym_COLON,
    STATE(254), 1,
      sym_line_comment,
    STATE(258), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(938), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(105), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [14843] = 18,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(145), 1,
      anon_sym_true,
    ACTIONS(147), 1,
      anon_sym_u22a4,
    ACTIONS(149), 1,
      anon_sym_false,
    ACTIONS(151), 1,
      anon_sym_u22a5,
    ACTIONS(252), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(255), 1,
      sym_line_comment,
    STATE(528), 1,
      sym_restriction_value,
    STATE(1122), 1,
      sym_identifier_reference,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    STATE(497), 2,
      sym_simple_value,
      sym_value_constructor,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [14908] = 6,
    STATE(256), 1,
      sym_line_comment,
    STATE(257), 1,
      aux_sym_function_composition_repeat1,
    ACTIONS(938), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(175), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(173), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [14948] = 5,
    ACTIONS(940), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    STATE(257), 2,
      sym_line_comment,
      aux_sym_function_composition_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(179), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(177), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [14986] = 6,
    STATE(257), 1,
      aux_sym_function_composition_repeat1,
    STATE(258), 1,
      sym_line_comment,
    ACTIONS(938), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(175), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(173), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15026] = 6,
    STATE(256), 1,
      aux_sym_function_composition_repeat1,
    STATE(259), 1,
      sym_line_comment,
    ACTIONS(938), 2,
      anon_sym_DOT,
      anon_sym_u2218,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(186), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(184), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15066] = 3,
    STATE(260), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(943), 22,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_LBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [15099] = 7,
    ACTIONS(947), 1,
      anon_sym_AT,
    ACTIONS(950), 1,
      anon_sym_assert,
    STATE(356), 1,
      sym_annotation,
    STATE(261), 2,
      sym_line_comment,
      aux_sym_module_body_repeat2,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(945), 16,
      anon_sym_end,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [15140] = 4,
    STATE(262), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(190), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(188), 14,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DOT,
      anon_sym_u2218,
  [15175] = 4,
    STATE(263), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(194), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(192), 14,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DOT,
      anon_sym_u2218,
  [15210] = 5,
    ACTIONS(953), 1,
      sym_language_tag,
    STATE(264), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(13), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(11), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15246] = 3,
    STATE(265), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(955), 21,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [15278] = 4,
    STATE(266), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(111), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(113), 13,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15312] = 4,
    STATE(267), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(117), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(119), 13,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15346] = 4,
    STATE(268), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(230), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(228), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15379] = 5,
    ACTIONS(957), 1,
      sym_language_tag,
    STATE(269), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(13), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(11), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [15414] = 5,
    ACTIONS(963), 1,
      anon_sym_LPAREN,
    STATE(270), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(961), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(959), 11,
      anon_sym_LBRACK,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15449] = 4,
    STATE(271), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(186), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(184), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15482] = 4,
    STATE(272), 1,
      sym_line_comment,
    ACTIONS(967), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(965), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [15515] = 4,
    STATE(273), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(226), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(224), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15548] = 4,
    STATE(274), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(75), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(73), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15581] = 4,
    STATE(275), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(79), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(77), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15614] = 4,
    STATE(276), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(83), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(81), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15647] = 4,
    STATE(277), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(87), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(85), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15680] = 4,
    STATE(278), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(250), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(248), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15713] = 4,
    STATE(279), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(198), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(196), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15746] = 4,
    STATE(280), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(91), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(89), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15779] = 4,
    STATE(281), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(95), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(93), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15812] = 4,
    STATE(282), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(202), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(200), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15845] = 4,
    STATE(283), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(99), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(97), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15878] = 4,
    STATE(284), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(246), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(244), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15911] = 4,
    STATE(285), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(214), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(212), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15944] = 4,
    STATE(286), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(210), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(208), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [15977] = 4,
    STATE(287), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(218), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(216), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16010] = 4,
    STATE(288), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(206), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(204), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16043] = 4,
    STATE(289), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(222), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(220), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16076] = 4,
    STATE(290), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(234), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(232), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16109] = 4,
    STATE(291), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(238), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(236), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16142] = 4,
    STATE(292), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(242), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(240), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16175] = 4,
    STATE(293), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(9), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(7), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16208] = 4,
    STATE(294), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(105), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16241] = 4,
    STATE(295), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(101), 12,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16274] = 7,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(169), 1,
      anon_sym_COLON,
    ACTIONS(971), 1,
      anon_sym_as,
    STATE(296), 1,
      sym_line_comment,
    STATE(405), 1,
      sym_iri,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(969), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [16313] = 10,
    ACTIONS(973), 1,
      anon_sym_end,
    STATE(23), 1,
      sym_logical_connective,
    STATE(297), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
  [16357] = 4,
    STATE(298), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(83), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(81), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [16389] = 4,
    STATE(299), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(99), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(97), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [16421] = 10,
    ACTIONS(975), 1,
      anon_sym_RBRACE,
    ACTIONS(986), 1,
      anon_sym_explicitTimezone,
    ACTIONS(989), 1,
      anon_sym_pattern,
    STATE(491), 1,
      sym_restriction_facet,
    ACTIONS(980), 2,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
    STATE(300), 2,
      sym_line_comment,
      aux_sym_datatype_type_restrictions_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(977), 3,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
    ACTIONS(983), 4,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
    STATE(496), 5,
      sym_length_restriction_facet,
      sym_digit_restriction_facet,
      sym_value_restriction_facet,
      sym_tz_restriction_facet,
      sym_pattern_restriction_facet,
  [16465] = 4,
    ACTIONS(992), 1,
      anon_sym_LPAREN,
    STATE(301), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(965), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [16497] = 12,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    STATE(302), 1,
      sym_line_comment,
    STATE(1039), 1,
      sym_simple_value,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    ACTIONS(147), 2,
      anon_sym_true,
      anon_sym_u22a4,
    ACTIONS(151), 2,
      anon_sym_false,
      anon_sym_u22a5,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [16545] = 10,
    ACTIONS(994), 1,
      anon_sym_end,
    STATE(23), 1,
      sym_logical_connective,
    STATE(303), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
  [16589] = 5,
    ACTIONS(996), 1,
      anon_sym_COLON,
    STATE(304), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(105), 12,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16623] = 4,
    STATE(305), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(9), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(7), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [16655] = 12,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    STATE(306), 1,
      sym_line_comment,
    STATE(1132), 1,
      sym_simple_value,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    ACTIONS(147), 2,
      anon_sym_true,
      anon_sym_u22a4,
    ACTIONS(151), 2,
      anon_sym_false,
      anon_sym_u22a5,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [16703] = 5,
    STATE(307), 1,
      sym_line_comment,
    ACTIONS(998), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(198), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(196), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16737] = 10,
    ACTIONS(1000), 1,
      anon_sym_RPAREN,
    STATE(23), 1,
      sym_logical_connective,
    STATE(308), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
  [16781] = 5,
    STATE(309), 1,
      sym_line_comment,
    ACTIONS(998), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(965), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [16815] = 4,
    STATE(310), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(87), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(85), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [16847] = 4,
    STATE(311), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(79), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(77), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [16879] = 5,
    ACTIONS(1006), 1,
      anon_sym_AT,
    ACTIONS(1008), 1,
      sym_controlled_language_tag,
    STATE(312), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1004), 17,
      anon_sym_end,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [16913] = 4,
    STATE(313), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(91), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(89), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [16945] = 11,
    ACTIONS(1010), 1,
      anon_sym_RBRACE,
    ACTIONS(1018), 1,
      anon_sym_explicitTimezone,
    ACTIONS(1020), 1,
      anon_sym_pattern,
    STATE(300), 1,
      aux_sym_datatype_type_restrictions_repeat1,
    STATE(314), 1,
      sym_line_comment,
    STATE(491), 1,
      sym_restriction_facet,
    ACTIONS(1014), 2,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1012), 3,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
    ACTIONS(1016), 4,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
    STATE(496), 5,
      sym_length_restriction_facet,
      sym_digit_restriction_facet,
      sym_value_restriction_facet,
      sym_tz_restriction_facet,
      sym_pattern_restriction_facet,
  [16991] = 10,
    ACTIONS(1022), 1,
      anon_sym_end,
    STATE(23), 1,
      sym_logical_connective,
    STATE(315), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
  [17035] = 7,
    ACTIONS(1026), 1,
      anon_sym_import,
    ACTIONS(1029), 1,
      anon_sym_from,
    STATE(415), 1,
      sym_import_statement,
    STATE(1131), 1,
      sym_from_clause,
    STATE(316), 2,
      sym_line_comment,
      aux_sym_module_body_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1024), 14,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [17073] = 10,
    ACTIONS(1032), 1,
      anon_sym_end,
    STATE(23), 1,
      sym_logical_connective,
    STATE(317), 1,
      sym_line_comment,
    ACTIONS(501), 2,
      anon_sym_and,
      anon_sym_u2227,
    ACTIONS(503), 2,
      anon_sym_or,
      anon_sym_u2228,
    ACTIONS(505), 2,
      anon_sym_xor,
      anon_sym_u22bb,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(507), 3,
      anon_sym_implies,
      anon_sym_EQ_EQ_GT,
      anon_sym_u21d2,
    ACTIONS(509), 3,
      anon_sym_iff,
      anon_sym_LT_EQ_EQ_GT,
      anon_sym_u21d4,
    STATE(211), 5,
      sym_logical_conjunction,
      sym_logical_disjunction,
      sym_logical_exclusive_disjunction,
      sym_logical_implication,
      sym_logical_biconditional,
  [17117] = 4,
    STATE(318), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(95), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(93), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [17149] = 4,
    STATE(319), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(75), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(73), 13,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [17181] = 12,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    STATE(320), 1,
      sym_line_comment,
    STATE(1013), 1,
      sym_simple_value,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    ACTIONS(147), 2,
      anon_sym_true,
      anon_sym_u22a4,
    ACTIONS(151), 2,
      anon_sym_false,
      anon_sym_u22a5,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [17229] = 12,
    ACTIONS(135), 1,
      sym_quoted_string,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(139), 1,
      anon_sym_POUND_LBRACK,
    STATE(321), 1,
      sym_line_comment,
    STATE(1006), 1,
      sym_simple_value,
    ACTIONS(141), 2,
      sym_double,
      sym_rational,
    ACTIONS(147), 2,
      anon_sym_true,
      anon_sym_u22a4,
    ACTIONS(151), 2,
      anon_sym_false,
      anon_sym_u22a5,
    STATE(8), 2,
      sym_boolean_truth,
      sym_boolean_falsity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(143), 3,
      sym_decimal,
      sym_integer,
      sym_unsigned,
    STATE(5), 4,
      sym_string,
      sym_iri,
      sym_binary,
      sym_boolean,
  [17277] = 3,
    STATE(322), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1034), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [17306] = 4,
    STATE(323), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1038), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1036), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17337] = 4,
    STATE(324), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1042), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1040), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17368] = 4,
    STATE(325), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1046), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1044), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17399] = 4,
    STATE(326), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1050), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1048), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17430] = 4,
    STATE(327), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1054), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1052), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17461] = 5,
    ACTIONS(1056), 1,
      anon_sym_LPAREN,
    STATE(328), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(965), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17494] = 4,
    STATE(329), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(105), 12,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17525] = 4,
    STATE(330), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1060), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1058), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17556] = 4,
    STATE(331), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1064), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1062), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17587] = 4,
    STATE(332), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1068), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1066), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17618] = 3,
    STATE(333), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1070), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [17647] = 4,
    STATE(334), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1074), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1072), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17678] = 3,
    STATE(335), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1076), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [17707] = 3,
    STATE(336), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1070), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [17736] = 4,
    STATE(337), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1080), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1078), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17767] = 4,
    STATE(338), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1084), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1082), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17798] = 4,
    STATE(339), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1088), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1086), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17829] = 5,
    ACTIONS(1056), 1,
      anon_sym_LPAREN,
    STATE(340), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1092), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1090), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17862] = 3,
    STATE(341), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(965), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [17891] = 4,
    STATE(342), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1096), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1094), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17922] = 4,
    STATE(343), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1100), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1098), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17953] = 4,
    STATE(344), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1104), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1102), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [17984] = 4,
    STATE(345), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1108), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1106), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18015] = 4,
    STATE(346), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1112), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1110), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18046] = 3,
    STATE(347), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1114), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18075] = 4,
    STATE(348), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1118), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1116), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18106] = 4,
    STATE(349), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1122), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1120), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18137] = 4,
    STATE(350), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1126), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1124), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18168] = 3,
    STATE(351), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1128), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18197] = 4,
    STATE(352), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1132), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1130), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18228] = 3,
    STATE(353), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1134), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18257] = 4,
    STATE(354), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1138), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1136), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18288] = 4,
    STATE(355), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(101), 12,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18319] = 3,
    STATE(356), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1140), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18348] = 3,
    STATE(357), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1142), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18377] = 3,
    STATE(358), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1144), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18406] = 3,
    STATE(359), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1146), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [18435] = 3,
    STATE(360), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1146), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [18464] = 3,
    STATE(361), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1148), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18493] = 3,
    STATE(362), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1150), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18522] = 3,
    STATE(363), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1152), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18551] = 3,
    STATE(364), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1154), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [18580] = 10,
    ACTIONS(1018), 1,
      anon_sym_explicitTimezone,
    ACTIONS(1020), 1,
      anon_sym_pattern,
    STATE(314), 1,
      aux_sym_datatype_type_restrictions_repeat1,
    STATE(365), 1,
      sym_line_comment,
    STATE(491), 1,
      sym_restriction_facet,
    ACTIONS(1014), 2,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1012), 3,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
    ACTIONS(1016), 4,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
    STATE(496), 5,
      sym_length_restriction_facet,
      sym_digit_restriction_facet,
      sym_value_restriction_facet,
      sym_tz_restriction_facet,
      sym_pattern_restriction_facet,
  [18623] = 3,
    STATE(366), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1156), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18652] = 3,
    STATE(367), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1158), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18681] = 3,
    STATE(368), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1160), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18710] = 3,
    STATE(369), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1162), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18739] = 3,
    STATE(370), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1164), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18768] = 3,
    STATE(371), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1166), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18797] = 3,
    STATE(372), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1168), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18826] = 3,
    STATE(373), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1170), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18855] = 4,
    STATE(374), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1174), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1172), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18886] = 3,
    STATE(375), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1176), 18,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
  [18915] = 4,
    STATE(376), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1180), 8,
      sym_identifier,
      sym_reserved_self,
      anon_sym_event,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1178), 10,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [18946] = 3,
    STATE(377), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1182), 18,
      anon_sym_end,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_source,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_identity,
  [18975] = 4,
    STATE(378), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1184), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(699), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19005] = 4,
    STATE(379), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(965), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19035] = 4,
    STATE(380), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(250), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(248), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19065] = 4,
    STATE(381), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(198), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(196), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19095] = 4,
    STATE(382), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1186), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1142), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19125] = 4,
    STATE(383), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(202), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(200), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19155] = 4,
    STATE(384), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1188), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1152), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19185] = 4,
    STATE(385), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1190), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1144), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19215] = 4,
    STATE(386), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(246), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(244), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19245] = 4,
    STATE(387), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1192), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1148), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19275] = 4,
    STATE(388), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1194), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1164), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19305] = 4,
    STATE(389), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(214), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(212), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19335] = 4,
    STATE(390), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(210), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(208), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19365] = 4,
    STATE(391), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1196), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1182), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19395] = 4,
    STATE(392), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1198), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1128), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19425] = 4,
    STATE(393), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1200), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1134), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19455] = 4,
    STATE(394), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1202), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1160), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19485] = 4,
    STATE(395), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(206), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(204), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19515] = 4,
    STATE(396), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(222), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(220), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19545] = 4,
    STATE(397), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1204), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1156), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19575] = 4,
    STATE(398), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1206), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1158), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19605] = 4,
    STATE(399), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1208), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1162), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19635] = 4,
    STATE(400), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(230), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(228), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19665] = 4,
    STATE(401), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(234), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(232), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19695] = 4,
    STATE(402), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(238), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(236), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19725] = 4,
    STATE(403), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1210), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1166), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19755] = 4,
    STATE(404), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(242), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(240), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19785] = 4,
    ACTIONS(1214), 1,
      anon_sym_as,
    STATE(405), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1212), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [19815] = 7,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(406), 1,
      sym_line_comment,
    STATE(409), 1,
      sym_function_body,
    STATE(435), 1,
      sym_annotation_only_body,
    ACTIONS(1220), 2,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1218), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [19851] = 4,
    STATE(407), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1092), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1090), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [19881] = 4,
    ACTIONS(1224), 1,
      anon_sym_as,
    STATE(408), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1222), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [19911] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(409), 1,
      sym_line_comment,
    STATE(436), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1226), 15,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [19943] = 8,
    ACTIONS(1228), 1,
      anon_sym_is,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    STATE(410), 1,
      sym_line_comment,
    STATE(440), 1,
      sym_from_definition_clause,
    STATE(507), 1,
      sym_type_class_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1230), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [19981] = 4,
    STATE(411), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(218), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(216), 11,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
      sym_value_empty_sequence,
  [20011] = 3,
    STATE(412), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1236), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20038] = 7,
    ACTIONS(1228), 1,
      anon_sym_is,
    ACTIONS(1232), 1,
      anon_sym_from,
    STATE(413), 1,
      sym_line_comment,
    STATE(447), 1,
      sym_from_definition_clause,
    STATE(481), 1,
      sym_type_class_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1238), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20073] = 3,
    STATE(414), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1240), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20100] = 3,
    STATE(415), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1242), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20127] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1244), 1,
      anon_sym_is,
    STATE(416), 1,
      sym_line_comment,
    STATE(446), 1,
      sym_from_definition_clause,
    STATE(513), 1,
      sym_dimension_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1246), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20162] = 5,
    STATE(417), 1,
      sym_line_comment,
    ACTIONS(998), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1250), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1248), 8,
      anon_sym_RBRACK,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
  [20193] = 3,
    STATE(418), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1252), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20220] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1254), 1,
      anon_sym_is,
    STATE(419), 1,
      sym_line_comment,
    STATE(455), 1,
      sym_from_definition_clause,
    STATE(518), 1,
      sym_entity_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1256), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20255] = 3,
    STATE(420), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1258), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20282] = 3,
    STATE(421), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1260), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20309] = 7,
    ACTIONS(1216), 1,
      anon_sym_is,
    ACTIONS(1264), 1,
      anon_sym_LBRACE,
    STATE(422), 1,
      sym_line_comment,
    STATE(456), 1,
      sym_datatype_type_restrictions,
    STATE(529), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1262), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20344] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1268), 1,
      anon_sym_of,
    STATE(423), 1,
      sym_line_comment,
    STATE(450), 1,
      sym_from_definition_clause,
    STATE(532), 1,
      sym_enum_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1266), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20379] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1270), 1,
      anon_sym_is,
    STATE(424), 1,
      sym_line_comment,
    STATE(461), 1,
      sym_from_definition_clause,
    STATE(546), 1,
      sym_metric_group_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1272), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20414] = 6,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(425), 1,
      sym_line_comment,
    STATE(479), 1,
      sym_annotation_only_body,
    ACTIONS(1276), 2,
      anon_sym_a,
      anon_sym_type,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1274), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20447] = 3,
    STATE(426), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1278), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20474] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1282), 1,
      anon_sym_of,
    STATE(427), 1,
      sym_line_comment,
    STATE(445), 1,
      sym_from_definition_clause,
    STATE(516), 1,
      sym_union_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1280), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20509] = 3,
    STATE(428), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1284), 16,
      anon_sym_is,
      anon_sym_end,
      anon_sym_RPAREN,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
      anon_sym_DASH_GT,
      anon_sym_u2192,
  [20536] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1286), 1,
      anon_sym_is,
    STATE(429), 1,
      sym_line_comment,
    STATE(462), 1,
      sym_from_definition_clause,
    STATE(482), 1,
      sym_structure_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1288), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20571] = 7,
    ACTIONS(1216), 1,
      anon_sym_is,
    ACTIONS(1264), 1,
      anon_sym_LBRACE,
    STATE(430), 1,
      sym_line_comment,
    STATE(441), 1,
      sym_datatype_type_restrictions,
    STATE(485), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1290), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20606] = 7,
    ACTIONS(1232), 1,
      anon_sym_from,
    ACTIONS(1292), 1,
      anon_sym_is,
    STATE(431), 1,
      sym_line_comment,
    STATE(454), 1,
      sym_from_definition_clause,
    STATE(541), 1,
      sym_event_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1294), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20641] = 3,
    STATE(432), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1296), 16,
      anon_sym_end,
      anon_sym_import,
      anon_sym_from,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20668] = 5,
    ACTIONS(1056), 1,
      anon_sym_LPAREN,
    STATE(433), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1250), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1248), 8,
      anon_sym_RBRACK,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
  [20698] = 4,
    ACTIONS(1300), 1,
      anon_sym_as,
    STATE(434), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1298), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20726] = 3,
    STATE(435), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1302), 15,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20752] = 3,
    STATE(436), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1304), 15,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20778] = 3,
    STATE(437), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1306), 15,
      anon_sym_end,
      anon_sym_event,
      anon_sym_for,
      anon_sym_def,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20804] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(438), 1,
      sym_line_comment,
    STATE(470), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1308), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20833] = 3,
    STATE(439), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1310), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20858] = 5,
    ACTIONS(1228), 1,
      anon_sym_is,
    STATE(440), 1,
      sym_line_comment,
    STATE(475), 1,
      sym_type_class_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1312), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20887] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(441), 1,
      sym_line_comment,
    STATE(548), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1314), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20916] = 3,
    STATE(442), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1316), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20941] = 3,
    STATE(443), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1318), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20966] = 3,
    STATE(444), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1320), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [20991] = 5,
    ACTIONS(1282), 1,
      anon_sym_of,
    STATE(445), 1,
      sym_line_comment,
    STATE(477), 1,
      sym_union_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1322), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21020] = 5,
    ACTIONS(1244), 1,
      anon_sym_is,
    STATE(446), 1,
      sym_line_comment,
    STATE(487), 1,
      sym_dimension_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1324), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21049] = 5,
    ACTIONS(1228), 1,
      anon_sym_is,
    STATE(447), 1,
      sym_line_comment,
    STATE(471), 1,
      sym_type_class_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1326), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21078] = 3,
    STATE(448), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1328), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_LBRACE,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21103] = 3,
    STATE(449), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1330), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21128] = 5,
    ACTIONS(1268), 1,
      anon_sym_of,
    STATE(450), 1,
      sym_line_comment,
    STATE(490), 1,
      sym_enum_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1332), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21157] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(451), 1,
      sym_line_comment,
    STATE(473), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1334), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21186] = 3,
    STATE(452), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1336), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21211] = 4,
    STATE(453), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1250), 6,
      sym_identifier,
      sym_decimal,
      sym_integer,
      sym_unsigned,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1248), 8,
      anon_sym_RBRACK,
      sym_quoted_string,
      anon_sym_LT,
      anon_sym_POUND_LBRACK,
      sym_double,
      sym_rational,
      anon_sym_u22a4,
      anon_sym_u22a5,
  [21238] = 5,
    ACTIONS(1292), 1,
      anon_sym_is,
    STATE(454), 1,
      sym_line_comment,
    STATE(545), 1,
      sym_event_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1338), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21267] = 5,
    ACTIONS(1254), 1,
      anon_sym_is,
    STATE(455), 1,
      sym_line_comment,
    STATE(488), 1,
      sym_entity_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1340), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21296] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(456), 1,
      sym_line_comment,
    STATE(504), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1342), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21325] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(457), 1,
      sym_line_comment,
    STATE(483), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1344), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21354] = 5,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(458), 1,
      sym_line_comment,
    STATE(480), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1346), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21383] = 3,
    STATE(459), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1348), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_from,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21408] = 3,
    STATE(460), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1350), 14,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_of,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21433] = 5,
    ACTIONS(1270), 1,
      anon_sym_is,
    STATE(461), 1,
      sym_line_comment,
    STATE(531), 1,
      sym_metric_group_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1352), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21462] = 5,
    ACTIONS(1286), 1,
      anon_sym_is,
    STATE(462), 1,
      sym_line_comment,
    STATE(542), 1,
      sym_structure_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1354), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21491] = 3,
    STATE(463), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1356), 13,
      anon_sym_is,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21515] = 13,
    ACTIONS(1358), 1,
      anon_sym_end,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1362), 1,
      anon_sym_AT,
    ACTIONS(1364), 1,
      anon_sym_assert,
    ACTIONS(1366), 1,
      anon_sym_ref,
    STATE(464), 1,
      sym_line_comment,
    STATE(560), 1,
      aux_sym_module_body_repeat2,
    STATE(579), 1,
      aux_sym_dimension_body_repeat2,
    STATE(702), 1,
      sym_annotation,
    STATE(889), 1,
      sym_member,
    STATE(703), 2,
      sym_annotation_property,
      sym_constraint,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [21559] = 3,
    STATE(465), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1368), 13,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_ref,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21583] = 13,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1362), 1,
      anon_sym_AT,
    ACTIONS(1364), 1,
      anon_sym_assert,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1370), 1,
      anon_sym_end,
    STATE(464), 1,
      aux_sym_module_body_repeat2,
    STATE(466), 1,
      sym_line_comment,
    STATE(581), 1,
      aux_sym_dimension_body_repeat2,
    STATE(702), 1,
      sym_annotation,
    STATE(889), 1,
      sym_member,
    STATE(703), 2,
      sym_annotation_property,
      sym_constraint,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [21627] = 12,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1372), 1,
      anon_sym_end,
    ACTIONS(1374), 1,
      anon_sym_metric,
    ACTIONS(1376), 1,
      anon_sym_ref,
    STATE(356), 1,
      sym_annotation,
    STATE(467), 1,
      sym_line_comment,
    STATE(519), 1,
      aux_sym_module_body_repeat2,
    STATE(647), 1,
      aux_sym_metric_group_body_repeat1,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    STATE(820), 2,
      sym_metric_def,
      sym_metric_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [21668] = 13,
    ACTIONS(1378), 1,
      anon_sym_end,
    ACTIONS(1380), 1,
      sym_identifier,
    ACTIONS(1382), 1,
      anon_sym_AT,
    ACTIONS(1384), 1,
      anon_sym_assert,
    STATE(468), 1,
      sym_line_comment,
    STATE(568), 1,
      aux_sym_module_body_repeat2,
    STATE(625), 1,
      aux_sym_union_body_repeat1,
    STATE(691), 1,
      sym_identifier_reference,
    STATE(749), 1,
      sym_qualified_identifier,
    STATE(808), 1,
      sym_annotation,
    STATE(919), 1,
      sym_type_variant,
    STATE(767), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [21711] = 3,
    STATE(469), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1386), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21734] = 3,
    STATE(470), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1388), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21757] = 3,
    STATE(471), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1390), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21780] = 3,
    STATE(472), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1392), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21803] = 3,
    STATE(473), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1394), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21826] = 3,
    STATE(474), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1396), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21849] = 3,
    STATE(475), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1398), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21872] = 3,
    STATE(476), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1400), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21895] = 3,
    STATE(477), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1402), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21918] = 3,
    STATE(478), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1404), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21941] = 3,
    STATE(479), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1406), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21964] = 3,
    STATE(480), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1408), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [21987] = 3,
    STATE(481), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1410), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22010] = 3,
    STATE(482), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1412), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22033] = 3,
    STATE(483), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1414), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22056] = 3,
    STATE(484), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1416), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22079] = 3,
    STATE(485), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1418), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22102] = 3,
    STATE(486), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1420), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22125] = 3,
    STATE(487), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1422), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22148] = 3,
    STATE(488), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1424), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22171] = 3,
    STATE(489), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1426), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22194] = 3,
    STATE(490), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1428), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22217] = 3,
    STATE(491), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1430), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22240] = 3,
    STATE(492), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1432), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22263] = 3,
    STATE(493), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1434), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22286] = 3,
    STATE(494), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1436), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22309] = 3,
    STATE(495), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1438), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22332] = 3,
    STATE(496), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1440), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22355] = 3,
    STATE(497), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1442), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22378] = 3,
    STATE(498), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1444), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22401] = 3,
    STATE(499), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1446), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22424] = 3,
    STATE(500), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1448), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22447] = 3,
    STATE(501), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1450), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [22470] = 3,
    STATE(502), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1452), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22493] = 3,
    STATE(503), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1454), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22516] = 3,
    STATE(504), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1456), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22539] = 3,
    STATE(505), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1458), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22562] = 3,
    STATE(506), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1460), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22585] = 3,
    STATE(507), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1462), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22608] = 3,
    STATE(508), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1464), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22631] = 13,
    ACTIONS(1380), 1,
      sym_identifier,
    ACTIONS(1382), 1,
      anon_sym_AT,
    ACTIONS(1384), 1,
      anon_sym_assert,
    ACTIONS(1466), 1,
      anon_sym_end,
    STATE(468), 1,
      aux_sym_module_body_repeat2,
    STATE(509), 1,
      sym_line_comment,
    STATE(616), 1,
      aux_sym_union_body_repeat1,
    STATE(691), 1,
      sym_identifier_reference,
    STATE(749), 1,
      sym_qualified_identifier,
    STATE(808), 1,
      sym_annotation,
    STATE(919), 1,
      sym_type_variant,
    STATE(767), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [22674] = 3,
    STATE(510), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1468), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22697] = 3,
    STATE(511), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1470), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22720] = 9,
    STATE(512), 1,
      sym_line_comment,
    STATE(872), 1,
      sym_logical_quantifier,
    STATE(1022), 1,
      sym_quantified_variable_binding,
    STATE(1093), 1,
      sym_quantified_sentence,
    ACTIONS(161), 2,
      anon_sym_forall,
      anon_sym_u2200,
    ACTIONS(163), 2,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(165), 2,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(1129), 3,
      sym_logical_quantifier_universal,
      sym_logical_quantifier_existential,
      sym_logical_quantifier_existential_one,
  [22755] = 3,
    STATE(513), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1472), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22778] = 3,
    STATE(514), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1474), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22801] = 3,
    STATE(515), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1476), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22824] = 3,
    STATE(516), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1478), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22847] = 3,
    STATE(517), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1480), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22870] = 3,
    STATE(518), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1482), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22893] = 12,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1374), 1,
      anon_sym_metric,
    ACTIONS(1376), 1,
      anon_sym_ref,
    ACTIONS(1484), 1,
      anon_sym_end,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(519), 1,
      sym_line_comment,
    STATE(641), 1,
      aux_sym_metric_group_body_repeat1,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    STATE(820), 2,
      sym_metric_def,
      sym_metric_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [22934] = 3,
    STATE(520), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1486), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22957] = 3,
    STATE(521), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1488), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [22980] = 3,
    STATE(522), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1490), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23003] = 3,
    STATE(523), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1492), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23026] = 3,
    STATE(524), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1494), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23049] = 3,
    STATE(525), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1496), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23072] = 3,
    STATE(526), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1498), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [23095] = 3,
    STATE(527), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1500), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [23118] = 3,
    STATE(528), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1502), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [23141] = 3,
    STATE(529), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1504), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23164] = 3,
    STATE(530), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1506), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [23187] = 3,
    STATE(531), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1508), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23210] = 3,
    STATE(532), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1510), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23233] = 3,
    STATE(533), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1512), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23256] = 3,
    STATE(534), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1514), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23279] = 3,
    STATE(535), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1516), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23302] = 3,
    STATE(536), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1518), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23325] = 3,
    STATE(537), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1520), 12,
      anon_sym_RBRACE,
      anon_sym_length,
      anon_sym_maxLength,
      anon_sym_minLength,
      anon_sym_fractionDigits,
      anon_sym_totalDigits,
      anon_sym_maxExclusive,
      anon_sym_minExclusive,
      anon_sym_maxInclusive,
      anon_sym_minInclusive,
      anon_sym_explicitTimezone,
      anon_sym_pattern,
  [23348] = 3,
    STATE(538), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1522), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23371] = 3,
    STATE(539), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1524), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23394] = 9,
    STATE(540), 1,
      sym_line_comment,
    STATE(872), 1,
      sym_logical_quantifier,
    STATE(1022), 1,
      sym_quantified_variable_binding,
    STATE(1043), 1,
      sym_quantified_sentence,
    ACTIONS(161), 2,
      anon_sym_forall,
      anon_sym_u2200,
    ACTIONS(163), 2,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(165), 2,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(1129), 3,
      sym_logical_quantifier_universal,
      sym_logical_quantifier_existential,
      sym_logical_quantifier_existential_one,
  [23429] = 3,
    STATE(541), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1526), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23452] = 3,
    STATE(542), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1528), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23475] = 9,
    STATE(543), 1,
      sym_line_comment,
    STATE(872), 1,
      sym_logical_quantifier,
    STATE(1015), 1,
      sym_quantified_sentence,
    STATE(1022), 1,
      sym_quantified_variable_binding,
    ACTIONS(161), 2,
      anon_sym_forall,
      anon_sym_u2200,
    ACTIONS(163), 2,
      anon_sym_exists,
      anon_sym_u2203,
    ACTIONS(165), 2,
      anon_sym_exists_BANG,
      anon_sym_u2203_BANG,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    STATE(1129), 3,
      sym_logical_quantifier_universal,
      sym_logical_quantifier_existential,
      sym_logical_quantifier_existential_one,
  [23510] = 3,
    STATE(544), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1530), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23533] = 3,
    STATE(545), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1532), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23556] = 3,
    STATE(546), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1534), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23579] = 3,
    STATE(547), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1536), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23602] = 3,
    STATE(548), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1538), 12,
      anon_sym_end,
      anon_sym_event,
      anon_sym_datatype,
      anon_sym_dimension,
      anon_sym_entity,
      anon_sym_enum,
      anon_sym_metric,
      anon_sym_property,
      anon_sym_rdf,
      anon_sym_structure,
      anon_sym_class,
      anon_sym_union,
  [23625] = 11,
    ACTIONS(1382), 1,
      anon_sym_AT,
    ACTIONS(1384), 1,
      anon_sym_assert,
    ACTIONS(1540), 1,
      anon_sym_end,
    ACTIONS(1542), 1,
      sym_identifier,
    STATE(549), 1,
      sym_line_comment,
    STATE(559), 1,
      aux_sym_module_body_repeat2,
    STATE(808), 1,
      sym_annotation,
    STATE(812), 1,
      aux_sym_enum_body_repeat1,
    STATE(979), 1,
      sym_value_variant,
    STATE(767), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23662] = 11,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1544), 1,
      anon_sym_end,
    ACTIONS(1546), 1,
      anon_sym_def,
    STATE(356), 1,
      sym_annotation,
    STATE(550), 1,
      sym_line_comment,
    STATE(558), 1,
      aux_sym_module_body_repeat2,
    STATE(730), 1,
      aux_sym_type_class_body_repeat1,
    STATE(915), 1,
      sym_keyword_function_def,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23699] = 10,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1548), 1,
      anon_sym_source,
    ACTIONS(1550), 1,
      anon_sym_identity,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(551), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    STATE(554), 2,
      sym_source_entity,
      sym_entity_identity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23734] = 10,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1548), 1,
      anon_sym_source,
    ACTIONS(1550), 1,
      anon_sym_identity,
    STATE(356), 1,
      sym_annotation,
    STATE(551), 1,
      aux_sym_module_body_repeat2,
    STATE(552), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    STATE(556), 2,
      sym_source_entity,
      sym_entity_identity,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23769] = 11,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1552), 1,
      anon_sym_end,
    ACTIONS(1554), 1,
      anon_sym_parent,
    STATE(553), 1,
      sym_line_comment,
    STATE(583), 1,
      aux_sym_dimension_body_repeat2,
    STATE(623), 1,
      aux_sym_dimension_body_repeat1,
    STATE(741), 1,
      sym_dimension_parent,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23806] = 11,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1554), 1,
      anon_sym_parent,
    ACTIONS(1556), 1,
      anon_sym_end,
    STATE(554), 1,
      sym_line_comment,
    STATE(557), 1,
      aux_sym_dimension_body_repeat1,
    STATE(584), 1,
      aux_sym_dimension_body_repeat2,
    STATE(741), 1,
      sym_dimension_parent,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23843] = 7,
    ACTIONS(1216), 1,
      anon_sym_is,
    STATE(409), 1,
      sym_function_body,
    STATE(435), 1,
      sym_annotation_only_body,
    STATE(555), 1,
      sym_line_comment,
    ACTIONS(1558), 2,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1218), 5,
      anon_sym_end,
      anon_sym_for,
      anon_sym_def,
      anon_sym_metric,
      anon_sym_ref,
  [23872] = 11,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1554), 1,
      anon_sym_parent,
    ACTIONS(1560), 1,
      anon_sym_end,
    STATE(553), 1,
      aux_sym_dimension_body_repeat1,
    STATE(556), 1,
      sym_line_comment,
    STATE(590), 1,
      aux_sym_dimension_body_repeat2,
    STATE(741), 1,
      sym_dimension_parent,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23909] = 11,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1554), 1,
      anon_sym_parent,
    ACTIONS(1562), 1,
      anon_sym_end,
    STATE(557), 1,
      sym_line_comment,
    STATE(600), 1,
      aux_sym_dimension_body_repeat2,
    STATE(623), 1,
      aux_sym_dimension_body_repeat1,
    STATE(741), 1,
      sym_dimension_parent,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23946] = 11,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1546), 1,
      anon_sym_def,
    ACTIONS(1564), 1,
      anon_sym_end,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(558), 1,
      sym_line_comment,
    STATE(786), 1,
      aux_sym_type_class_body_repeat1,
    STATE(915), 1,
      sym_keyword_function_def,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [23983] = 11,
    ACTIONS(1382), 1,
      anon_sym_AT,
    ACTIONS(1384), 1,
      anon_sym_assert,
    ACTIONS(1542), 1,
      sym_identifier,
    ACTIONS(1566), 1,
      anon_sym_end,
    STATE(559), 1,
      sym_line_comment,
    STATE(568), 1,
      aux_sym_module_body_repeat2,
    STATE(798), 1,
      aux_sym_enum_body_repeat1,
    STATE(808), 1,
      sym_annotation,
    STATE(979), 1,
      sym_value_variant,
    STATE(767), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24020] = 7,
    ACTIONS(1570), 1,
      anon_sym_AT,
    ACTIONS(1573), 1,
      anon_sym_assert,
    STATE(702), 1,
      sym_annotation,
    STATE(560), 2,
      sym_line_comment,
      aux_sym_module_body_repeat2,
    STATE(703), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1568), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [24048] = 8,
    ACTIONS(59), 1,
      anon_sym_u2209,
    STATE(115), 1,
      sym_set_membership,
    STATE(561), 1,
      sym_line_comment,
    STATE(883), 1,
      sym_logical_op_negation,
    ACTIONS(33), 2,
      anon_sym_not,
      anon_sym_u00ac,
    ACTIONS(57), 2,
      anon_sym_in,
      anon_sym_u2208,
    STATE(354), 2,
      sym_set_op_is_member,
      sym_set_op_is_not_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24078] = 9,
    ACTIONS(1576), 1,
      sym_identifier,
    STATE(562), 1,
      sym_line_comment,
    STATE(799), 1,
      sym_sequence_ordering,
    STATE(807), 1,
      aux_sym_sequence_builder_repeat1,
    STATE(914), 1,
      sym_variable,
    STATE(1102), 1,
      sym_sequence_uniqueness,
    ACTIONS(1578), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1580), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24110] = 9,
    ACTIONS(1576), 1,
      sym_identifier,
    STATE(563), 1,
      sym_line_comment,
    STATE(721), 1,
      aux_sym_sequence_builder_repeat1,
    STATE(806), 1,
      sym_sequence_ordering,
    STATE(914), 1,
      sym_variable,
    STATE(1002), 1,
      sym_sequence_uniqueness,
    ACTIONS(1578), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1580), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24142] = 9,
    ACTIONS(1576), 1,
      sym_identifier,
    STATE(564), 1,
      sym_line_comment,
    STATE(804), 1,
      sym_sequence_ordering,
    STATE(810), 1,
      aux_sym_sequence_builder_repeat1,
    STATE(914), 1,
      sym_variable,
    STATE(988), 1,
      sym_sequence_uniqueness,
    ACTIONS(1578), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1580), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24174] = 9,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1582), 1,
      anon_sym_identity,
    STATE(356), 1,
      sym_annotation,
    STATE(565), 1,
      sym_line_comment,
    STATE(569), 1,
      aux_sym_module_body_repeat2,
    STATE(604), 1,
      sym_entity_identity,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24205] = 9,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1584), 1,
      anon_sym_source,
    STATE(356), 1,
      sym_annotation,
    STATE(566), 1,
      sym_line_comment,
    STATE(570), 1,
      aux_sym_module_body_repeat2,
    STATE(602), 1,
      sym_source_entity,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24236] = 6,
    ACTIONS(1588), 1,
      anon_sym_with,
    ACTIONS(1590), 1,
      anon_sym_without,
    STATE(567), 1,
      sym_line_comment,
    STATE(738), 2,
      sym_mixin_with_members,
      sym_mixin_without_members,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1586), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [24261] = 7,
    ACTIONS(1592), 1,
      anon_sym_AT,
    ACTIONS(1595), 1,
      anon_sym_assert,
    STATE(808), 1,
      sym_annotation,
    ACTIONS(1568), 2,
      anon_sym_end,
      sym_identifier,
    STATE(568), 2,
      sym_line_comment,
      aux_sym_module_body_repeat2,
    STATE(767), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24288] = 9,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1582), 1,
      anon_sym_identity,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(569), 1,
      sym_line_comment,
    STATE(576), 1,
      sym_entity_identity,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24319] = 9,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1584), 1,
      anon_sym_source,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(570), 1,
      sym_line_comment,
    STATE(607), 1,
      sym_source_entity,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24350] = 5,
    ACTIONS(1598), 1,
      sym_language_tag,
    STATE(571), 1,
      sym_line_comment,
    ACTIONS(11), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(13), 5,
      anon_sym_end,
      sym_identifier,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_ref,
  [24373] = 8,
    ACTIONS(1600), 1,
      anon_sym_RBRACE,
    ACTIONS(1602), 1,
      sym_unsigned,
    STATE(572), 1,
      sym_line_comment,
    STATE(662), 1,
      sym_sequence_ordering,
    STATE(909), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24402] = 8,
    ACTIONS(1608), 1,
      anon_sym_RBRACE,
    ACTIONS(1610), 1,
      sym_unsigned,
    STATE(573), 1,
      sym_line_comment,
    STATE(699), 1,
      sym_sequence_ordering,
    STATE(921), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24431] = 4,
    STATE(574), 1,
      sym_line_comment,
    ACTIONS(1614), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1612), 5,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_in,
      anon_sym_u2208,
      anon_sym_u2209,
  [24451] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1616), 1,
      anon_sym_end,
    STATE(575), 1,
      sym_line_comment,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24479] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1618), 1,
      anon_sym_end,
    STATE(576), 1,
      sym_line_comment,
    STATE(585), 1,
      aux_sym_dimension_body_repeat2,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24507] = 8,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1620), 1,
      anon_sym_end,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(577), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24535] = 7,
    ACTIONS(1622), 1,
      sym_unsigned,
    STATE(578), 1,
      sym_line_comment,
    STATE(752), 1,
      sym_sequence_ordering,
    STATE(1098), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24561] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1624), 1,
      anon_sym_end,
    STATE(579), 1,
      sym_line_comment,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24589] = 7,
    ACTIONS(1626), 1,
      anon_sym_end,
    ACTIONS(1628), 1,
      sym_identifier,
    ACTIONS(1631), 1,
      anon_sym_ref,
    STATE(889), 1,
      sym_member,
    STATE(580), 2,
      sym_line_comment,
      aux_sym_dimension_body_repeat2,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24615] = 8,
    ACTIONS(1358), 1,
      anon_sym_end,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(581), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24643] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1634), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(582), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24671] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1636), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(583), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24699] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1562), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(584), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24727] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1638), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(585), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24755] = 5,
    ACTIONS(1640), 1,
      sym_language_tag,
    STATE(586), 1,
      sym_line_comment,
    ACTIONS(11), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(13), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_AT,
      anon_sym_assert,
  [24777] = 5,
    ACTIONS(965), 1,
      anon_sym_AT,
    STATE(587), 1,
      sym_line_comment,
    ACTIONS(1642), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24799] = 4,
    ACTIONS(1644), 1,
      anon_sym_COLON,
    STATE(588), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 6,
      anon_sym_end,
      sym_identifier,
      anon_sym_with,
      anon_sym_without,
      anon_sym_parent,
      anon_sym_ref,
  [24819] = 4,
    STATE(589), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(73), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(75), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24839] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1552), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(590), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [24867] = 4,
    STATE(591), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(81), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(83), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24887] = 4,
    STATE(592), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(85), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(87), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24907] = 4,
    STATE(593), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(89), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(91), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24927] = 4,
    STATE(594), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(93), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(95), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24947] = 4,
    STATE(595), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(97), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(99), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [24967] = 4,
    ACTIONS(1646), 1,
      anon_sym_COLON,
    STATE(596), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(105), 6,
      anon_sym_RBRACK,
      sym_identifier,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_seq_builder_separator,
      anon_sym_PLUS,
  [24987] = 4,
    STATE(597), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(7), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(9), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [25007] = 5,
    ACTIONS(1648), 1,
      anon_sym_COLON,
    STATE(598), 1,
      sym_line_comment,
    ACTIONS(105), 2,
      anon_sym_AT,
      anon_sym_LPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [25029] = 8,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1650), 1,
      anon_sym_end,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(599), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25057] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1652), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(600), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25085] = 8,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1654), 1,
      anon_sym_end,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(601), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25113] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1656), 1,
      anon_sym_end,
    STATE(602), 1,
      sym_line_comment,
    STATE(606), 1,
      aux_sym_dimension_body_repeat2,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25141] = 8,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    ACTIONS(1658), 1,
      anon_sym_end,
    STATE(261), 1,
      aux_sym_module_body_repeat2,
    STATE(356), 1,
      sym_annotation,
    STATE(603), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25169] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1660), 1,
      anon_sym_end,
    STATE(575), 1,
      aux_sym_dimension_body_repeat2,
    STATE(604), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25197] = 6,
    ACTIONS(1662), 1,
      anon_sym_with,
    ACTIONS(1664), 1,
      anon_sym_without,
    STATE(605), 1,
      sym_line_comment,
    STATE(838), 2,
      sym_mixin_with_members,
      sym_mixin_without_members,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1586), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [25221] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1666), 1,
      anon_sym_end,
    STATE(580), 1,
      aux_sym_dimension_body_repeat2,
    STATE(606), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25249] = 8,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    ACTIONS(1668), 1,
      anon_sym_end,
    STATE(582), 1,
      aux_sym_dimension_body_repeat2,
    STATE(607), 1,
      sym_line_comment,
    STATE(889), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25277] = 4,
    STATE(608), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(77), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(79), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [25297] = 4,
    STATE(609), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(7), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(9), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25316] = 5,
    ACTIONS(965), 1,
      anon_sym_AT,
    ACTIONS(1670), 1,
      anon_sym_LPAREN,
    STATE(610), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [25337] = 6,
    ACTIONS(1672), 1,
      anon_sym_RBRACK,
    ACTIONS(1674), 1,
      sym_identifier,
    STATE(853), 1,
      sym_qualified_identifier,
    STATE(611), 2,
      sym_line_comment,
      aux_sym_import_statement_repeat1,
    STATE(960), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25360] = 4,
    STATE(612), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(85), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(87), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25379] = 4,
    STATE(613), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(89), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(91), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25398] = 4,
    STATE(614), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(93), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(95), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25417] = 4,
    STATE(615), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(97), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(99), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25436] = 8,
    ACTIONS(1378), 1,
      anon_sym_end,
    ACTIONS(1380), 1,
      sym_identifier,
    STATE(616), 1,
      sym_line_comment,
    STATE(633), 1,
      aux_sym_union_body_repeat1,
    STATE(691), 1,
      sym_identifier_reference,
    STATE(749), 1,
      sym_qualified_identifier,
    STATE(919), 1,
      sym_type_variant,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25463] = 3,
    STATE(617), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 6,
      anon_sym_end,
      sym_identifier,
      anon_sym_with,
      anon_sym_without,
      anon_sym_parent,
      anon_sym_ref,
  [25480] = 6,
    STATE(618), 1,
      sym_line_comment,
    STATE(803), 1,
      sym_sequence_ordering,
    STATE(1112), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25503] = 7,
    ACTIONS(1677), 1,
      anon_sym_RBRACK,
    ACTIONS(1679), 1,
      sym_identifier,
    STATE(611), 1,
      aux_sym_import_statement_repeat1,
    STATE(619), 1,
      sym_line_comment,
    STATE(853), 1,
      sym_qualified_identifier,
    STATE(960), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25528] = 6,
    STATE(620), 1,
      sym_line_comment,
    STATE(781), 1,
      sym_sequence_ordering,
    STATE(999), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25551] = 8,
    ACTIONS(969), 1,
      anon_sym_RBRACK,
    ACTIONS(1681), 1,
      anon_sym_as,
    ACTIONS(1683), 1,
      sym_identifier,
    ACTIONS(1685), 1,
      anon_sym_COLON,
    ACTIONS(1687), 1,
      anon_sym_LT,
    STATE(621), 1,
      sym_line_comment,
    STATE(858), 1,
      sym_iri,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25578] = 5,
    ACTIONS(965), 1,
      anon_sym_AT,
    STATE(622), 1,
      sym_line_comment,
    ACTIONS(1689), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25599] = 5,
    ACTIONS(1693), 1,
      anon_sym_parent,
    STATE(741), 1,
      sym_dimension_parent,
    STATE(623), 2,
      sym_line_comment,
      aux_sym_dimension_body_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1691), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [25620] = 8,
    ACTIONS(1696), 1,
      sym_identifier,
    ACTIONS(1698), 1,
      anon_sym_LBRACE,
    STATE(624), 1,
      sym_line_comment,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(684), 1,
      sym_type_parameter_restriction,
    STATE(722), 1,
      sym_identifier_reference,
    STATE(836), 1,
      sym_cardinality_reference_expression,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25647] = 8,
    ACTIONS(1380), 1,
      sym_identifier,
    ACTIONS(1700), 1,
      anon_sym_end,
    STATE(625), 1,
      sym_line_comment,
    STATE(633), 1,
      aux_sym_union_body_repeat1,
    STATE(691), 1,
      sym_identifier_reference,
    STATE(749), 1,
      sym_qualified_identifier,
    STATE(919), 1,
      sym_type_variant,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25674] = 4,
    ACTIONS(1702), 1,
      sym_controlled_language_tag,
    STATE(626), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1006), 5,
      anon_sym_end,
      sym_identifier,
      anon_sym_AT,
      anon_sym_assert,
      anon_sym_ref,
  [25693] = 3,
    STATE(627), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(105), 6,
      anon_sym_RBRACK,
      sym_identifier,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_seq_builder_separator,
      anon_sym_PLUS,
  [25710] = 5,
    ACTIONS(1704), 1,
      anon_sym_is,
    STATE(628), 1,
      sym_line_comment,
    STATE(743), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1706), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [25731] = 4,
    STATE(629), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(81), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(83), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25750] = 3,
    STATE(630), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(101), 6,
      anon_sym_RBRACK,
      sym_identifier,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_seq_builder_separator,
      anon_sym_PLUS,
  [25767] = 4,
    ACTIONS(1708), 1,
      anon_sym_COLON,
    STATE(631), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 5,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [25786] = 5,
    ACTIONS(1710), 1,
      anon_sym_COLON,
    STATE(632), 1,
      sym_line_comment,
    ACTIONS(105), 2,
      anon_sym_AT,
      anon_sym_LPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [25807] = 7,
    ACTIONS(1712), 1,
      anon_sym_end,
    ACTIONS(1714), 1,
      sym_identifier,
    STATE(691), 1,
      sym_identifier_reference,
    STATE(749), 1,
      sym_qualified_identifier,
    STATE(919), 1,
      sym_type_variant,
    STATE(633), 2,
      sym_line_comment,
      aux_sym_union_body_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25832] = 4,
    STATE(634), 1,
      sym_line_comment,
    ACTIONS(105), 2,
      anon_sym_AT,
      anon_sym_LPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [25851] = 4,
    STATE(635), 1,
      sym_line_comment,
    ACTIONS(101), 2,
      anon_sym_AT,
      anon_sym_LPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [25870] = 6,
    STATE(636), 1,
      sym_line_comment,
    STATE(793), 1,
      sym_sequence_ordering,
    STATE(1087), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25893] = 7,
    ACTIONS(1679), 1,
      sym_identifier,
    ACTIONS(1717), 1,
      anon_sym_RBRACK,
    STATE(611), 1,
      aux_sym_import_statement_repeat1,
    STATE(637), 1,
      sym_line_comment,
    STATE(853), 1,
      sym_qualified_identifier,
    STATE(960), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25918] = 5,
    ACTIONS(1704), 1,
      anon_sym_is,
    STATE(638), 1,
      sym_line_comment,
    STATE(768), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1719), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [25939] = 8,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(1721), 1,
      anon_sym_version,
    ACTIONS(1723), 1,
      anon_sym_is,
    STATE(639), 1,
      sym_line_comment,
    STATE(769), 1,
      sym_iri,
    STATE(911), 1,
      sym_module_version,
    STATE(1057), 1,
      sym_module_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [25966] = 5,
    STATE(640), 1,
      sym_line_comment,
    STATE(841), 1,
      sym_function_body,
    ACTIONS(1558), 2,
      anon_sym_COLON_EQ,
      anon_sym_u2254,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1725), 3,
      anon_sym_end,
      anon_sym_metric,
      anon_sym_ref,
  [25987] = 7,
    ACTIONS(1374), 1,
      anon_sym_metric,
    ACTIONS(1376), 1,
      anon_sym_ref,
    ACTIONS(1727), 1,
      anon_sym_end,
    STATE(641), 1,
      sym_line_comment,
    STATE(642), 1,
      aux_sym_metric_group_body_repeat1,
    STATE(820), 2,
      sym_metric_def,
      sym_metric_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26012] = 6,
    ACTIONS(1729), 1,
      anon_sym_end,
    ACTIONS(1731), 1,
      anon_sym_metric,
    ACTIONS(1734), 1,
      anon_sym_ref,
    STATE(642), 2,
      sym_line_comment,
      aux_sym_metric_group_body_repeat1,
    STATE(820), 2,
      sym_metric_def,
      sym_metric_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26035] = 4,
    ACTIONS(1737), 1,
      anon_sym_COLON,
    STATE(643), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 5,
      anon_sym_end,
      sym_identifier,
      anon_sym_with,
      anon_sym_without,
      anon_sym_ref,
  [26054] = 7,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    STATE(356), 1,
      sym_annotation,
    STATE(577), 1,
      aux_sym_module_body_repeat2,
    STATE(644), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26079] = 8,
    ACTIONS(1696), 1,
      sym_identifier,
    ACTIONS(1698), 1,
      anon_sym_LBRACE,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(645), 1,
      sym_line_comment,
    STATE(722), 1,
      sym_identifier_reference,
    STATE(836), 1,
      sym_cardinality_reference_expression,
    STATE(875), 1,
      sym_type_parameter_restriction,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26106] = 4,
    ACTIONS(511), 1,
      anon_sym_COLON,
    STATE(646), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(105), 5,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_in,
      anon_sym_u2208,
      anon_sym_u2209,
  [26125] = 7,
    ACTIONS(1374), 1,
      anon_sym_metric,
    ACTIONS(1376), 1,
      anon_sym_ref,
    ACTIONS(1484), 1,
      anon_sym_end,
    STATE(642), 1,
      aux_sym_metric_group_body_repeat1,
    STATE(647), 1,
      sym_line_comment,
    STATE(820), 2,
      sym_metric_def,
      sym_metric_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26150] = 7,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    STATE(356), 1,
      sym_annotation,
    STATE(599), 1,
      aux_sym_module_body_repeat2,
    STATE(648), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26175] = 5,
    ACTIONS(1704), 1,
      anon_sym_is,
    STATE(649), 1,
      sym_line_comment,
    STATE(770), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1739), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [26196] = 7,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    STATE(356), 1,
      sym_annotation,
    STATE(601), 1,
      aux_sym_module_body_repeat2,
    STATE(650), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26221] = 4,
    STATE(651), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(73), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(75), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [26240] = 4,
    STATE(652), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(77), 3,
      anon_sym_AT,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(79), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [26259] = 7,
    ACTIONS(374), 1,
      anon_sym_AT,
    ACTIONS(376), 1,
      anon_sym_assert,
    STATE(356), 1,
      sym_annotation,
    STATE(603), 1,
      aux_sym_module_body_repeat2,
    STATE(653), 1,
      sym_line_comment,
    STATE(373), 2,
      sym_annotation_property,
      sym_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26284] = 6,
    STATE(654), 1,
      sym_line_comment,
    STATE(791), 1,
      sym_sequence_ordering,
    STATE(1084), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26307] = 6,
    STATE(655), 1,
      sym_line_comment,
    STATE(797), 1,
      sym_sequence_ordering,
    STATE(1099), 1,
      sym_sequence_uniqueness,
    ACTIONS(1604), 2,
      anon_sym_ordered,
      anon_sym_unordered,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26330] = 3,
    STATE(656), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 6,
      anon_sym_end,
      sym_identifier,
      anon_sym_with,
      anon_sym_without,
      anon_sym_parent,
      anon_sym_ref,
  [26347] = 6,
    ACTIONS(1743), 1,
      anon_sym_RPAREN,
    STATE(657), 1,
      sym_line_comment,
    STATE(711), 1,
      aux_sym_type_parameter_restriction_repeat1,
    STATE(867), 1,
      sym_type_restriction_argument,
    ACTIONS(1741), 2,
      sym_identifier,
      sym_wildcard,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26369] = 4,
    ACTIONS(1745), 1,
      anon_sym_COLON,
    STATE(658), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_is,
      anon_sym_end,
      anon_sym_as,
      sym_identifier,
  [26387] = 4,
    ACTIONS(1168), 1,
      anon_sym_AT,
    STATE(659), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1747), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26405] = 4,
    ACTIONS(1150), 1,
      anon_sym_AT,
    STATE(660), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1749), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26423] = 3,
    STATE(661), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 5,
      anon_sym_end,
      sym_identifier,
      anon_sym_with,
      anon_sym_without,
      anon_sym_ref,
  [26439] = 6,
    ACTIONS(1751), 1,
      anon_sym_RBRACE,
    ACTIONS(1753), 1,
      sym_unsigned,
    STATE(662), 1,
      sym_line_comment,
    STATE(983), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26461] = 4,
    ACTIONS(1160), 1,
      anon_sym_AT,
    STATE(663), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1202), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26479] = 5,
    ACTIONS(1757), 1,
      anon_sym_PLUS,
    STATE(645), 1,
      sym_type_op_combiner,
    ACTIONS(1755), 2,
      sym_identifier,
      anon_sym_RPAREN,
    STATE(664), 2,
      sym_line_comment,
      aux_sym_type_parameter_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26499] = 4,
    ACTIONS(1760), 1,
      sym_controlled_language_tag,
    STATE(665), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1006), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_AT,
      anon_sym_assert,
  [26517] = 4,
    ACTIONS(965), 1,
      anon_sym_AT,
    STATE(666), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26535] = 3,
    STATE(667), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1762), 5,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [26551] = 4,
    ACTIONS(1142), 1,
      anon_sym_AT,
    STATE(668), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1186), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26569] = 5,
    ACTIONS(1764), 1,
      anon_sym_is,
    STATE(669), 1,
      sym_line_comment,
    STATE(822), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1706), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [26589] = 4,
    ACTIONS(1152), 1,
      anon_sym_AT,
    STATE(670), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1188), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26607] = 4,
    ACTIONS(1144), 1,
      anon_sym_AT,
    STATE(671), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1190), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26625] = 4,
    ACTIONS(1768), 1,
      anon_sym_as,
    STATE(672), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1766), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [26643] = 4,
    ACTIONS(1148), 1,
      anon_sym_AT,
    STATE(673), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1192), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26661] = 4,
    ACTIONS(1164), 1,
      anon_sym_AT,
    STATE(674), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1194), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26679] = 4,
    ACTIONS(699), 1,
      anon_sym_AT,
    STATE(675), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1184), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26697] = 4,
    ACTIONS(1182), 1,
      anon_sym_AT,
    STATE(676), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1196), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26715] = 4,
    ACTIONS(1128), 1,
      anon_sym_AT,
    STATE(677), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1198), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26733] = 4,
    ACTIONS(1134), 1,
      anon_sym_AT,
    STATE(678), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1200), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26751] = 6,
    ACTIONS(1770), 1,
      anon_sym_RBRACK,
    ACTIONS(1772), 1,
      sym_identifier,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(936), 1,
      sym_identifier_reference,
    STATE(679), 2,
      sym_line_comment,
      aux_sym__rdf_types_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26773] = 3,
    STATE(680), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1775), 5,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [26789] = 4,
    ACTIONS(1156), 1,
      anon_sym_AT,
    STATE(681), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1204), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26807] = 4,
    ACTIONS(1158), 1,
      anon_sym_AT,
    STATE(682), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1206), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26825] = 4,
    ACTIONS(1162), 1,
      anon_sym_AT,
    STATE(683), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1208), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [26843] = 6,
    ACTIONS(1779), 1,
      anon_sym_PLUS,
    STATE(645), 1,
      sym_type_op_combiner,
    STATE(684), 1,
      sym_line_comment,
    STATE(707), 1,
      aux_sym_type_parameter_repeat1,
    ACTIONS(1777), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26865] = 6,
    ACTIONS(1781), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1783), 1,
      sym_identifier,
    STATE(685), 1,
      sym_line_comment,
    STATE(1050), 1,
      sym__module_path_actual,
    STATE(1024), 2,
      sym_module_path_absolute,
      sym_module_path_relative,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26887] = 3,
    STATE(686), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 5,
      anon_sym_end,
      sym_identifier,
      anon_sym_with,
      anon_sym_without,
      anon_sym_ref,
  [26903] = 5,
    ACTIONS(1764), 1,
      anon_sym_is,
    STATE(687), 1,
      sym_line_comment,
    STATE(833), 1,
      sym_annotation_only_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1719), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [26923] = 4,
    ACTIONS(1785), 1,
      anon_sym_COLON,
    STATE(688), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [26941] = 4,
    ACTIONS(1787), 1,
      anon_sym_COLON,
    STATE(689), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [26959] = 3,
    STATE(690), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1789), 5,
      anon_sym_not,
      anon_sym_u00ac,
      anon_sym_in,
      anon_sym_u2208,
      anon_sym_u2209,
  [26975] = 6,
    ACTIONS(1791), 1,
      anon_sym_is,
    ACTIONS(1795), 1,
      anon_sym_as,
    STATE(691), 1,
      sym_line_comment,
    STATE(953), 1,
      sym_annotation_only_body,
    ACTIONS(1793), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [26997] = 3,
    STATE(692), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 5,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27013] = 4,
    STATE(693), 1,
      sym_line_comment,
    ACTIONS(105), 2,
      anon_sym_AT,
      anon_sym_LPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27031] = 3,
    STATE(694), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 5,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27047] = 4,
    STATE(695), 1,
      sym_line_comment,
    ACTIONS(101), 2,
      anon_sym_AT,
      anon_sym_LPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27065] = 6,
    ACTIONS(1797), 1,
      anon_sym_LBRACK,
    ACTIONS(1799), 1,
      sym_identifier,
    STATE(408), 1,
      sym_qualified_identifier,
    STATE(696), 1,
      sym_line_comment,
    STATE(426), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27087] = 6,
    ACTIONS(1360), 1,
      sym_identifier,
    ACTIONS(1366), 1,
      anon_sym_ref,
    STATE(697), 1,
      sym_line_comment,
    STATE(821), 1,
      sym_member,
    STATE(895), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27109] = 7,
    ACTIONS(1696), 1,
      sym_identifier,
    ACTIONS(1801), 1,
      anon_sym_RBRACK,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(679), 1,
      aux_sym__rdf_types_repeat1,
    STATE(698), 1,
      sym_line_comment,
    STATE(936), 1,
      sym_identifier_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27133] = 6,
    ACTIONS(1803), 1,
      anon_sym_RBRACE,
    ACTIONS(1805), 1,
      sym_unsigned,
    STATE(699), 1,
      sym_line_comment,
    STATE(926), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27155] = 5,
    ACTIONS(1809), 1,
      sym_kw_is_fixed,
    STATE(500), 1,
      sym_tz_restriction_value,
    STATE(700), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1807), 3,
      anon_sym_required,
      anon_sym_prohibited,
      anon_sym_optional,
  [27175] = 6,
    ACTIONS(1679), 1,
      sym_identifier,
    STATE(637), 1,
      aux_sym_import_statement_repeat1,
    STATE(701), 1,
      sym_line_comment,
    STATE(853), 1,
      sym_qualified_identifier,
    STATE(960), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27197] = 4,
    ACTIONS(1140), 1,
      anon_sym_AT,
    STATE(702), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1811), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [27215] = 4,
    ACTIONS(1170), 1,
      anon_sym_AT,
    STATE(703), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1813), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [27233] = 4,
    ACTIONS(1114), 1,
      anon_sym_AT,
    STATE(704), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1815), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [27251] = 6,
    ACTIONS(1817), 1,
      sym_identifier,
    ACTIONS(1819), 1,
      anon_sym_ref,
    STATE(705), 1,
      sym_line_comment,
    STATE(814), 1,
      sym_member,
    STATE(779), 2,
      sym_member_def,
      sym_property_ref,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27273] = 5,
    ACTIONS(965), 1,
      anon_sym_AT,
    ACTIONS(1821), 1,
      anon_sym_LPAREN,
    STATE(706), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27293] = 6,
    ACTIONS(1779), 1,
      anon_sym_PLUS,
    STATE(645), 1,
      sym_type_op_combiner,
    STATE(664), 1,
      aux_sym_type_parameter_repeat1,
    STATE(707), 1,
      sym_line_comment,
    ACTIONS(1823), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27315] = 4,
    ACTIONS(1076), 1,
      anon_sym_AT,
    STATE(708), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1825), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [27333] = 6,
    ACTIONS(1679), 1,
      sym_identifier,
    STATE(619), 1,
      aux_sym_import_statement_repeat1,
    STATE(709), 1,
      sym_line_comment,
    STATE(853), 1,
      sym_qualified_identifier,
    STATE(960), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27355] = 6,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1827), 1,
      anon_sym_LBRACK,
    STATE(408), 1,
      sym_qualified_identifier,
    STATE(710), 1,
      sym_line_comment,
    STATE(420), 2,
      sym_member_import,
      sym_module_import,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27377] = 5,
    ACTIONS(1832), 1,
      anon_sym_RPAREN,
    STATE(867), 1,
      sym_type_restriction_argument,
    ACTIONS(1829), 2,
      sym_identifier,
      sym_wildcard,
    STATE(711), 2,
      sym_line_comment,
      aux_sym_type_parameter_restriction_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27397] = 6,
    ACTIONS(1834), 1,
      anon_sym_RPAREN,
    STATE(711), 1,
      aux_sym_type_parameter_restriction_repeat1,
    STATE(712), 1,
      sym_line_comment,
    STATE(867), 1,
      sym_type_restriction_argument,
    ACTIONS(1741), 2,
      sym_identifier,
      sym_wildcard,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27419] = 3,
    STATE(713), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1836), 5,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27435] = 4,
    ACTIONS(1166), 1,
      anon_sym_AT,
    STATE(714), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1210), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
      anon_sym_ref,
  [27453] = 4,
    ACTIONS(1838), 1,
      anon_sym_COLON,
    STATE(715), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [27470] = 3,
    STATE(716), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1840), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27485] = 4,
    ACTIONS(1156), 1,
      anon_sym_AT,
    STATE(717), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1204), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27502] = 4,
    ACTIONS(1158), 1,
      anon_sym_AT,
    STATE(718), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1206), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27519] = 4,
    ACTIONS(1162), 1,
      anon_sym_AT,
    STATE(719), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1208), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27536] = 4,
    ACTIONS(1166), 1,
      anon_sym_AT,
    STATE(720), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1210), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27553] = 6,
    ACTIONS(1842), 1,
      sym_identifier,
    ACTIONS(1844), 1,
      sym_seq_builder_separator,
    STATE(721), 1,
      sym_line_comment,
    STATE(737), 1,
      aux_sym_sequence_builder_repeat1,
    STATE(914), 1,
      sym_variable,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27574] = 4,
    ACTIONS(1848), 1,
      anon_sym_LPAREN,
    STATE(722), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1846), 3,
      sym_identifier,
      anon_sym_RPAREN,
      anon_sym_PLUS,
  [27591] = 4,
    ACTIONS(1114), 1,
      anon_sym_AT,
    STATE(723), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1815), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27608] = 5,
    ACTIONS(1850), 1,
      sym_identifier,
    ACTIONS(1853), 1,
      anon_sym_RPAREN,
    STATE(948), 1,
      sym_function_parameter,
    STATE(724), 2,
      sym_line_comment,
      aux_sym_function_signature_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27627] = 5,
    STATE(712), 1,
      aux_sym_type_parameter_restriction_repeat1,
    STATE(725), 1,
      sym_line_comment,
    STATE(867), 1,
      sym_type_restriction_argument,
    ACTIONS(1741), 2,
      sym_identifier,
      sym_wildcard,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27646] = 4,
    STATE(530), 1,
      sym_tz_restriction_value,
    STATE(726), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1807), 3,
      anon_sym_required,
      anon_sym_prohibited,
      anon_sym_optional,
  [27663] = 6,
    ACTIONS(1855), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      anon_sym_group,
    STATE(406), 1,
      sym_function_signature,
    STATE(465), 1,
      sym_function_def,
    STATE(727), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27684] = 6,
    ACTIONS(1696), 1,
      sym_identifier,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(698), 1,
      aux_sym__rdf_types_repeat1,
    STATE(728), 1,
      sym_line_comment,
    STATE(936), 1,
      sym_identifier_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27705] = 5,
    ACTIONS(1859), 1,
      sym_identifier,
    ACTIONS(1861), 1,
      anon_sym_AT,
    STATE(729), 1,
      sym_line_comment,
    STATE(534), 2,
      sym_member_def,
      sym_annotation_member_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27724] = 6,
    ACTIONS(1546), 1,
      anon_sym_def,
    ACTIONS(1863), 1,
      anon_sym_end,
    STATE(730), 1,
      sym_line_comment,
    STATE(734), 1,
      aux_sym_type_class_body_repeat1,
    STATE(915), 1,
      sym_keyword_function_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27745] = 4,
    ACTIONS(1168), 1,
      anon_sym_AT,
    STATE(731), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1747), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27762] = 4,
    ACTIONS(1150), 1,
      anon_sym_AT,
    STATE(732), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1749), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [27779] = 5,
    ACTIONS(1865), 1,
      anon_sym_for,
    ACTIONS(1867), 1,
      anon_sym_def,
    STATE(945), 1,
      sym_keyword_function_def,
    STATE(733), 2,
      sym_line_comment,
      aux_sym_sentence_with_environment_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27798] = 5,
    ACTIONS(1870), 1,
      anon_sym_end,
    ACTIONS(1872), 1,
      anon_sym_def,
    STATE(915), 1,
      sym_keyword_function_def,
    STATE(734), 2,
      sym_line_comment,
      aux_sym_type_class_body_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27817] = 3,
    STATE(735), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1875), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27832] = 4,
    STATE(736), 1,
      sym_line_comment,
    ACTIONS(1877), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(1879), 2,
      anon_sym_LT_DASH,
      anon_sym_u2190,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27849] = 5,
    ACTIONS(1881), 1,
      sym_identifier,
    ACTIONS(1884), 1,
      sym_seq_builder_separator,
    STATE(914), 1,
      sym_variable,
    STATE(737), 2,
      sym_line_comment,
      aux_sym_sequence_builder_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27868] = 3,
    STATE(738), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1886), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27883] = 4,
    ACTIONS(1888), 1,
      anon_sym_as,
    STATE(739), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1766), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [27900] = 5,
    ACTIONS(1791), 1,
      anon_sym_is,
    STATE(740), 1,
      sym_line_comment,
    STATE(969), 1,
      sym_annotation_only_body,
    ACTIONS(1890), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27919] = 3,
    STATE(741), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1892), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27934] = 6,
    ACTIONS(1894), 1,
      sym_identifier,
    ACTIONS(1896), 1,
      anon_sym_RPAREN,
    STATE(742), 1,
      sym_line_comment,
    STATE(744), 1,
      aux_sym__class_type_parameters_repeat1,
    STATE(952), 1,
      sym_type_parameter,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27955] = 3,
    STATE(743), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1898), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [27970] = 5,
    ACTIONS(1900), 1,
      sym_identifier,
    ACTIONS(1903), 1,
      anon_sym_RPAREN,
    STATE(952), 1,
      sym_type_parameter,
    STATE(744), 2,
      sym_line_comment,
      aux_sym__class_type_parameters_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [27989] = 3,
    STATE(745), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28004] = 4,
    ACTIONS(1076), 1,
      anon_sym_AT,
    STATE(746), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1825), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28021] = 3,
    STATE(747), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [28036] = 3,
    STATE(748), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28051] = 3,
    STATE(749), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 4,
      anon_sym_is,
      anon_sym_end,
      anon_sym_as,
      sym_identifier,
  [28066] = 5,
    ACTIONS(1905), 1,
      anon_sym_RBRACK,
    ACTIONS(1907), 1,
      sym_identifier,
    STATE(927), 1,
      sym_mixin_member,
    STATE(750), 2,
      sym_line_comment,
      aux_sym__mixin_member_or_sequence_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28085] = 3,
    STATE(751), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 4,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [28100] = 5,
    ACTIONS(1910), 1,
      sym_unsigned,
    STATE(752), 1,
      sym_line_comment,
    STATE(1061), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28119] = 4,
    ACTIONS(1152), 1,
      anon_sym_AT,
    STATE(753), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1188), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28136] = 4,
    ACTIONS(1144), 1,
      anon_sym_AT,
    STATE(754), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1190), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28153] = 6,
    ACTIONS(1912), 1,
      sym_identifier,
    ACTIONS(1914), 1,
      anon_sym_RPAREN,
    STATE(724), 1,
      aux_sym_function_signature_repeat1,
    STATE(755), 1,
      sym_line_comment,
    STATE(948), 1,
      sym_function_parameter,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28174] = 3,
    STATE(756), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1916), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28189] = 6,
    ACTIONS(1546), 1,
      anon_sym_def,
    ACTIONS(1918), 1,
      anon_sym_for,
    STATE(733), 1,
      aux_sym_sentence_with_environment_repeat1,
    STATE(757), 1,
      sym_line_comment,
    STATE(945), 1,
      sym_keyword_function_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28210] = 5,
    ACTIONS(1920), 1,
      anon_sym_is,
    ACTIONS(1922), 1,
      anon_sym_EQ,
    STATE(758), 1,
      sym_line_comment,
    STATE(723), 2,
      sym_informal_constraint,
      sym_formal_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28229] = 4,
    STATE(759), 1,
      sym_line_comment,
    ACTIONS(1612), 2,
      sym_identifier,
      sym_seq_builder_separator,
    ACTIONS(1924), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28246] = 3,
    STATE(760), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1926), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28261] = 6,
    ACTIONS(1928), 1,
      anon_sym_LBRACK,
    ACTIONS(1930), 1,
      sym_identifier,
    ACTIONS(1932), 1,
      sym_wildcard,
    STATE(761), 1,
      sym_line_comment,
    STATE(830), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28282] = 3,
    STATE(762), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1934), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28297] = 6,
    ACTIONS(1936), 1,
      anon_sym_RBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(750), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(763), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28318] = 6,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(1940), 1,
      anon_sym_RBRACK,
    STATE(750), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(764), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28339] = 6,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(1942), 1,
      anon_sym_RBRACK,
    STATE(750), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(765), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28360] = 6,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(1944), 1,
      anon_sym_RBRACK,
    STATE(750), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(766), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28381] = 4,
    ACTIONS(1170), 1,
      anon_sym_AT,
    STATE(767), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1813), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28398] = 3,
    STATE(768), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1946), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28413] = 6,
    ACTIONS(1721), 1,
      anon_sym_version,
    ACTIONS(1723), 1,
      anon_sym_is,
    STATE(769), 1,
      sym_line_comment,
    STATE(917), 1,
      sym_module_version,
    STATE(1046), 1,
      sym_module_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28434] = 3,
    STATE(770), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1948), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28449] = 5,
    ACTIONS(1950), 1,
      anon_sym_is,
    ACTIONS(1952), 1,
      anon_sym_EQ,
    STATE(771), 1,
      sym_line_comment,
    STATE(704), 2,
      sym_informal_constraint,
      sym_formal_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28468] = 3,
    STATE(772), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1954), 4,
      anon_sym_RBRACE,
      sym_unsigned,
      anon_sym_unique,
      anon_sym_nonunique,
  [28483] = 3,
    STATE(773), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1956), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28498] = 6,
    ACTIONS(1958), 1,
      anon_sym_LBRACK,
    ACTIONS(1960), 1,
      sym_identifier,
    ACTIONS(1962), 1,
      sym_wildcard,
    STATE(760), 1,
      sym_mixin_member,
    STATE(774), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28519] = 5,
    ACTIONS(1964), 1,
      anon_sym_with,
    ACTIONS(1966), 1,
      anon_sym_without,
    STATE(775), 1,
      sym_line_comment,
    STATE(452), 2,
      sym_mixin_with_members,
      sym_mixin_without_members,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28538] = 6,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(1968), 1,
      anon_sym_RBRACK,
    STATE(750), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(776), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28559] = 6,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(1970), 1,
      anon_sym_RBRACK,
    STATE(750), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(777), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28580] = 4,
    ACTIONS(1148), 1,
      anon_sym_AT,
    STATE(778), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1192), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28597] = 3,
    STATE(779), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1972), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28612] = 3,
    STATE(780), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 4,
      anon_sym_is,
      anon_sym_end,
      anon_sym_as,
      sym_identifier,
  [28627] = 5,
    ACTIONS(1974), 1,
      anon_sym_RBRACE,
    STATE(781), 1,
      sym_line_comment,
    STATE(1097), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28646] = 4,
    ACTIONS(1164), 1,
      anon_sym_AT,
    STATE(782), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1194), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28663] = 4,
    ACTIONS(1160), 1,
      anon_sym_AT,
    STATE(783), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1202), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28680] = 5,
    STATE(657), 1,
      aux_sym_type_parameter_restriction_repeat1,
    STATE(784), 1,
      sym_line_comment,
    STATE(867), 1,
      sym_type_restriction_argument,
    ACTIONS(1741), 2,
      sym_identifier,
      sym_wildcard,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28699] = 4,
    ACTIONS(699), 1,
      anon_sym_AT,
    STATE(785), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1184), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28716] = 6,
    ACTIONS(1546), 1,
      anon_sym_def,
    ACTIONS(1976), 1,
      anon_sym_end,
    STATE(734), 1,
      aux_sym_type_class_body_repeat1,
    STATE(786), 1,
      sym_line_comment,
    STATE(915), 1,
      sym_keyword_function_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28737] = 4,
    ACTIONS(1980), 1,
      anon_sym_LPAREN,
    STATE(787), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1978), 3,
      sym_identifier,
      anon_sym_RPAREN,
      anon_sym_PLUS,
  [28754] = 4,
    ACTIONS(1182), 1,
      anon_sym_AT,
    STATE(788), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1196), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28771] = 4,
    ACTIONS(1128), 1,
      anon_sym_AT,
    STATE(789), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1198), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28788] = 4,
    ACTIONS(965), 1,
      anon_sym_AT,
    STATE(790), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1002), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28805] = 5,
    ACTIONS(1982), 1,
      anon_sym_RBRACE,
    STATE(791), 1,
      sym_line_comment,
    STATE(1088), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28824] = 4,
    ACTIONS(1134), 1,
      anon_sym_AT,
    STATE(792), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1200), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [28841] = 5,
    ACTIONS(1984), 1,
      anon_sym_RBRACE,
    STATE(793), 1,
      sym_line_comment,
    STATE(1089), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28860] = 3,
    STATE(794), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1986), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28875] = 5,
    ACTIONS(1988), 1,
      anon_sym_is,
    ACTIONS(1990), 1,
      anon_sym_EQ,
    STATE(795), 1,
      sym_line_comment,
    STATE(347), 2,
      sym_informal_constraint,
      sym_formal_constraint,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28894] = 3,
    STATE(796), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1992), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [28909] = 5,
    ACTIONS(1994), 1,
      anon_sym_RBRACE,
    STATE(797), 1,
      sym_line_comment,
    STATE(1103), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28928] = 6,
    ACTIONS(1542), 1,
      sym_identifier,
    ACTIONS(1996), 1,
      anon_sym_end,
    STATE(798), 1,
      sym_line_comment,
    STATE(800), 1,
      aux_sym_enum_body_repeat1,
    STATE(979), 1,
      sym_value_variant,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28949] = 5,
    ACTIONS(1998), 1,
      anon_sym_RBRACE,
    STATE(799), 1,
      sym_line_comment,
    STATE(1104), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28968] = 5,
    ACTIONS(2000), 1,
      anon_sym_end,
    ACTIONS(2002), 1,
      sym_identifier,
    STATE(979), 1,
      sym_value_variant,
    STATE(800), 2,
      sym_line_comment,
      aux_sym_enum_body_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [28987] = 6,
    ACTIONS(2005), 1,
      anon_sym_LBRACK,
    ACTIONS(2007), 1,
      sym_identifier,
    ACTIONS(2009), 1,
      sym_wildcard,
    STATE(460), 1,
      sym_mixin_member,
    STATE(801), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29008] = 3,
    STATE(802), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1762), 4,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29023] = 5,
    ACTIONS(2011), 1,
      anon_sym_RBRACE,
    STATE(803), 1,
      sym_line_comment,
    STATE(1116), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29042] = 5,
    ACTIONS(2013), 1,
      anon_sym_RBRACE,
    STATE(804), 1,
      sym_line_comment,
    STATE(1117), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29061] = 5,
    ACTIONS(1791), 1,
      anon_sym_is,
    STATE(805), 1,
      sym_line_comment,
    STATE(944), 1,
      sym_annotation_only_body,
    ACTIONS(2015), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29080] = 5,
    ACTIONS(2017), 1,
      anon_sym_RBRACE,
    STATE(806), 1,
      sym_line_comment,
    STATE(1127), 1,
      sym_sequence_uniqueness,
    ACTIONS(1606), 2,
      anon_sym_unique,
      anon_sym_nonunique,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29099] = 6,
    ACTIONS(1842), 1,
      sym_identifier,
    ACTIONS(2019), 1,
      sym_seq_builder_separator,
    STATE(737), 1,
      aux_sym_sequence_builder_repeat1,
    STATE(807), 1,
      sym_line_comment,
    STATE(914), 1,
      sym_variable,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29120] = 4,
    ACTIONS(1140), 1,
      anon_sym_AT,
    STATE(808), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1811), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [29137] = 6,
    ACTIONS(2021), 1,
      anon_sym_LBRACK,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(457), 1,
      sym_identifier_reference,
    STATE(809), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29158] = 6,
    ACTIONS(1842), 1,
      sym_identifier,
    ACTIONS(2025), 1,
      sym_seq_builder_separator,
    STATE(737), 1,
      aux_sym_sequence_builder_repeat1,
    STATE(810), 1,
      sym_line_comment,
    STATE(914), 1,
      sym_variable,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29179] = 4,
    ACTIONS(1142), 1,
      anon_sym_AT,
    STATE(811), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1186), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_assert,
  [29196] = 6,
    ACTIONS(1542), 1,
      sym_identifier,
    ACTIONS(1566), 1,
      anon_sym_end,
    STATE(800), 1,
      aux_sym_enum_body_repeat1,
    STATE(812), 1,
      sym_line_comment,
    STATE(979), 1,
      sym_value_variant,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29217] = 3,
    STATE(813), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1775), 4,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29232] = 3,
    STATE(814), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2027), 4,
      anon_sym_end,
      sym_identifier,
      anon_sym_parent,
      anon_sym_ref,
  [29247] = 3,
    STATE(815), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1836), 4,
      anon_sym_is,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29262] = 4,
    ACTIONS(2029), 1,
      anon_sym_LPAREN,
    STATE(816), 1,
      sym_line_comment,
    ACTIONS(2031), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29278] = 5,
    ACTIONS(2033), 1,
      sym_identifier,
    STATE(555), 1,
      sym_function_signature,
    STATE(817), 1,
      sym_line_comment,
    STATE(834), 1,
      sym_function_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29296] = 5,
    ACTIONS(2033), 1,
      sym_identifier,
    STATE(465), 1,
      sym_function_def,
    STATE(555), 1,
      sym_function_signature,
    STATE(818), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29314] = 5,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(640), 1,
      sym_identifier_reference,
    STATE(819), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29332] = 3,
    STATE(820), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2035), 3,
      anon_sym_end,
      anon_sym_metric,
      anon_sym_ref,
  [29346] = 3,
    STATE(821), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2027), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29360] = 3,
    STATE(822), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1898), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29374] = 5,
    ACTIONS(2037), 1,
      anon_sym_RBRACK,
    ACTIONS(2039), 1,
      sym_quoted_string,
    STATE(823), 1,
      sym_line_comment,
    STATE(837), 1,
      aux_sym_pattern_restriction_facet_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29392] = 5,
    ACTIONS(1696), 1,
      sym_identifier,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(824), 1,
      sym_line_comment,
    STATE(978), 1,
      sym_identifier_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29410] = 3,
    STATE(825), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1875), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29424] = 5,
    ACTIONS(2041), 1,
      sym_identifier,
    STATE(107), 1,
      sym_qualified_identifier,
    STATE(690), 1,
      sym_identifier_reference,
    STATE(826), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29442] = 5,
    ACTIONS(2043), 1,
      anon_sym_RBRACE,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    STATE(827), 1,
      sym_line_comment,
    STATE(1042), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29460] = 3,
    STATE(828), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2047), 3,
      sym_identifier,
      anon_sym_RPAREN,
      anon_sym_PLUS,
  [29474] = 3,
    STATE(829), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1916), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29488] = 3,
    STATE(830), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1926), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29502] = 3,
    STATE(831), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1934), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29516] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2049), 1,
      anon_sym_RBRACE,
    STATE(832), 1,
      sym_line_comment,
    STATE(1047), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29534] = 3,
    STATE(833), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1946), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29548] = 3,
    STATE(834), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2051), 3,
      anon_sym_end,
      anon_sym_for,
      anon_sym_def,
  [29562] = 4,
    ACTIONS(2053), 1,
      anon_sym_import,
    ACTIONS(2055), 1,
      anon_sym_COLON_COLON,
    STATE(835), 2,
      sym_line_comment,
      aux_sym__module_path_actual_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29578] = 5,
    ACTIONS(1696), 1,
      sym_identifier,
    STATE(627), 1,
      sym_qualified_identifier,
    STATE(787), 1,
      sym_identifier_reference,
    STATE(836), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29596] = 4,
    ACTIONS(2058), 1,
      anon_sym_RBRACK,
    ACTIONS(2060), 1,
      sym_quoted_string,
    STATE(837), 2,
      sym_line_comment,
      aux_sym_pattern_restriction_facet_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29612] = 3,
    STATE(838), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1886), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29626] = 3,
    STATE(839), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2063), 3,
      sym_identifier,
      anon_sym_RPAREN,
      anon_sym_PLUS,
  [29640] = 5,
    ACTIONS(1546), 1,
      anon_sym_def,
    STATE(757), 1,
      aux_sym_sentence_with_environment_repeat1,
    STATE(840), 1,
      sym_line_comment,
    STATE(945), 1,
      sym_keyword_function_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29658] = 3,
    STATE(841), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2065), 3,
      anon_sym_end,
      anon_sym_metric,
      anon_sym_ref,
  [29672] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2067), 1,
      anon_sym_RBRACE,
    STATE(842), 1,
      sym_line_comment,
    STATE(1111), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29690] = 3,
    STATE(843), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1986), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29704] = 3,
    STATE(844), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1840), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29718] = 3,
    STATE(845), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1992), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29732] = 5,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(763), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(846), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29750] = 5,
    ACTIONS(2069), 1,
      anon_sym_RBRACK,
    ACTIONS(2071), 1,
      sym_hex_byte,
    STATE(847), 1,
      sym_line_comment,
    STATE(870), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29768] = 5,
    ACTIONS(2073), 1,
      anon_sym_import,
    ACTIONS(2075), 1,
      anon_sym_COLON_COLON,
    STATE(835), 1,
      aux_sym__module_path_actual_repeat1,
    STATE(848), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29786] = 5,
    ACTIONS(137), 1,
      anon_sym_LT,
    ACTIONS(2077), 1,
      sym_quoted_string,
    STATE(849), 1,
      sym_line_comment,
    STATE(1000), 1,
      sym_iri,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29804] = 3,
    STATE(850), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(107), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29818] = 5,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(764), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(851), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29836] = 3,
    STATE(852), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(103), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [29850] = 5,
    ACTIONS(1222), 1,
      anon_sym_RBRACK,
    ACTIONS(2079), 1,
      anon_sym_as,
    ACTIONS(2081), 1,
      sym_identifier,
    STATE(853), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29868] = 4,
    ACTIONS(7), 1,
      anon_sym_RBRACK,
    STATE(854), 1,
      sym_line_comment,
    ACTIONS(9), 2,
      anon_sym_as,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29884] = 5,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(775), 1,
      sym_identifier_reference,
    STATE(855), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29902] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2083), 1,
      anon_sym_RBRACK,
    STATE(856), 1,
      sym_line_comment,
    STATE(860), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29920] = 5,
    ACTIONS(2085), 1,
      sym_identifier,
    STATE(567), 1,
      sym_identifier_reference,
    STATE(617), 1,
      sym_qualified_identifier,
    STATE(857), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29938] = 5,
    ACTIONS(1212), 1,
      anon_sym_RBRACK,
    ACTIONS(2087), 1,
      anon_sym_as,
    ACTIONS(2089), 1,
      sym_identifier,
    STATE(858), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29956] = 5,
    ACTIONS(2091), 1,
      sym_identifier,
    STATE(735), 1,
      sym_identifier_reference,
    STATE(745), 1,
      sym_qualified_identifier,
    STATE(859), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29974] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2093), 1,
      anon_sym_RBRACK,
    STATE(860), 1,
      sym_line_comment,
    STATE(904), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [29992] = 5,
    ACTIONS(1894), 1,
      sym_identifier,
    STATE(742), 1,
      aux_sym__class_type_parameters_repeat1,
    STATE(861), 1,
      sym_line_comment,
    STATE(952), 1,
      sym_type_parameter,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30010] = 5,
    ACTIONS(1930), 1,
      sym_identifier,
    ACTIONS(2095), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_mixin_member,
    STATE(862), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30028] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2097), 1,
      anon_sym_RBRACE,
    STATE(863), 1,
      sym_line_comment,
    STATE(1023), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30046] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2099), 1,
      anon_sym_RBRACE,
    STATE(864), 1,
      sym_line_comment,
    STATE(1025), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30064] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2101), 1,
      anon_sym_RBRACE,
    STATE(865), 1,
      sym_line_comment,
    STATE(1080), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30082] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2103), 1,
      anon_sym_RBRACE,
    STATE(866), 1,
      sym_line_comment,
    STATE(1092), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30100] = 3,
    STATE(867), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2105), 3,
      sym_identifier,
      anon_sym_RPAREN,
      sym_wildcard,
  [30114] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2107), 1,
      anon_sym_RBRACE,
    STATE(868), 1,
      sym_line_comment,
    STATE(1125), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30132] = 4,
    ACTIONS(101), 1,
      anon_sym_RBRACK,
    STATE(869), 1,
      sym_line_comment,
    ACTIONS(103), 2,
      anon_sym_as,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30148] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2109), 1,
      anon_sym_RBRACK,
    STATE(870), 1,
      sym_line_comment,
    STATE(904), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30166] = 5,
    ACTIONS(2111), 1,
      sym_identifier,
    STATE(649), 1,
      sym_identifier_reference,
    STATE(692), 1,
      sym_qualified_identifier,
    STATE(871), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30184] = 5,
    ACTIONS(2113), 1,
      sym_identifier,
    STATE(561), 1,
      sym_variable,
    STATE(872), 1,
      sym_line_comment,
    STATE(1081), 1,
      sym_quantified_variable,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30202] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2115), 1,
      anon_sym_RBRACK,
    STATE(873), 1,
      sym_line_comment,
    STATE(876), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30220] = 5,
    ACTIONS(2117), 1,
      sym_identifier,
    STATE(605), 1,
      sym_identifier_reference,
    STATE(661), 1,
      sym_qualified_identifier,
    STATE(874), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30238] = 3,
    STATE(875), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1755), 3,
      sym_identifier,
      anon_sym_RPAREN,
      anon_sym_PLUS,
  [30252] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2119), 1,
      anon_sym_RBRACK,
    STATE(876), 1,
      sym_line_comment,
    STATE(904), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30270] = 5,
    ACTIONS(1960), 1,
      sym_identifier,
    ACTIONS(2121), 1,
      anon_sym_LBRACK,
    STATE(762), 1,
      sym_mixin_member,
    STATE(877), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30288] = 5,
    ACTIONS(2123), 1,
      anon_sym_import,
    ACTIONS(2125), 1,
      sym_identifier,
    STATE(878), 1,
      sym_line_comment,
    STATE(1055), 1,
      sym__module_path_actual,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30306] = 5,
    ACTIONS(2075), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2127), 1,
      anon_sym_import,
    STATE(848), 1,
      aux_sym__module_path_actual_repeat1,
    STATE(879), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30324] = 5,
    ACTIONS(1912), 1,
      sym_identifier,
    STATE(755), 1,
      aux_sym_function_signature_repeat1,
    STATE(880), 1,
      sym_line_comment,
    STATE(948), 1,
      sym_function_parameter,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30342] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2129), 1,
      anon_sym_RBRACE,
    STATE(881), 1,
      sym_line_comment,
    STATE(995), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30360] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2131), 1,
      anon_sym_RBRACK,
    STATE(882), 1,
      sym_line_comment,
    STATE(884), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30378] = 4,
    STATE(345), 1,
      sym_set_op_is_member,
    STATE(883), 1,
      sym_line_comment,
    ACTIONS(57), 2,
      anon_sym_in,
      anon_sym_u2208,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30394] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2133), 1,
      anon_sym_RBRACK,
    STATE(884), 1,
      sym_line_comment,
    STATE(904), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30412] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2135), 1,
      anon_sym_RBRACE,
    STATE(885), 1,
      sym_line_comment,
    STATE(1041), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30430] = 5,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(459), 1,
      sym_identifier_reference,
    STATE(886), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30448] = 5,
    ACTIONS(2137), 1,
      sym_identifier,
    STATE(825), 1,
      sym_identifier_reference,
    STATE(850), 1,
      sym_qualified_identifier,
    STATE(887), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30466] = 3,
    STATE(888), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1956), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [30480] = 3,
    STATE(889), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2139), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [30494] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2141), 1,
      anon_sym_RBRACK,
    STATE(890), 1,
      sym_line_comment,
    STATE(891), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30512] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2143), 1,
      anon_sym_RBRACK,
    STATE(891), 1,
      sym_line_comment,
    STATE(904), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30530] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2145), 1,
      anon_sym_RBRACK,
    STATE(892), 1,
      sym_line_comment,
    STATE(893), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30548] = 5,
    ACTIONS(2071), 1,
      sym_hex_byte,
    ACTIONS(2147), 1,
      anon_sym_RBRACK,
    STATE(893), 1,
      sym_line_comment,
    STATE(904), 1,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30566] = 5,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(894), 1,
      sym_line_comment,
    STATE(1007), 1,
      sym_identifier_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30584] = 3,
    STATE(895), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(1972), 3,
      anon_sym_end,
      sym_identifier,
      anon_sym_ref,
  [30598] = 5,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(765), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(896), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30616] = 5,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(766), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(897), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30634] = 5,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(898), 1,
      sym_line_comment,
    STATE(1134), 1,
      sym_identifier_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30652] = 5,
    ACTIONS(1298), 1,
      anon_sym_RBRACK,
    ACTIONS(1766), 1,
      sym_identifier,
    ACTIONS(2149), 1,
      anon_sym_as,
    STATE(899), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30670] = 5,
    ACTIONS(2023), 1,
      sym_identifier,
    STATE(14), 1,
      sym_qualified_identifier,
    STATE(900), 1,
      sym_line_comment,
    STATE(1090), 1,
      sym_identifier_reference,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30688] = 5,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(776), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(901), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30706] = 5,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(777), 1,
      aux_sym__mixin_member_or_sequence_repeat1,
    STATE(902), 1,
      sym_line_comment,
    STATE(927), 1,
      sym_mixin_member,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30724] = 5,
    ACTIONS(2007), 1,
      sym_identifier,
    ACTIONS(2151), 1,
      anon_sym_LBRACK,
    STATE(439), 1,
      sym_mixin_member,
    STATE(903), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30742] = 4,
    ACTIONS(2153), 1,
      anon_sym_RBRACK,
    ACTIONS(2155), 1,
      sym_hex_byte,
    STATE(904), 2,
      sym_line_comment,
      aux_sym_binary_repeat1,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30758] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2158), 1,
      anon_sym_RBRACE,
    STATE(905), 1,
      sym_line_comment,
    STATE(1020), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30776] = 5,
    ACTIONS(2045), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2160), 1,
      anon_sym_RBRACE,
    STATE(906), 1,
      sym_line_comment,
    STATE(1021), 1,
      sym_cardinality_range,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30794] = 3,
    STATE(907), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
    ACTIONS(2162), 3,
      sym_identifier,
      anon_sym_RPAREN,
      sym_wildcard,
  [30808] = 4,
    ACTIONS(2164), 1,
      sym_unsigned,
    ACTIONS(2166), 1,
      sym_kw_is_fixed,
    STATE(908), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30823] = 4,
    ACTIONS(2168), 1,
      anon_sym_RBRACE,
    ACTIONS(2170), 1,
      sym_unsigned,
    STATE(909), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30838] = 4,
    ACTIONS(1859), 1,
      sym_identifier,
    STATE(478), 1,
      sym_member_def,
    STATE(910), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30853] = 4,
    ACTIONS(1723), 1,
      anon_sym_is,
    STATE(911), 1,
      sym_line_comment,
    STATE(993), 1,
      sym_module_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30868] = 3,
    STATE(912), 1,
      sym_line_comment,
    ACTIONS(2172), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30881] = 4,
    ACTIONS(2174), 1,
      anon_sym_LBRACK,
    ACTIONS(2176), 1,
      sym_value_empty_sequence,
    STATE(913), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30896] = 3,
    STATE(914), 1,
      sym_line_comment,
    ACTIONS(2178), 2,
      sym_identifier,
      sym_seq_builder_separator,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30909] = 3,
    STATE(915), 1,
      sym_line_comment,
    ACTIONS(2180), 2,
      anon_sym_end,
      anon_sym_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30922] = 3,
    STATE(916), 1,
      sym_line_comment,
    ACTIONS(2182), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30935] = 4,
    ACTIONS(1723), 1,
      anon_sym_is,
    STATE(917), 1,
      sym_line_comment,
    STATE(1026), 1,
      sym_module_body,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30950] = 4,
    ACTIONS(963), 1,
      anon_sym_LPAREN,
    ACTIONS(2184), 1,
      anon_sym_COMMA,
    STATE(918), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30965] = 3,
    STATE(919), 1,
      sym_line_comment,
    ACTIONS(2186), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30978] = 3,
    STATE(920), 1,
      sym_line_comment,
    ACTIONS(2188), 2,
      anon_sym_import,
      anon_sym_COLON_COLON,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [30991] = 4,
    ACTIONS(2190), 1,
      anon_sym_RBRACE,
    ACTIONS(2192), 1,
      sym_unsigned,
    STATE(921), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31006] = 4,
    ACTIONS(2194), 1,
      anon_sym_LBRACK,
    ACTIONS(2196), 1,
      sym_value_empty_sequence,
    STATE(922), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31021] = 4,
    ACTIONS(2198), 1,
      anon_sym_LBRACK,
    ACTIONS(2200), 1,
      sym_value_empty_sequence,
    STATE(923), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31036] = 4,
    ACTIONS(2202), 1,
      anon_sym_LBRACK,
    ACTIONS(2204), 1,
      sym_value_empty_sequence,
    STATE(924), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31051] = 4,
    ACTIONS(2206), 1,
      anon_sym_LBRACK,
    ACTIONS(2208), 1,
      sym_value_empty_sequence,
    STATE(925), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31066] = 4,
    ACTIONS(2210), 1,
      anon_sym_RBRACE,
    ACTIONS(2212), 1,
      sym_unsigned,
    STATE(926), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31081] = 3,
    STATE(927), 1,
      sym_line_comment,
    ACTIONS(2214), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31094] = 4,
    ACTIONS(2216), 1,
      anon_sym_LBRACK,
    ACTIONS(2218), 1,
      sym_value_empty_sequence,
    STATE(928), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31109] = 4,
    ACTIONS(2220), 1,
      anon_sym_LBRACK,
    ACTIONS(2222), 1,
      sym_value_empty_sequence,
    STATE(929), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31124] = 3,
    STATE(930), 1,
      sym_line_comment,
    ACTIONS(2224), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31137] = 3,
    STATE(931), 1,
      sym_line_comment,
    ACTIONS(1252), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31150] = 3,
    STATE(932), 1,
      sym_line_comment,
    ACTIONS(2226), 2,
      anon_sym_RBRACK,
      sym_hex_byte,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31163] = 4,
    ACTIONS(2228), 1,
      sym_unsigned,
    ACTIONS(2230), 1,
      sym_kw_is_fixed,
    STATE(933), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31178] = 4,
    ACTIONS(2232), 1,
      anon_sym_LBRACK,
    ACTIONS(2234), 1,
      sym_quoted_string,
    STATE(934), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31193] = 3,
    STATE(935), 1,
      sym_line_comment,
    ACTIONS(1236), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31206] = 3,
    STATE(936), 1,
      sym_line_comment,
    ACTIONS(2236), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31219] = 4,
    ACTIONS(2238), 1,
      anon_sym_LBRACK,
    ACTIONS(2240), 1,
      sym_value_empty_sequence,
    STATE(937), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31234] = 3,
    STATE(938), 1,
      sym_line_comment,
    ACTIONS(2242), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31247] = 4,
    ACTIONS(137), 1,
      anon_sym_LT,
    STATE(939), 1,
      sym_line_comment,
    STATE(1101), 1,
      sym_iri,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31262] = 3,
    STATE(940), 1,
      sym_line_comment,
    ACTIONS(2244), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31275] = 4,
    ACTIONS(2246), 1,
      anon_sym_LBRACK,
    ACTIONS(2248), 1,
      sym_value_empty_sequence,
    STATE(941), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31290] = 3,
    STATE(942), 1,
      sym_line_comment,
    ACTIONS(2250), 2,
      anon_sym_LT_DASH,
      anon_sym_u2190,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31303] = 4,
    ACTIONS(2252), 1,
      anon_sym_DASH_GT,
    ACTIONS(2254), 1,
      anon_sym_u2192,
    STATE(943), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31318] = 3,
    STATE(944), 1,
      sym_line_comment,
    ACTIONS(2256), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31331] = 3,
    STATE(945), 1,
      sym_line_comment,
    ACTIONS(2258), 2,
      anon_sym_for,
      anon_sym_def,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31344] = 4,
    ACTIONS(2260), 1,
      anon_sym_LBRACK,
    ACTIONS(2262), 1,
      sym_value_empty_sequence,
    STATE(946), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31359] = 4,
    ACTIONS(2264), 1,
      anon_sym_LBRACK,
    ACTIONS(2266), 1,
      sym_value_empty_sequence,
    STATE(947), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31374] = 3,
    STATE(948), 1,
      sym_line_comment,
    ACTIONS(2268), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31387] = 4,
    ACTIONS(2270), 1,
      anon_sym_LBRACK,
    ACTIONS(2272), 1,
      sym_value_empty_sequence,
    STATE(949), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31402] = 4,
    ACTIONS(2274), 1,
      anon_sym_LBRACK,
    ACTIONS(2276), 1,
      sym_value_empty_sequence,
    STATE(950), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31417] = 4,
    ACTIONS(2278), 1,
      anon_sym_LBRACK,
    ACTIONS(2280), 1,
      sym_value_empty_sequence,
    STATE(951), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31432] = 3,
    STATE(952), 1,
      sym_line_comment,
    ACTIONS(2282), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31445] = 3,
    STATE(953), 1,
      sym_line_comment,
    ACTIONS(2284), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31458] = 3,
    STATE(954), 1,
      sym_line_comment,
    ACTIONS(2286), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31471] = 3,
    STATE(955), 1,
      sym_line_comment,
    ACTIONS(2288), 2,
      anon_sym_RBRACK,
      sym_quoted_string,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31484] = 4,
    ACTIONS(2290), 1,
      anon_sym_on,
    STATE(424), 1,
      sym_metric_event_binding,
    STATE(956), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31499] = 4,
    ACTIONS(2039), 1,
      sym_quoted_string,
    STATE(823), 1,
      aux_sym_pattern_restriction_facet_repeat1,
    STATE(957), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31514] = 3,
    STATE(958), 1,
      sym_line_comment,
    ACTIONS(955), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31527] = 3,
    STATE(959), 1,
      sym_line_comment,
    ACTIONS(2292), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31540] = 3,
    STATE(960), 1,
      sym_line_comment,
    ACTIONS(2294), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31553] = 4,
    ACTIONS(2296), 1,
      anon_sym_LBRACK,
    ACTIONS(2298), 1,
      sym_value_empty_sequence,
    STATE(961), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31568] = 4,
    ACTIONS(2300), 1,
      anon_sym_LBRACK,
    ACTIONS(2302), 1,
      sym_value_empty_sequence,
    STATE(962), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31583] = 4,
    ACTIONS(2304), 1,
      anon_sym_LBRACK,
    ACTIONS(2306), 1,
      sym_value_empty_sequence,
    STATE(963), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31598] = 4,
    ACTIONS(2308), 1,
      anon_sym_LBRACK,
    ACTIONS(2310), 1,
      sym_value_empty_sequence,
    STATE(964), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31613] = 4,
    ACTIONS(2312), 1,
      anon_sym_LBRACK,
    ACTIONS(2314), 1,
      sym_value_empty_sequence,
    STATE(965), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31628] = 4,
    ACTIONS(2316), 1,
      anon_sym_LBRACK,
    ACTIONS(2318), 1,
      sym_value_empty_sequence,
    STATE(966), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31643] = 3,
    STATE(967), 1,
      sym_line_comment,
    ACTIONS(2320), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31656] = 3,
    STATE(968), 1,
      sym_line_comment,
    ACTIONS(2322), 2,
      sym_identifier,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31669] = 3,
    STATE(969), 1,
      sym_line_comment,
    ACTIONS(2324), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31682] = 4,
    ACTIONS(2326), 1,
      anon_sym_LBRACK,
    ACTIONS(2328), 1,
      sym_value_empty_sequence,
    STATE(970), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31697] = 4,
    ACTIONS(2330), 1,
      anon_sym_LBRACK,
    ACTIONS(2332), 1,
      sym_value_empty_sequence,
    STATE(971), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31712] = 3,
    STATE(972), 1,
      sym_line_comment,
    ACTIONS(797), 2,
      anon_sym_in,
      anon_sym_u2208,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31725] = 3,
    STATE(973), 1,
      sym_line_comment,
    ACTIONS(1956), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31738] = 4,
    ACTIONS(2334), 1,
      anon_sym_RBRACE,
    ACTIONS(2336), 1,
      sym_unsigned,
    STATE(974), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31753] = 4,
    ACTIONS(2338), 1,
      anon_sym_LBRACK,
    ACTIONS(2340), 1,
      sym_value_empty_sequence,
    STATE(975), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31768] = 3,
    STATE(976), 1,
      sym_line_comment,
    ACTIONS(943), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31781] = 3,
    STATE(977), 1,
      sym_line_comment,
    ACTIONS(1034), 2,
      sym_identifier,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31794] = 3,
    STATE(978), 1,
      sym_line_comment,
    ACTIONS(1789), 2,
      sym_identifier,
      sym_seq_builder_separator,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31807] = 3,
    STATE(979), 1,
      sym_line_comment,
    ACTIONS(2342), 2,
      anon_sym_end,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31820] = 3,
    STATE(980), 1,
      sym_line_comment,
    ACTIONS(1260), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31833] = 3,
    STATE(981), 1,
      sym_line_comment,
    ACTIONS(2344), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31846] = 3,
    STATE(982), 1,
      sym_line_comment,
    ACTIONS(2346), 2,
      anon_sym_RBRACE,
      sym_unsigned,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31859] = 4,
    ACTIONS(2348), 1,
      anon_sym_RBRACE,
    ACTIONS(2350), 1,
      sym_unsigned,
    STATE(983), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31874] = 3,
    STATE(984), 1,
      sym_line_comment,
    ACTIONS(2352), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31887] = 3,
    STATE(985), 1,
      sym_line_comment,
    ACTIONS(2354), 2,
      anon_sym_DASH_GT,
      anon_sym_u2192,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31900] = 3,
    STATE(986), 1,
      sym_line_comment,
    ACTIONS(1318), 2,
      anon_sym_RBRACK,
      sym_identifier,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31913] = 4,
    ACTIONS(2356), 1,
      anon_sym_LBRACK,
    ACTIONS(2358), 1,
      sym_value_empty_sequence,
    STATE(987), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31928] = 3,
    ACTIONS(2360), 1,
      anon_sym_RBRACE,
    STATE(988), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31940] = 3,
    ACTIONS(2362), 1,
      sym_unsigned,
    STATE(989), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31952] = 3,
    ACTIONS(2364), 1,
      sym_identifier,
    STATE(990), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31964] = 3,
    ACTIONS(2366), 1,
      sym_identifier,
    STATE(991), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31976] = 3,
    ACTIONS(2368), 1,
      sym_quoted_string,
    STATE(992), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [31988] = 3,
    ACTIONS(2370), 1,
      ts_builtin_sym_end,
    STATE(993), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32000] = 3,
    ACTIONS(2372), 1,
      ts_builtin_sym_end,
    STATE(994), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32012] = 3,
    ACTIONS(2374), 1,
      anon_sym_RBRACE,
    STATE(995), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32024] = 3,
    ACTIONS(2376), 1,
      aux_sym_iri_token1,
    STATE(996), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32036] = 3,
    ACTIONS(2380), 1,
      sym_identifier,
    STATE(997), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32048] = 3,
    ACTIONS(2382), 1,
      ts_builtin_sym_end,
    STATE(998), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32060] = 3,
    ACTIONS(2384), 1,
      anon_sym_RBRACE,
    STATE(999), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32072] = 3,
    ACTIONS(2386), 1,
      anon_sym_is,
    STATE(1000), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32084] = 3,
    ACTIONS(2388), 1,
      sym_identifier,
    STATE(1001), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32096] = 3,
    ACTIONS(2390), 1,
      anon_sym_RBRACE,
    STATE(1002), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32108] = 3,
    ACTIONS(2392), 1,
      sym_identifier,
    STATE(1003), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32120] = 3,
    ACTIONS(2394), 1,
      anon_sym_EQ,
    STATE(1004), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32132] = 3,
    ACTIONS(2396), 1,
      anon_sym_GT,
    STATE(1005), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32144] = 3,
    ACTIONS(2398), 1,
      anon_sym_RPAREN,
    STATE(1006), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32156] = 3,
    ACTIONS(2400), 1,
      anon_sym_EQ,
    STATE(1007), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32168] = 3,
    ACTIONS(2402), 1,
      sym_identifier,
    STATE(1008), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32180] = 3,
    ACTIONS(2404), 1,
      sym_identifier,
    STATE(1009), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32192] = 3,
    ACTIONS(2406), 1,
      anon_sym_EQ,
    STATE(1010), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32204] = 3,
    ACTIONS(2408), 1,
      anon_sym_RPAREN,
    STATE(1011), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32216] = 3,
    ACTIONS(2410), 1,
      anon_sym_RBRACE,
    STATE(1012), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32228] = 3,
    ACTIONS(2412), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32240] = 3,
    ACTIONS(2414), 1,
      sym_identifier,
    STATE(1014), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32252] = 3,
    ACTIONS(2416), 1,
      anon_sym_RBRACE,
    STATE(1015), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32264] = 3,
    ACTIONS(2418), 1,
      ts_builtin_sym_end,
    STATE(1016), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32276] = 3,
    ACTIONS(2420), 1,
      sym_identifier,
    STATE(1017), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32288] = 3,
    ACTIONS(2422), 1,
      sym_identifier,
    STATE(1018), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32300] = 3,
    ACTIONS(2424), 1,
      anon_sym_RPAREN,
    STATE(1019), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32312] = 3,
    ACTIONS(2426), 1,
      anon_sym_RBRACE,
    STATE(1020), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32324] = 3,
    ACTIONS(2428), 1,
      anon_sym_RBRACE,
    STATE(1021), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32336] = 3,
    ACTIONS(2430), 1,
      anon_sym_COMMA,
    STATE(1022), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32348] = 3,
    ACTIONS(2432), 1,
      anon_sym_RBRACE,
    STATE(1023), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32360] = 3,
    ACTIONS(2434), 1,
      anon_sym_import,
    STATE(1024), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32372] = 3,
    ACTIONS(2436), 1,
      anon_sym_RBRACE,
    STATE(1025), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32384] = 3,
    ACTIONS(2438), 1,
      ts_builtin_sym_end,
    STATE(1026), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32396] = 3,
    ACTIONS(338), 1,
      sym_identifier,
    STATE(1027), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32408] = 3,
    ACTIONS(2440), 1,
      anon_sym_GT,
    STATE(1028), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32420] = 3,
    ACTIONS(2442), 1,
      sym_identifier,
    STATE(1029), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32432] = 3,
    ACTIONS(2444), 1,
      sym_identifier,
    STATE(1030), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32444] = 3,
    ACTIONS(342), 1,
      sym_identifier,
    STATE(1031), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32456] = 3,
    ACTIONS(2446), 1,
      sym_identifier,
    STATE(1032), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32468] = 3,
    ACTIONS(2448), 1,
      sym_quoted_string,
    STATE(1033), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32480] = 3,
    ACTIONS(354), 1,
      sym_identifier,
    STATE(1034), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32492] = 3,
    ACTIONS(350), 1,
      sym_identifier,
    STATE(1035), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32504] = 3,
    ACTIONS(2450), 1,
      anon_sym_EQ,
    STATE(1036), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32516] = 3,
    ACTIONS(346), 1,
      sym_identifier,
    STATE(1037), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32528] = 3,
    ACTIONS(2452), 1,
      sym_identifier,
    STATE(1038), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32540] = 3,
    ACTIONS(2454), 1,
      anon_sym_RPAREN,
    STATE(1039), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32552] = 3,
    ACTIONS(2456), 1,
      sym_identifier,
    STATE(1040), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32564] = 3,
    ACTIONS(2458), 1,
      anon_sym_RBRACE,
    STATE(1041), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32576] = 3,
    ACTIONS(2460), 1,
      anon_sym_RBRACE,
    STATE(1042), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32588] = 3,
    ACTIONS(2462), 1,
      anon_sym_RBRACE,
    STATE(1043), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32600] = 3,
    ACTIONS(358), 1,
      sym_identifier,
    STATE(1044), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32612] = 3,
    ACTIONS(2464), 1,
      sym_unsigned,
    STATE(1045), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32624] = 3,
    ACTIONS(2466), 1,
      ts_builtin_sym_end,
    STATE(1046), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32636] = 3,
    ACTIONS(2468), 1,
      anon_sym_RBRACE,
    STATE(1047), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32648] = 3,
    ACTIONS(2470), 1,
      sym_identifier,
    STATE(1048), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32660] = 3,
    ACTIONS(2472), 1,
      anon_sym_RPAREN,
    STATE(1049), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32672] = 3,
    ACTIONS(2474), 1,
      anon_sym_import,
    STATE(1050), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32684] = 3,
    ACTIONS(2476), 1,
      anon_sym_EQ,
    STATE(1051), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32696] = 3,
    ACTIONS(2478), 1,
      sym_identifier,
    STATE(1052), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32708] = 3,
    ACTIONS(2480), 1,
      ts_builtin_sym_end,
    STATE(1053), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32720] = 3,
    ACTIONS(2482), 1,
      sym_identifier,
    STATE(1054), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32732] = 3,
    ACTIONS(2484), 1,
      anon_sym_import,
    STATE(1055), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32744] = 3,
    ACTIONS(2486), 1,
      ts_builtin_sym_end,
    STATE(1056), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32756] = 3,
    ACTIONS(2488), 1,
      ts_builtin_sym_end,
    STATE(1057), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32768] = 3,
    ACTIONS(2490), 1,
      anon_sym_GT,
    STATE(1058), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32780] = 3,
    ACTIONS(2492), 1,
      sym_identifier,
    STATE(1059), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32792] = 3,
    ACTIONS(366), 1,
      sym_identifier,
    STATE(1060), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32804] = 3,
    ACTIONS(2494), 1,
      sym_unsigned,
    STATE(1061), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32816] = 3,
    ACTIONS(2496), 1,
      anon_sym_GT,
    STATE(1062), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32828] = 3,
    ACTIONS(2498), 1,
      sym_identifier,
    STATE(1063), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32840] = 3,
    ACTIONS(2500), 1,
      sym_identifier,
    STATE(1064), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32852] = 3,
    ACTIONS(2502), 1,
      anon_sym_RPAREN,
    STATE(1065), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32864] = 3,
    ACTIONS(2504), 1,
      anon_sym_GT,
    STATE(1066), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32876] = 3,
    ACTIONS(2506), 1,
      sym_identifier,
    STATE(1067), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32888] = 3,
    ACTIONS(326), 1,
      sym_identifier,
    STATE(1068), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32900] = 3,
    ACTIONS(2508), 1,
      sym_identifier,
    STATE(1069), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32912] = 3,
    ACTIONS(2510), 1,
      sym_identifier,
    STATE(1070), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32924] = 3,
    ACTIONS(2512), 1,
      sym_identifier,
    STATE(1071), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32936] = 3,
    ACTIONS(2514), 1,
      sym_identifier,
    STATE(1072), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32948] = 3,
    ACTIONS(2516), 1,
      sym_identifier,
    STATE(1073), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32960] = 3,
    ACTIONS(2518), 1,
      sym_identifier,
    STATE(1074), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32972] = 3,
    ACTIONS(2520), 1,
      sym_identifier,
    STATE(1075), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32984] = 3,
    ACTIONS(2522), 1,
      aux_sym_iri_token1,
    STATE(1076), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [32996] = 3,
    ACTIONS(2524), 1,
      sym_identifier,
    STATE(1077), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33008] = 3,
    ACTIONS(2526), 1,
      sym_identifier,
    STATE(1078), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33020] = 3,
    ACTIONS(2528), 1,
      sym_identifier,
    STATE(1079), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33032] = 3,
    ACTIONS(2530), 1,
      anon_sym_RBRACE,
    STATE(1080), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33044] = 3,
    ACTIONS(2532), 1,
      anon_sym_COMMA,
    STATE(1081), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33056] = 3,
    ACTIONS(2534), 1,
      sym_identifier,
    STATE(1082), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33068] = 3,
    ACTIONS(330), 1,
      sym_identifier,
    STATE(1083), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33080] = 3,
    ACTIONS(2536), 1,
      anon_sym_RBRACE,
    STATE(1084), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33092] = 3,
    ACTIONS(2538), 1,
      sym_identifier,
    STATE(1085), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33104] = 3,
    ACTIONS(2540), 1,
      sym_identifier,
    STATE(1086), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33116] = 3,
    ACTIONS(2542), 1,
      anon_sym_RBRACE,
    STATE(1087), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33128] = 3,
    ACTIONS(2544), 1,
      anon_sym_RBRACE,
    STATE(1088), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33140] = 3,
    ACTIONS(2546), 1,
      anon_sym_RBRACE,
    STATE(1089), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33152] = 3,
    ACTIONS(2548), 1,
      anon_sym_EQ,
    STATE(1090), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33164] = 3,
    ACTIONS(2550), 1,
      sym_identifier,
    STATE(1091), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33176] = 3,
    ACTIONS(2552), 1,
      anon_sym_RBRACE,
    STATE(1092), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33188] = 3,
    ACTIONS(2554), 1,
      anon_sym_RBRACE,
    STATE(1093), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33200] = 3,
    ACTIONS(2556), 1,
      aux_sym_iri_token1,
    STATE(1094), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33212] = 3,
    ACTIONS(2558), 1,
      sym_identifier,
    STATE(1095), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33224] = 3,
    ACTIONS(322), 1,
      sym_identifier,
    STATE(1096), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33236] = 3,
    ACTIONS(2560), 1,
      anon_sym_RBRACE,
    STATE(1097), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33248] = 3,
    ACTIONS(2562), 1,
      sym_unsigned,
    STATE(1098), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33260] = 3,
    ACTIONS(2564), 1,
      anon_sym_RBRACE,
    STATE(1099), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33272] = 3,
    ACTIONS(2566), 1,
      sym_identifier,
    STATE(1100), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33284] = 3,
    ACTIONS(2568), 1,
      anon_sym_is,
    STATE(1101), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33296] = 3,
    ACTIONS(2570), 1,
      anon_sym_RBRACE,
    STATE(1102), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33308] = 3,
    ACTIONS(2572), 1,
      anon_sym_RBRACE,
    STATE(1103), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33320] = 3,
    ACTIONS(2574), 1,
      anon_sym_RBRACE,
    STATE(1104), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33332] = 3,
    ACTIONS(2576), 1,
      sym_identifier,
    STATE(1105), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33344] = 3,
    ACTIONS(2578), 1,
      sym_quoted_string,
    STATE(1106), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33356] = 3,
    ACTIONS(334), 1,
      sym_identifier,
    STATE(1107), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33368] = 3,
    ACTIONS(2580), 1,
      sym_identifier,
    STATE(1108), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33380] = 3,
    ACTIONS(2582), 1,
      aux_sym_iri_token1,
    STATE(1109), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33392] = 3,
    ACTIONS(2584), 1,
      sym_identifier,
    STATE(1110), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33404] = 3,
    ACTIONS(2586), 1,
      anon_sym_RBRACE,
    STATE(1111), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33416] = 3,
    ACTIONS(2588), 1,
      anon_sym_RBRACE,
    STATE(1112), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33428] = 3,
    ACTIONS(2590), 1,
      sym_identifier,
    STATE(1113), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33440] = 3,
    ACTIONS(2592), 1,
      sym_identifier,
    STATE(1114), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33452] = 3,
    ACTIONS(2594), 1,
      sym_identifier,
    STATE(1115), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33464] = 3,
    ACTIONS(2596), 1,
      anon_sym_RBRACE,
    STATE(1116), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33476] = 3,
    ACTIONS(2598), 1,
      anon_sym_RBRACE,
    STATE(1117), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33488] = 3,
    ACTIONS(362), 1,
      sym_identifier,
    STATE(1118), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33500] = 3,
    ACTIONS(2600), 1,
      aux_sym_iri_token1,
    STATE(1119), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33512] = 3,
    ACTIONS(2602), 1,
      aux_sym_iri_token1,
    STATE(1120), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33524] = 3,
    ACTIONS(2604), 1,
      aux_sym_iri_token1,
    STATE(1121), 1,
      sym_line_comment,
    ACTIONS(2378), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33536] = 3,
    ACTIONS(992), 1,
      anon_sym_LPAREN,
    STATE(1122), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33548] = 3,
    ACTIONS(2606), 1,
      sym_identifier,
    STATE(1123), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33560] = 3,
    ACTIONS(2608), 1,
      anon_sym_GT,
    STATE(1124), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33572] = 3,
    ACTIONS(2610), 1,
      anon_sym_RBRACE,
    STATE(1125), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33584] = 3,
    ACTIONS(2612), 1,
      sym_identifier,
    STATE(1126), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33596] = 3,
    ACTIONS(2614), 1,
      anon_sym_RBRACE,
    STATE(1127), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33608] = 3,
    ACTIONS(2616), 1,
      sym_identifier,
    STATE(1128), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33620] = 3,
    ACTIONS(2618), 1,
      sym_identifier,
    STATE(1129), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33632] = 3,
    ACTIONS(2620), 1,
      anon_sym_EQ,
    STATE(1130), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33644] = 3,
    ACTIONS(2622), 1,
      anon_sym_import,
    STATE(1131), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33656] = 3,
    ACTIONS(2624), 1,
      anon_sym_RPAREN,
    STATE(1132), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33668] = 3,
    ACTIONS(2626), 1,
      anon_sym_GT,
    STATE(1133), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33680] = 3,
    ACTIONS(2628), 1,
      anon_sym_EQ,
    STATE(1134), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33692] = 3,
    ACTIONS(2630), 1,
      sym_identifier,
    STATE(1135), 1,
      sym_line_comment,
    ACTIONS(3), 3,
      sym_comment_aligned,
      sym_comment_local,
      sym_comment_page,
  [33704] = 1,
    ACTIONS(2632), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(30)] = 0,
  [SMALL_STATE(31)] = 78,
  [SMALL_STATE(32)] = 153,
  [SMALL_STATE(33)] = 228,
  [SMALL_STATE(34)] = 301,
  [SMALL_STATE(35)] = 376,
  [SMALL_STATE(36)] = 446,
  [SMALL_STATE(37)] = 516,
  [SMALL_STATE(38)] = 584,
  [SMALL_STATE(39)] = 652,
  [SMALL_STATE(40)] = 720,
  [SMALL_STATE(41)] = 788,
  [SMALL_STATE(42)] = 856,
  [SMALL_STATE(43)] = 924,
  [SMALL_STATE(44)] = 992,
  [SMALL_STATE(45)] = 1060,
  [SMALL_STATE(46)] = 1128,
  [SMALL_STATE(47)] = 1196,
  [SMALL_STATE(48)] = 1264,
  [SMALL_STATE(49)] = 1332,
  [SMALL_STATE(50)] = 1400,
  [SMALL_STATE(51)] = 1468,
  [SMALL_STATE(52)] = 1536,
  [SMALL_STATE(53)] = 1616,
  [SMALL_STATE(54)] = 1696,
  [SMALL_STATE(55)] = 1776,
  [SMALL_STATE(56)] = 1853,
  [SMALL_STATE(57)] = 1930,
  [SMALL_STATE(58)] = 2007,
  [SMALL_STATE(59)] = 2084,
  [SMALL_STATE(60)] = 2158,
  [SMALL_STATE(61)] = 2232,
  [SMALL_STATE(62)] = 2306,
  [SMALL_STATE(63)] = 2380,
  [SMALL_STATE(64)] = 2454,
  [SMALL_STATE(65)] = 2528,
  [SMALL_STATE(66)] = 2602,
  [SMALL_STATE(67)] = 2676,
  [SMALL_STATE(68)] = 2750,
  [SMALL_STATE(69)] = 2824,
  [SMALL_STATE(70)] = 2898,
  [SMALL_STATE(71)] = 2969,
  [SMALL_STATE(72)] = 3040,
  [SMALL_STATE(73)] = 3111,
  [SMALL_STATE(74)] = 3182,
  [SMALL_STATE(75)] = 3252,
  [SMALL_STATE(76)] = 3319,
  [SMALL_STATE(77)] = 3378,
  [SMALL_STATE(78)] = 3437,
  [SMALL_STATE(79)] = 3496,
  [SMALL_STATE(80)] = 3555,
  [SMALL_STATE(81)] = 3614,
  [SMALL_STATE(82)] = 3673,
  [SMALL_STATE(83)] = 3732,
  [SMALL_STATE(84)] = 3791,
  [SMALL_STATE(85)] = 3849,
  [SMALL_STATE(86)] = 3907,
  [SMALL_STATE(87)] = 3965,
  [SMALL_STATE(88)] = 4023,
  [SMALL_STATE(89)] = 4081,
  [SMALL_STATE(90)] = 4139,
  [SMALL_STATE(91)] = 4197,
  [SMALL_STATE(92)] = 4255,
  [SMALL_STATE(93)] = 4313,
  [SMALL_STATE(94)] = 4371,
  [SMALL_STATE(95)] = 4429,
  [SMALL_STATE(96)] = 4487,
  [SMALL_STATE(97)] = 4583,
  [SMALL_STATE(98)] = 4679,
  [SMALL_STATE(99)] = 4772,
  [SMALL_STATE(100)] = 4865,
  [SMALL_STATE(101)] = 4956,
  [SMALL_STATE(102)] = 5049,
  [SMALL_STATE(103)] = 5100,
  [SMALL_STATE(104)] = 5193,
  [SMALL_STATE(105)] = 5286,
  [SMALL_STATE(106)] = 5379,
  [SMALL_STATE(107)] = 5426,
  [SMALL_STATE(108)] = 5473,
  [SMALL_STATE(109)] = 5566,
  [SMALL_STATE(110)] = 5659,
  [SMALL_STATE(111)] = 5710,
  [SMALL_STATE(112)] = 5771,
  [SMALL_STATE(113)] = 5823,
  [SMALL_STATE(114)] = 5910,
  [SMALL_STATE(115)] = 5997,
  [SMALL_STATE(116)] = 6084,
  [SMALL_STATE(117)] = 6171,
  [SMALL_STATE(118)] = 6220,
  [SMALL_STATE(119)] = 6307,
  [SMALL_STATE(120)] = 6356,
  [SMALL_STATE(121)] = 6403,
  [SMALL_STATE(122)] = 6452,
  [SMALL_STATE(123)] = 6511,
  [SMALL_STATE(124)] = 6555,
  [SMALL_STATE(125)] = 6599,
  [SMALL_STATE(126)] = 6680,
  [SMALL_STATE(127)] = 6761,
  [SMALL_STATE(128)] = 6808,
  [SMALL_STATE(129)] = 6853,
  [SMALL_STATE(130)] = 6895,
  [SMALL_STATE(131)] = 6939,
  [SMALL_STATE(132)] = 7023,
  [SMALL_STATE(133)] = 7065,
  [SMALL_STATE(134)] = 7107,
  [SMALL_STATE(135)] = 7191,
  [SMALL_STATE(136)] = 7275,
  [SMALL_STATE(137)] = 7317,
  [SMALL_STATE(138)] = 7401,
  [SMALL_STATE(139)] = 7485,
  [SMALL_STATE(140)] = 7569,
  [SMALL_STATE(141)] = 7611,
  [SMALL_STATE(142)] = 7653,
  [SMALL_STATE(143)] = 7695,
  [SMALL_STATE(144)] = 7737,
  [SMALL_STATE(145)] = 7779,
  [SMALL_STATE(146)] = 7821,
  [SMALL_STATE(147)] = 7863,
  [SMALL_STATE(148)] = 7905,
  [SMALL_STATE(149)] = 7947,
  [SMALL_STATE(150)] = 7989,
  [SMALL_STATE(151)] = 8031,
  [SMALL_STATE(152)] = 8073,
  [SMALL_STATE(153)] = 8115,
  [SMALL_STATE(154)] = 8157,
  [SMALL_STATE(155)] = 8199,
  [SMALL_STATE(156)] = 8241,
  [SMALL_STATE(157)] = 8283,
  [SMALL_STATE(158)] = 8325,
  [SMALL_STATE(159)] = 8367,
  [SMALL_STATE(160)] = 8409,
  [SMALL_STATE(161)] = 8453,
  [SMALL_STATE(162)] = 8537,
  [SMALL_STATE(163)] = 8621,
  [SMALL_STATE(164)] = 8705,
  [SMALL_STATE(165)] = 8789,
  [SMALL_STATE(166)] = 8873,
  [SMALL_STATE(167)] = 8957,
  [SMALL_STATE(168)] = 9041,
  [SMALL_STATE(169)] = 9125,
  [SMALL_STATE(170)] = 9209,
  [SMALL_STATE(171)] = 9293,
  [SMALL_STATE(172)] = 9375,
  [SMALL_STATE(173)] = 9459,
  [SMALL_STATE(174)] = 9543,
  [SMALL_STATE(175)] = 9627,
  [SMALL_STATE(176)] = 9711,
  [SMALL_STATE(177)] = 9795,
  [SMALL_STATE(178)] = 9879,
  [SMALL_STATE(179)] = 9963,
  [SMALL_STATE(180)] = 10047,
  [SMALL_STATE(181)] = 10131,
  [SMALL_STATE(182)] = 10215,
  [SMALL_STATE(183)] = 10299,
  [SMALL_STATE(184)] = 10383,
  [SMALL_STATE(185)] = 10467,
  [SMALL_STATE(186)] = 10551,
  [SMALL_STATE(187)] = 10635,
  [SMALL_STATE(188)] = 10719,
  [SMALL_STATE(189)] = 10763,
  [SMALL_STATE(190)] = 10807,
  [SMALL_STATE(191)] = 10849,
  [SMALL_STATE(192)] = 10890,
  [SMALL_STATE(193)] = 10931,
  [SMALL_STATE(194)] = 10972,
  [SMALL_STATE(195)] = 11013,
  [SMALL_STATE(196)] = 11054,
  [SMALL_STATE(197)] = 11095,
  [SMALL_STATE(198)] = 11136,
  [SMALL_STATE(199)] = 11177,
  [SMALL_STATE(200)] = 11218,
  [SMALL_STATE(201)] = 11259,
  [SMALL_STATE(202)] = 11300,
  [SMALL_STATE(203)] = 11378,
  [SMALL_STATE(204)] = 11456,
  [SMALL_STATE(205)] = 11534,
  [SMALL_STATE(206)] = 11612,
  [SMALL_STATE(207)] = 11690,
  [SMALL_STATE(208)] = 11768,
  [SMALL_STATE(209)] = 11846,
  [SMALL_STATE(210)] = 11887,
  [SMALL_STATE(211)] = 11928,
  [SMALL_STATE(212)] = 11969,
  [SMALL_STATE(213)] = 12010,
  [SMALL_STATE(214)] = 12051,
  [SMALL_STATE(215)] = 12092,
  [SMALL_STATE(216)] = 12133,
  [SMALL_STATE(217)] = 12204,
  [SMALL_STATE(218)] = 12273,
  [SMALL_STATE(219)] = 12344,
  [SMALL_STATE(220)] = 12415,
  [SMALL_STATE(221)] = 12486,
  [SMALL_STATE(222)] = 12551,
  [SMALL_STATE(223)] = 12614,
  [SMALL_STATE(224)] = 12679,
  [SMALL_STATE(225)] = 12750,
  [SMALL_STATE(226)] = 12821,
  [SMALL_STATE(227)] = 12892,
  [SMALL_STATE(228)] = 12963,
  [SMALL_STATE(229)] = 13034,
  [SMALL_STATE(230)] = 13105,
  [SMALL_STATE(231)] = 13176,
  [SMALL_STATE(232)] = 13247,
  [SMALL_STATE(233)] = 13318,
  [SMALL_STATE(234)] = 13389,
  [SMALL_STATE(235)] = 13460,
  [SMALL_STATE(236)] = 13531,
  [SMALL_STATE(237)] = 13602,
  [SMALL_STATE(238)] = 13673,
  [SMALL_STATE(239)] = 13744,
  [SMALL_STATE(240)] = 13815,
  [SMALL_STATE(241)] = 13886,
  [SMALL_STATE(242)] = 13957,
  [SMALL_STATE(243)] = 14028,
  [SMALL_STATE(244)] = 14099,
  [SMALL_STATE(245)] = 14170,
  [SMALL_STATE(246)] = 14241,
  [SMALL_STATE(247)] = 14312,
  [SMALL_STATE(248)] = 14383,
  [SMALL_STATE(249)] = 14454,
  [SMALL_STATE(250)] = 14525,
  [SMALL_STATE(251)] = 14596,
  [SMALL_STATE(252)] = 14667,
  [SMALL_STATE(253)] = 14732,
  [SMALL_STATE(254)] = 14800,
  [SMALL_STATE(255)] = 14843,
  [SMALL_STATE(256)] = 14908,
  [SMALL_STATE(257)] = 14948,
  [SMALL_STATE(258)] = 14986,
  [SMALL_STATE(259)] = 15026,
  [SMALL_STATE(260)] = 15066,
  [SMALL_STATE(261)] = 15099,
  [SMALL_STATE(262)] = 15140,
  [SMALL_STATE(263)] = 15175,
  [SMALL_STATE(264)] = 15210,
  [SMALL_STATE(265)] = 15246,
  [SMALL_STATE(266)] = 15278,
  [SMALL_STATE(267)] = 15312,
  [SMALL_STATE(268)] = 15346,
  [SMALL_STATE(269)] = 15379,
  [SMALL_STATE(270)] = 15414,
  [SMALL_STATE(271)] = 15449,
  [SMALL_STATE(272)] = 15482,
  [SMALL_STATE(273)] = 15515,
  [SMALL_STATE(274)] = 15548,
  [SMALL_STATE(275)] = 15581,
  [SMALL_STATE(276)] = 15614,
  [SMALL_STATE(277)] = 15647,
  [SMALL_STATE(278)] = 15680,
  [SMALL_STATE(279)] = 15713,
  [SMALL_STATE(280)] = 15746,
  [SMALL_STATE(281)] = 15779,
  [SMALL_STATE(282)] = 15812,
  [SMALL_STATE(283)] = 15845,
  [SMALL_STATE(284)] = 15878,
  [SMALL_STATE(285)] = 15911,
  [SMALL_STATE(286)] = 15944,
  [SMALL_STATE(287)] = 15977,
  [SMALL_STATE(288)] = 16010,
  [SMALL_STATE(289)] = 16043,
  [SMALL_STATE(290)] = 16076,
  [SMALL_STATE(291)] = 16109,
  [SMALL_STATE(292)] = 16142,
  [SMALL_STATE(293)] = 16175,
  [SMALL_STATE(294)] = 16208,
  [SMALL_STATE(295)] = 16241,
  [SMALL_STATE(296)] = 16274,
  [SMALL_STATE(297)] = 16313,
  [SMALL_STATE(298)] = 16357,
  [SMALL_STATE(299)] = 16389,
  [SMALL_STATE(300)] = 16421,
  [SMALL_STATE(301)] = 16465,
  [SMALL_STATE(302)] = 16497,
  [SMALL_STATE(303)] = 16545,
  [SMALL_STATE(304)] = 16589,
  [SMALL_STATE(305)] = 16623,
  [SMALL_STATE(306)] = 16655,
  [SMALL_STATE(307)] = 16703,
  [SMALL_STATE(308)] = 16737,
  [SMALL_STATE(309)] = 16781,
  [SMALL_STATE(310)] = 16815,
  [SMALL_STATE(311)] = 16847,
  [SMALL_STATE(312)] = 16879,
  [SMALL_STATE(313)] = 16913,
  [SMALL_STATE(314)] = 16945,
  [SMALL_STATE(315)] = 16991,
  [SMALL_STATE(316)] = 17035,
  [SMALL_STATE(317)] = 17073,
  [SMALL_STATE(318)] = 17117,
  [SMALL_STATE(319)] = 17149,
  [SMALL_STATE(320)] = 17181,
  [SMALL_STATE(321)] = 17229,
  [SMALL_STATE(322)] = 17277,
  [SMALL_STATE(323)] = 17306,
  [SMALL_STATE(324)] = 17337,
  [SMALL_STATE(325)] = 17368,
  [SMALL_STATE(326)] = 17399,
  [SMALL_STATE(327)] = 17430,
  [SMALL_STATE(328)] = 17461,
  [SMALL_STATE(329)] = 17494,
  [SMALL_STATE(330)] = 17525,
  [SMALL_STATE(331)] = 17556,
  [SMALL_STATE(332)] = 17587,
  [SMALL_STATE(333)] = 17618,
  [SMALL_STATE(334)] = 17647,
  [SMALL_STATE(335)] = 17678,
  [SMALL_STATE(336)] = 17707,
  [SMALL_STATE(337)] = 17736,
  [SMALL_STATE(338)] = 17767,
  [SMALL_STATE(339)] = 17798,
  [SMALL_STATE(340)] = 17829,
  [SMALL_STATE(341)] = 17862,
  [SMALL_STATE(342)] = 17891,
  [SMALL_STATE(343)] = 17922,
  [SMALL_STATE(344)] = 17953,
  [SMALL_STATE(345)] = 17984,
  [SMALL_STATE(346)] = 18015,
  [SMALL_STATE(347)] = 18046,
  [SMALL_STATE(348)] = 18075,
  [SMALL_STATE(349)] = 18106,
  [SMALL_STATE(350)] = 18137,
  [SMALL_STATE(351)] = 18168,
  [SMALL_STATE(352)] = 18197,
  [SMALL_STATE(353)] = 18228,
  [SMALL_STATE(354)] = 18257,
  [SMALL_STATE(355)] = 18288,
  [SMALL_STATE(356)] = 18319,
  [SMALL_STATE(357)] = 18348,
  [SMALL_STATE(358)] = 18377,
  [SMALL_STATE(359)] = 18406,
  [SMALL_STATE(360)] = 18435,
  [SMALL_STATE(361)] = 18464,
  [SMALL_STATE(362)] = 18493,
  [SMALL_STATE(363)] = 18522,
  [SMALL_STATE(364)] = 18551,
  [SMALL_STATE(365)] = 18580,
  [SMALL_STATE(366)] = 18623,
  [SMALL_STATE(367)] = 18652,
  [SMALL_STATE(368)] = 18681,
  [SMALL_STATE(369)] = 18710,
  [SMALL_STATE(370)] = 18739,
  [SMALL_STATE(371)] = 18768,
  [SMALL_STATE(372)] = 18797,
  [SMALL_STATE(373)] = 18826,
  [SMALL_STATE(374)] = 18855,
  [SMALL_STATE(375)] = 18886,
  [SMALL_STATE(376)] = 18915,
  [SMALL_STATE(377)] = 18946,
  [SMALL_STATE(378)] = 18975,
  [SMALL_STATE(379)] = 19005,
  [SMALL_STATE(380)] = 19035,
  [SMALL_STATE(381)] = 19065,
  [SMALL_STATE(382)] = 19095,
  [SMALL_STATE(383)] = 19125,
  [SMALL_STATE(384)] = 19155,
  [SMALL_STATE(385)] = 19185,
  [SMALL_STATE(386)] = 19215,
  [SMALL_STATE(387)] = 19245,
  [SMALL_STATE(388)] = 19275,
  [SMALL_STATE(389)] = 19305,
  [SMALL_STATE(390)] = 19335,
  [SMALL_STATE(391)] = 19365,
  [SMALL_STATE(392)] = 19395,
  [SMALL_STATE(393)] = 19425,
  [SMALL_STATE(394)] = 19455,
  [SMALL_STATE(395)] = 19485,
  [SMALL_STATE(396)] = 19515,
  [SMALL_STATE(397)] = 19545,
  [SMALL_STATE(398)] = 19575,
  [SMALL_STATE(399)] = 19605,
  [SMALL_STATE(400)] = 19635,
  [SMALL_STATE(401)] = 19665,
  [SMALL_STATE(402)] = 19695,
  [SMALL_STATE(403)] = 19725,
  [SMALL_STATE(404)] = 19755,
  [SMALL_STATE(405)] = 19785,
  [SMALL_STATE(406)] = 19815,
  [SMALL_STATE(407)] = 19851,
  [SMALL_STATE(408)] = 19881,
  [SMALL_STATE(409)] = 19911,
  [SMALL_STATE(410)] = 19943,
  [SMALL_STATE(411)] = 19981,
  [SMALL_STATE(412)] = 20011,
  [SMALL_STATE(413)] = 20038,
  [SMALL_STATE(414)] = 20073,
  [SMALL_STATE(415)] = 20100,
  [SMALL_STATE(416)] = 20127,
  [SMALL_STATE(417)] = 20162,
  [SMALL_STATE(418)] = 20193,
  [SMALL_STATE(419)] = 20220,
  [SMALL_STATE(420)] = 20255,
  [SMALL_STATE(421)] = 20282,
  [SMALL_STATE(422)] = 20309,
  [SMALL_STATE(423)] = 20344,
  [SMALL_STATE(424)] = 20379,
  [SMALL_STATE(425)] = 20414,
  [SMALL_STATE(426)] = 20447,
  [SMALL_STATE(427)] = 20474,
  [SMALL_STATE(428)] = 20509,
  [SMALL_STATE(429)] = 20536,
  [SMALL_STATE(430)] = 20571,
  [SMALL_STATE(431)] = 20606,
  [SMALL_STATE(432)] = 20641,
  [SMALL_STATE(433)] = 20668,
  [SMALL_STATE(434)] = 20698,
  [SMALL_STATE(435)] = 20726,
  [SMALL_STATE(436)] = 20752,
  [SMALL_STATE(437)] = 20778,
  [SMALL_STATE(438)] = 20804,
  [SMALL_STATE(439)] = 20833,
  [SMALL_STATE(440)] = 20858,
  [SMALL_STATE(441)] = 20887,
  [SMALL_STATE(442)] = 20916,
  [SMALL_STATE(443)] = 20941,
  [SMALL_STATE(444)] = 20966,
  [SMALL_STATE(445)] = 20991,
  [SMALL_STATE(446)] = 21020,
  [SMALL_STATE(447)] = 21049,
  [SMALL_STATE(448)] = 21078,
  [SMALL_STATE(449)] = 21103,
  [SMALL_STATE(450)] = 21128,
  [SMALL_STATE(451)] = 21157,
  [SMALL_STATE(452)] = 21186,
  [SMALL_STATE(453)] = 21211,
  [SMALL_STATE(454)] = 21238,
  [SMALL_STATE(455)] = 21267,
  [SMALL_STATE(456)] = 21296,
  [SMALL_STATE(457)] = 21325,
  [SMALL_STATE(458)] = 21354,
  [SMALL_STATE(459)] = 21383,
  [SMALL_STATE(460)] = 21408,
  [SMALL_STATE(461)] = 21433,
  [SMALL_STATE(462)] = 21462,
  [SMALL_STATE(463)] = 21491,
  [SMALL_STATE(464)] = 21515,
  [SMALL_STATE(465)] = 21559,
  [SMALL_STATE(466)] = 21583,
  [SMALL_STATE(467)] = 21627,
  [SMALL_STATE(468)] = 21668,
  [SMALL_STATE(469)] = 21711,
  [SMALL_STATE(470)] = 21734,
  [SMALL_STATE(471)] = 21757,
  [SMALL_STATE(472)] = 21780,
  [SMALL_STATE(473)] = 21803,
  [SMALL_STATE(474)] = 21826,
  [SMALL_STATE(475)] = 21849,
  [SMALL_STATE(476)] = 21872,
  [SMALL_STATE(477)] = 21895,
  [SMALL_STATE(478)] = 21918,
  [SMALL_STATE(479)] = 21941,
  [SMALL_STATE(480)] = 21964,
  [SMALL_STATE(481)] = 21987,
  [SMALL_STATE(482)] = 22010,
  [SMALL_STATE(483)] = 22033,
  [SMALL_STATE(484)] = 22056,
  [SMALL_STATE(485)] = 22079,
  [SMALL_STATE(486)] = 22102,
  [SMALL_STATE(487)] = 22125,
  [SMALL_STATE(488)] = 22148,
  [SMALL_STATE(489)] = 22171,
  [SMALL_STATE(490)] = 22194,
  [SMALL_STATE(491)] = 22217,
  [SMALL_STATE(492)] = 22240,
  [SMALL_STATE(493)] = 22263,
  [SMALL_STATE(494)] = 22286,
  [SMALL_STATE(495)] = 22309,
  [SMALL_STATE(496)] = 22332,
  [SMALL_STATE(497)] = 22355,
  [SMALL_STATE(498)] = 22378,
  [SMALL_STATE(499)] = 22401,
  [SMALL_STATE(500)] = 22424,
  [SMALL_STATE(501)] = 22447,
  [SMALL_STATE(502)] = 22470,
  [SMALL_STATE(503)] = 22493,
  [SMALL_STATE(504)] = 22516,
  [SMALL_STATE(505)] = 22539,
  [SMALL_STATE(506)] = 22562,
  [SMALL_STATE(507)] = 22585,
  [SMALL_STATE(508)] = 22608,
  [SMALL_STATE(509)] = 22631,
  [SMALL_STATE(510)] = 22674,
  [SMALL_STATE(511)] = 22697,
  [SMALL_STATE(512)] = 22720,
  [SMALL_STATE(513)] = 22755,
  [SMALL_STATE(514)] = 22778,
  [SMALL_STATE(515)] = 22801,
  [SMALL_STATE(516)] = 22824,
  [SMALL_STATE(517)] = 22847,
  [SMALL_STATE(518)] = 22870,
  [SMALL_STATE(519)] = 22893,
  [SMALL_STATE(520)] = 22934,
  [SMALL_STATE(521)] = 22957,
  [SMALL_STATE(522)] = 22980,
  [SMALL_STATE(523)] = 23003,
  [SMALL_STATE(524)] = 23026,
  [SMALL_STATE(525)] = 23049,
  [SMALL_STATE(526)] = 23072,
  [SMALL_STATE(527)] = 23095,
  [SMALL_STATE(528)] = 23118,
  [SMALL_STATE(529)] = 23141,
  [SMALL_STATE(530)] = 23164,
  [SMALL_STATE(531)] = 23187,
  [SMALL_STATE(532)] = 23210,
  [SMALL_STATE(533)] = 23233,
  [SMALL_STATE(534)] = 23256,
  [SMALL_STATE(535)] = 23279,
  [SMALL_STATE(536)] = 23302,
  [SMALL_STATE(537)] = 23325,
  [SMALL_STATE(538)] = 23348,
  [SMALL_STATE(539)] = 23371,
  [SMALL_STATE(540)] = 23394,
  [SMALL_STATE(541)] = 23429,
  [SMALL_STATE(542)] = 23452,
  [SMALL_STATE(543)] = 23475,
  [SMALL_STATE(544)] = 23510,
  [SMALL_STATE(545)] = 23533,
  [SMALL_STATE(546)] = 23556,
  [SMALL_STATE(547)] = 23579,
  [SMALL_STATE(548)] = 23602,
  [SMALL_STATE(549)] = 23625,
  [SMALL_STATE(550)] = 23662,
  [SMALL_STATE(551)] = 23699,
  [SMALL_STATE(552)] = 23734,
  [SMALL_STATE(553)] = 23769,
  [SMALL_STATE(554)] = 23806,
  [SMALL_STATE(555)] = 23843,
  [SMALL_STATE(556)] = 23872,
  [SMALL_STATE(557)] = 23909,
  [SMALL_STATE(558)] = 23946,
  [SMALL_STATE(559)] = 23983,
  [SMALL_STATE(560)] = 24020,
  [SMALL_STATE(561)] = 24048,
  [SMALL_STATE(562)] = 24078,
  [SMALL_STATE(563)] = 24110,
  [SMALL_STATE(564)] = 24142,
  [SMALL_STATE(565)] = 24174,
  [SMALL_STATE(566)] = 24205,
  [SMALL_STATE(567)] = 24236,
  [SMALL_STATE(568)] = 24261,
  [SMALL_STATE(569)] = 24288,
  [SMALL_STATE(570)] = 24319,
  [SMALL_STATE(571)] = 24350,
  [SMALL_STATE(572)] = 24373,
  [SMALL_STATE(573)] = 24402,
  [SMALL_STATE(574)] = 24431,
  [SMALL_STATE(575)] = 24451,
  [SMALL_STATE(576)] = 24479,
  [SMALL_STATE(577)] = 24507,
  [SMALL_STATE(578)] = 24535,
  [SMALL_STATE(579)] = 24561,
  [SMALL_STATE(580)] = 24589,
  [SMALL_STATE(581)] = 24615,
  [SMALL_STATE(582)] = 24643,
  [SMALL_STATE(583)] = 24671,
  [SMALL_STATE(584)] = 24699,
  [SMALL_STATE(585)] = 24727,
  [SMALL_STATE(586)] = 24755,
  [SMALL_STATE(587)] = 24777,
  [SMALL_STATE(588)] = 24799,
  [SMALL_STATE(589)] = 24819,
  [SMALL_STATE(590)] = 24839,
  [SMALL_STATE(591)] = 24867,
  [SMALL_STATE(592)] = 24887,
  [SMALL_STATE(593)] = 24907,
  [SMALL_STATE(594)] = 24927,
  [SMALL_STATE(595)] = 24947,
  [SMALL_STATE(596)] = 24967,
  [SMALL_STATE(597)] = 24987,
  [SMALL_STATE(598)] = 25007,
  [SMALL_STATE(599)] = 25029,
  [SMALL_STATE(600)] = 25057,
  [SMALL_STATE(601)] = 25085,
  [SMALL_STATE(602)] = 25113,
  [SMALL_STATE(603)] = 25141,
  [SMALL_STATE(604)] = 25169,
  [SMALL_STATE(605)] = 25197,
  [SMALL_STATE(606)] = 25221,
  [SMALL_STATE(607)] = 25249,
  [SMALL_STATE(608)] = 25277,
  [SMALL_STATE(609)] = 25297,
  [SMALL_STATE(610)] = 25316,
  [SMALL_STATE(611)] = 25337,
  [SMALL_STATE(612)] = 25360,
  [SMALL_STATE(613)] = 25379,
  [SMALL_STATE(614)] = 25398,
  [SMALL_STATE(615)] = 25417,
  [SMALL_STATE(616)] = 25436,
  [SMALL_STATE(617)] = 25463,
  [SMALL_STATE(618)] = 25480,
  [SMALL_STATE(619)] = 25503,
  [SMALL_STATE(620)] = 25528,
  [SMALL_STATE(621)] = 25551,
  [SMALL_STATE(622)] = 25578,
  [SMALL_STATE(623)] = 25599,
  [SMALL_STATE(624)] = 25620,
  [SMALL_STATE(625)] = 25647,
  [SMALL_STATE(626)] = 25674,
  [SMALL_STATE(627)] = 25693,
  [SMALL_STATE(628)] = 25710,
  [SMALL_STATE(629)] = 25731,
  [SMALL_STATE(630)] = 25750,
  [SMALL_STATE(631)] = 25767,
  [SMALL_STATE(632)] = 25786,
  [SMALL_STATE(633)] = 25807,
  [SMALL_STATE(634)] = 25832,
  [SMALL_STATE(635)] = 25851,
  [SMALL_STATE(636)] = 25870,
  [SMALL_STATE(637)] = 25893,
  [SMALL_STATE(638)] = 25918,
  [SMALL_STATE(639)] = 25939,
  [SMALL_STATE(640)] = 25966,
  [SMALL_STATE(641)] = 25987,
  [SMALL_STATE(642)] = 26012,
  [SMALL_STATE(643)] = 26035,
  [SMALL_STATE(644)] = 26054,
  [SMALL_STATE(645)] = 26079,
  [SMALL_STATE(646)] = 26106,
  [SMALL_STATE(647)] = 26125,
  [SMALL_STATE(648)] = 26150,
  [SMALL_STATE(649)] = 26175,
  [SMALL_STATE(650)] = 26196,
  [SMALL_STATE(651)] = 26221,
  [SMALL_STATE(652)] = 26240,
  [SMALL_STATE(653)] = 26259,
  [SMALL_STATE(654)] = 26284,
  [SMALL_STATE(655)] = 26307,
  [SMALL_STATE(656)] = 26330,
  [SMALL_STATE(657)] = 26347,
  [SMALL_STATE(658)] = 26369,
  [SMALL_STATE(659)] = 26387,
  [SMALL_STATE(660)] = 26405,
  [SMALL_STATE(661)] = 26423,
  [SMALL_STATE(662)] = 26439,
  [SMALL_STATE(663)] = 26461,
  [SMALL_STATE(664)] = 26479,
  [SMALL_STATE(665)] = 26499,
  [SMALL_STATE(666)] = 26517,
  [SMALL_STATE(667)] = 26535,
  [SMALL_STATE(668)] = 26551,
  [SMALL_STATE(669)] = 26569,
  [SMALL_STATE(670)] = 26589,
  [SMALL_STATE(671)] = 26607,
  [SMALL_STATE(672)] = 26625,
  [SMALL_STATE(673)] = 26643,
  [SMALL_STATE(674)] = 26661,
  [SMALL_STATE(675)] = 26679,
  [SMALL_STATE(676)] = 26697,
  [SMALL_STATE(677)] = 26715,
  [SMALL_STATE(678)] = 26733,
  [SMALL_STATE(679)] = 26751,
  [SMALL_STATE(680)] = 26773,
  [SMALL_STATE(681)] = 26789,
  [SMALL_STATE(682)] = 26807,
  [SMALL_STATE(683)] = 26825,
  [SMALL_STATE(684)] = 26843,
  [SMALL_STATE(685)] = 26865,
  [SMALL_STATE(686)] = 26887,
  [SMALL_STATE(687)] = 26903,
  [SMALL_STATE(688)] = 26923,
  [SMALL_STATE(689)] = 26941,
  [SMALL_STATE(690)] = 26959,
  [SMALL_STATE(691)] = 26975,
  [SMALL_STATE(692)] = 26997,
  [SMALL_STATE(693)] = 27013,
  [SMALL_STATE(694)] = 27031,
  [SMALL_STATE(695)] = 27047,
  [SMALL_STATE(696)] = 27065,
  [SMALL_STATE(697)] = 27087,
  [SMALL_STATE(698)] = 27109,
  [SMALL_STATE(699)] = 27133,
  [SMALL_STATE(700)] = 27155,
  [SMALL_STATE(701)] = 27175,
  [SMALL_STATE(702)] = 27197,
  [SMALL_STATE(703)] = 27215,
  [SMALL_STATE(704)] = 27233,
  [SMALL_STATE(705)] = 27251,
  [SMALL_STATE(706)] = 27273,
  [SMALL_STATE(707)] = 27293,
  [SMALL_STATE(708)] = 27315,
  [SMALL_STATE(709)] = 27333,
  [SMALL_STATE(710)] = 27355,
  [SMALL_STATE(711)] = 27377,
  [SMALL_STATE(712)] = 27397,
  [SMALL_STATE(713)] = 27419,
  [SMALL_STATE(714)] = 27435,
  [SMALL_STATE(715)] = 27453,
  [SMALL_STATE(716)] = 27470,
  [SMALL_STATE(717)] = 27485,
  [SMALL_STATE(718)] = 27502,
  [SMALL_STATE(719)] = 27519,
  [SMALL_STATE(720)] = 27536,
  [SMALL_STATE(721)] = 27553,
  [SMALL_STATE(722)] = 27574,
  [SMALL_STATE(723)] = 27591,
  [SMALL_STATE(724)] = 27608,
  [SMALL_STATE(725)] = 27627,
  [SMALL_STATE(726)] = 27646,
  [SMALL_STATE(727)] = 27663,
  [SMALL_STATE(728)] = 27684,
  [SMALL_STATE(729)] = 27705,
  [SMALL_STATE(730)] = 27724,
  [SMALL_STATE(731)] = 27745,
  [SMALL_STATE(732)] = 27762,
  [SMALL_STATE(733)] = 27779,
  [SMALL_STATE(734)] = 27798,
  [SMALL_STATE(735)] = 27817,
  [SMALL_STATE(736)] = 27832,
  [SMALL_STATE(737)] = 27849,
  [SMALL_STATE(738)] = 27868,
  [SMALL_STATE(739)] = 27883,
  [SMALL_STATE(740)] = 27900,
  [SMALL_STATE(741)] = 27919,
  [SMALL_STATE(742)] = 27934,
  [SMALL_STATE(743)] = 27955,
  [SMALL_STATE(744)] = 27970,
  [SMALL_STATE(745)] = 27989,
  [SMALL_STATE(746)] = 28004,
  [SMALL_STATE(747)] = 28021,
  [SMALL_STATE(748)] = 28036,
  [SMALL_STATE(749)] = 28051,
  [SMALL_STATE(750)] = 28066,
  [SMALL_STATE(751)] = 28085,
  [SMALL_STATE(752)] = 28100,
  [SMALL_STATE(753)] = 28119,
  [SMALL_STATE(754)] = 28136,
  [SMALL_STATE(755)] = 28153,
  [SMALL_STATE(756)] = 28174,
  [SMALL_STATE(757)] = 28189,
  [SMALL_STATE(758)] = 28210,
  [SMALL_STATE(759)] = 28229,
  [SMALL_STATE(760)] = 28246,
  [SMALL_STATE(761)] = 28261,
  [SMALL_STATE(762)] = 28282,
  [SMALL_STATE(763)] = 28297,
  [SMALL_STATE(764)] = 28318,
  [SMALL_STATE(765)] = 28339,
  [SMALL_STATE(766)] = 28360,
  [SMALL_STATE(767)] = 28381,
  [SMALL_STATE(768)] = 28398,
  [SMALL_STATE(769)] = 28413,
  [SMALL_STATE(770)] = 28434,
  [SMALL_STATE(771)] = 28449,
  [SMALL_STATE(772)] = 28468,
  [SMALL_STATE(773)] = 28483,
  [SMALL_STATE(774)] = 28498,
  [SMALL_STATE(775)] = 28519,
  [SMALL_STATE(776)] = 28538,
  [SMALL_STATE(777)] = 28559,
  [SMALL_STATE(778)] = 28580,
  [SMALL_STATE(779)] = 28597,
  [SMALL_STATE(780)] = 28612,
  [SMALL_STATE(781)] = 28627,
  [SMALL_STATE(782)] = 28646,
  [SMALL_STATE(783)] = 28663,
  [SMALL_STATE(784)] = 28680,
  [SMALL_STATE(785)] = 28699,
  [SMALL_STATE(786)] = 28716,
  [SMALL_STATE(787)] = 28737,
  [SMALL_STATE(788)] = 28754,
  [SMALL_STATE(789)] = 28771,
  [SMALL_STATE(790)] = 28788,
  [SMALL_STATE(791)] = 28805,
  [SMALL_STATE(792)] = 28824,
  [SMALL_STATE(793)] = 28841,
  [SMALL_STATE(794)] = 28860,
  [SMALL_STATE(795)] = 28875,
  [SMALL_STATE(796)] = 28894,
  [SMALL_STATE(797)] = 28909,
  [SMALL_STATE(798)] = 28928,
  [SMALL_STATE(799)] = 28949,
  [SMALL_STATE(800)] = 28968,
  [SMALL_STATE(801)] = 28987,
  [SMALL_STATE(802)] = 29008,
  [SMALL_STATE(803)] = 29023,
  [SMALL_STATE(804)] = 29042,
  [SMALL_STATE(805)] = 29061,
  [SMALL_STATE(806)] = 29080,
  [SMALL_STATE(807)] = 29099,
  [SMALL_STATE(808)] = 29120,
  [SMALL_STATE(809)] = 29137,
  [SMALL_STATE(810)] = 29158,
  [SMALL_STATE(811)] = 29179,
  [SMALL_STATE(812)] = 29196,
  [SMALL_STATE(813)] = 29217,
  [SMALL_STATE(814)] = 29232,
  [SMALL_STATE(815)] = 29247,
  [SMALL_STATE(816)] = 29262,
  [SMALL_STATE(817)] = 29278,
  [SMALL_STATE(818)] = 29296,
  [SMALL_STATE(819)] = 29314,
  [SMALL_STATE(820)] = 29332,
  [SMALL_STATE(821)] = 29346,
  [SMALL_STATE(822)] = 29360,
  [SMALL_STATE(823)] = 29374,
  [SMALL_STATE(824)] = 29392,
  [SMALL_STATE(825)] = 29410,
  [SMALL_STATE(826)] = 29424,
  [SMALL_STATE(827)] = 29442,
  [SMALL_STATE(828)] = 29460,
  [SMALL_STATE(829)] = 29474,
  [SMALL_STATE(830)] = 29488,
  [SMALL_STATE(831)] = 29502,
  [SMALL_STATE(832)] = 29516,
  [SMALL_STATE(833)] = 29534,
  [SMALL_STATE(834)] = 29548,
  [SMALL_STATE(835)] = 29562,
  [SMALL_STATE(836)] = 29578,
  [SMALL_STATE(837)] = 29596,
  [SMALL_STATE(838)] = 29612,
  [SMALL_STATE(839)] = 29626,
  [SMALL_STATE(840)] = 29640,
  [SMALL_STATE(841)] = 29658,
  [SMALL_STATE(842)] = 29672,
  [SMALL_STATE(843)] = 29690,
  [SMALL_STATE(844)] = 29704,
  [SMALL_STATE(845)] = 29718,
  [SMALL_STATE(846)] = 29732,
  [SMALL_STATE(847)] = 29750,
  [SMALL_STATE(848)] = 29768,
  [SMALL_STATE(849)] = 29786,
  [SMALL_STATE(850)] = 29804,
  [SMALL_STATE(851)] = 29818,
  [SMALL_STATE(852)] = 29836,
  [SMALL_STATE(853)] = 29850,
  [SMALL_STATE(854)] = 29868,
  [SMALL_STATE(855)] = 29884,
  [SMALL_STATE(856)] = 29902,
  [SMALL_STATE(857)] = 29920,
  [SMALL_STATE(858)] = 29938,
  [SMALL_STATE(859)] = 29956,
  [SMALL_STATE(860)] = 29974,
  [SMALL_STATE(861)] = 29992,
  [SMALL_STATE(862)] = 30010,
  [SMALL_STATE(863)] = 30028,
  [SMALL_STATE(864)] = 30046,
  [SMALL_STATE(865)] = 30064,
  [SMALL_STATE(866)] = 30082,
  [SMALL_STATE(867)] = 30100,
  [SMALL_STATE(868)] = 30114,
  [SMALL_STATE(869)] = 30132,
  [SMALL_STATE(870)] = 30148,
  [SMALL_STATE(871)] = 30166,
  [SMALL_STATE(872)] = 30184,
  [SMALL_STATE(873)] = 30202,
  [SMALL_STATE(874)] = 30220,
  [SMALL_STATE(875)] = 30238,
  [SMALL_STATE(876)] = 30252,
  [SMALL_STATE(877)] = 30270,
  [SMALL_STATE(878)] = 30288,
  [SMALL_STATE(879)] = 30306,
  [SMALL_STATE(880)] = 30324,
  [SMALL_STATE(881)] = 30342,
  [SMALL_STATE(882)] = 30360,
  [SMALL_STATE(883)] = 30378,
  [SMALL_STATE(884)] = 30394,
  [SMALL_STATE(885)] = 30412,
  [SMALL_STATE(886)] = 30430,
  [SMALL_STATE(887)] = 30448,
  [SMALL_STATE(888)] = 30466,
  [SMALL_STATE(889)] = 30480,
  [SMALL_STATE(890)] = 30494,
  [SMALL_STATE(891)] = 30512,
  [SMALL_STATE(892)] = 30530,
  [SMALL_STATE(893)] = 30548,
  [SMALL_STATE(894)] = 30566,
  [SMALL_STATE(895)] = 30584,
  [SMALL_STATE(896)] = 30598,
  [SMALL_STATE(897)] = 30616,
  [SMALL_STATE(898)] = 30634,
  [SMALL_STATE(899)] = 30652,
  [SMALL_STATE(900)] = 30670,
  [SMALL_STATE(901)] = 30688,
  [SMALL_STATE(902)] = 30706,
  [SMALL_STATE(903)] = 30724,
  [SMALL_STATE(904)] = 30742,
  [SMALL_STATE(905)] = 30758,
  [SMALL_STATE(906)] = 30776,
  [SMALL_STATE(907)] = 30794,
  [SMALL_STATE(908)] = 30808,
  [SMALL_STATE(909)] = 30823,
  [SMALL_STATE(910)] = 30838,
  [SMALL_STATE(911)] = 30853,
  [SMALL_STATE(912)] = 30868,
  [SMALL_STATE(913)] = 30881,
  [SMALL_STATE(914)] = 30896,
  [SMALL_STATE(915)] = 30909,
  [SMALL_STATE(916)] = 30922,
  [SMALL_STATE(917)] = 30935,
  [SMALL_STATE(918)] = 30950,
  [SMALL_STATE(919)] = 30965,
  [SMALL_STATE(920)] = 30978,
  [SMALL_STATE(921)] = 30991,
  [SMALL_STATE(922)] = 31006,
  [SMALL_STATE(923)] = 31021,
  [SMALL_STATE(924)] = 31036,
  [SMALL_STATE(925)] = 31051,
  [SMALL_STATE(926)] = 31066,
  [SMALL_STATE(927)] = 31081,
  [SMALL_STATE(928)] = 31094,
  [SMALL_STATE(929)] = 31109,
  [SMALL_STATE(930)] = 31124,
  [SMALL_STATE(931)] = 31137,
  [SMALL_STATE(932)] = 31150,
  [SMALL_STATE(933)] = 31163,
  [SMALL_STATE(934)] = 31178,
  [SMALL_STATE(935)] = 31193,
  [SMALL_STATE(936)] = 31206,
  [SMALL_STATE(937)] = 31219,
  [SMALL_STATE(938)] = 31234,
  [SMALL_STATE(939)] = 31247,
  [SMALL_STATE(940)] = 31262,
  [SMALL_STATE(941)] = 31275,
  [SMALL_STATE(942)] = 31290,
  [SMALL_STATE(943)] = 31303,
  [SMALL_STATE(944)] = 31318,
  [SMALL_STATE(945)] = 31331,
  [SMALL_STATE(946)] = 31344,
  [SMALL_STATE(947)] = 31359,
  [SMALL_STATE(948)] = 31374,
  [SMALL_STATE(949)] = 31387,
  [SMALL_STATE(950)] = 31402,
  [SMALL_STATE(951)] = 31417,
  [SMALL_STATE(952)] = 31432,
  [SMALL_STATE(953)] = 31445,
  [SMALL_STATE(954)] = 31458,
  [SMALL_STATE(955)] = 31471,
  [SMALL_STATE(956)] = 31484,
  [SMALL_STATE(957)] = 31499,
  [SMALL_STATE(958)] = 31514,
  [SMALL_STATE(959)] = 31527,
  [SMALL_STATE(960)] = 31540,
  [SMALL_STATE(961)] = 31553,
  [SMALL_STATE(962)] = 31568,
  [SMALL_STATE(963)] = 31583,
  [SMALL_STATE(964)] = 31598,
  [SMALL_STATE(965)] = 31613,
  [SMALL_STATE(966)] = 31628,
  [SMALL_STATE(967)] = 31643,
  [SMALL_STATE(968)] = 31656,
  [SMALL_STATE(969)] = 31669,
  [SMALL_STATE(970)] = 31682,
  [SMALL_STATE(971)] = 31697,
  [SMALL_STATE(972)] = 31712,
  [SMALL_STATE(973)] = 31725,
  [SMALL_STATE(974)] = 31738,
  [SMALL_STATE(975)] = 31753,
  [SMALL_STATE(976)] = 31768,
  [SMALL_STATE(977)] = 31781,
  [SMALL_STATE(978)] = 31794,
  [SMALL_STATE(979)] = 31807,
  [SMALL_STATE(980)] = 31820,
  [SMALL_STATE(981)] = 31833,
  [SMALL_STATE(982)] = 31846,
  [SMALL_STATE(983)] = 31859,
  [SMALL_STATE(984)] = 31874,
  [SMALL_STATE(985)] = 31887,
  [SMALL_STATE(986)] = 31900,
  [SMALL_STATE(987)] = 31913,
  [SMALL_STATE(988)] = 31928,
  [SMALL_STATE(989)] = 31940,
  [SMALL_STATE(990)] = 31952,
  [SMALL_STATE(991)] = 31964,
  [SMALL_STATE(992)] = 31976,
  [SMALL_STATE(993)] = 31988,
  [SMALL_STATE(994)] = 32000,
  [SMALL_STATE(995)] = 32012,
  [SMALL_STATE(996)] = 32024,
  [SMALL_STATE(997)] = 32036,
  [SMALL_STATE(998)] = 32048,
  [SMALL_STATE(999)] = 32060,
  [SMALL_STATE(1000)] = 32072,
  [SMALL_STATE(1001)] = 32084,
  [SMALL_STATE(1002)] = 32096,
  [SMALL_STATE(1003)] = 32108,
  [SMALL_STATE(1004)] = 32120,
  [SMALL_STATE(1005)] = 32132,
  [SMALL_STATE(1006)] = 32144,
  [SMALL_STATE(1007)] = 32156,
  [SMALL_STATE(1008)] = 32168,
  [SMALL_STATE(1009)] = 32180,
  [SMALL_STATE(1010)] = 32192,
  [SMALL_STATE(1011)] = 32204,
  [SMALL_STATE(1012)] = 32216,
  [SMALL_STATE(1013)] = 32228,
  [SMALL_STATE(1014)] = 32240,
  [SMALL_STATE(1015)] = 32252,
  [SMALL_STATE(1016)] = 32264,
  [SMALL_STATE(1017)] = 32276,
  [SMALL_STATE(1018)] = 32288,
  [SMALL_STATE(1019)] = 32300,
  [SMALL_STATE(1020)] = 32312,
  [SMALL_STATE(1021)] = 32324,
  [SMALL_STATE(1022)] = 32336,
  [SMALL_STATE(1023)] = 32348,
  [SMALL_STATE(1024)] = 32360,
  [SMALL_STATE(1025)] = 32372,
  [SMALL_STATE(1026)] = 32384,
  [SMALL_STATE(1027)] = 32396,
  [SMALL_STATE(1028)] = 32408,
  [SMALL_STATE(1029)] = 32420,
  [SMALL_STATE(1030)] = 32432,
  [SMALL_STATE(1031)] = 32444,
  [SMALL_STATE(1032)] = 32456,
  [SMALL_STATE(1033)] = 32468,
  [SMALL_STATE(1034)] = 32480,
  [SMALL_STATE(1035)] = 32492,
  [SMALL_STATE(1036)] = 32504,
  [SMALL_STATE(1037)] = 32516,
  [SMALL_STATE(1038)] = 32528,
  [SMALL_STATE(1039)] = 32540,
  [SMALL_STATE(1040)] = 32552,
  [SMALL_STATE(1041)] = 32564,
  [SMALL_STATE(1042)] = 32576,
  [SMALL_STATE(1043)] = 32588,
  [SMALL_STATE(1044)] = 32600,
  [SMALL_STATE(1045)] = 32612,
  [SMALL_STATE(1046)] = 32624,
  [SMALL_STATE(1047)] = 32636,
  [SMALL_STATE(1048)] = 32648,
  [SMALL_STATE(1049)] = 32660,
  [SMALL_STATE(1050)] = 32672,
  [SMALL_STATE(1051)] = 32684,
  [SMALL_STATE(1052)] = 32696,
  [SMALL_STATE(1053)] = 32708,
  [SMALL_STATE(1054)] = 32720,
  [SMALL_STATE(1055)] = 32732,
  [SMALL_STATE(1056)] = 32744,
  [SMALL_STATE(1057)] = 32756,
  [SMALL_STATE(1058)] = 32768,
  [SMALL_STATE(1059)] = 32780,
  [SMALL_STATE(1060)] = 32792,
  [SMALL_STATE(1061)] = 32804,
  [SMALL_STATE(1062)] = 32816,
  [SMALL_STATE(1063)] = 32828,
  [SMALL_STATE(1064)] = 32840,
  [SMALL_STATE(1065)] = 32852,
  [SMALL_STATE(1066)] = 32864,
  [SMALL_STATE(1067)] = 32876,
  [SMALL_STATE(1068)] = 32888,
  [SMALL_STATE(1069)] = 32900,
  [SMALL_STATE(1070)] = 32912,
  [SMALL_STATE(1071)] = 32924,
  [SMALL_STATE(1072)] = 32936,
  [SMALL_STATE(1073)] = 32948,
  [SMALL_STATE(1074)] = 32960,
  [SMALL_STATE(1075)] = 32972,
  [SMALL_STATE(1076)] = 32984,
  [SMALL_STATE(1077)] = 32996,
  [SMALL_STATE(1078)] = 33008,
  [SMALL_STATE(1079)] = 33020,
  [SMALL_STATE(1080)] = 33032,
  [SMALL_STATE(1081)] = 33044,
  [SMALL_STATE(1082)] = 33056,
  [SMALL_STATE(1083)] = 33068,
  [SMALL_STATE(1084)] = 33080,
  [SMALL_STATE(1085)] = 33092,
  [SMALL_STATE(1086)] = 33104,
  [SMALL_STATE(1087)] = 33116,
  [SMALL_STATE(1088)] = 33128,
  [SMALL_STATE(1089)] = 33140,
  [SMALL_STATE(1090)] = 33152,
  [SMALL_STATE(1091)] = 33164,
  [SMALL_STATE(1092)] = 33176,
  [SMALL_STATE(1093)] = 33188,
  [SMALL_STATE(1094)] = 33200,
  [SMALL_STATE(1095)] = 33212,
  [SMALL_STATE(1096)] = 33224,
  [SMALL_STATE(1097)] = 33236,
  [SMALL_STATE(1098)] = 33248,
  [SMALL_STATE(1099)] = 33260,
  [SMALL_STATE(1100)] = 33272,
  [SMALL_STATE(1101)] = 33284,
  [SMALL_STATE(1102)] = 33296,
  [SMALL_STATE(1103)] = 33308,
  [SMALL_STATE(1104)] = 33320,
  [SMALL_STATE(1105)] = 33332,
  [SMALL_STATE(1106)] = 33344,
  [SMALL_STATE(1107)] = 33356,
  [SMALL_STATE(1108)] = 33368,
  [SMALL_STATE(1109)] = 33380,
  [SMALL_STATE(1110)] = 33392,
  [SMALL_STATE(1111)] = 33404,
  [SMALL_STATE(1112)] = 33416,
  [SMALL_STATE(1113)] = 33428,
  [SMALL_STATE(1114)] = 33440,
  [SMALL_STATE(1115)] = 33452,
  [SMALL_STATE(1116)] = 33464,
  [SMALL_STATE(1117)] = 33476,
  [SMALL_STATE(1118)] = 33488,
  [SMALL_STATE(1119)] = 33500,
  [SMALL_STATE(1120)] = 33512,
  [SMALL_STATE(1121)] = 33524,
  [SMALL_STATE(1122)] = 33536,
  [SMALL_STATE(1123)] = 33548,
  [SMALL_STATE(1124)] = 33560,
  [SMALL_STATE(1125)] = 33572,
  [SMALL_STATE(1126)] = 33584,
  [SMALL_STATE(1127)] = 33596,
  [SMALL_STATE(1128)] = 33608,
  [SMALL_STATE(1129)] = 33620,
  [SMALL_STATE(1130)] = 33632,
  [SMALL_STATE(1131)] = 33644,
  [SMALL_STATE(1132)] = 33656,
  [SMALL_STATE(1133)] = 33668,
  [SMALL_STATE(1134)] = 33680,
  [SMALL_STATE(1135)] = 33692,
  [SMALL_STATE(1136)] = 33704,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1123),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_iri, 3, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_iri, 3, 0, 0),
  [11] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 1, 0, 0),
  [13] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 1, 0, 0),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [17] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_body, 2, 0, 30),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(330),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [51] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [73] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_value, 1, 0, 0),
  [75] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_simple_value, 1, 0, 0),
  [77] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_truth, 1, 0, 0),
  [79] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean_truth, 1, 0, 0),
  [81] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_falsity, 1, 0, 0),
  [83] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean_falsity, 1, 0, 0),
  [85] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [87] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [89] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 35),
  [91] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 35),
  [93] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary, 2, 0, 0),
  [95] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary, 2, 0, 0),
  [97] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary, 3, 0, 58),
  [99] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary, 3, 0, 58),
  [101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_identifier, 3, 0, 19),
  [103] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_qualified_identifier, 3, 0, 19),
  [105] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier_reference, 1, 0, 0),
  [107] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_identifier_reference, 1, 0, 0),
  [109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_atomic_sentence, 3, 0, 65),
  [111] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functional_term, 3, 0, 32),
  [113] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functional_term, 3, 0, 32),
  [115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_atomic_sentence, 4, 0, 90),
  [117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functional_term, 4, 0, 91),
  [119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functional_term, 4, 0, 91),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(30),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(36),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(840),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6),
  [147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(215),
  [157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1064),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1077),
  [165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_composition, 2, 0, 36),
  [175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_composition, 2, 0, 36),
  [177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 61),
  [179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 61),
  [181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 61), SHIFT_REPEAT(1086),
  [184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_term, 1, 0, 0),
  [186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_term, 1, 0, 0),
  [188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 9),
  [190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 9),
  [192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reserved_event, 1, 0, 0),
  [194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reserved_event, 1, 0, 0),
  [196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_predicate_value, 1, 0, 0),
  [198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_predicate_value, 1, 0, 0),
  [200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 2, 0, 0),
  [202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 2, 0, 0),
  [204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 5, 0, 106),
  [206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 5, 0, 106),
  [208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 4, 0, 87),
  [210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 4, 0, 87),
  [212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 4, 0, 86),
  [214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 4, 0, 86),
  [216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 5, 0, 76),
  [218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 5, 0, 76),
  [220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 5, 0, 77),
  [222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 5, 0, 77),
  [224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_builder, 5, 0, 108),
  [226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_builder, 5, 0, 108),
  [228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 6, 0, 121),
  [230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 6, 0, 121),
  [232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 6, 0, 99),
  [234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 6, 0, 99),
  [236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 6, 0, 122),
  [238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 6, 0, 122),
  [240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 7, 0, 128),
  [242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 7, 0, 128),
  [244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 3, 0, 58),
  [246] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 3, 0, 58),
  [248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_predicate_values, 1, 0, 22),
  [250] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_predicate_values, 1, 0, 22),
  [252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(127),
  [254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(260),
  [260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(428),
  [262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(689),
  [264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(802),
  [268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(815),
  [270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(631),
  [272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(667),
  [276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(596),
  [280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(976),
  [286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(75),
  [288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 6, 0, 126),
  [290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 6, 0, 126),
  [292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 3, 0, 75),
  [294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 3, 0, 75),
  [296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 4, 0, 97),
  [298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 4, 0, 97),
  [300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 4, 0, 98),
  [302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 4, 0, 98),
  [304] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 4, 0, 100),
  [306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 4, 0, 100),
  [308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 5, 0, 117),
  [310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 5, 0, 117),
  [312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 5, 0, 119),
  [314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 5, 0, 119),
  [316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_expression, 5, 0, 118),
  [318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_expression, 5, 0, 118),
  [320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 3, 0, 75),
  [322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 3, 0, 75),
  [324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 3, 0, 77),
  [326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 3, 0, 77),
  [328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 100),
  [330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 100),
  [332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 3, 0, 76),
  [334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 3, 0, 76),
  [336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 97),
  [338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 97),
  [340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 98),
  [342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 98),
  [344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 99),
  [346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 4, 0, 99),
  [348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 5, 0, 118),
  [350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 5, 0, 118),
  [352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 5, 0, 117),
  [354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 5, 0, 117),
  [356] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 5, 0, 119),
  [358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 5, 0, 119),
  [360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 6, 0, 126),
  [362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 6, 0, 126),
  [364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cardinality_reference_expression, 2, 0, 0),
  [366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_reference_expression, 2, 0, 0),
  [368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(727),
  [390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(254),
  [406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(263),
  [412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(275),
  [426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(276),
  [430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [436] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(172),
  [439] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(254),
  [442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92),
  [444] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(259),
  [447] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(263),
  [450] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(562),
  [453] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(264),
  [456] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(1109),
  [459] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(873),
  [462] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(274),
  [465] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(274),
  [468] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(275),
  [471] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(275),
  [474] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(276),
  [477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(276),
  [480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 2, 0, 92), SHIFT_REPEAT(278),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_boolean_sentence, 2, 0, 38),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_boolean_sentence, 3, 0, 63),
  [499] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quantified_sentence, 3, 0, 64),
  [501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [545] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 61), SHIFT_REPEAT(1001),
  [548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(998),
  [550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1085),
  [552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inequation, 3, 0, 68),
  [556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(304),
  [564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(636),
  [566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(319),
  [576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(311),
  [578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_equation, 3, 0, 67),
  [598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [618] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(162),
  [621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59),
  [623] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(304),
  [626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(636),
  [629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(269),
  [632] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(1094),
  [635] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(856),
  [638] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(319),
  [641] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(319),
  [644] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(311),
  [647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(311),
  [650] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(298),
  [653] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(298),
  [656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 2, 0, 59), SHIFT_REPEAT(380),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [691] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_expression_sentence, 3, 0, 63),
  [693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arithmetic_expression_sentence, 3, 0, 63),
  [695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_sentence, 1, 0, 0),
  [697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constraint_sentence, 1, 0, 0),
  [699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_constructor, 4, 0, 88),
  [701] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sentence_with_environment, 5, 0, 107),
  [703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_sentence, 1, 0, 0),
  [705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_sentence, 1, 0, 0),
  [707] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constraint_sentence, 3, 0, 0),
  [709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(632),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(651),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(629),
  [731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(620),
  [745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(589),
  [763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(608),
  [765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(591),
  [769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_disjunction, 1, 0, 0),
  [775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_disjunction, 1, 0, 0),
  [777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_conjunction, 1, 0, 0),
  [779] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_conjunction, 1, 0, 0),
  [781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_connective, 1, 0, 0),
  [783] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_connective, 1, 0, 0),
  [785] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_implication, 1, 0, 0),
  [787] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_implication, 1, 0, 0),
  [789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_biconditional, 1, 0, 0),
  [791] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_biconditional, 1, 0, 0),
  [793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_exclusive_disjunction, 1, 0, 0),
  [795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_exclusive_disjunction, 1, 0, 0),
  [797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_op_negation, 1, 0, 0),
  [799] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_logical_op_negation, 1, 0, 0),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59),
  [805] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(304),
  [808] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(269),
  [811] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(1094),
  [814] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(856),
  [817] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(319),
  [820] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(319),
  [823] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(311),
  [826] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(311),
  [829] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(298),
  [832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 2, 0, 59), SHIFT_REPEAT(298),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0),
  [843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1003),
  [846] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1114),
  [849] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1115),
  [852] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1032),
  [855] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1108),
  [858] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(727),
  [861] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(729),
  [864] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(990),
  [867] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1091),
  [870] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1100),
  [873] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 2, 0, 0), SHIFT_REPEAT(1008),
  [876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(255),
  [936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [940] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_composition_repeat1, 2, 0, 61), SHIFT_REPEAT(997),
  [943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_builtin_types, 1, 0, 0),
  [945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0),
  [947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0), SHIFT_REPEAT(900),
  [950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0), SHIFT_REPEAT(1105),
  [953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mapping_type, 5, 0, 116),
  [957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__actual_arguments_repeat1, 1, 0, 66),
  [961] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__actual_arguments_repeat1, 1, 0, 66),
  [963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value, 1, 0, 0),
  [967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_import, 1, 0, 6),
  [971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1082),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 2, 0, 0),
  [977] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 2, 0, 0), SHIFT_REPEAT(1130),
  [980] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 2, 0, 0), SHIFT_REPEAT(1051),
  [983] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 2, 0, 0), SHIFT_REPEAT(1004),
  [986] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 2, 0, 0), SHIFT_REPEAT(1010),
  [989] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 2, 0, 0), SHIFT_REPEAT(1036),
  [992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [1000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [1002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value, 1, 0, 0),
  [1004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_informal_constraint, 2, 0, 24),
  [1006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_informal_constraint, 2, 0, 24),
  [1008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [1010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [1012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1130),
  [1014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [1016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [1018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [1020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [1022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [1024] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_body_repeat1, 2, 0, 0),
  [1026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat1, 2, 0, 0), SHIFT_REPEAT(710),
  [1029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat1, 2, 0, 0), SHIFT_REPEAT(685),
  [1032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [1034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type_reference, 1, 0, 0),
  [1036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_op_inequality, 1, 0, 0),
  [1038] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_op_inequality, 1, 0, 0),
  [1040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inequality_relation, 1, 0, 0),
  [1042] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inequality_relation, 1, 0, 0),
  [1044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_op_less_than_or_equal, 1, 0, 0),
  [1046] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_op_less_than_or_equal, 1, 0, 0),
  [1048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_op_greater_than_or_equal, 1, 0, 0),
  [1050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_op_greater_than_or_equal, 1, 0, 0),
  [1052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_intersection, 1, 0, 0),
  [1054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_intersection, 1, 0, 0),
  [1056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [1058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_subset, 1, 0, 0),
  [1060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_subset, 1, 0, 0),
  [1062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_subset_or_equal, 1, 0, 0),
  [1064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_subset_or_equal, 1, 0, 0),
  [1066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_supset, 1, 0, 0),
  [1068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_supset, 1, 0, 0),
  [1070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature, 6, 0, 95),
  [1072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_supset_or_equal, 1, 0, 0),
  [1074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_supset_or_equal, 1, 0, 0),
  [1076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation_property, 4, 0, 23),
  [1078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_complement, 1, 0, 0),
  [1080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_complement, 1, 0, 0),
  [1082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_product, 1, 0, 0),
  [1084] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_product, 1, 0, 0),
  [1086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_is_member, 1, 0, 0),
  [1088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_is_member, 1, 0, 0),
  [1090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 1, 0, 34),
  [1092] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_sequence_of_predicate_values_repeat1, 1, 0, 34),
  [1094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_is_not_member, 1, 0, 0),
  [1096] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_is_not_member, 1, 0, 0),
  [1098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_math_op_multiply, 1, 0, 0),
  [1100] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_math_op_multiply, 1, 0, 0),
  [1102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_math_op_divide, 1, 0, 0),
  [1104] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_math_op_divide, 1, 0, 0),
  [1106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_is_not_member, 2, 0, 0),
  [1108] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_is_not_member, 2, 0, 0),
  [1110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_op_union, 1, 0, 0),
  [1112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_op_union, 1, 0, 0),
  [1114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constraint, 3, 0, 1),
  [1116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_math_op_modulo, 1, 0, 0),
  [1118] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_math_op_modulo, 1, 0, 0),
  [1120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_math_op_add, 1, 0, 0),
  [1122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_math_op_add, 1, 0, 0),
  [1124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_math_operator, 1, 0, 0),
  [1126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_math_operator, 1, 0, 0),
  [1128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 5, 0, 106),
  [1130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_operator, 1, 0, 0),
  [1132] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_operator, 1, 0, 0),
  [1134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 5, 0, 77),
  [1136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_membership, 1, 0, 0),
  [1138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_membership, 1, 0, 0),
  [1140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_body_repeat2, 1, 0, 0),
  [1142] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 2, 0, 0),
  [1144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mapping_value, 3, 0, 60),
  [1146] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature, 7, 0, 115),
  [1148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 4, 0, 86),
  [1150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_informal_constraint, 3, 0, 40),
  [1152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 3, 0, 58),
  [1154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature, 3, 0, 28),
  [1156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 6, 0, 121),
  [1158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 6, 0, 99),
  [1160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 1, 0, 22),
  [1162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 6, 0, 122),
  [1164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 4, 0, 87),
  [1166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 7, 0, 128),
  [1168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_constraint, 3, 0, 30),
  [1170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation, 1, 0, 0),
  [1172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_op_equality, 1, 0, 0),
  [1174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_op_equality, 1, 0, 0),
  [1176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature, 4, 0, 48),
  [1178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_op_less_than, 1, 0, 0),
  [1180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_op_less_than, 1, 0, 0),
  [1182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_of_values, 5, 0, 76),
  [1184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_constructor, 4, 0, 88),
  [1186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 2, 0, 0),
  [1188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 3, 0, 58),
  [1190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mapping_value, 3, 0, 60),
  [1192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 4, 0, 86),
  [1194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 4, 0, 87),
  [1196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 5, 0, 76),
  [1198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 5, 0, 106),
  [1200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 5, 0, 77),
  [1202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 1, 0, 22),
  [1204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 6, 0, 121),
  [1206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 6, 0, 99),
  [1208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 6, 0, 122),
  [1210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sequence_of_values, 7, 0, 128),
  [1212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_import, 2, 0, 13),
  [1214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1126),
  [1216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [1218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_def, 1, 0, 10),
  [1220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [1222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_import, 1, 0, 6),
  [1224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1128),
  [1226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_def, 2, 0, 17),
  [1228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [1230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 2, 0, 9),
  [1232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [1234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [1236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_import, 3, 0, 18),
  [1238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 5, 0, 57),
  [1240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 5, 0, 0),
  [1242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_body_repeat1, 1, 0, 0),
  [1244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [1246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_def, 2, 0, 9),
  [1248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_of_values_repeat1, 1, 0, 34),
  [1250] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_sequence_of_values_repeat1, 1, 0, 34),
  [1252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_import, 3, 0, 18),
  [1254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [1256] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_def, 2, 0, 9),
  [1258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 2, 0, 0),
  [1260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_import, 4, 0, 33),
  [1262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 4, 0, 26),
  [1264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [1266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_def, 2, 0, 9),
  [1268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [1270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [1272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_def, 4, 0, 29),
  [1274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rdf_def, 2, 0, 9),
  [1276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [1278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 3, 0, 0),
  [1280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_def, 2, 0, 9),
  [1282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [1284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_reference, 1, 0, 0),
  [1286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [1288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_def, 2, 0, 9),
  [1290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 5, 0, 43),
  [1292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [1294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_def, 2, 0, 9),
  [1296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 4, 0, 0),
  [1298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_member, 1, 0, 70),
  [1300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [1302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_def, 2, 0, 10),
  [1304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_def, 3, 0, 17),
  [1306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation_only_body, 3, 0, 0),
  [1308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rdf_def, 6, 0, 80),
  [1310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_without_members, 2, 0, 11),
  [1312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 3, 0, 16),
  [1314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 6, 0, 73),
  [1316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_with_members, 4, 0, 109),
  [1318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_member, 3, 0, 111),
  [1320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_without_members, 4, 0, 109),
  [1322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_def, 3, 0, 16),
  [1324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_def, 3, 0, 16),
  [1326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 6, 0, 84),
  [1328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_base_type_reference, 1, 0, 0),
  [1330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_with_members, 2, 0, 71),
  [1332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_def, 3, 0, 16),
  [1334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_def, 3, 0, 28),
  [1336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_from_definition_clause, 3, 0, 41),
  [1338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_def, 3, 0, 16),
  [1340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_def, 3, 0, 16),
  [1342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 5, 0, 45),
  [1344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rdf_def, 4, 0, 31),
  [1346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_def, 4, 0, 48),
  [1348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_event_binding, 2, 0, 0),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mixin_with_members, 2, 0, 11),
  [1352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_def, 5, 0, 49),
  [1354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_def, 3, 0, 16),
  [1356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_type_restrictions, 3, 0, 0),
  [1358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(484),
  [1360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [1362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [1364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [1366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [1368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_def, 2, 0, 0),
  [1370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(535),
  [1372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [1374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [1376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [1378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(476),
  [1380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [1382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(894),
  [1384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1078),
  [1386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_definition, 1, 0, 0),
  [1388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rdf_def, 7, 0, 102),
  [1390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 7, 0, 104),
  [1392] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_body, 4, 0, 82),
  [1394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_def, 4, 0, 51),
  [1396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_body, 3, 0, 55),
  [1398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 4, 0, 25),
  [1400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_body, 3, 0, 0),
  [1402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_def, 4, 0, 25),
  [1404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation_member_def, 2, 0, 0),
  [1406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rdf_def, 3, 0, 1),
  [1408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_def, 5, 0, 79),
  [1410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 6, 0, 85),
  [1412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_def, 3, 0, 1),
  [1414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rdf_def, 5, 0, 53),
  [1416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_body, 3, 0, 0),
  [1418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 6, 0, 72),
  [1420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_body, 4, 0, 0),
  [1422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_def, 4, 0, 25),
  [1424] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_def, 4, 0, 25),
  [1426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_body, 2, 0, 0),
  [1428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_def, 4, 0, 25),
  [1430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_datatype_type_restrictions_repeat1, 1, 0, 0),
  [1432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_body, 2, 0, 0),
  [1434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_body, 2, 0, 0),
  [1436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_length_restriction_facet, 3, 0, 112),
  [1438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_digit_restriction_facet, 3, 0, 112),
  [1440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_restriction_facet, 1, 0, 0),
  [1442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_restriction_value, 1, 0, 0),
  [1444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_restriction_facet, 3, 0, 112),
  [1446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tz_restriction_value, 1, 0, 0),
  [1448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tz_restriction_facet, 3, 0, 113),
  [1450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_restriction_facet, 3, 0, 113),
  [1452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_body, 6, 0, 69),
  [1454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_body_repeat3, 1, 0, 0),
  [1456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 6, 0, 74),
  [1458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_body, 4, 0, 42),
  [1460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_body, 4, 0, 69),
  [1462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_def, 3, 0, 1),
  [1464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_body, 4, 0, 0),
  [1466] = {.entry = {.count = 1, .reusable = false}}, SHIFT(492),
  [1468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_body, 5, 0, 69),
  [1470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_body, 4, 0, 69),
  [1472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_def, 3, 0, 1),
  [1474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_body, 4, 0, 42),
  [1476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_body, 4, 0, 69),
  [1478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_def, 3, 0, 1),
  [1480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_body, 4, 0, 0),
  [1482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_def, 3, 0, 1),
  [1484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [1486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_body, 3, 0, 0),
  [1488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_body, 5, 0, 42),
  [1490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_body, 3, 0, 42),
  [1492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_body, 5, 0, 69),
  [1494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_body, 4, 0, 0),
  [1496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_body, 5, 0, 69),
  [1498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_length_restriction_facet, 4, 0, 123),
  [1500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_digit_restriction_facet, 4, 0, 123),
  [1502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_restriction_facet, 4, 0, 123),
  [1504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 5, 0, 44),
  [1506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tz_restriction_facet, 4, 0, 124),
  [1508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_def, 6, 0, 78),
  [1510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_def, 3, 0, 1),
  [1512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dimension_body, 3, 0, 42),
  [1514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_def, 2, 0, 11),
  [1516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_body, 2, 0, 0),
  [1518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_body, 3, 0, 42),
  [1520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_restriction_facet, 5, 0, 129),
  [1522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_body, 3, 0, 0),
  [1524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_class_body, 2, 0, 0),
  [1526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_def, 3, 0, 1),
  [1528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure_def, 4, 0, 25),
  [1530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_body, 4, 0, 42),
  [1532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_event_def, 4, 0, 25),
  [1534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_def, 5, 0, 50),
  [1536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_group_body, 3, 0, 0),
  [1538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype_def, 7, 0, 94),
  [1540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(489),
  [1542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [1544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [1546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [1548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [1550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [1552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [1554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1113),
  [1556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [1558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [1560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(533),
  [1562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(523),
  [1564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [1566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [1568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0),
  [1570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0), SHIFT_REPEAT(898),
  [1573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0), SHIFT_REPEAT(1095),
  [1576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(759),
  [1578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(772),
  [1580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(982),
  [1582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [1584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [1586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_source_entity, 2, 0, 41),
  [1588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [1590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(877),
  [1592] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0), SHIFT_REPEAT(894),
  [1595] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_body_repeat2, 2, 0, 0), SHIFT_REPEAT(1078),
  [1598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [1600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [1602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [1604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [1606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [1608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [1610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [1612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable, 1, 0, 6),
  [1614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [1616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(514),
  [1618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(515),
  [1620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [1622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [1624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(524),
  [1626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat2, 2, 0, 0),
  [1628] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat2, 2, 0, 0), SHIFT_REPEAT(912),
  [1631] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat2, 2, 0, 0), SHIFT_REPEAT(887),
  [1634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(510),
  [1636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(521),
  [1638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [1640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [1642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [1644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [1646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [1648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [1650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [1652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(502),
  [1654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [1656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(522),
  [1658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [1660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [1662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(761),
  [1664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(862),
  [1666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(544),
  [1668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(511),
  [1670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [1672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 0),
  [1674] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(621),
  [1677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [1679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [1681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1014),
  [1683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_import, 1, 0, 6),
  [1685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [1687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1076),
  [1689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [1691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat1, 2, 0, 0),
  [1693] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1113),
  [1696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [1698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [1700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(486),
  [1702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [1704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(653),
  [1706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_def, 3, 0, 28),
  [1708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1073),
  [1710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [1712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_union_body_repeat1, 2, 0, 0),
  [1714] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_union_body_repeat1, 2, 0, 0), SHIFT_REPEAT(658),
  [1717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [1719] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_def, 4, 0, 48),
  [1721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [1723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [1725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_ref, 2, 0, 101),
  [1727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [1729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_metric_group_body_repeat1, 2, 0, 0),
  [1731] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_metric_group_body_repeat1, 2, 0, 0), SHIFT_REPEAT(818),
  [1734] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_metric_group_body_repeat1, 2, 0, 0), SHIFT_REPEAT(819),
  [1737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [1739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dimension_parent, 4, 0, 114),
  [1741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [1743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [1745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [1747] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_formal_constraint, 3, 0, 30),
  [1749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_informal_constraint, 3, 0, 40),
  [1751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [1753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [1755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameter_repeat1, 2, 0, 0),
  [1757] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_parameter_repeat1, 2, 0, 0), SHIFT_REPEAT(968),
  [1760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [1762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_builtin_types, 1, 0, 0),
  [1764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(650),
  [1766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_member, 1, 0, 70),
  [1768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1048),
  [1770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__rdf_types_repeat1, 2, 0, 81),
  [1772] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__rdf_types_repeat1, 2, 0, 81), SHIFT_REPEAT(596),
  [1775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mapping_type, 5, 0, 116),
  [1777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter, 3, 0, 6),
  [1779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [1781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [1783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [1785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [1787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [1789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable, 3, 0, 89),
  [1791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(648),
  [1793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variant, 1, 0, 6),
  [1795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1009),
  [1797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [1799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [1801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [1803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [1805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [1807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [1809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [1811] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_body_repeat2, 1, 0, 0),
  [1813] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_annotation, 1, 0, 0),
  [1815] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constraint, 3, 0, 1),
  [1817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [1819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(859),
  [1821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [1823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter, 4, 0, 6),
  [1825] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_annotation_property, 4, 0, 23),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [1829] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_parameter_restriction_repeat1, 2, 0, 92), SHIFT_REPEAT(907),
  [1832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameter_restriction_repeat1, 2, 0, 92),
  [1834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [1836] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_reference, 1, 0, 0),
  [1838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [1840] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_member, 3, 0, 111),
  [1842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [1844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [1846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_restriction, 1, 0, 83),
  [1848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [1850] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_signature_repeat1, 2, 0, 47), SHIFT_REPEAT(967),
  [1853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_signature_repeat1, 2, 0, 47),
  [1855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(816),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1110),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [1863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [1865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sentence_with_environment_repeat1, 2, 0, 0),
  [1867] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sentence_with_environment_repeat1, 2, 0, 0), SHIFT_REPEAT(817),
  [1870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_class_body_repeat1, 2, 0, 56),
  [1872] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_class_body_repeat1, 2, 0, 56), SHIFT_REPEAT(817),
  [1875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_ref, 2, 0, 54),
  [1877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter, 1, 0, 6),
  [1879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [1881] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_builder_repeat1, 2, 0, 62), SHIFT_REPEAT(759),
  [1884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_builder_repeat1, 2, 0, 62),
  [1886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_source_entity, 3, 0, 41),
  [1888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1052),
  [1890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variant, 3, 0, 18),
  [1892] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat1, 1, 0, 0),
  [1894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [1896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_def, 4, 0, 51),
  [1900] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__class_type_parameters_repeat1, 2, 0, 47), SHIFT_REPEAT(736),
  [1903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__class_type_parameters_repeat1, 2, 0, 47),
  [1905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__mixin_member_or_sequence_repeat1, 2, 0, 110),
  [1907] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__mixin_member_or_sequence_repeat1, 2, 0, 110), SHIFT_REPEAT(899),
  [1910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [1912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [1914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [1916] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_with_members, 2, 0, 71),
  [1918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [1920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [1922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [1924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [1926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_with_members, 2, 0, 11),
  [1928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [1930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [1932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [1934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_without_members, 2, 0, 11),
  [1936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [1938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [1940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [1942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [1944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [1946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_def, 5, 0, 79),
  [1948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dimension_parent, 5, 0, 125),
  [1950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [1952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [1954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_ordering, 1, 0, 0),
  [1956] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_annotation_only_body, 3, 0, 0),
  [1958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [1960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [1962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [1964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(801),
  [1966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [1968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [1970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [1972] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member, 1, 0, 0),
  [1974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [1976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [1978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_restriction, 2, 0, 103),
  [1980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [1982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [1984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [1986] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_with_members, 4, 0, 109),
  [1988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [1990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [1992] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mixin_without_members, 4, 0, 109),
  [1994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [1996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(517),
  [1998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [2000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_enum_body_repeat1, 2, 0, 0),
  [2002] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_enum_body_repeat1, 2, 0, 0), SHIFT_REPEAT(805),
  [2005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [2007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [2009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [2011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [2013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [2015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_variant, 1, 0, 6),
  [2017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [2019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [2021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [2023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [2025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [2027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_entity_identity, 2, 0, 42),
  [2029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [2031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [2033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [2035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_metric_group_body_repeat1, 1, 0, 0),
  [2037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [2039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [2041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [2043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [2045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [2047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_restriction, 4, 0, 127),
  [2049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [2051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyword_function_def, 2, 0, 0),
  [2053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__module_path_actual_repeat1, 2, 0, 21),
  [2055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__module_path_actual_repeat1, 2, 0, 21), SHIFT_REPEAT(1054),
  [2058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pattern_restriction_facet_repeat1, 2, 0, 0),
  [2060] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pattern_restriction_facet_repeat1, 2, 0, 0), SHIFT_REPEAT(955),
  [2063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_restriction, 5, 0, 130),
  [2065] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_metric_ref, 3, 0, 120),
  [2067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [2069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [2071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [2073] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_path_actual, 2, 0, 15),
  [2075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1054),
  [2077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [2079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1030),
  [2081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_import, 1, 0, 6),
  [2083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [2085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1040),
  [2089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_import, 2, 0, 13),
  [2091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [2093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [2095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [2097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [2099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [2101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [2103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [2105] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameter_restriction_repeat1, 1, 0, 66),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [2109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [2111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [2113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [2117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [2119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [2121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [2123] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path_absolute, 1, 0, 0),
  [2125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(879),
  [2127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_path_actual, 1, 0, 7),
  [2129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [2131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [2133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [2135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [2137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [2139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_dimension_body_repeat2, 1, 0, 0),
  [2141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [2143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [2145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [2147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [2149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [2151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [2153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_binary_repeat1, 2, 0, 59),
  [2155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_binary_repeat1, 2, 0, 59), SHIFT_REPEAT(932),
  [2158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [2160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [2162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_restriction_argument, 1, 0, 0),
  [2164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [2166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [2168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [2170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [2172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [2174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [2176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [2178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_builder_repeat1, 1, 0, 37),
  [2180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_class_body_repeat1, 1, 0, 32),
  [2182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [2184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quantified_variable, 3, 0, 93),
  [2186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_union_body_repeat1, 1, 0, 0),
  [2188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__module_path_actual_repeat1, 2, 0, 20),
  [2190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [2192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [2194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [2196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [2198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [2200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [2202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [2204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [2206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [2208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [2210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [2212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [2214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__mixin_member_or_sequence_repeat1, 1, 0, 70),
  [2216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [2218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [2220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [2222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [2224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [2226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_binary_repeat1, 1, 0, 34),
  [2228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [2230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [2232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [2234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [2236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__rdf_types_repeat1, 1, 0, 52),
  [2238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [2240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [2242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [2244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter, 4, 0, 48),
  [2246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [2248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [2250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [2254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [2256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_variant, 2, 0, 46),
  [2258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sentence_with_environment_repeat1, 1, 0, 0),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [2268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_signature_repeat1, 1, 0, 27),
  [2270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [2272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [2274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [2276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [2278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [2280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [2282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__class_type_parameters_repeat1, 1, 0, 27),
  [2284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variant, 2, 0, 46),
  [2286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pattern_restriction_facet_repeat1, 1, 0, 0),
  [2290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [2292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter, 3, 0, 28),
  [2294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 1, 0, 0),
  [2296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [2298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [2300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [2302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [2304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [2306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [2308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [2310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [2312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [2314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [2316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [2318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [2320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [2322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_op_combiner, 1, 0, 0),
  [2324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variant, 4, 0, 105),
  [2326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [2328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [2330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [2332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [2334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_range, 1, 0, 0),
  [2336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [2338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [2340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [2342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_enum_body_repeat1, 1, 0, 0),
  [2344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [2346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_uniqueness, 1, 0, 0),
  [2348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [2350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [2352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [2354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [2356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [2358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [2360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [2362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [2366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [2370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 4, 0, 3),
  [2372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_body, 5, 0, 0),
  [2374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [2376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1124),
  [2378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1136),
  [2380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [2382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_body, 4, 0, 0),
  [2384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [2386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_version, 2, 0, 2),
  [2388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [2390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(941),
  [2392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [2394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [2396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [2398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [2400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [2402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [2404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [2406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [2408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [2410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cardinality_range, 2, 0, 96),
  [2412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [2414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [2418] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [2428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [2432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [2434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_from_clause, 2, 0, 0),
  [2436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [2438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 5, 0, 12),
  [2440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [2442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [2444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [2446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [2448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [2450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [2452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [2454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [2456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [2458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [2460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [2462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [2464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [2466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 4, 0, 4),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [2470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [2472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [2474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path_relative, 1, 0, 8),
  [2476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [2478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [2480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_body, 3, 0, 0),
  [2482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [2484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path_absolute, 2, 0, 14),
  [2486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_body, 2, 0, 0),
  [2488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 3, 0, 1),
  [2490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [2492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [2494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [2498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [2500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_quantifier_universal, 1, 0, 0),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [2504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [2506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [2508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [2510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [2512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [2514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [2518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [2520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [2522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1005),
  [2524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_quantifier_existential, 1, 0, 0),
  [2526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [2528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_quantifier_existential_one, 1, 0, 0),
  [2530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [2532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quantified_variable_binding, 2, 0, 39),
  [2534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [2536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [2538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [2540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [2542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [2544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [2546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [2548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [2550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [2552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [2554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [2556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1133),
  [2558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [2560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [2562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [2564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [2568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_version, 3, 0, 5),
  [2570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [2572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [2574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [2582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1028),
  [2584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [2586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [2588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(942),
  [2594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [2598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [2600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1058),
  [2602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1062),
  [2604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1066),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [2608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [2610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [2612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [2616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [2618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_quantifier, 1, 0, 0),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [2622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [2632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_line_comment, 1, 0, 0),
};

#ifdef __cplusplus
extern "C" {
#endif
#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_sdml(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .primary_state_ids = ts_primary_state_ids,
    .name = "sdml",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 4,
      .patch_version = 11,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
